module psflib


!------------------------------------------------------------------------------- 
! Routines to access and manipulate 
! OGIP compliant PSF calibration files 
! 
! 
! Author: GLamer, AIP,  date: 22-JUN-2010
! 
!
!------------------------------------------------------------------------------- 

  use param 
  use numkind
  use cfitsio_ifaces  
  use fitsimage
  
  
  implicit none

  public


  type psf2dhdr_type
    !sequence
    integer(kind=int32)   :: bitpix
    integer(kind=int32)   :: naxis
    integer(kind=int32)   :: naxis1
    integer(kind=int32)   :: naxis2
    character(len=68)     :: extname  
    character(len=68)      :: hduclass  
    character(len=68)      :: hdudoc  
    character(len=68)      :: hduclas1  
    character(len=68)      :: hduclas2     
    character(len=68)      :: hduclas3     
    character(len=68)      :: hduclas4
    character(len=68)      :: hduvers 
    character(len=68)      :: longstr
    character(len=68)      :: origin
    character(len=68)      :: creator
    character(len=68)      :: mission    
    character(len=8)      :: telescop
    character(len=8)      :: instrume
    character(len=8)      :: obs_mode 
    character(len=8)      :: ctype1
    real(kind=real64)     :: crpix1
    real(kind=real64)     :: crval1
    real(kind=real64)     :: cdelt1 
    character(len=8)      :: cunit1  
    character(len=8)      :: ctype2
    real(kind=real64)     :: crpix2 
    real(kind=real64)     :: crval2
    real(kind=real64)     :: cdelt2 
    character(len=8)      :: cunit2  
    character(len=8)      :: filter
    real(kind=real32)     :: backgrnd
    real(kind=real32)     :: sumrcts
    real(kind=real32)     :: energy
    real(kind=real32)     :: theta
    real(kind=real32)     :: phi    
    real(kind=real32)     :: energ_lo
    real(kind=real32)     :: energ_hi
    integer(kind=int32)   :: chanmin
    integer(kind=int32)   :: chanmax
    character(len=8)      :: chantype
    character(len=8)      :: ccls0001
    character(len=8)      :: cdtp0001 
    character(len=8)      :: ccnm0001 
    character(len=68)     :: cbd10001  
    character(len=68)     :: cbd20001 
    character(len=68)     :: cbd30001    
    character(len=10)     :: cvsd0001    
    character(len=8)      :: cvst0001  
    character(len=68)      :: cdes0001 
    character(len=1024)     :: history
    character(len=1024)     :: comment
  

    character(len=47) :: c_bitpix 
    character(len=47) :: c_naxis
    character(len=47) :: c_naxis1
    character(len=47) :: c_naxis2
    character(len=47) :: c_extname
    character(len=47) :: c_hduclass
    character(len=47) :: c_hdudoc  
    character(len=47) :: c_hduclas1
    character(len=47) :: c_hduclas2
    character(len=47) :: c_hduclas3
    character(len=47) :: c_hduclas4
    character(len=47) :: c_hduvers 
    character(len=47) :: c_longstr
    character(len=47) :: c_origin
    character(len=47) :: c_creator
    character(len=47) :: c_mission    
    character(len=47) :: c_telescop
    character(len=47) :: c_instrume
    character(len=47) :: c_obs_mode
    character(len=47) :: c_ctype1
    character(len=47) :: c_crpix1
    character(len=47) :: c_crval1
    character(len=47) :: c_cdelt1
    character(len=47) :: c_cunit1
    character(len=47) :: c_ctype2
    character(len=47) :: c_crpix2
    character(len=47) :: c_crval2
    character(len=47) :: c_cdelt2
    character(len=47) :: c_cunit2
    character(len=47) :: c_filter  
    character(len=47) :: c_backgrnd
    character(len=47) :: c_sumrcts
    character(len=47) :: c_energy
    character(len=47) :: c_theta
    character(len=47) :: c_phi
    character(len=47) :: c_energ_lo
    character(len=47) :: c_energ_hi
    character(len=47) :: c_chanmin 
    character(len=47) :: c_chanmax 
    character(len=47) :: c_chantype
    character(len=47) :: c_ccls0001
    character(len=47) :: c_cdtp0001
    character(len=47) :: c_ccnm0001
    character(len=47) :: c_cbd10001
    character(len=47) :: c_cbd20001
    character(len=47) :: c_cbd30001
    character(len=47) :: c_cvsd0001
    character(len=47) :: c_cvst0001
    character(len=47) :: c_cdes0001

    logical(kind=log32)   :: e_bitpix
    logical(kind=log32)   :: e_naxis
    logical(kind=log32)   :: e_naxis1
    logical(kind=log32)   :: e_naxis2
    logical(kind=log32)   :: e_extname
    logical(kind=log32)   :: e_hduclass
    logical(kind=log32)   :: e_hdudoc  
    logical(kind=log32)   :: e_hduclas1
    logical(kind=log32)   :: e_hduclas2
    logical(kind=log32)   :: e_hduclas3
    logical(kind=log32)   :: e_hduclas4
    logical(kind=log32)   :: e_hduvers 
    logical(kind=log32)   :: e_longstr
    logical(kind=log32)   :: e_origin
    logical(kind=log32)   :: e_creator
    logical(kind=log32)   :: e_mission
    logical(kind=log32)   :: e_telescop
    logical(kind=log32)   :: e_instrume
    logical(kind=log32)   :: e_obs_mode
    logical(kind=log32)   :: e_ctype1
    logical(kind=log32)   :: e_crpix1
    logical(kind=log32)   :: e_crval1
    logical(kind=log32)   :: e_cdelt1
    logical(kind=log32)   :: e_cunit1
    logical(kind=log32)   :: e_ctype2
    logical(kind=log32)   :: e_crpix2
    logical(kind=log32)   :: e_crval2
    logical(kind=log32)   :: e_cdelt2
    logical(kind=log32)   :: e_cunit2
    logical(kind=log32)   :: e_filter  
    logical(kind=log32)   :: e_backgrnd
    logical(kind=log32)   :: e_sumrcts
    logical(kind=log32)   :: e_energy
    logical(kind=log32)   :: e_theta
    logical(kind=log32)   :: e_phi
    logical(kind=log32)   :: e_energ_lo
    logical(kind=log32)   :: e_energ_hi
    logical(kind=log32)   :: e_chanmin 
    logical(kind=log32)   :: e_chanmax 
    logical(kind=log32)   :: e_chantype
    logical(kind=log32)   :: e_ccls0001
    logical(kind=log32)   :: e_cdtp0001
    logical(kind=log32)   :: e_ccnm0001
    logical(kind=log32)   :: e_cbd10001
    logical(kind=log32)   :: e_cbd20001
    logical(kind=log32)   :: e_cbd30001
    logical(kind=log32)   :: e_cvsd0001
    logical(kind=log32)   :: e_cvst0001
    logical(kind=log32)   :: e_cdes0001
    logical(kind=log32)   :: e_history
    logical(kind=log32)   :: e_comment
  end type psf2dhdr_type


  type psf2d_type
    !sequence
    real(kind=real32),  pointer, dimension(:,:) :: arr => NULL()
    type(psf2dhdr_type)   :: psfhdr
  end type  psf2d_type

!-----------------------------------------------------------------------------
! This type is only needed temporarily, the eROSITA CAL PSF will be real32

  type psf2d64_type
    !sequence
    real(kind=real64),  pointer, dimension(:,:) :: arr => NULL()
    type(psf2dhdr_type)   :: psfhdr
  end type  psf2d64_type
!-----------------------------------------------------------------------------



contains

  subroutine fwrite_psf2dhdr(unit,header,ierr)


    type(psf2dhdr_type)               :: header 
    integer(kind=int32),intent(in)    :: unit
    integer(kind=int32)               :: status,ierr 
    character(len=*), parameter       :: rtname = 'fwrite_psfheader'
 
  
    ierr=0
    status=0
   status=0

    ! shape keywords:
!    call FTPKYJ(unit,'BITPIX',header%bitpix ,header%c_bitpix,status)
!    call FTPKYJ(unit,'NAXIS',header%naxis   ,header%c_naxis,status)
!    call FTPKYJ(unit,'NAXIS1',header%naxis1 ,header%c_naxis1,status)
!    call FTPKYJ(unit,'NAXIS2',header%naxis2 ,header%c_naxis2,status)
    call FTPKYS(unit,'EXTNAME',header%extname,header%c_extname,status)
    ! standard:
    call FTPKYS(unit,'HDUCLASS',header%hduclass,header%c_hduclass,status) 
    call FTPKYS(unit,'HDUDOC',header%hdudoc,header%c_hdudoc,status) 
    call FTPKYS(unit,'HDUVERS',header%hduvers,header%c_hduvers,status) 
    call FTPKYS(unit,'HDUCLAS1',header%hduclas1,header%c_hduclas1,status) 
    call FTPKYS(unit,'HDUCLAS2',header%hduclas2,header%c_hduclas2,status) 
    call FTPKYS(unit,'HDUCLAS3',header%hduclas3,header%c_hduclas3,status) 
    call FTPKYS(unit,'HDUCLAS4',header%hduclas4,header%c_hduclas4,status) 
    call FTPKYS(unit,'LONGSTR',header%longstr,header%c_longstr,status)
    call FTPKYS(unit,'ORIGIN',header%origin,header%c_origin,status)
    call FTPKYS(unit,'CREATOR',header%creator,header%c_creator,status)
    ! write system date to DATE keyword:
    call FTPDAT(unit,status)    

    call FTPKYS(unit,'MISSION',header%mission,header%c_mission,status)
    call FTPKYS(unit,'TELESCOP',header%telescop,header%c_telescop,status)
    call FTPKYS(unit,'INSTRUME',header%instrume,header%c_instrume,status)
    call FTPKYS(unit,'FILTER',header%filter,header%c_filter,status)
    call FTPKYS(unit,'OBS_MODE',header%obs_mode,header%c_obs_mode,status)
    !wcs:
    call FTPKYS(unit,'CTYPE1',header%ctype1 ,header%c_ctype1 ,status)
    call FTPKYD(unit,'CRPIX1',header%crpix1 ,decimals,   header%c_crpix1,status)
    call FTPKYD(unit,'CRVAL1',header%crval1 ,decimals,   header%c_crval1 ,status)
    call FTPKYD(unit,'CDELT1',header%cdelt1 ,decimals,   header%c_cdelt1 ,status)
    call FTPKYS(unit,'CUNIT1',header%cunit1 ,header%c_cunit1 ,status)
    call FTPKYS(unit,'CTYPE2',header%ctype2 ,header%c_ctype2 ,status)
    call FTPKYD(unit,'CRPIX2',header%crpix2 ,decimals,   header%c_crpix2,status)
    call FTPKYD(unit,'CRVAL2',header%crval2 ,decimals,   header%c_crval2 ,status)
    call FTPKYD(unit,'CDELT2',header%cdelt2 ,decimals,   header%c_cdelt2 ,status)
    call FTPKYS(unit,'CUNIT2',header%cunit2 ,header%c_cunit2 ,status)
    ! caldb specific:  
    call FTPKYE(unit,'BACKGRND',header%backgrnd ,decimals,   header%c_backgrnd,status)
    call FTPKYE(unit,'SUMRCTS',header%sumrcts ,decimals,   header%c_sumrcts,status)
    call FTPKYE(unit,'ENERG_LO',header%energ_lo ,decimals,   header%c_energ_lo,status)
    call FTPKYE(unit,'ENERG_HI',header%energ_hi ,decimals,   header%c_energ_hi,status)
    call FTPKYJ(unit,'CHANMIN',header%chanmin ,header%c_chanmin,status)
    call FTPKYJ(unit,'CHANMAX',header%chanmax ,header%c_chanmax,status)
    call FTPKYS(unit,'CCLS0001',header%ccls0001,header%c_ccls0001,status)
    call FTPKYS(unit,'CDTP0001',header%cdtp0001,header%c_cdtp0001 ,status)
    call FTPKYS(unit,'CCNM0001',header%ccnm0001 ,header%c_ccnm0001 ,status)
    call FTPKYS(unit,'CBD10001',header%cbd10001 ,header%c_cbd10001 ,status)
    call FTPKYS(unit,'CBD20001',header%cbd20001 ,header%c_cbd20001 ,status)
    call FTPKYS(unit,'CBD30001',header%cbd30001 ,header%c_cbd30001 ,status)
    call FTPKYS(unit,'CVSD0001',header%cvsd0001 ,header%c_cvsd0001 ,status)
    call FTPKYS(unit,'CVST0001',header%cvst0001 ,header%c_cvst0001 ,status)
    call FTPKYS(unit,'CDES0001',header%cdes0001 ,header%c_cdes0001 ,status)

   ! write HISTORY, COMMENT:
    call FTPHIS(unit,trim(header%history),status)
    call FTPCOM(unit,trim(header%comment),status)


  end subroutine fwrite_psf2dhdr


!-------------------------------------------------------------------------------
  
  subroutine fread_psf2dhdr(unit,header,ierr)


    type(psf2dhdr_type)               :: header 
    integer(kind=int32),intent(in)    :: unit
    integer(kind=int32)               :: status,ierr 
    character(len=15)                 :: rtname = 'fread_psfheader'
 
    
    ierr=0
    status=0

    ! shape keywords:
    call FTGKYJ(unit,'BITPIX',header%bitpix ,header%c_bitpix,status)
    call freadkey_err(rtname,'BITPIX',status,header%e_bitpix,ierr)  
    call FTGKYJ(unit,'NAXIS',header%naxis   ,header%c_naxis,status)
    call freadkey_err(rtname,'NAXIS',status ,header%e_naxis,ierr)  
    call FTGKYJ(unit,'NAXIS1',header%naxis1 ,header%c_naxis1,status)
    call freadkey_err(rtname,'NAXIS1',status,header%e_naxis1,ierr)
    call FTGKYJ(unit,'NAXIS2',header%naxis2 ,header%c_naxis2,status)
    call freadkey_err(rtname,'NAXIS2',status,header%e_naxis2,ierr) 
    call FTGKYS(unit,'EXTNAME',header%extname ,header%c_extname,status)
    call freadkey_err(rtname,'EXTNAME',status,header%e_extname,ierr) 
    call FTGKYS(unit,'HDUCLASS',header%hduclass ,header%c_hduclass,status)
    call freadkey_err(rtname,'HDUCLASS',status,header%e_hduclass,ierr) 
    call FTGKYS(unit,'HDUDOC',header%hdudoc ,header%c_hdudoc,status)
    call freadkey_err(rtname,'HDUDOC',status,header%e_hdudoc,ierr) 
    call FTGKYS(unit,'HDUVERS',header%hduvers ,header%c_hduvers,status)
    call freadkey_err(rtname,'HDUVERS',status,header%e_hduvers,ierr) 
    call FTGKYS(unit,'HDUCLAS1',header%hduclas1 ,header%c_hduclas1,status)
    call freadkey_err(rtname,'HDUCLAS1',status,header%e_hduclas1,ierr) 
    call FTGKYS(unit,'HDUCLAS2',header%hduclas2 ,header%c_hduclas2,status)
    call freadkey_err(rtname,'HDUCLAS2',status,header%e_hduclas2,ierr) 
    call FTGKYS(unit,'HDUCLAS3',header%hduclas3 ,header%c_hduclas3,status)
    call freadkey_err(rtname,'HDUCLAS3',status,header%e_hduclas3,ierr) 
    call FTGKYS(unit,'HDUCLAS4',header%hduclas4 ,header%c_hduclas4,status)
    call freadkey_err(rtname,'HDUCLAS4',status,header%e_hduclas4,ierr)
    call FTGKYS(unit,'LONGSTR',header%longstr ,header%c_longstr,status)
    call freadkey_err(rtname,'LONGSTR',status,header%e_longstr,ierr)
    call FTGKYS(unit,'ORIGIN',header%origin ,header%c_origin,status)
    call freadkey_err(rtname,'ORIGIN',status,header%e_origin,ierr)
    call FTGKYS(unit,'CREATOR',header%creator ,header%c_creator,status)
    call freadkey_err(rtname,'CREATOR',status,header%e_creator,ierr)
    ! standard:
    call FTGKYS(unit,'MISSION',header%mission,header%c_mission,status)
    call freadkey_err(rtname,'MISSION',status ,header%e_mission,ierr) 
    call FTGKYS(unit,'TELESCOP',header%telescop,header%c_telescop,status)
    call freadkey_err(rtname,'TELESCOP',status ,header%e_telescop,ierr) 
    call FTGKYS(unit,'INSTRUME',header%instrume,header%c_instrume,status)
    call freadkey_err(rtname,'INSTRUME',status ,header%e_instrume,ierr) 
    call FTGKYS(unit,'FILTER',header%filter,header%c_filter,status)
    call freadkey_err(rtname,'FILTER',status ,header%e_filter,ierr) 
    call FTGKYS(unit,'OBS_MODE',header%obs_mode,header%c_obs_mode,status)
    call freadkey_err(rtname,'OBS_MODE',status ,header%e_obs_mode,ierr) 
    !wcs
    call FTGKYS(unit,'CTYPE1',header%ctype1 ,header%c_ctype1 ,status)
    call freadkey_err(rtname,'CTYPE1',status,header%e_ctype1 ,ierr)
    call FTGKYD(unit,'CRPIX1',header%crpix1 ,header%c_crpix1,status)
    call freadkey_err(rtname,'CRPIX1',status,header%e_crpix1,ierr)
    call FTGKYD(unit,'CRVAL1',header%crval1 ,header%c_crval1 ,status)
    call freadkey_err(rtname,'CRVAL1',status,header%e_crval1 ,ierr)
    call FTGKYD(unit,'CDELT1',header%cdelt1 ,header%c_cdelt1 ,status)
    call freadkey_err(rtname,'CDELT1',status,header%e_cdelt1, ierr)
    call FTGKYS(unit,'CUNIT1',header%cunit1 ,header%c_cunit1 ,status)
    call freadkey_err(rtname,'CUNIT1',status,header%e_cunit1 ,ierr)
    call FTGKYS(unit,'CTYPE2',header%ctype2 ,header%c_ctype2 ,status)
    call freadkey_err(rtname,'CTYPE2',status,header%e_ctype2 ,ierr)
    call FTGKYD(unit,'CRPIX2',header%crpix2 ,header%c_crpix2,status)
    call freadkey_err(rtname,'CRPIX2',status,header%e_crpix2,ierr)
    call FTGKYD(unit,'CRVAL2',header%crval2 ,header%c_crval2 ,status)
    call freadkey_err(rtname,'CRVAL2',status,header%e_crval2 ,ierr)
    call FTGKYD(unit,'CDELT2',header%cdelt2 ,header%c_cdelt2 ,status)
    call freadkey_err(rtname,'CDELT2',status,header%e_cdelt2, ierr)
    call FTGKYS(unit,'CUNIT2',header%cunit2 ,header%c_cunit2 ,status)
    call freadkey_err(rtname,'CUNIT2',status,header%e_cunit2 ,ierr)
    ! caldb specific:  
    call FTGKYE(unit,'ENERGY',header%energy ,header%c_energy,status)
    call freadkey_err(rtname,'ENERGY',status,header%e_energy,ierr)
    call FTGKYE(unit,'THETA',header%theta ,header%c_theta,status)
    call freadkey_err(rtname,'THETA',status,header%e_theta,ierr)
    call FTGKYE(unit,'PHI',header%phi      ,header%c_phi  ,status)
    call freadkey_err(rtname,'PHI'  ,status,header%e_phi  ,ierr)

    call FTGKYE(unit,'BACKGRND',header%backgrnd ,header%c_backgrnd,status)
    call freadkey_err(rtname,'BACKGRND',status,header%e_backgrnd,ierr)
    call FTGKYE(unit,'SUMRCTS',header%sumrcts ,header%c_sumrcts,status)
    call freadkey_err(rtname,'SUMRCTS',status,header%e_sumrcts,ierr)

    call FTGKYE(unit,'ENERG_LO',header%energ_lo ,header%c_energ_lo,status)
    call freadkey_err(rtname,'ENERG_LO',status,header%e_energ_lo,ierr)
    call FTGKYE(unit,'ENERG_HI',header%energ_hi ,header%c_energ_hi,status)
    call freadkey_err(rtname,'ENERG_HI',status,header%e_energ_hi,ierr)
    call FTGKYJ(unit,'CHANMIN',header%chanmin ,header%c_chanmin,status)
    call freadkey_err(rtname,'CHANMIN',status,header%e_chanmin,ierr)
    call FTGKYJ(unit,'CHANMAX',header%chanmax ,header%c_chanmax,status)
    call freadkey_err(rtname,'CHANMAX',status,header%e_chanmax,ierr)

  
  
    call FTGKYS(unit,'CCLS0001',header%ccls0001,header%c_ccls0001,status)
    call freadkey_err(rtname,'CCLS0001',status,header%e_ccls0001,ierr)   
    call FTGKYS(unit,'CDTP0001',header%cdtp0001,header%c_cdtp0001 ,status)
    call freadkey_err(rtname,'CDTP0001',status,header%e_cdtp0001 ,ierr)    
    call FTGKYS(unit,'CCNM0001',header%ccnm0001 ,header%c_ccnm0001 ,status)
    call freadkey_err(rtname,'CCNM0001',status,header%e_ccnm0001 ,ierr)    
    call FTGKYS(unit,'CBD10001',header%cbd10001 ,header%c_cbd10001 ,status)
    call freadkey_err(rtname,'CBD10001',status,header%e_cbd10001 ,ierr) 
    call FTGKYS(unit,'CBD20001',header%cbd20001 ,header%c_cbd20001 ,status)
    call freadkey_err(rtname,'CBD20001',status,header%e_cbd20001 ,ierr) 
    call FTGKYS(unit,'CBD30001',header%cbd30001 ,header%c_cbd30001 ,status)
    call freadkey_err(rtname,'CBD30001',status,header%e_cbd30001 ,ierr) 
    call FTGKYS(unit,'CVSD0001',header%cvsd0001 ,header%c_cvsd0001 ,status)
    call freadkey_err(rtname,'CVSD0001',status,header%e_cvsd0001 ,ierr)    
    call FTGKYS(unit,'CVST0001',header%cvst0001 ,header%c_cvst0001 ,status)
    call freadkey_err(rtname,'CVST0001',status,header%e_cvst0001 ,ierr)    
    call FTGKYS(unit,'CDES0001',header%cdes0001 ,header%c_cdes0001 ,status)
    call freadkey_err(rtname,'CDES0001',status,header%e_cdes0001 ,ierr)   

    call FTGCRD(unit,'HISTORY,',header%history,status)
    call FTGCRD(unit,'COMMENT,',header%comment,status)    


  end subroutine fread_psf2dhdr

!-------------------------------------------------------------------------------
  
  subroutine get_2dpsf(filename,ext,phi,psf,status)

 !------------------------------------------------------------------------------
 ! Extract 2 dimensional PSF image from given fits extension (FILENAME, EXT)
 ! Rotation of PSF image to position angle PHI not yet implemented
 !------------------------------------------------------------------------------

    type(psf2d_type)           :: psf
    character(len=*)           :: filename
    integer(kind=int32)        :: ext   
    real(kind=real32)          :: phi
    integer(kind=int32)        :: naxis,naxis1,naxis2
    integer(kind=int32)        :: group,unit,status,readwrite,blocksize,ierr
    integer(kind=int32)        :: nhdu, hdutype
    real(kind=real32)          :: null_r4 = -9.9E9, epsilon=1E-8
    character(len=*),parameter :: rtname = 'GET_2DPSF'
    logical                    :: anyf    

    nhdu=ext+1
    status=0
    readwrite=0
    ierr=0    
    group=0

!    write(*,*) rtname, ' ',filename

    call ftgiou(unit,status)
    call ftopen(unit,filename,readwrite,blocksize,status)

    if (status > 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error opening FITS file')
      write(*,*) 'filename: ',filename
    else
      call ftmahd(unit,nhdu,hdutype,status)
      call fread_psf2dhdr(unit,psf%psfhdr,status)
      if (status > 2 .or. psf%psfhdr%naxis /= 2) then
        ierr=3
        call wrerr(rtname,ierr,'FITS extension is not a 2D image')
      else
        if (trim(psf%psfhdr%ccnm0001) /= '2D_PSF') then
          ierr=1
          call wrerr(rtname,ierr,'FITS extension is not a 2D-PSF image')
        endif
        naxis1 = psf%psfhdr%naxis1
        naxis2 = psf%psfhdr%naxis2
        status=0
        if (associated(psf%arr)) deallocate(psf%arr)
        allocate(psf%arr(naxis1,naxis2))    
        call ftg2de(unit,group,null_r4,naxis1,naxis1,naxis2,psf%arr,anyf,status)
        if (status > 0) then
          ierr=3
          call wrerr(rtname,ierr,'Error reading FITS image extension')
        endif 
      end if 
      call ftclos(unit,status)
    end if
   
    ! check if PSF conforms to eROSITA SASS standards:
    ! 1. square pixels?
    if (ABS(psf%psfhdr%cdelt1 - psf%psfhdr%cdelt2) > epsilon) then
      ierr = max(ierr,3)
      call wrerr(rtname,ierr,'PSF pixels not square')
    endif

    ! 2. pixel size given in deg or arcsec?
    if (psf%psfhdr%cunit1 == psf%psfhdr%cunit2) then
      if ((psf%psfhdr%cunit1 /= 'DEG') .and. (psf%psfhdr%cunit1 /= 'deg')) then 
         ! change to deg:
         if ((psf%psfhdr%cunit1 == 'ARCSEC') .or. (psf%psfhdr%cunit1 == 'arcsec')) then
           psf%psfhdr%cdelt1 = psf%psfhdr%cdelt1/3600.D0
           psf%psfhdr%cdelt2 = psf%psfhdr%cdelt2/3600.D0
           psf%psfhdr%cunit1 = 'deg'
           psf%psfhdr%cunit2 = 'deg'           
         else
           ierr=3
           call wrerr(rtname,ierr,'PSF pixel size not given in units of deg')
         endif
      endif
    else
      ierr=3
      call wrerr(rtname,ierr,'Different units for PSF X and Y sizes')
    endif    
    
    call ftfiou(unit, status)

  end subroutine get_2dpsf



!-------------------------------------------------------------------------------

  subroutine psf_surfrad(psf,surf,exact,radius,ierr)

  !-------------------------------------------------------------
  !
  ! Calculate radius, where the PSF falls below a given 
  ! surface brightness (in counts/(arcsec^2).  
  ! It is assumed that CRPIX1 and CRPIX2 define the origin of 
  ! the  PSF.
  ! If the logical parameter 'exact' is set to .true.,
  ! the weight of fractional pixels will be calculated 
  ! (to be implemented).
  ! Input parameters:
  !   PSF : PSF image of type PSF2D_TYPE
  !   SURF: real32 value for surface brightness limit 
  !         unit: [ normalized PSF flux / PSF image pixel ]
  !   EXACT : logical parameter switching between normal (fast) and 
  !            exact calculation (TBD)
  ! Output parameters:
  !    RADIUS: radius in arcsec
  !    IERR  : SASS error status
  ! Algorithm: 
  !  - Sort PSF image according to distance from PSF centre 
  !  - Integrate PSF surface brightness in radial intervals of 1 pixel width,
  !    until value of 'SURF' is reached.
  !-------------------------------------------------------------


    type(psf2d_type),    intent(in)  :: psf  
    real(kind=real32), intent(in)    :: surf   
    logical(kind=log32), intent(in)  :: exact       
    real(kind=real32), intent(out)   :: radius
    integer(kind=int32), intent(out) :: ierr

    real(kind=real32), parameter     :: eps = 1E-6     
    integer(kind=real32)             :: ndata,sla_err
    real(kind=real32)                :: x0,y0,surfn, area,rad2
    real(kind=real32), dimension(:),allocatable,save    :: radarr,psfarr
    integer(kind=int32),dimension(:),allocatable,save   :: iperm
!    logical, save                        :: first=.true.
    integer(kind=int32), dimension(1)    :: arrshape
    integer(kind=int32)                  :: kflag,index,i,j,arrsize
    real(kind=real32)                    :: xdist,ydist,ydist2,psfsum,psfsurf,dist2
    real(kind=real32),save               :: crpix1old,crpix2old
    real(kind=real32)                    ::  rad_low, rad_hi,rad_low2,rad_hi2,rad_step 
    real(kind=real32)                    ::  pix2arcsec 
    character(len=*),parameter           :: rtname = 'psf_surfrad'

    ! system time variables
    integer(kind=real32),dimension(8)    :: timeval1,timeval2

    ierr=0
    sla_err=0
    ndata=psf%psfhdr%naxis1*psf%psfhdr%naxis2

    

    ! PSF pixel size, assume square PSF pixels:
    if (psf%psfhdr%e_cunit1 .and. psf%psfhdr%e_cdelt1 .and. psf%psfhdr%cdelt1 /= 0.0) then
      if ( trim(psf%psfhdr%cunit1) == 'arcsec' .or. trim(psf%psfhdr%cunit1) == 'ARCSEC') then   
        pix2arcsec=abs(psf%psfhdr%cdelt1)
      else if ( trim(psf%psfhdr%cunit1) == 'deg' .or. trim(psf%psfhdr%cunit1) == 'DEG') then
        pix2arcsec=abs(psf%psfhdr%cdelt1)*3600.0
      else
        pix2arcsec=abs(psf%psfhdr%cdelt1)
        ierr=3
        call wrerr(rtname,ierr,'Unknown CUNIT1 value in PSF file, assume arcsec')
      endif
    else
      pix2arcsec=1.0
      ierr=3
      call wrerr(rtname,ierr,'Unknown PSF pixel size, assume 1 arcsec')
    endif



    if (exact) then
      ierr=1
      call wrerr(rtname,ierr,'Warning: Exact mode not implemented yet')
    endif 

    if (surf  <= 0.0 .or. surf > 1.0) then
      ierr=3
      call wrerr(rtname,ierr,'Error: Surface brightness limit  not in (0.0:1.0)')
    else
      !re-normalize surface brightness limit to PSF normalization
      surfn = surf * psf%psfhdr%sumrcts  
      ! transform surface brightness limit from  counts/arcsec^2 to counts/(PSF pixel)^2
 
      surfn = surfn * pix2arcsec*pix2arcsec




      arrshape(1)=psf%psfhdr%naxis1*psf%psfhdr%naxis2

      if (.not. allocated(radarr))  allocate(radarr(arrshape(1)))
      if (.not. allocated(psfarr))  allocate(psfarr(arrshape(1)))
      if (.not. allocated(iperm))  allocate(iperm(arrshape(1)))


      if ((abs(psf%psfhdr%crpix1-crpix1old) > eps).or.(abs(psf%psfhdr%crpix2-crpix2old) > eps)) then

        index=0
        do j=1,psf%psfhdr%naxis2
          ydist=real(j)-psf%psfhdr%crpix2
          ydist2=ydist*ydist
          do i=1,psf%psfhdr%naxis1
            index=index+1
            xdist=real(i)-psf%psfhdr%crpix1
            radarr(index)=ydist2+xdist*xdist
           enddo  
        enddo 


        kflag=1 ! kflag=1 : ascending sort, don't sort psfarr
        call spsort(radarr,ndata,iperm,kflag,sla_err)


        if (sla_err /= 0) then
          ierr = 3
          call wrerr(rtname,ierr,'Error: SLATEC error in SPSORT)')
        endif

      endif
      sla_err = 0

     
      psfarr=reshape(psf%arr,arrshape)  
   



      rad_low=0.0
      rad_hi=1.5    
      radius = 0.0
      psfsurf = 1.0 
      area=0.0
      index=1
      rad_step=1.0
      
      do while (index < ndata .and. psfsurf > surfn)
        rad2=rad_hi*rad_hi
        psfsum=0.0
        do while (radarr(iperm(index)) < rad2)
         psfsum = psfsum+psfarr(iperm(index))
         area=area+1.0
         index=index+1
        end do
        rad_low=rad_hi
        rad_hi=rad_hi+rad_step
        psfsurf=psfsum/area
      end do
      radius=(rad_low+rad_hi)/2.

      radius=radius*pix2arcsec


      if (index == 1 .or. index == ndata) then
        ierr=1
        call wrerr(rtname,ierr,'Warning: surface brightness value not reached)')
      endif
   
!      deallocate(radarr)
!      deallocate(psfarr)
!      deallocate(iperm)

    endif 



      ! save for next call
      crpix1old=psf%psfhdr%crpix1
      crpix2old=psf%psfhdr%crpix2


  end subroutine psf_surfrad



!-------------------------------------------------------------------------------

  subroutine psf_eef(psf,radius,exact,eef,ierr)

  !-------------------------------------------------------------
  !
  ! Calculate PSF encircled energy fraction in a radius
  ! given in units of arcsec.
  ! 
  ! It is assumed that CRPIX1 and CRPIX2 define the origin of the 
  ! PSF
  ! If the logical parameter 'exact' is set to .true.,
  ! the weight of fractional pixels will be calculated (to be implemented).
  ! 
  !-------------------------------------------------------------
     
    type(psf2d_type),            intent(in)  :: psf  
    logical(kind=log32), intent(in)          :: exact
    real(kind=real32),   intent(in)          :: radius
    real(kind=real32),   intent(out)         :: eef
    integer(kind=int32), intent(out)         :: ierr
    integer(kind=int32)                      :: xlow,xhi,ylow,yhi
    integer(kind=int32)                      :: xmax,ymax,i,j            
    real(kind=real32)                        :: x0,y0,radius_pix
    real(kind=real32)                        :: xpix,ypix,pix2arcsec
    real(kind=real32)                        :: rad2,xdist,ydist,ydist2,dist2       
    character(len=*), parameter              :: rtname = 'psf_eef'    

    ierr = 0

    ! PSF pixel size, assume square PSF pixels:
    if (psf%psfhdr%e_cunit1 .and. psf%psfhdr%e_cdelt1 .and. psf%psfhdr%cdelt1 /= 0.0) then
      if ( trim(psf%psfhdr%cunit1) == 'arcsec' .or. trim(psf%psfhdr%cunit1) == 'ARCSEC') then   
        pix2arcsec=abs(psf%psfhdr%cdelt1)
      else if ( trim(psf%psfhdr%cunit1) == 'deg' .or. trim(psf%psfhdr%cunit1) == 'DEG') then
        pix2arcsec=abs(psf%psfhdr%cdelt1)*3600.0
      else
        pix2arcsec=abs(psf%psfhdr%cdelt1)
        ierr=3
        call wrerr(rtname,ierr,'Unknown CUNIT1 value in PSF file, assume arcsec')
      endif
    else
      pix2arcsec=1.0
      ierr=3
      call wrerr(rtname,ierr,'Unknown PSF pixel size, assume 1 arcsec')
    endif

    radius_pix = radius / pix2arcsec

    rad2=radius_pix*radius_pix

    xmax=psf%psfhdr%naxis1
    ymax=psf%psfhdr%naxis2
    x0 = psf%psfhdr%crpix1
    y0 = psf%psfhdr%crpix1

    xlow=floor(x0-radius_pix)
    xhi=ceiling(x0+radius_pix)
    ylow=floor(y0-radius_pix)    
    yhi=ceiling(y0+radius_pix)

    if (xlow < 1) then
     xlow=1
     ierr=1
     call wrerr(rtname,ierr,'Warning: Pixel out of PSF image)')
    endif
    if (ylow < 1) then
     ylow=1
     ierr=1
     call wrerr(rtname,ierr,'Warning: Pixel out of PSF image)')
    endif
    if (xhi > xmax) then
     xhi=xmax
     ierr=1
     call wrerr(rtname,ierr,'Warning: Pixel out of PSF image)')
    endif
    if (yhi > ymax) then
     yhi=ymax
     ierr=1
     call wrerr(rtname,ierr,'Warning: Pixel out of PSF image)')
    endif 
 
    eef=0.0
    do j=ylow,yhi
      ydist=real(j)-y0
      ydist2=ydist*ydist
      do i=xlow,xhi
        xdist=real(i)-x0
        dist2=xdist*xdist+ydist2
        if (dist2 < rad2) then
          eef=eef+psf%arr(i,j)
        endif 
      enddo
    enddo
  
    eef=eef / psf%psfhdr%sumrcts

        

  end subroutine psf_eef

!-------------------------------------------------------------------------------

  subroutine psf_eefrad(psf,eef,radius,ierr)

  !-------------------------------------------------------------
  !
  ! Calculate radius containing a given encircled energy fraction
  ! (eef) of the psf.
  ! 
  ! The radius is returned in units of arcsec. 
  !
  ! 
  ! It is assumed that CRPIX1 and CRPIX2 define the origin of the 
  ! PSF.
  ! Uses SLATEC routine SPSORT to sort PSF array according to radius 
  ! from center. Then pixel values are integrated from r=0 until
  ! EEF is reached at r=radius.
  !-------------------------------------------------------------
 
    type(psf2d_type),    intent(in)  :: psf  
    real(kind=real32), intent(in)    :: eef       
    real(kind=real32), intent(out)   :: radius
    integer(kind=int32), intent(out) :: ierr
   
    real(kind=real32), parameter     :: eps = 1E-6     
    integer(kind=real32)             :: ndata,sla_err
    real(kind=real32)                :: x0,y0,eefn   
    real(kind=4), dimension(:), allocatable ,save  :: radarr,psfarr
    integer(kind=4),dimension(:), allocatable, save :: iperm
    logical, save                        :: first=.true.
    integer(kind=4), dimension(1)             :: arrshape
    integer(kind=4)                  :: kflag,index,i,j,arrsize
    real(kind=4)                     :: xdist,ydist,ydist2,psfsum,dist2,pix2arcsec
    real(kind=real32),save           :: crpix1old,crpix2old
    character(len=*), parameter      :: rtname = 'psf_eefrad'

    ierr=0
    sla_err=0
    ndata=psf%psfhdr%naxis1*psf%psfhdr%naxis2

    ! PSF pixel size, assume square PSF pixels:
    if (psf%psfhdr%e_cunit1 .and. psf%psfhdr%e_cdelt1  .and. psf%psfhdr%cdelt1 /= 0.0) then
      if ( trim(psf%psfhdr%cunit1) == 'arcsec' .or. trim(psf%psfhdr%cunit1) == 'ARCSEC') then   
        pix2arcsec=abs(psf%psfhdr%cdelt1)
      else if ( trim(psf%psfhdr%cunit1) == 'deg' .or. trim(psf%psfhdr%cunit1) == 'DEG') then
        pix2arcsec=abs(psf%psfhdr%cdelt1)*3600.0
      else
        pix2arcsec=abs(psf%psfhdr%cdelt1)
        ierr=3
        call wrerr(rtname,ierr,'Unknown CUNIT1 value in PSF file, assume arcsec')
      endif
    else
      pix2arcsec=1.0
      ierr=3
      call wrerr(rtname,ierr,'Unknown PSF pixel size, assume 1 arcsec')
    endif
 


    if (eef  < 0.0 .or. eef > 1.0) then
      ierr=3
      call wrerr(rtname,ierr,'Error: EEF not in (0.0:1.0)')
    else
        !re-normalize EEF to PSF normalization
        eefn = eef * psf%psfhdr%sumrcts
  
      if (first) then 
        arrshape(1)=psf%psfhdr%naxis1*psf%psfhdr%naxis2
        allocate(radarr(arrshape(1)))
        allocate(psfarr(arrshape(1)))
        allocate(iperm(arrshape(1)))
      endif 
    
      psfarr=reshape(psf%arr,arrshape)   

      ! sort pixels by distance from PSF center on first call, only recalculate
      !  in case PSF center changes:
      if (first.or. (abs(psf%psfhdr%crpix1-crpix1old) > eps).or.(abs(psf%psfhdr%crpix2-crpix2old) > eps)) then
        first=.false.
        index=0
        do j=1,psf%psfhdr%naxis2
          ydist=real(j)-psf%psfhdr%crpix2
          ydist2=ydist*ydist
          do i=1,psf%psfhdr%naxis1
            index=index+1
            xdist=real(j)-psf%psfhdr%crpix1
            radarr(index)=ydist2+xdist*xdist
          enddo  
        enddo 

        kflag=1 ! kflag=1 : ascending sort, don't permutate radarr
        call spsort(radarr,ndata,iperm,kflag,sla_err)
        if (sla_err /= 0) then
          ierr = 3
          call wrerr(rtname,ierr,'Error: SLATEC error in SPSORT)')
        endif
        sla_err = 0

      endif

    
      radius = 0.0
      psfsum = 0.0 
      index=0
      do index=0,ndata
        if (psfsum >= eefn) exit
        psfsum = psfsum+psfarr(iperm(index))
      end do
      if (index > 1 .and. index <= ndata) then
        radius=sqrt(radarr(index-1))
      else
        ierr=1
        call wrerr(rtname,ierr,'Warning: EEF not reached)')
      endif
   
      radius = radius * pix2arcsec

    endif 


    ! save for next call
    crpix1old=psf%psfhdr%crpix1
    crpix2old=psf%psfhdr%crpix2



  end subroutine psf_eefrad

!-------------------------------------------------------------------------------



  subroutine psf_boxfrac(psf,boxhw,eef,ierr)

  !-------------------------------------------------------------------------
  !
  ! - Calculate PSF enboxed energy fraction EEF in a square box with half 
  ! width BOXHW (given in units of arcseconds).
  ! It is assumed that keywords CRPIX1 and CRPIX2 define the origin of the 
  ! PSF image.
  ! - Fractional pixels are treated correctly.
  !
  !-------------------------------------------------------------------------
  
  
    type(psf2d_type),    intent(in)          :: psf  
    real(kind=real32),   intent(in)          :: boxhw
    real(kind=real32),   intent(out)         :: eef
    integer(kind=int32), intent(out)         :: ierr
   
    real(kind=real32)                        :: xboxmin,xboxmax,yboxmin,yboxmax
    real(kind=real32)                        :: x0,y0,pix2arcsec,boxhw_pix
    real(kind=real32)                        :: psfsum
    integer(kind=int32)                      :: ix1,ix2,iy1,iy2
    integer(kind=int32)                      :: i,j
    real(kind=real32)                        :: x1frac, x2frac, y1frac, y2frac
    character(len=*), parameter              :: rtname = 'psf_boxfrac'
 

    ierr=0
    
    ! PSF pixel size, assume square PSF pixels:
    if (psf%psfhdr%e_cunit1 .and. psf%psfhdr%e_cdelt1 .and. psf%psfhdr%cdelt1 /= 0.0) then
      if ( trim(psf%psfhdr%cunit1) == 'arcsec' .or. trim(psf%psfhdr%cunit1) == 'ARCSEC') then   
        pix2arcsec=abs(psf%psfhdr%cdelt1)
      else if ( trim(psf%psfhdr%cunit1) == 'deg' .or. trim(psf%psfhdr%cunit1) == 'DEG') then
        pix2arcsec=abs(psf%psfhdr%cdelt1)*3600.0
      else
        pix2arcsec=abs(psf%psfhdr%cdelt1)
        ierr=3
        call wrerr(rtname,ierr,'Unknown CUNIT1 value in PSF file, assume arcsec')
      endif
    else
      pix2arcsec=1.0
      ierr=3
      call wrerr(rtname,ierr,'Unknown PSF pixel size, assume 1 arcsec')
    endif


    boxhw_pix = boxhw / pix2arcsec
  
!   calculate borders of box:
    xboxmin= psf%psfhdr%crpix1-boxhw_pix
    xboxmax= psf%psfhdr%crpix1+boxhw_pix
    yboxmin= psf%psfhdr%crpix2-boxhw_pix
    yboxmax= psf%psfhdr%crpix2+boxhw_pix




!   calculate min/max  coordinates of pixels fully contained in box:
    ix1=ceiling(xboxmin-0.5)
    ix2=floor(xboxmax-0.5)
    iy1=ceiling(yboxmin-0.5)
    iy2=floor(yboxmax-0.5)
    x1frac=abs(real(ix1)-xboxmin-0.5)
    x2frac=abs(xboxmax-real(ix2)-0.5)  
    y1frac=abs(real(iy1)-yboxmin-0.5)
    y2frac=abs(yboxmax-real(iy2)-0.5)



    if ((ix1 <= 1) .or. (iy1 <= 1) .or. (ix2 >  psf%psfhdr%naxis1-1)  .or. (iy2 >  psf%psfhdr%naxis2-1)) then  
      ierr=3
      call wrerr(rtname,ierr,'Box size exceeds PSF image size)')
    else
      psfsum=0.0
!     calculate  sum
      do j=iy1,iy2 
        do i=ix1,ix2 
          psfsum=psfsum+psf%arr(i,j) 
        enddo
      enddo   
      do j=iy1,iy2
        psfsum=psfsum+psf%arr(ix1-1,j)*x1frac
      enddo
      do j=iy1,iy2
        psfsum=psfsum+psf%arr(ix2+1,j)*x2frac
      enddo
      do i=ix1,ix2
        psfsum=psfsum+psf%arr(i,iy1-1)*y1frac
      enddo
      do i=ix1,ix2
        psfsum=psfsum+psf%arr(i,iy2+1)*y2frac
      enddo
      psfsum=psfsum+psf%arr(ix1-1,iy1-1)*x1frac*y1frac
      psfsum=psfsum+psf%arr(ix1-1,iy2+1)*x1frac*y2frac 
      psfsum=psfsum+psf%arr(ix2+1,iy1-1)*x2frac*y1frac  
      psfsum=psfsum+psf%arr(ix2+1,iy2+2)*x2frac*y2frac  

      eef=psfsum/psf%psfhdr%sumrcts

    endif  
 
  end subroutine psf_boxfrac

!-------------------------------------------------------------------------------


  subroutine psf_eefbox(psf,eef,boxhw,ierr)


  !-------------------------------------------------------------
  !
  ! Calculate half-width of a box containing a given energy fraction
  ! (eef) of the psf.
  ! 
  ! The half-width is returned in units of arcsec. 
  !
  ! 
  ! It is assumed that CRPIX1 and CRPIX2 define the origin of the 
  ! PSF.
  ! Uses the routine  psf_boxfrac, boxhw is calculated iteratively
  !
  !
  !-------------------------------------------------------------


  
    type(psf2d_type),    intent(in)         :: psf  
    real(kind=real32),   intent(in)         :: eef
    real(kind=real32),   intent(out)        :: boxhw
    integer(kind=int32), intent(out)        :: ierr

    real(kind=real32)                       :: eps,eediff,ee_old,ee
    real(kind=real32)                       :: hw_old,hw,dh
    integer(kind=int32)                     :: i
    integer(kind=int32)                     :: status
   
    character(len=*), parameter             :: rtname = 'psf_eefbox'  
    character(len=80)                       :: errstr

    ierr=0
    status=0
    eps=1.E-6

    hw_old=2.0
    dh=1.0
    call  psf_boxfrac(psf,hw_old,ee_old,status)

    do i=1,100
      boxhw = hw_old+dh
      call  psf_boxfrac(psf,boxhw,ee,status)
      eediff = eef-ee
      if (abs(eediff) <  eps) exit
      dh = dh * eediff/(ee-ee_old)
      ee_old = ee 
      hw_old = hw    

    enddo
 
  
    if (i > 99) then
      ierr=3
      errstr='iteration failed'
      call wrerr(rtname,ierr,errstr)
    endif
   
    


  end subroutine psf_eefbox

end module psflib
