! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
program catprep_main
  use param
  use erolib_general_mod
  use erolib_str_mod

  use catprep_mod
  use catprep_srclist_mod
  !    use catprep_srcconf_mod


  implicit none

  character(*), parameter :: functionName = "catprep_main"

  integer :: status,ierr,n_sources

  type(tCatprep) :: task
  type(t_catprep_file) :: catprep_file ! product
!  character(buffsize) :: pipeline_mode
  real(r8), dimension(:), pointer :: ptr_tstart=>null(),ptr_tstop=>null()

  status = ok

  ! read/open the .par file
  call opast(status)
  if(status /= ok) call wrstop(functionName, "error read par file")

  ! read the input paramters
  call wrlog(functionName, "reading parameters")
  call task%read_params(status)
  ! validate the input parameters
  call wrlog(functionName, "validate parameters")
  call task%validate(status)
  ! sanity check
  if(status /= ok) then
    call wrstop(functionName, "input parameters are invalid")
  else
    call wrlog(functionName, "parameters - valid")
  end if
  
  ! fill the CatSrcList type with the provided input file
  call wrlog(functionName, "import infile")
  call catprep_file%src_list%load_mllist(task%infile, task%det_alg_str&
    , task%detuid_prefix, status, catprep_file%info)

  n_sources = catprep_file%src_list%getNumRows()
  if(n_sources > 0) then
    call wrlog(functionName, "set tstart/tstop columns")

    call catprep_file%src_list%get_col_ptr("TSTART", ptr_tstart)
    call catprep_file%src_list%get_col_ptr("TSTOP",  ptr_tstop)
    ptr_tstart(:) = catprep_file%src_list%std_header%tstart
    ptr_tstop( :) = catprep_file%src_list%std_header%tstop
    nullify(ptr_tstart,ptr_tstop)

    ! calculate the distance_NN and distance_X
    call catprep_file%src_list%calc_srcdensity(task%p_srcdensity, status)
  end if

  ! write the CatSrcList to HDD
  ! create the new CatSrcList File
!  call wrlog(functionName, "write / close new catalog")
  call catprep_file%write(status, filename=task%outfile, clobber=task%clobber)

  if(task%do_outfile2) then
    call wrlog(functionName, "write / close new optional ownerflag filtered catalog")

    if(task%do_owner_filter) then
      call catprep_file%src_list%apply_filter_ownerflag(task%owner, status)
    end if
    if(task%do_flux_filter) then
      call catprep_file%src_list%apply_filter_flux(task%flux_threshold, status)
    end if

    call catprep_file%write(status, filename=task%outfile2, clobber=task%clobber)
  end if

  call catprep_file%src_list%destroy(status)

  call cpast (status, ierr)

end program catprep_main
