! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module catprep_mllist_mod

  ! *** NOTE! This module is used also by varichck and specfit.

  use param
  use erolib_general_mod, only : buffsize,FF_READONLY,FF_CREATE
  use erolib_bintable_mod
!  use erolib_str_mod
!  use erolib_sort_mod
!  use erolib_ffits_mod
!  use erolib_math_mod
  use fitsimage, only : stdhdr_type, fread_stdhdr, fwrite_stdhdr
!  use detect_io, only : fget_mllist, mllist_type

  implicit none

logical, private, parameter :: doTest=.false.

  type, extends(tBinTable) :: t_ml_list
    type(stdhdr_type) :: std_header
    character(buffsize) :: default_table_name = "SRCLIST"
  contains
    procedure :: init => init_t_ml_list
    procedure :: catprep_setup => catprep_setup_t_ml_list
  end type t_ml_list

  type, extends(tFitsFile) :: t_mllist_file
    type(t_ml_list) :: srclist
  contains
!    procedure :: read_n_bands => read_n_bands_t_catprep_file
    procedure :: read => read_t_mllist_file
    procedure :: write => write_t_mllist_file
  end type t_mllist_file

contains

  !.....................................................................
  subroutine init_t_ml_list(this, status, fileInfo)
    implicit none

    class(t_ml_list)                                 :: this
    integer,                           intent(inout) :: status
    type(tFitsInfo), optional, target, intent(in)    :: fileInfo

    character(*), parameter :: functionName = "init_t_ml_list"
!    integer :: ii
!    character(buffsize) :: i_str

    if(present(fileInfo)) this%info%fileInfo => fileInfo
    this%info%tableName = this%default_table_name

    call this%initColumn("J", "ID_SRC")
    call this%initColumn("J", "BOX_ID_SRC")
    call this%initColumn("J", "ID_INST")
    call this%initColumn("J", "ID_BAND")
    call this%initColumn("J", "ID_CLUSTER")
    call this%initColumn("E", "ML_CTS",        "count")
    call this%initColumn("E", "ML_CTS_ERR",    "count")
    call this%initColumn("E", "ML_CTS_LOWERR", "count")
    call this%initColumn("E", "ML_CTS_UPERR",  "count")
    call this%initColumn("E", "X_IMA",        "pixel")
    call this%initColumn("E", "X_IMA_ERR",    "pixel")
    call this%initColumn("E", "X_IMA_LOWERR", "pixel")
    call this%initColumn("E", "X_IMA_UPERR",  "pixel")
    call this%initColumn("E", "Y_IMA",        "pixel")
    call this%initColumn("E", "Y_IMA_ERR",    "pixel")
    call this%initColumn("E", "Y_IMA_LOWERR", "pixel")
    call this%initColumn("E", "Y_IMA_UPERR",  "pixel")
    call this%initColumn("E", "EXT",        "arcsec")
    call this%initColumn("E", "EXT_ERR",    "arcsec")
    call this%initColumn("E", "EXT_LOWERR", "arcsec")
    call this%initColumn("E", "EXT_UPERR",  "arcsec")
    call this%initColumn("E", "DET_LIKE")
    call this%initColumn("E", "EXT_LIKE")
    call this%initColumn("E", "ML_BKG", "arcmin**(-2)")
    call this%initColumn("E", "ML_EXP", "s")
    call this%initColumn("E", "ML_FLUX",        "erg cm**(-2) s**(-1)")
    call this%initColumn("E", "ML_FLUX_ERR",    "erg cm**(-2) s**(-1)")
    call this%initColumn("E", "ML_FLUX_LOWERR", "erg cm**(-2) s**(-1)")
    call this%initColumn("E", "ML_FLUX_UPERR",  "erg cm**(-2) s**(-1)")
    call this%initColumn("E", "ML_RATE",        "count s**(-1)")
    call this%initColumn("E", "ML_RATE_ERR",    "count s**(-1)")
    call this%initColumn("E", "ML_RATE_LOWERR", "count s**(-1)")
    call this%initColumn("E", "ML_RATE_UPERR",  "count s**(-1)")
    call this%initColumn("D", "RA",        "deg")
    call this%initColumn("E", "RA_LOWERR", "arcsec")
    call this%initColumn("E", "RA_UPERR",  "arcsec")
    call this%initColumn("D", "DEC",       "deg")
    call this%initColumn("E", "DEC_LOWERR","arcsec")
    call this%initColumn("E", "DEC_UPERR", "arcsec")
    call this%initColumn("E", "RADEC_ERR", "arcsec")
    call this%initColumn("D", "LII", "deg")
    call this%initColumn("D", "BII", "deg")
    call this%initColumn("J", "RAWX", "pixel")
    call this%initColumn("J", "RAWY", "pixel")
    call this%initColumn("E", "OFFAX", "arcmin")
    call this%initColumn("J", "TM_NR")
    call this%initColumn("E", "HR1")
    call this%initColumn("E", "HR1_ERR")
    call this%initColumn("E", "HR2")
    call this%initColumn("E", "HR2_ERR")
    call this%initColumn("E", "HR3")
    call this%initColumn("E", "HR3_ERR")
    call this%initColumn("E", "ML_RADIUS", "arcsec")
    call this%initColumn("E", "MASKFRAC")
    call this%initColumn("E", "ML_EEF")
    call this%initColumn("E", "VIGNET")
    call this%initColumn("E", "ONTIME", "s")
    call this%initColumn("E", "DIST_NN", "arcsec")
    call this%initColumn("B", "OWNER")
    ! For the ownership code, see
    !
    !	https://erosita.mpe.mpg.de/eROdoc/products/EventFiles_doc.html
    !
    ! In short, only the 2 least-significant bits have meaning. Bit 0
    ! indicates whether MPE owns the source location; bit 1, likewise
    ! for IKI. Somewhat confusingly, a zero value means 'yes', a value
    ! of unity means 'no'. A table of the four combinations is therefore:
    !
    !	OWNER	Meaning
    !	---------------
    !	0	Both MPE and IKI own it (both bits set to 0).
    !	1	IKI owns it.
    !	2	MPE owns it.
    !	3	Nobody owns it.
    !	---------------
    !
    call this%initColumn("J", "APE_CTS", "count")
    call this%initColumn("E", "APE_BKG", "count")
    call this%initColumn("E", "APE_EXP", "s")
    call this%initColumn("E", "APE_EEF")
    call this%initColumn("E", "APE_RADIUS", "pixel")
    call this%initColumn("E", "APE_POIS")

    call this%initColumns(status)

  end subroutine init_t_ml_list

  !.....................................................................
!  subroutine catprep_setup_t_ml_list(this, status)
  subroutine catprep_setup_t_ml_list(this)
    ! Sets column useage flags according to the needs of catprep.
    implicit none

    class(t_ml_list)       :: this
!    integer, intent(inout) :: status

    character(*), parameter :: functionName = "catprep_setup_t_ml_list"

    if(.not.this%isInitialized)&
call wrstop(functionName, "You must initialize the t_ml_list object before calling me.")

    call this%ibset_all_useage(TCOL_READ_BIT)

    ! Don't need these:
    !
    call this%ibclr_col_useage("BOX_ID_SRC",   TCOL_READ_BIT)
    call this%ibclr_col_useage("X_IMA",        TCOL_READ_BIT)
    call this%ibclr_col_useage("X_IMA_ERR",    TCOL_READ_BIT)
    call this%ibclr_col_useage("X_IMA_LOWERR", TCOL_READ_BIT)
    call this%ibclr_col_useage("X_IMA_UPERR",  TCOL_READ_BIT)
    call this%ibclr_col_useage("Y_IMA",        TCOL_READ_BIT)
    call this%ibclr_col_useage("Y_IMA_ERR",    TCOL_READ_BIT)
    call this%ibclr_col_useage("Y_IMA_LOWERR", TCOL_READ_BIT)
    call this%ibclr_col_useage("Y_IMA_UPERR",  TCOL_READ_BIT)
    call this%ibclr_col_useage("OFFAX",        TCOL_READ_BIT)
    call this%ibclr_col_useage("TM_NR",        TCOL_READ_BIT)
    call this%ibclr_col_useage("ONTIME",       TCOL_READ_BIT)

    ! Make the following new columns optional for backward compatibility:
    !
    call this%ibset_col_useage("RA_LOWERR",      TCOL_ISOPT_BIT)
    call this%ibset_col_useage("RA_UPERR",       TCOL_ISOPT_BIT)
    call this%ibset_col_useage("DEC_LOWERR",     TCOL_ISOPT_BIT)
    call this%ibset_col_useage("DEC_UPERR",      TCOL_ISOPT_BIT)
    call this%ibset_col_useage("EXT_LOWERR",     TCOL_ISOPT_BIT)
    call this%ibset_col_useage("EXT_UPERR",      TCOL_ISOPT_BIT)
    call this%ibset_col_useage("ML_CTS_LOWERR",  TCOL_ISOPT_BIT)
    call this%ibset_col_useage("ML_CTS_UPERR",   TCOL_ISOPT_BIT)
    call this%ibset_col_useage("ML_FLUX_LOWERR", TCOL_ISOPT_BIT)
    call this%ibset_col_useage("ML_FLUX_UPERR",  TCOL_ISOPT_BIT)
    call this%ibset_col_useage("ML_RATE_LOWERR", TCOL_ISOPT_BIT)
    call this%ibset_col_useage("ML_RATE_UPERR",  TCOL_ISOPT_BIT)

    this%useageIsSet = .true.

  end subroutine catprep_setup_t_ml_list

  !.....................................................................
  subroutine read_t_mllist_file(this, status, filename)
    implicit none
    class(t_mllist_file) :: this
    integer,                intent(inout) :: status
    character(*), optional, intent(in)    :: filename

    character(*), parameter :: functionName = "read_t_mllist_file"

    if(present(filename)) call this%open(FF_READONLY, status, filename)

!*** could check extension exists. See analogous method of tEventFile.

    call this%srclist%read(status)
    call fread_stdhdr(this%info%unitNo, this%srclist%std_header, status)

    if(present(filename)) call this%close(status)

  end subroutine read_t_mllist_file

  !.....................................................................
  subroutine write_t_mllist_file(this, status, filename, clobber)
    implicit none
    class(t_mllist_file) :: this
    integer,                intent(inout) :: status
    character(*), optional, intent(in)    :: filename
    logical,      optional, intent(in)    :: clobber

    character(*), parameter :: functionName = "write_t_mllist_file"

    if(present(filename)) call this%open(FF_CREATE, status, filename, clobber)

    call this%srclist%write(status)
    call fwrite_stdhdr(this%info%unitNo, this%srclist%std_header, status)

    if(present(filename)) call this%close(status)

  end subroutine write_t_mllist_file

end module catprep_mllist_mod
