!-----------------------------------------------------------------------
! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------

program erbackmap

!------------------------------------------------------------------------------
! eROSITA background map task.
! Author : GLamer, AIP   
! 
!
!
!------------------------------------------------------------------------------
  

  use erbackmap_mod
  use numkind
  use fitsimage
  use detect_io
  use cal

  implicit none

  integer(kind=int32)         :: ierr,status
  type(param_type)            :: par
  type(i4image_type)          :: image
  type(i1image_type)          :: detmask
  type(i1image_type)          :: usermask
  type(i1image_type)          :: cheesemask
  type(r4image_type)          :: expimage,expimage2
  type(r4image_type)          :: backmap 
  type(boxlist_type)          :: boxlist 
  type(cal_par_type)          :: cal_par

  real(kind=real32), dimension(:,:), allocatable :: grid   
  integer(kind=int32)         :: ngrid 
  integer(kind=int32)         :: nsplit=4
  integer(kind=int32)         :: iit  
  real(kind=real32)           :: chisq   


  character(len=*), parameter :: rtname='ERBACKMAP'

  
  ierr=0
  
  !test:
  !write(*,*) rtname, ' ',boxlist%cname(1), boxlist%cname(33)


  ! open parameter and log files:
  call erbackmap_ini(par,status)
  ierr=max(ierr,status)  

  ! get task parameters:
  if (ierr < 3) then
    call erbackmap_par(par,status)
    ierr=max(ierr,status)
  endif

  
  ! get input images and source list:
  if (ierr < 3) then
    call erbackmap_in(par,image,boxlist,expimage,expimage2,detmask,usermask,status)
    ierr=max(ierr,status)
  endif
 

  if (ierr < 3) then  
    call erbackmap_blind(par,cal_par,boxlist,image,detmask,usermask,cheesemask,status)
    ierr=max(ierr,status)
  endif

!  allocate(grid(4,par%nsplinenodes*par%nsplinenodes))
  ! split option (GL 04/11/2015:)
  allocate(grid(4,nsplit*par%nsplinenodes*par%nsplinenodes))




  if (ierr < 3) then  
    call erbackmap_create(image,backmap,status)
    ierr=max(ierr,status)
  endif


  if (ierr < 3) then 

    if ((par%fitmethod == "smooth") .or. (par%fitmethod == "SMOOTH")) then

       ! call erbackmap_smooth(par,image,cheesemask,expimage,backmap,status)
        call erbackmap_expand(par,image,cheesemask,expimage,backmap,status)


    else

      call erbackmap_bin(par,image,cheesemask,grid,ngrid,status)
      ierr=max(ierr,status)


      do iit=1,par%nfitrun
        call erbackmap_spline(par,grid,ngrid,backmap,status)
        ierr=max(ierr,status)
        status=0

        call erbackmap_clip(par,backmap,grid,ngrid,chisq,status)
  
        backmap%dethdr%chisq=chisq
        backmap%dethdr%e_chisq=.true.
        ierr=max(ierr,status)
        if (status > 0) exit

 
      enddo
      

    endif

  endif


  if (ierr < 3) then  
    if (par%detmask_flag) then
      call erbackmap_mask(backmap,detmask,status)
      ierr=max(ierr,status)
    endif
  endif


  if (ierr < 3) then  
    call erbackmap_out(par,backmap,cheesemask,status)
    ierr=max(ierr,status)
  endif

      
  ! close parameter and log files:   
  call erbackmap_cls(par,ierr,status)
  

end program erbackmap
