!-----------------------------------------------------------------------
! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------

program erbox

!------------------------------------------------------------------------------
! eROSITA background map task.
! Author : GLamer, AIP   
! 
!
!
!------------------------------------------------------------------------------
  
  use param
  use erbox_mod
  use numkind
  use fitsimage
  use detect_io
  use cal
  use eboxdet

  implicit none


  integer(kind=int32)                      :: ierr,status
  type(erbox_param_type)                   :: par
  type(i4cube_type)                        :: image
  type(l1cube_type)                        :: detmask
  type(r4cube_type)                        :: expimage
  type(r4cube_type)                        :: bkgimage
  type(boxlist_type)                       :: boxlist
!  type(boxrow_type),dimension(:,:),pointer :: srclist => NULL()
  type(srcrow_type),dimension(:,:),pointer :: srclist => NULL()
!  type(cal_par_type)                       :: cal_par
!  integer(kind=int32)                      :: ximbuf,yimbuf
  integer(kind=int32)                      :: nsrc
  

  character(len=*), parameter :: rtname='ERBOX'


  status = 0
  nsrc = 0
  ierr = 0

 
  ! open parameter and log files:
  call erbox_ini(par,status)
  ierr=max(ierr,status)  

  ! get task parameters:
  if (ierr < 3) then
    call erbox_par(par,status)
    ierr=max(ierr,status)
  endif


  ! get input images and source list:
  if (ierr < 3) then
    call erbox_in(par,image,expimage,bkgimage,detmask,status)
    ierr=max(ierr,status)
  endif


  nullify(srclist)
  allocate(srclist(par%n_imagesets,max_sou)) 


  if (ierr < 3) then
    call mdetec(par,image,expimage,bkgimage,detmask,srclist,nsrc,status)
    ierr=max(ierr,status)
  endif

  
  if (associated(boxlist%data))  deallocate(boxlist%data) 


  allocate(boxlist%data(1:max_sou),STAT=status)


  boxlist%stdhdr = image%stdhdr(1)

  if (par%bkgima_flag)  boxlist%stdhdr%history = bkgimage%stdhdr(1)%history
   !  (inherit history from background map)


  if (ierr < 3) then
    call format_boxlist(par,image,nsrc,srclist,boxlist,status)
    ierr=max(ierr,status)
  endif

        
  ! close parameter and log files:   
  call  cpast(rtname,par%lun,ierr,status)


    if(status /= 0) then
      call wrerr(rtname,status,'Error closing parameter/log files')
    endif    


end program erbox
