program ermask

!------------------------------------------------------------------------------
! eROSITA mask creation task.
! Author : GLamer, AIP 
! Calculates detection mask based on exposure mask 
! and thresholds for relative exposure (parameter threshold1)
! and normalized gradient of exposure map  (parameter threshold2).
! 
! Determines min, max  boundaries of valid mask area and writes
! keywords xdatamin, xdatamax, ydatamin, ydatamax to header of detection mask.
!------------------------------------------------------------------------------
  
  use ermask_mod
  use numkind
  use fitsimage
  use detect_io



  implicit none


  integer(kind=int32)                      :: ierr,status
  type(ermask_param_type)                  :: par
  type(i1image_type)                       :: detmask
  type(r4image_type)                       :: expimage
  

  character(len=*), parameter :: rtname='ERMASK'


  status = 0
  ierr   = 0
  
 
  ! open parameter and log files:
  call ermask_ini(par,status)
  ierr=max(ierr,status)  

  ! get task parameters:
  if (ierr < 3) then
    call ermask_par(par,status)
    ierr=max(ierr,status)
  endif


  ! get input images and source list:
  if (ierr < 3) then
    call ermask_in(par,expimage,status)
    ierr=max(ierr,status)
  endif


  if (ierr < 3) then
    call calc_mask(par,expimage,detmask,status)
    ierr=max(ierr,status)
  endif

  
  if (ierr < 3) then
   call ermask_out(par,detmask,status)
   ierr=max(ierr,status)
  endif

       
  ! close parameter and log files:   
  call  cpast(rtname,par%lun,ierr,status)


    if(status /= 0) then
      call wrerr(rtname,status,'Error closing parameter/log files')
    endif    


end program ermask

