module ermask_mod

  use param
  use numkind  
  use fitsimage
  use detect_io

  implicit none


    
contains

!-------------------------------------------------------------------------------

  subroutine ermask_ini(par,ierr)

  !---------------------------------------------------------
  ! Calls subroutine OPAST to open parameter and log files
  !---------------------------------------------------------


    type(ermask_param_type)    , intent(out) :: par
    integer(kind=int32) , intent(out) :: ierr
    integer(kind=int32)               :: status
    character(len=*),parameter        :: rtname='ERMASK_INI'
    character(len=*),parameter        :: taskname='ERMASK' 

    ierr=0
    status=0

    call opast(taskname,par%lun,status)

    if (status /= 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error opening parameter file')
    endif 

  end subroutine ermask_ini


!-------------------------------------------------------------------------------

  subroutine ermask_par(par,ierr)
 
  !---------------------------------------------------------------------
  ! Get task parameters 
  !
  !---------------------------------------------------------------------    

    type(ermask_param_type) , intent(out)  :: par
    integer(kind=int32)      , intent(out)  :: ierr
    integer(kind=int32)                     :: status

    character(len=*),parameter         :: rtname='ERMASK_PAR'
    
    ierr = 0

    call rgetc(par%lun,'EXPIMAGE',par%expimage,status)
    if (status /= 0) ierr = max(ierr,status)   
    call rgetc(par%lun,'DETMASK',par%detmask,status)
    if (status /= 0) ierr = max(ierr,status)   
    call rgetf(par%lun,'THRESHOLD1',par%threshold1,status) 
    if (status /= 0) ierr = max(ierr,status)  
    call rgetf(par%lun,'THRESHOLD2',par%threshold2,status) 
    if (status /= 0) ierr = max(ierr,status)  
    call rgetl(par%lun,'REGIONFILE_FLAG',par%regionfile_flag ,status)  
    if (status /= 0) ierr = max(ierr,status)
    if(par%regionfile_flag .eqv. .true.) then
      call rgetc(par%lun,'REGIONFILE',par%regionfile,status)
      if (status /= 0) ierr = max(ierr,status)
    end if
    call rgetl(par%lun,'CLOBBER',par%clobber,status)  
    if (status /= 0) ierr = max(ierr,status) 
     
    if (ierr /= 0)  call wrerr(rtname,ierr,'Error getting parameters')


    if (par%clobber) par%detmask="!"//trim(par%detmask)
                       
  end subroutine ermask_par

!-------------------------------------------------------------------------------

  subroutine ermask_in(par,expimage,ierr)

  !--------------------------------------------------------------------------------
  ! Read input exposure map 
  ! and region file (tbd)
  !--------------------------------------------------------------------------------

    type(ermask_param_type), intent(in)   :: par
    type(r4image_type)                     :: expimage
!    type(i1image_type)                     :: detmask
    integer(kind=int32), intent(out)       :: ierr

    integer(kind=int32)                    :: status

    character(len=*), parameter            :: rtname='ERMASK_IN'
!    character(len=80)                      :: errstr

    status=0
    ierr=0


   call fget_r4image(par%expimage,expimage,status)

   
    if (status > 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error reading exposure map')
    endif    


  end subroutine ermask_in

!-------------------------------------------------------------------------------

  subroutine ermask_out(par,detmask,ierr)

  !--------------------------------------------------------------------------------
  ! Write detection mask
  ! 
  !--------------------------------------------------------------------------------

    type(ermask_param_type), intent(in)    :: par
    type(i1image_type)                     :: detmask
    integer(kind=int32), intent(out)       :: ierr

    integer(kind=int32)                    :: status

    character(len=*), parameter            :: rtname='ERMASK_OUT'
!    character(len=80)                      :: errstr

    status=0
    ierr=0

   call fput_i1image(par%detmask,detmask,ierr)

   if (status > 0) then
     ierr=3
     call wrerr(rtname,ierr,'Error writing detection mask')
   endif

  
  end subroutine ermask_out


!-------------------------------------------------------------------------------------------

  subroutine calc_mask(par,expimage,detmask,ierr)

  !--------------------------------------------------------------------------------
  ! Create detection mask
  ! Calculate  xdatamin, xdatamax, ydatamin, ydatamax 
  !--------------------------------------------------------------------------------


    type(ermask_param_type)               :: par
    type(r4image_type)                     :: expimage
    type(i1image_type)                     :: detmask
    integer(kind=int32)                    :: ierr

    integer(kind=int32)                    :: xsize, ysize, ix, iy
    integer(kind=int32)                    :: xdatamin, xdatamax, ydatamin, ydatamax

!    integer(kind=int32), dimension(:,:), pointer :: mask
!    real(kind=real32), dimension(:,:),  pointer  :: expo  

    real(kind=real32)                            :: emax, fracth,  gradth2, mean2
    character(len=*), parameter :: rtname='CALC_MASK'
    ierr=0

    
    xsize=expimage%shapehdr%naxis1
    ysize=expimage%shapehdr%naxis2
    
    
    allocate(detmask%arr(xsize,ysize))
    

    ! copy and update header keywords: 
   
    detmask%shapehdr = expimage%shapehdr
    detmask%shapehdr%bitpix=8    

    detmask%stdhdr   = expimage%stdhdr
    detmask%stdhdr%creator  = "ERMASK  "
    detmask%stdhdr%hduclas1 = "IMAGE   "
    detmask%stdhdr%hduclas2 = "MASK    "

    
    detmask%refhdr   = expimage%refhdr
    detmask%wcshdr   = expimage%wcshdr
    detmask%dethdr   = expimage%dethdr

   
    
    emax  = maxval(expimage%arr(1:xsize,1:ysize))

    
    
    fracth = par%threshold1*emax
    gradth2 = (par%threshold2*emax)**2
    
    detmask%arr(1:xsize,1:ysize) = 1

    xdatamin = xsize
    xdatamax = 0
    ydatamin = ysize
    ydatamax = 0

    ! calculate mask: 


    if (emax == 0.0) then
      detmask%arr(1:xsize,1:ysize) = 0
    else

      do iy = 1,ysize
        do ix = 1,xsize
          if (expimage%arr(ix,iy) < fracth) detmask%arr(ix,iy) = 0
          ! fix xsize -> ysize (GL 15/04/2018): 
          if (ix > 1 .and. ix < xsize  .and. iy > 1 .and. iy < ysize) then
            mean2=(expimage%arr(ix,iy)+expimage%arr(ix+1,iy)+expimage%arr(ix-1,iy)+expimage%arr(ix,iy+1)+expimage%arr(ix,iy-1))/5.
            if  ((( expimage%arr(ix+1,iy)-expimage%arr(ix-1,iy) )**2 + &
                 ( expimage%arr(ix,iy+1)-expimage%arr(ix,iy-1) )**2 ) > (par%threshold2*mean2)**2) detmask%arr(ix,iy) = 0 
          endif
        enddo
      enddo
 
    endif

    ! calculate x,y  boundaries of useful data:   

    xdatamin = xsize
    xdatamax = 0
    ydatamin = ysize
    ydatamax = 0

    

    do iy = 1,ysize
      do ix = 1,xsize
        if (ix < xdatamin .and. detmask%arr(ix,iy) > 0) xdatamin = ix
        if (iy < ydatamin .and. detmask%arr(ix,iy) > 0) ydatamin = iy
        if (ix > xdatamax .and. detmask%arr(ix,iy) > 0) xdatamax = ix
        if (iy > ydatamax .and. detmask%arr(ix,iy) > 0) ydatamax = iy
      enddo
    enddo

    detmask%dethdr%xdatamin = xdatamin
    detmask%dethdr%xdatamax = xdatamax
    detmask%dethdr%ydatamin = ydatamin
    detmask%dethdr%ydatamax = ydatamax

    if (xdatamax == 0 .or. ydatamax == 0) then 
      ierr = 1
      call wrerr(rtname,ierr,'No valid data in exposure map')
    endif
   
 
  end subroutine calc_mask

!---------------------------------------------------------------------------------------------

end module  ermask_mod

