! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.

program ermldet

!------------------------------------------------------------------------------
! eROSITA ML PSF fitting task.
! Author : GLamer, AIP   
! 
!
!
!------------------------------------------------------------------------------
  

  use param
  use ermldet_mod
  use numkind
  use fitsimage
  use detect_io
  use cal
  use erolib_eventfile_mod
  use slib_constants


  ! AGE ADDED THAT
  use slib_data_psf, only: slib_dealloc_FPSF


  implicit none


  integer(kind=int32)                           :: ierr,status
  type(ermldet_param_type)                      :: par
  type(i4cube_type)                             :: image
  type(tEventFile), dimension(:), allocatable   :: eventlist
  type(pixlist_type), dimension(:), allocatable :: pixlist
  
  type(FPSF)                                    :: allStackPSF,allStackPSF2
  type(FPSF)                                    :: cluPSF, cluPSF2


  type(l1cube_type)                        :: detmask
  type(r4cube_type)                        :: expimage
  type(r4cube_type)                        :: bkgimage
!  type(r4cube_type)                        :: srcimage
  type(boxlist2d_type)                     :: boxlist
  type(mllist_type)                        :: mllist
  type(srcrow_type),dimension(:,:),pointer :: srclist => NULL()  
!  type(cal_par_type)                       :: cal_par
  
  integer(kind=int32)                      :: nsrc,iim
  integer(kind=int32),parameter            :: srcmodrad=120 

  character(len=*), parameter :: rtname='ERMLDET'

  status = 0
  nsrc = 0
  ierr=0


 
  ! open parameter and log files:
  call ermldet_ini(par,status)
  ierr=max(ierr,status)  


  ! get task parameters:
  if (ierr < 3) then
    call ermldet_par(par,status)
    ierr=max(ierr,status)
  else
    
  endif


  ! get input images and source list:
  if (ierr < 3) then
    if (par%shapelet_flag) then
      call alloc_eventfiles(eventlist, par%n_imagesets, status)
      if (status > 0) then
        ierr=3
        call  wrerr(rtname,ierr,'Cannot allocate memory for event lists')
      endif
      allocate(pixlist(par%n_imagesets))
    endif
    call ermldet_in(par,boxlist,image,eventlist,pixlist,expimage,bkgimage,detmask,status)
    ierr=max(ierr,status)
  endif


  if  (par%shapelet_flag) then
    if (ierr < 3) then
      ! shape v1.8, allocated in 'slib_run'
      !allocate(allStackPSF(nim))
      !allocate(allStackPSF2(nim))
      call ermldet_shapelets(par,boxlist,image,eventlist,allStackPSF,0,cluPSF,status)
      call ermldet_shapelets(par,boxlist,image,eventlist,allStackPSF2,srcmodrad,cluPSF2,status)
      do iim=1,par%n_imagesets
        call init_pixlist(eventlist(iim)%events,image%wcshdr(iim),pixlist(iim),status)
      enddo  
    endif
    if (status > 0) then
       ierr=3
       call  wrerr(rtname,status,'Error preparing shapelet PSF')
    endif   
  endif 


  nullify(srclist)
  allocate(srclist(0:par%n_imagesets,max_sou)) 



  if (ierr < 3) then
    call mldetect(par,image,eventlist,pixlist,detmask,expimage,bkgimage,allStackPSF,cluPSF,allStackPSF2,cluPSF2,boxlist,srclist,srcmodrad,nsrc,status)
    ierr=max(ierr,status)
  endif
  



  if (associated(mllist%data))  deallocate(mllist%data) 
  allocate(mllist%data(1:max_sou),STAT=status)
  mllist%stdhdr = image%stdhdr(1)
  
  
  if (ierr < 3) then
     call format_mllist2(par,image,nsrc,srclist,mllist,status)
     ierr=max(ierr,status)
    status=0
  endif

 
  mllist%stdhdr%history = boxlist%stdhdr%history
 
    !  call fput_mllist(par%mllistset,mllist,status)  
  if (ierr < 3) then
   call ermldet_out(par,mllist,bkgimage,status)
   ierr=max(ierr,status)
   status=0
  endif

        
  ! close parameter and log files:   
  call  cpast(rtname,par%lun,ierr,status)


 !AGE 28-06-2020
  ! deallocate MyPSFSIZE structure
  ! closing down ....
  if (par%shapelet_flag) then
    call slib_dealloc_FPSF(allStackPSF)
    call slib_dealloc_FPSF(allStackPSF2)
    call slib_dealloc_FPSF(cluPSF)
    call slib_dealloc_FPSF(cluPSF2)
    !deallocate events via erolib
    call dealloc_eventfiles(eventlist, status) ! IMS: this does a complete free of each eventfile.
    if (allocated(pixlist)) deallocate(pixlist)
  endif  
  ! AGE2020


    if(status /= 0) then
      call wrerr(rtname,status,'Error closing parameter/log files')
    endif    


end program ermldet

