! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.

module ermldet_mod

  use param
  !  use numkind
  use ISO_FORTRAN_ENV
  use fitsimage
  use detect_io
  use cal
  use psflib
  use mulfit
!  use ISO_FORTRAN_ENV, only: REAL64
  use erolib_bintable_mod, only : TCOL_READ_BIT
  use erolib_eventfile_mod
  use erolib_events_mod
  use slib_constants
  use slib_engine
  use slib_basic_calcs
  use slib_data_evt
  use slib_data_psf
  use slib_calc_space


  implicit none

 

    
contains

!-------------------------------------------------------------------------------

  subroutine ermldet_ini(par,ierr)

  !---------------------------------------------------------
  ! Calls subroutine OPAST to open parameter and log files
  !---------------------------------------------------------


    type(ermldet_param_type), intent(out) :: par
    integer(kind=int32) , intent(out) :: ierr
    integer(kind=int32)               :: status
    character(len=*),parameter        :: rtname='ERMLDET_INI'
    character(len=*),parameter        :: taskname='ERMLDET' 

    ierr=0
    status=0


    call opast(taskname,par%lun,status)

    if (status /= 0) then
      ierr=3
      call wrerr(rtname,ierr,'Error opening parameter file')
    endif 

  end subroutine ermldet_ini


!-------------------------------------------------------------------------------

  subroutine ermldet_par(par,ierr)
 
  !---------------------------------------------------------------------
  ! Get task parameters 
  !
  !---------------------------------------------------------------------    

    type(ermldet_param_type) , intent(out)  :: par
    integer(kind=int32)      , intent(out)  :: ierr
    integer(kind=int32)                     :: status
    integer(kind=int32)                     :: i
    character(len=*),parameter         :: rtname='ERMLDET_PAR'
    
    

    ierr = 0

    call rgetc(par%lun,'BOXLIST'   ,par%boxlistset,status)
    if (status /= 0) ierr = max(ierr,status)   
    call rgetc(par%lun,'MLLIST'   ,par%mllistset,status)
    if (status /= 0) ierr = max(ierr,status)   
    call rgetca(par%lun,'IMAGES',max_ima ,par%imagesets,par%n_imagesets,status)
    if (status /= 0) ierr = max(ierr,status)
!CG    call rgetca(par%lun,'EXPIMAGES',max_ima,par%expimagesets,par%n_expimagesets,status)
!CG    if (status /= 0) ierr = max(ierr,status)   
!CG    call rgetca(par%lun,'DETMASKS',max_ima,par%detmasksets,par%n_detmasksets,status)
!CG    if (status /= 0) ierr = max(ierr,status)           
    call rgetca(par%lun,'BKGIMAGES',max_ima,par%bkgimagesets,par%n_bkgimagesets,status)
    if (status /= 0) ierr = max(ierr,status)  
!CG    call rgetca(par%lun,'SRCIMAGES',max_ima,par%srcimagesets,par%n_srcimagesets,status)
!CG    if (status /= 0) ierr = max(ierr,status)  
    call rgetfa(par%lun,'EMIN', max_ima,par%emin,par%n_emin,status)  
    if (status /= 0) ierr = max(ierr,status)       
    call rgetfa(par%lun,'EMAX', max_ima,par%emax,par%n_emax,status)  
    if (status /= 0) ierr = max(ierr,status)
    call rgetfa(par%lun,'ECF', max_ima,par%ecf,par%n_ecf,status)  
    if (status /= 0) ierr = max(ierr,status)
    call rgetia(par%lun,'HRDEF',2*max_hr,par%hrdef,par%n_hrdef,status)  
    if (status /= 0) ierr = max(ierr,status)
    call rgetf(par%lun,'LIKEMIN',par%likemin,status) 
    if (status /= 0) ierr = max(ierr,status)  
    call rgetf(par%lun,'EXTLIKEMIN',par%extlikemin,status) 
    if (status /= 0) ierr = max(ierr,status)  
    call rgetfa(par%lun,'EXTLIKE_SLOPE',2,par%extlike_slope,par%n_extlike_slope,status) 
    if (status /= 0) ierr = max(ierr,status)  
    call rgetf(par%lun,'EXTMIN',par%extmin,status) 
    if (status /= 0) ierr = max(ierr,status) 
    call rgetf(par%lun,'EXTMAX',par%extmax,status) 
    if (status /= 0) ierr = max(ierr,status) 
    call rgetc(par%lun,'EXTENTMODEL',par%extentmodel,status)
    if (status /= 0) ierr = max(ierr,status)  
    call rgeti(par%lun,'NMAXFIT',par%nmaxfit,status)  
    if (status /= 0) ierr = max(ierr,status) 
    call rgeti(par%lun,'NMULSOU',par%nmulsou,status)  
    if (status /= 0) ierr = max(ierr,status) 
    if (par%nmulsou > max_nmul) then
      ierr= max(ierr,1)
      par%nmulsou = max_nmul
      call wrerr(rtname,ierr,'Parameter nmulsou out of range')
    endif
    call rgetl(par%lun,'THRES_FLAG',par%thres_flag ,status)  
    if (status /= 0) ierr = max(ierr,status) 
    call rgetf(par%lun,'THRES_VAL',par%thres_val,status) 
    if (status /= 0) ierr = max(ierr,status) 
    call rgetc(par%lun,'THRES_COL',par%thres_col,status)
    if (status /= 0) ierr = max(ierr,status)   
    call rgetf(par%lun,'MULTRAD',par%multrad,status) 
    if (status /= 0) ierr = max(ierr,status) 
    call rgetf(par%lun,'CUTRAD',par%cutrad,status) 
    if (status /= 0) ierr = max(ierr,status) 
    call rgetl(par%lun,'SHAPELET_FLAG',par%shapelet_flag,status)
    if (status /= 0) ierr = max(ierr,status)   
    call rgetl(par%lun,'PHOTON_FLAG',par%photon_flag,status)
    if (status /= 0) ierr = max(ierr,status)   
    call rgetl(par%lun,'FITEXT_FLAG',par%fitext_flag ,status)  
    if (status /= 0) ierr = max(ierr,status)
    call rgetl(par%lun,'FITPOS_FLAG',par%fitpos_flag ,status)  
    if (status /= 0) ierr = max(ierr,status)
    call rgetl(par%lun,'COMPRESS_FLAG',par%compress_flag ,status)  
    if (status /= 0) ierr = max(ierr,status) 

    call rgetl(par%lun,'SRCIMA_FLAG',par%srcima_flag ,status)  
    if (status /= 0) ierr = max(ierr,status) 
    if(par%srcima_flag .eqv. .true.) then
      call rgetca(par%lun,'SRCIMAGES',max_ima,par%srcimagesets,par%n_srcimagesets,status)
      if (status /= 0) ierr = max(ierr,status)  
    end if
    call rgetl(par%lun,'EXPIMA_FLAG',par%expima_flag ,status)  
    if (status /= 0) ierr = max(ierr,status) 
    if(par%expima_flag .eqv. .true.) then
      call rgetca(par%lun,'EXPIMAGES',max_ima,par%expimagesets,par%n_expimagesets,status)
      if (status /= 0) ierr = max(ierr,status)
    end if

    call rgetl(par%lun,'DETMASK_FLAG',par%detmask_flag,status)  
    if (status /= 0) ierr = max(ierr,status)
    if(par%detmask_flag .eqv. .true.) then
      call rgetca(par%lun,'DETMASKS',max_ima,par%detmasksets,par%n_detmasksets,status)
      if (status /= 0) ierr = max(ierr,status)
    end if

    call rgetl(par%lun,'TWOSTAGE_FLAG',par%twostage_flag ,status)  
    if (status /= 0) ierr = max(ierr,status) 


    call rgetl(par%lun,'SIXTE_FLAG',par%sixte_flag ,status)  
    if (status /= 0) ierr = max(ierr,status)
    
    call rgetl(par%lun,'CLOBBER',par%clobber,status)  
    if (status /= 0) ierr = max(ierr,status) 

    if (par%photon_flag) then
      if (.not. par%shapelet_flag) then
         par%shapelet_flag=.true.
         ierr=1
         call wrerr(rtname,ierr,'Warning: photon mode requires shapelet PSF, set shapelet_flag=yes')
      endif    
    endif
 

    !(GL 07/12/2016)
    if (.not. par%fitext_flag) then
      par%twostage_flag=.false.
    endif

    if (par%clobber) then
       par%mllistset="!"//trim(par%mllistset)
       do i=1,par%n_srcimagesets
          par%srcimagesets(i)="!"//trim(par%srcimagesets(i)) 
       enddo
    endif
    
    if (ierr /= 0)  call wrerr(rtname,ierr,'Error getting parameters')
                       
  end subroutine ermldet_par




!-------------------------------------------------------------------------------

  subroutine setup_evlist(imageset, eventlist, ierr)

  !--------------------------------------------------------------------------------
  ! This sets up the tEventFile object, in two steps. The first just sets up an
  ! empty object, with the correct extension names and correct column names and
  ! types for each extension. The second step is to flag the small subset of
  ! extensions/columns required for the present task.
  !--------------------------------------------------------------------------------


    character(*),     intent(in)    :: imageset
    type(tEventFile), intent(inout) :: eventlist
    integer,          intent(out)   :: ierr

    integer, dimension(:), allocatable :: ccdnrs
    integer :: corratt_i,kk,ci

    ierr = 0

    ! IMS: set up the extensions expected in the eventlist file, plus the columns expected for each extension.
    !
    call eventlist%read_ccd_nrs(imageset, ccdnrs, ierr)
    call eventlist%init(ccdnrs, ierr)
    deallocate(ccdnrs)

    ! IMS: we only want to read the following subset of the EVENT-extension columns. Note that only X and Y columns seem directly to be used in ermldet, the others are needed by shapelib:slib_evttool_2_myevent().
    !
    call eventlist%events%ibset_col_useage("TIME",  TCOL_READ_BIT)
    call eventlist%events%ibset_col_useage("RA",    TCOL_READ_BIT)
    call eventlist%events%ibset_col_useage("DEC",   TCOL_READ_BIT)
    call eventlist%events%ibset_col_useage("PI",    TCOL_READ_BIT)
    call eventlist%events%ibset_col_useage("RAWX",  TCOL_READ_BIT)
    call eventlist%events%ibset_col_useage("RAWY",  TCOL_READ_BIT)
    call eventlist%events%ibset_col_useage("TM_NR", TCOL_READ_BIT)
    call eventlist%events%ibset_col_useage("X",     TCOL_READ_BIT)
    call eventlist%events%ibset_col_useage("Y",     TCOL_READ_BIT)
    eventlist%events%useageIsSet = .true.

    ! IMS: set up to read all CORRATT columns.
    !
    corratt_i = eventlist%get_ext_index("ext", extCORRATT)
    do kk=1,size(eventlist%ccdnr_infile)
      ci = eventlist%ccdnr_infile(kk) ! TM number
      call eventlist%extList(corratt_i)%extPerTM(ci)%ibset_all_useage(TCOL_READ_BIT) ! read all cols.
      eventlist%extList(corratt_i)%extPerTM(ci)%useageIsSet = .true.
    end do
    eventlist%extList(corratt_i)%useageFlags&
      = ibset(eventlist%extList(corratt_i)%useageFlags, TCOL_READ_BIT)

  end subroutine setup_evlist


!-------------------------------------------------------------------------------

  subroutine ermldet_in(par,boxsrclist,image,eventlist,pixlist,expimage,bkgimage,detmask,ierr)

  !--------------------------------------------------------------------------------
  ! Read input boxlist:
  !  - If instrument summary exists, select only rows id_inst=0,
  !    else if band summary rows exist, select id_band=0   
  !  - Sort accordning to column 'like' in descending order
  ! Read fits images:
  ! If par%compress_flag == true, only relevant section of fits images 
  ! as specified in detection mask header, is applied:
  !   - use fget_<type>compress for reading
  !   - set global variables ximbuf,yimbuf, ixb0, iyb0
  ! 
  !--------------------------------------------------------------------------------


    type(ermldet_param_type), intent(inout)            :: par
    type(boxlist_type)                              :: boxlist,tmplist
    type(boxlist2d_type)                            :: boxsrclist
    type(i4cube_type)                               :: image
    type(tEventFile), dimension(:)                  :: eventlist
    type(pixlist_type), dimension(:)                :: pixlist
    type(l1cube_type)                               :: detmask
    type(r4cube_type)                               :: expimage
    type(r4cube_type)                               :: bkgimage
    type(mllist_type)                               :: mllist
    type(stdhdr_type)                               :: header
    
    integer(kind=int32), intent(out)                :: ierr
    integer(kind=int32)                             :: status,ext,sla_err
    integer(kind=int32)                             :: i,j,k,ipoint,nrows,ord=1,nbands,nmask
    integer(kind=int32),dimension(:),allocatable    :: isort
    type(dethdr_type),dimension(max_exp)            :: mask_dethdr
    type(stdhdr_type),dimension(max_exp)            :: mask_stdhdr
    type(stdhdr_type),dimension(max_ima)            :: image_stdhdr
    logical(kind=4)                                 :: has_instsum,has_bandsum,is_mllist                              
    character(len=*), parameter                     :: rtname='ERMLDET_IN'
    character(len=80)                               :: errstr

    real(REAL64), dimension(:), pointer :: ptr_x=>null(),ptr_y=>null()

    status=0
    ierr=0
    ext=0
    is_mllist=.false.

    ! check input images
    nim=par%n_imagesets



    compress_flag=par%compress_flag

    if (par%expima_flag .and. par%n_expimagesets /= nim) then
      errstr = 'Number of exposure maps and images differ'
      call WRERR(rtname,ierr,errstr)
    endif


! get image std headers:
    if (ierr < 3) then
      do i=1,nim
        call fget_stdhdr(par%imagesets(i),image_stdhdr(i),status)   
      enddo
    endif


 !    call sort_ima(image_stdhdr,nim,status)
 !    -> moved after reading of detection masks headers

! get detection mask headers and determine image buffer size:


!    if (compress_flag) then
    nmask=0
    if (par%detmask_flag) then
!     ximbuf=0
!     yimbuf=0
      nmask=par%n_detmasksets
      do i=1,par%n_detmasksets
         call fget_dethdr(par%detmasksets(i),mask_dethdr(i),status)
         call fget_stdhdr(par%detmasksets(i),mask_stdhdr(i),status)
!         ximbuf=max(ximbuf,mask_dethdr(i)%xdatamax-mask_dethdr(i)%xdatamin+1)
!         yimbuf=max(yimbuf,mask_dethdr(i)%ydatamax-mask_dethdr(i)%ydatamin+1)
      enddo
      if (compress_flag) then
        ximbuf=0
        yimbuf=0
        do i=1,par%n_detmasksets
          ximbuf=max(ximbuf,mask_dethdr(i)%xdatamax-mask_dethdr(i)%xdatamin+1)
          yimbuf=max(yimbuf,mask_dethdr(i)%ydatamax-mask_dethdr(i)%ydatamin+1)
        enddo
      endif
    else
      if (compress_flag) then
        compress_flag=.false.
        ierr = 1
        errstr = 'Need detection masks for compressed reading'
        call WRERR(rtname,ierr,errstr) 
      endif
    endif
!    endif



 !  create index arrays IP and EXPIP   
    call sort_ima(image_stdhdr,mask_stdhdr,nim,nmask,status)

! read images:
    if (ierr < 3) then
      do i=1,num_exp
        ixb0(i)=0
        iyb0(i)=0
        if (par%detmask_flag) then
          call fget_l1compress(par%detmasksets(i),num_exp,ximbuf,yimbuf,mask_dethdr(i),compress_flag,detmask,ord,i,status) 
          if (compress_flag) then
             ixb0(i)=mask_dethdr(i)%xdatamin-1
             iyb0(i)=mask_dethdr(i)%ydatamin-1
             ! set borders to .false.  to avoid problems with ixb/iyb mechanism:
             detmask%arr(i,1,:)=.false.
             detmask%arr(i,ximbuf,:)=.false.
             detmask%arr(i,:,1)=.false.
             detmask%arr(i,:,yimbuf)=.false.
          else
          ! Make sure data boundaries are present in any case:
            detmask%dethdr(i)%xdatamin = 1
            detmask%dethdr(i)%xdatamax =  detmask%shapehdr(i)%naxis1
            detmask%dethdr(i)%ydatamin = 1
            detmask%dethdr(i)%ydatamax =  detmask%shapehdr(i)%naxis2
          endif
        endif  


        do j=1,num_image(i)
          ipoint=ip(i,j)
            ixb0(i)=mask_dethdr(i)%xdatamin-1
            iyb0(i)=mask_dethdr(i)%ydatamin-1
            call fget_i4compress(par%imagesets(ipoint),nim,ximbuf,yimbuf,mask_dethdr(i), &
                                 compress_flag,image,ord,ipoint,status)
            if (.not. compress_flag) then
              ximbuf =  image%shapehdr(1)%naxis1
              yimbuf =  image%shapehdr(1)%naxis2 
            endif
            if (status > 0) then
              ierr=3
              errstr = 'Cannot read science image'
              call WRERR(rtname,ierr,errstr) 
            endif
            
            if (par%expima_flag) then
              call fget_r4compress(par%expimagesets(ipoint),nim,ximbuf,yimbuf,mask_dethdr(i), &
                                   compress_flag,expimage,ord,ipoint,status)

              if (status > 0) then
                ierr=3
                errstr = 'Cannot read exposure map'
                call WRERR(rtname,ierr,errstr) 
              else if (.not. (match_wcs(expimage%wcshdr(ipoint),image%wcshdr(ipoint))  &
                       .and. match_size(expimage%shapehdr(ipoint),image%shapehdr(ipoint)))) then
                ierr=3
                errstr = 'Mismatch of image  and exposure map '
                call WRERR(rtname,ierr,errstr)   
              endif 
            endif
            call fget_r4compress(par%bkgimagesets(ipoint),nim,ximbuf,yimbuf,mask_dethdr(i), &
                                 compress_flag,bkgimage,ord,ipoint,status)
            if (status > 0) then
              ierr=3
              errstr = 'Cannot read background map'
              call WRERR(rtname,ierr,errstr)  
            else if (.not. (match_wcs(bkgimage%wcshdr(ipoint),image%wcshdr(ipoint))  &
                     .and. match_size(bkgimage%shapehdr(ipoint),image%shapehdr(ipoint)))) then
              ierr=3
              errstr = 'Mismatch of image and background map'
              call WRERR(rtname,ierr,errstr)   
            endif
!          else
!            ierr=3
!            errstr = 'Mismatch of image and detector mask'
!            call WRERR(rtname,ierr,errstr)
!          endif
        enddo
      enddo
    endif

!---------------------------------------------------------------------------------------------------- 


! read events and attitude extensions, store in erolib event file structure:

    if (par%shapelet_flag) then

      do i=1,nim
        call setup_evlist(par%imagesets(i), eventlist(i), status)

        ! IMS: now we can open the eventlist file and read the required data.
        !
        call eventlist(i)%open(READONLY, status, filename=par%imagesets(i))
        call eventlist(i)%read(status)
        call eventlist(i)%close(status)

        call eventlist(i)%events%get_col_ptr("X", ptr_x)
        call eventlist(i)%events%get_col_ptr("Y", ptr_y)

        ! calcluate image pixels for events:
!       write(*,*) 'CALL  skypix2ima *********'
      ! remove GL 10/07/2017
        call skypix2ima(image%wcshdr(nim),int(ptr_x(:),kind=int32),int(ptr_x(:),kind=int32),pixlist(i),status)
!       write(*,*) 'DONE  skypix2ima: ', minval(eventlist(i)%events%x), maxval(eventlist(i)%events%x),minval(pixlist(1)%pix(:)%x), maxval(pixlist(1)%pix(:)%x)
!       write(*,*) 'DONE  skypix2ima: ',image%wcshdr(1)%crpix1,image%wcshdr(1)%cdelt1

        nullify(ptr_x, ptr_y)
      enddo

    endif



! get boxlist:


    
!    call fget_boxlist(par%boxlistset,boxlist,status)



    status=0
    call fget_stdhdr(par%boxlistset,header,status)
    if (trim(header%creator) == "ERMLDET") then
      is_mllist=.true.
      call fget_mllist(par%boxlistset,mllist,status)
      allocate(boxlist%data(size(mllist%data)))
      call init_mllist1d(boxlist%data)
      ! copy rows required by erbackmap to boxlist structure:
      boxlist%stdhdr=mllist%stdhdr
      boxlist%shapehdr=mllist%shapehdr
      !boxlist%data(:)%id_src=mllist%data(:)%id_src
      !boxlist%data(:)%box_id_src=mllist%data(:)%id_src      
      !boxlist%data(:)%id_inst=mllist%data(:)%id_inst      
      !boxlist%data(:)%id_band=mllist%data(:)%id_band
      !boxlist%data(:)%x_ima=mllist%data(:)%x_ima
      !boxlist%data(:)%y_ima=mllist%data(:)%y_ima
      !boxlist%data(:)%ra=mllist%data(:)%ra
      !boxlist%data(:)%dec=mllist%data(:)%dec
      !boxlist%data(:)%like=mllist%data(:)%like
      !boxlist%data(:)%scts=mllist%data(:)%scts
      !boxlist%data(:)%box_size=1
      !boxlist%data(:)%ext=mllist%data(:)%ext
      boxlist%data(:)=mllist%data(:)
      boxlist%data(:)%box_id_src= boxlist%data(:)%id_src
      !call copy_mlrows(mllist%data,boxlist%data )
      
      boxlist%has_extentcol=.true.
      deallocate(mllist%data)
      ! set these columns to zero, since we don't want to copy these values to the output list:
      boxlist%data(:)%ccdnr=0
      boxlist%data(:)%rawx=0
      boxlist%data(:)%rawy=0
      boxlist%data(:)%offax=0.0
      boxlist%data(:)%dist_nn=0.0
      boxlist%data(:)%ape_cts  = 0
      boxlist%data(:)%ape_bkg  = 0.0
      boxlist%data(:)%ape_exp  = 0.0     
      boxlist%data(:)%ape_eef  = 0.0
      boxlist%data(:)%ape_radius  = 0.0
      boxlist%data(:)%ape_pois  = 0.0

      
   else
      call fget_boxlist(par%boxlistset,boxlist,status)
    endif   
    
    
    ierr=max(ierr,status) 
    par%fixed_ext_flag=.false.
    if (boxlist%has_extentcol .and. .not. par%fitext_flag) par%fixed_ext_flag=.true.
    
    nrows=boxlist%shapehdr%naxis2   
    
   ! check source list for summary rows:
    has_bandsum = .false.
    has_instsum = .false.
    nbands = maxval(boxlist%data%id_band) 


    if (nrows > 0) then
      !skip this if boxlist is empty (GL 16/04/2019):
      if  (minval(boxlist%data%id_band) /=  maxval(boxlist%data%id_band)) then
        if (minval(boxlist%data%id_band) == 0) then 
          has_bandsum = .true.
        else
          ierr=3
          errstr = 'Unsupported format of input source list'
          call WRERR(rtname,ierr,errstr)
        endif
      endif
      if  (minval(boxlist%data%id_inst) /=  maxval(boxlist%data%id_inst)) then
        if (minval(boxlist%data%id_inst) == 0) then 
          has_instsum = .true.
        else
          ierr=3
          errstr = 'Unsupported format of input source list'
          call WRERR(rtname,ierr,errstr)
        endif
      endif

      if (has_instsum) then
        allocate(tmplist%data(nrows))
        call init_boxlist1d(tmplist%data)
        tmplist%shapehdr=boxlist%shapehdr
        tmplist%stdhdr=boxlist%stdhdr
        j=0
        do i=1,nrows
          if (boxlist%data(i)%id_inst == 0) then
            j=j+1
            tmplist%data(j)=boxlist%data(i)
           endif
        enddo
        nrows=j
      else if (has_bandsum) then 
        allocate(tmplist%data(nrows))
        call init_boxlist1d(tmplist%data)
        tmplist%shapehdr=boxlist%shapehdr
        tmplist%stdhdr=boxlist%stdhdr
        j=0
        do i=1,nrows
          if (boxlist%data(i)%id_band == 0) then
            j=j+1
            tmplist%data(j)=boxlist%data(i)
           endif
        enddo
        nrows=j
      endif

      ! sort input list according to 'like' column:
      ! do not sort, if fitpos_flag=N (forced photometry)
      if (.not. par%fitpos_flag .and. is_mllist) then
        call wrchatty(rtname, "Forced photometry mode, skip sorting input list")
      else
        if (minval(tmplist%data%like) /= maxval(tmplist%data%like)) then
           allocate(isort(nrows))
           ! sort in descending order, don't permutate array  (kflag=-1)
           call spsort(tmplist%data%like,nrows,isort,-1,sla_err)
           tmplist%data=tmplist%data(isort)
           deallocate(isort)
        endif
      endif
  
    else
      ! nrows=0:
      allocate(tmplist%data(0))
    endif

    ! write 2-dimensional source list:
    allocate(boxsrclist%data(0:nim,nrows))
    call init_boxlist2d(boxsrclist%data)
    ! summary row gets first index 0:
    boxsrclist%data(0,:) = tmplist%data(1:nrows)
    deallocate(tmplist%data)
    boxsrclist%shapehdr=boxlist%shapehdr
    boxsrclist%stdhdr=boxlist%stdhdr
    boxsrclist%shapehdr%naxis2=nrows  
    

     ! now fill the other columns 
     do j=1,nrows
      k=0
      do i=1,boxlist%shapehdr%naxis2 
        if  (boxsrclist%data(0,j)%id_src  == boxlist%data(i)%id_src) then
          if (boxlist%data(i)%id_inst /= 0 .and. boxlist%data(i)%id_band /= 0) then 
            k=k+1
            boxsrclist%data(k,j)=boxlist%data(i) 
          endif
        endif
      enddo
    enddo   


    deallocate(boxlist%data)


  end subroutine ermldet_in

!--------------------------------------------------------------------------------



  subroutine ermldet_out(par,mllist,bkgimage,ierr)

  !----------------------------------------------------------------------------
  ! Write ermldet output files:
  !  - output source list
  !  - source map 
  !-----------------------------------------------------------------------------

    type(ermldet_param_type),intent(in)      :: par
    type(mllist_type)                        :: mllist
!    type(i4cube_type)                        :: image
    type(r4cube_type)                        :: bkgimage
    type(r4image_type)                       :: srcmap
    integer(kind=int32)                      :: ierr
    integer(kind=int32)                      :: status,i,xsize,ysize
 
    character(len=*), parameter :: rtname='ERMLDET_OUT'  

    ierr = 0
    status = 0

    call fput_mllist(par%mllistset,mllist,status)
    
 
    if (status >  0) then
      ierr = 3
      call wrerr(rtname,ierr,'Error writing source list')
    endif     

    xsize=size(bkgimage%arr,1)
    ysize=size(bkgimage%arr,2)
    allocate(srcmap%arr(xsize,ysize))
  

    status=0
    do i=1,par%n_srcimagesets
      srcmap%arr=bkgimage%arr(:,:,i)
      srcmap%stdhdr=bkgimage%stdhdr(i)
      srcmap%refhdr=bkgimage%refhdr(i)
      srcmap%wcshdr=bkgimage%wcshdr(i)
       

      srcmap%shapehdr%bitpix=-32
      srcmap%shapehdr%naxis=2
      srcmap%shapehdr%naxis1=xsize
      srcmap%shapehdr%naxis2=ysize  

      srcmap%stdhdr%hduclas1='IMAGE'
      srcmap%stdhdr%hduclas2='GENERIC'
      srcmap%stdhdr%creator='ERMLDET'

     
      call fput_r4image(par%srcimagesets(i),srcmap,status)
      if (status >  0) then
        ierr = 3
        call wrerr(rtname,ierr,'Error writing source-map file')
      endif     

    enddo

    deallocate(srcmap%arr)
    

  end subroutine ermldet_out

!---------------------------------------------------------------------------------

  subroutine ermldet_shapelets(par,boxlist,image,eventlist,allStackPSF,ipsfrad,cluPSF,ierr)

    !-------------------------------------------------------------------------------------------
    ! Copy relevant values from boxlist to  allSRC
    ! Call slib_run to calculate averaged shapelib PSF  for each source.
    ! Allocate arrays in cluPSF
    !-------------------------------------------------------------------------------------------

    ! AGE ADDED THAT: 
    use slib_data_src, only: slib_deallocate_src

   
    type(ermldet_param_type)                    :: par
    type(boxlist2d_type)                        :: boxlist
    type(i4cube_type)                           :: image
    type(tEventFile), dimension(:)              :: eventlist
    type(SRC)                                   :: allSRC
    type(FPSF)                                  :: allStackPSF  
    integer(kind=int32)                         :: ipsfrad
    type(FPSF)                                  :: cluPSF     
    integer(kind=int32)                         :: ierr
    real(kind=real64)                           :: pixscale
    real(kind=real64)                           :: IMG_PIXEL_MM
    real(kind=real64)                           :: ra,dec,xd,yd,zd
    integer(kind=int32)                         :: MAX_PSF_IMG_SIZE
    integer(kind=int32)                         :: nsrc,isrc,iima
    integer(kind=int32)                         :: nc0,nc1,nc2    
!    integer(kind=int32)                         :: status
    logical(kind=log32)                         :: HEALFLAG
    character(len=*),parameter                  :: rtname = 'ERMLDET_SHAPELETS'
    !test
!    real(kind=real32)                        :: t1,t2 

    ierr=0
    pixscale=(abs(image%wcshdr(1)%cdelt1)+abs(image%wcshdr(1)%cdelt2)) * 1800.D0
    IMG_PIXEL_MM =  pixscale/3600.D0 * myDEG2RAD *  1600.D0

    if (ipsfrad == 0) then
       MAX_PSF_IMG_SIZE = 2*nint(par%cutrad)+2
    else
       MAX_PSF_IMG_SIZE = 2*ipsfrad+2

    endif
    cluPSF%myCPSF%MAX_PSF_IMG_SIZE=MAX_PSF_IMG_SIZE
    
    nsrc=boxlist%shapehdr%naxis2
    pixscale=(abs(image%wcshdr(1)%cdelt1)+abs(image%wcshdr(1)%cdelt2)) * 1800.D0


    allocate(allSRC%RA(nsrc))
    allocate(allSRC%DEC(nsrc))
    allocate(allSRC%RADIUS(nsrc))
    allocate(allSRC%XD(nsrc))
    allocate(allSRC%YD(nsrc))
    allocate(allSRC%ZD(nsrc))


    do isrc=1,nsrc

       
      allSRC%ra(isrc)=boxlist%data(0,isrc)%ra   
      allSRC%dec(isrc)=boxlist%data(0,isrc)%dec   
      allSRC%radius(isrc)=par%cutrad*pixscale
      ra=boxlist%data(0,isrc)%ra*cal_deg2rad
      dec=boxlist%data(0,isrc)%dec*cal_deg2rad
      call  slib_src_pos2xyz ( ra, dec, xd, yd, zd)
      
      allSRC%xd(isrc)=xd  
      allSRC%yd(isrc)=yd   
      allSRC%zd(isrc)=zd

    enddo

    allSRC%total=nsrc

      

!    call cpu_time(t1) 
    HEALFLAG=.True.
!    call slib_dealloc_FPSF(allStackPSF)
    call slib_run_wrap(eventlist, allSRC, allStackPSF, MAX_PSF_IMG_SIZE, IMG_PIXEL_MM,HEALFLAG  )
!    call cpu_time(t2)

    
    ! initialize cluPSF: 
    nc0=size(allStackPSF%myCPSF%nx0)
    nc1=size(allStackPSF%myCPSF%nx1)
    nc2=size(allStackPSF%myCPSF%nx2)

    if (allocated(cluPSF%mySPSF)) deallocate(cluPSF%mySPSF)
    allocate(cluPSF%mySPSF(nim))


    do iima=1,nim
      allocate(cluPSF%mySPSF(iima)%c0(nc0,max_fit))
      allocate(cluPSF%mySPSF(iima)%c1(nc1,max_fit))
      allocate(cluPSF%mySPSF(iima)%c2(nc2,max_fit))
    enddo      




      ! copy common PSF data:
      cluPSF%myCPSF = allStackPSF%myCPSF
 
  

     
    ! AGE 28-06-2020
    call slib_deallocate_src(allSRC)    


  end subroutine ermldet_shapelets



!---------------------------------------------------------------------------------


  subroutine cut_radius (par,image,boxlist,ierr)

  !--------------------------------------------------------------
  ! Determine source cut-out radii from PSF 
  ! 
  !  par%multrad: max. radius for multiple source fitting
  !  par%cutrad : cut-out radius for sub-image used for fitting 
  !  rad < 1.0  : interpret as PSF encircled energy fraction
  !               -> calculate corresponding radius in image pixels
  !  rad >= 1.0 : interpret as radius in pixels
  !               -> calculate EEF 
  !  Results are written to boxlist columns cutrad, ecutrad, eef 
  !-------------------------------------------------------------- 


    type(ermldet_param_type)    :: par
    type(i4cube_type)   :: image
    type(cal_par_type)  :: cal_par
    type(boxlist2d_type)  :: boxlist
    integer(kind=int32)  :: ierr
    type(psf2d_type)    :: psf
    logical (kind=log32) :: survmode
    real(kind=real32)   :: energy,eef
    real(kind=real32),parameter   :: max_cal_offax=29.99    
    integer(kind=int32) :: ext,ext_old    
    character(len=clong) :: psf_file,psf_file_old
    integer(kind=int32) :: i,ipoint,status!,j,k
    real(kind=real32) ::  radius,offax,phi,deltax,deltay,pix2arcsec
    real(kind=real64) ::  xcen,ycen
!    real(kind=real64) :: E, theta, fraction

    character(len=*),parameter   :: rtname = 'CUT_RADIUS'


    ierr=0
    

!    write(*,*) '******** START  cut_radius *****'

    ! determine observation mode (POINTING/SURVEY)
    if (image%stdhdr(1)%obs_mode == 'SURVEY' .or. image%stdhdr(1)%obs_mode =='survey') then
      survmode = .true.
    else if  (image%stdhdr(1)%obs_mode == 'SCAN' .or. image%stdhdr(1)%obs_mode == 'scan') then
      survmode = .true. 
    else if  (image%stdhdr(1)%obs_mode == 'SLEW' .or. image%stdhdr(1)%obs_mode == 'slew') then
      survmode = .true. 
    else if  (image%stdhdr(1)%obs_mode == 'POINTING' .or. image%stdhdr(1)%obs_mode =='pointing') then
      survmode = .false. 
    else 
      survmode = .false. 
      ierr=3
      call wrerr(rtname,ierr,'Invalid observing mode')
    endif


    psf_file_old='NONE'
    ext_old=-99
    do ipoint=1,par%n_imagesets
      energy=0.5*(par%emin(ipoint)+par%emax(ipoint))
      pix2arcsec = real(0.5*(abs(image%wcshdr(1)%cdelt1)+abs(image%wcshdr(1)%cdelt2))*3600.)

      call cal_setpar(image%stdhdr(ipoint),cal_par,status) 
      cal_par%codenam='2D_PSF'
      cal_par%expression=''
             
      if (survmode) then
        phi=0.0
        ! get survey PSF
        !cal_par%codenam='2D_PSF'
        !cal_par%expression=''
        
        call cal_r4toex('ENERG',energy,cal_par%expression,status)
        call cal_stringtoex('OBSMODE','SURVEY',cal_par%expression,status)

        call cal_filename(cal_par,psf_file,ext,status)
        call get_2dpsf(psf_file,ext,phi,psf,status)
        ! use same radii for all sources:
        if (par%multrad < 1.0) then
          
          call  psf_eefrad(psf,par%multrad,radius,ierr)
          if (pix2arcsec > 0.0) radius = radius /  pix2arcsec
          ! write max. radius for all bands: 
          if (radius > boxlist%data(ipoint,1)%scutrad) then
            boxlist%data(:,:)%scutrad = radius
          endif
        else
          boxlist%data(:,:)%scutrad = par%multrad
          call  psf_eef(psf,par%multrad*pix2arcsec,.false.,eef,ierr)
          boxlist%data(ipoint,:)%eef=eef
        endif   
        if (par%cutrad < 1.0) then
          call  psf_eefrad(psf,par%multrad,radius,ierr)
          if (pix2arcsec > 0.0) radius = radius /  pix2arcsec 
          ! write max. radius for all bands: 
          if (radius > boxlist%data(ipoint,1)%ecutrad) then
            boxlist%data(:,:)%ecutrad = radius
          endif          
        else
          boxlist%data(:,:)%ecutrad = par%cutrad
          call psf_eef(psf,par%cutrad*pix2arcsec,.false.,eef,status) 
          boxlist%data(ipoint,:)%eef=eef 
        endif   
      else ! pointed mode:
        call  equ2pix(image%wcshdr(ipoint),image%stdhdr(ipoint)%ra_pnt,image%stdhdr(ipoint)%dec_pnt,xcen,ycen,status)
        do i=1,boxlist%shapehdr%naxis2
          deltax=abs(boxlist%data(0,i)%x_ima-real(xcen,real32))
          deltay=abs(boxlist%data(0,i)%y_ima-real(ycen,real32))
          offax = real(sqrt((deltax*image%wcshdr(1)%cdelt1)**2+(deltay*image%wcshdr(1)%cdelt2)**2)*60.D0,real32)
          if (offax > max_cal_offax) then
            ierr=2
            call wrerr(rtname,ierr,"Source off-axis angle exeeds calibrated range")
            offax = max_cal_offax
          endif
          if (deltax == 0.0 .and. deltay == 0.0) then
            phi = 0.0
          else
            phi = atan2(deltax,deltay)
          endif
          cal_par%expression=''
          call cal_r4toex('ENERG',energy,cal_par%expression,status)
          call cal_r4toex('THETA',offax,cal_par%expression,status)
          call cal_stringtoex('OBSMODE','POINTING',cal_par%expression,status)
          call cal_filename(cal_par,psf_file,ext,status)
          if (psf_file /= psf_file_old .or. ext /= ext_old) then
            call get_2dpsf(psf_file,ext,phi,psf,status)   
          endif
          psf_file_old=psf_file
          ext_old=ext   
             
          if (par%multrad < 1.0) then
            call  psf_eefrad(psf,par%multrad,radius,ierr)
            if (pix2arcsec > 0.0) radius = radius /  pix2arcsec 
            if (radius > boxlist%data(ipoint,i)%ecutrad) then
              boxlist%data(:,i)%ecutrad = radius
            endif
          else
             boxlist%data(:,i)%ecutrad = par%cutrad
             call psf_eef(psf,par%cutrad*pix2arcsec,.false.,eef,status)
             boxlist%data(ipoint,i)%eef=eef
          endif
          if (par%cutrad < 1.0) then
            call  psf_eefrad(psf,par%multrad,radius,ierr)
            if (pix2arcsec > 0.0) radius = radius /  pix2arcsec 
            if (radius > boxlist%data(ipoint,i)%ecutrad) then
              boxlist%data(:,i)%ecutrad = radius
            endif
          else
            boxlist%data(ipoint,i)%ecutrad = par%cutrad
            call psf_eef(psf,par%cutrad*pix2arcsec,.false.,eef,status)
            boxlist%data(ipoint,i)%eef=eef
          endif
        enddo
      endif



    end do

    ! 06/12/2016:
    if (associated(psf%arr)) deallocate(psf%arr)




  end subroutine cut_radius

  !---------------------------------------------------------------------

  subroutine sou_clus(par,boxlist,allStackPSF,allStackPSF2,isrc,firststage,clulist,cluPSF,cluPSF2,nclu,mask,ierr)  

  !--------------------------------------------------------------------
  ! Search boxlist (assumed to be sorted by brightness (or significance)
  ! for neighbouring sources within boxlist%data%scutrad.
  ! Parameters of sources in cluster are copied from boxlist to clulist
  ! Split sources into nmulsou new sources.
  ! Set boxlist%data(0,isrc)%cluflag = .true., if source isrc was added
  ! to cluster. 
  !--------------------------------------------------------------------

    type(ermldet_param_type),intent(in)      :: par
    type(boxlist2d_type)                     :: boxlist
    type(FPSF)                               :: allStackPSF
    type(FPSF)                               :: allStackPSF2
    integer(kind=int32)                      :: isrc
    logical(kind=log32)                      :: firststage
!    type(srcrow_type),dimension(:,:),pointer :: clulist
    type(srcrow_type),dimension(:,:),allocatable :: clulist
    
    type(FPSF)                               :: cluPSF
    type(FPSF)                               :: cluPSF2
!    type(SRC),dimension(1)                   :: dummySRC
    type(mask_type)                          :: mask
    integer(kind=int32)                      :: nclu,ierr

    integer(kind=int32)                      :: nsrc,nmul,nmul_act
    integer(kind=int32)                      :: status!,ixm,iym
    integer(kind=int32)                      :: i,j,iclu,iima!,isave
    real(kind=real32)                        :: r_off,scut2,rad2
    real(kind=real32),dimension(max_nmul)    :: xoff_nmul,yoff_nmul
    real(kind=real32),parameter              :: pi=3.1415927
    real(kind=real32),parameter              :: epsrad2=1.E-6 
    real(kind=real64)                        :: dxima,dyima
    logical(kind=log32)                      :: splitflag
!    real(kind=real64)                        :: IMG_PIXEL_MM
!    integer(kind=int32)                      :: MAX_PSF_IMG_SIZE

    character(len=*), parameter              :: rtname='SOU_CLUS'


    ierr=0


    
    nsrc=boxlist%shapehdr%naxis2
    if (par%twostage_flag .and. firststage) then
      nmul=1
    else
      nmul=par%nmulsou
    endif

!   
    if (.not. par%twostage_flag .or. firststage) then
      ! check whether source is already assigned to cluster
      ! and search for neighbours within scutrad
       nclu=0
      scut2=boxlist%data(0,isrc)%scutrad**2 
      nclu=nclu+1
      
      clulist(:,nclu)=boxlist%data(:,isrc)
      boxlist%data(0,isrc)%cluflag=.true.
      if (par%shapelet_flag) then
        if (.not. allocated(cluPSF%myModelPSF)) then
          allocate(cluPSF%myModelPSF(size(allStackPSF%myModelPSF)))
        endif 
        do iima=1,nim
          cluPSF%myModelPSF(:) = allStackPSF%myModelPSF(:)
          cluPSF%mySPSF(iima)%c0(:,nclu) = allStackPSF%mySPSF(iima)%c0(:,isrc)
          cluPSF%mySPSF(iima)%c1(:,nclu) = allStackPSF%mySPSF(iima)%c1(:,isrc)
          cluPSF%mySPSF(iima)%c2(:,nclu) = allStackPSF%mySPSF(iima)%c2(:,isrc)

          cluPSF%mySPSF(iima)%nx0 = allStackPSF%mySPSF(iima)%nx0
          cluPSF%mySPSF(iima)%nx1 = allStackPSF%mySPSF(iima)%nx1
          cluPSF%mySPSF(iima)%nx2 = allStackPSF%mySPSF(iima)%nx2

          cluPSF%mySPSF(iima)%ny0 = allStackPSF%mySPSF(iima)%ny0
          cluPSF%mySPSF(iima)%ny1 = allStackPSF%mySPSF(iima)%ny1
          cluPSF%mySPSF(iima)%ny2 = allStackPSF%mySPSF(iima)%ny2

          cluPSF2%mySPSF(iima)%c0(:,nclu) = allStackPSF2%mySPSF(iima)%c0(:,isrc)
          cluPSF2%mySPSF(iima)%c1(:,nclu) = allStackPSF2%mySPSF(iima)%c1(:,isrc)
          cluPSF2%mySPSF(iima)%c2(:,nclu) = allStackPSF2%mySPSF(iima)%c2(:,isrc)

          cluPSF2%mySPSF(iima)%nx0 = allStackPSF2%mySPSF(iima)%nx0
          cluPSF2%mySPSF(iima)%nx1 = allStackPSF2%mySPSF(iima)%nx1
          cluPSF2%mySPSF(iima)%nx2 = allStackPSF2%mySPSF(iima)%nx2

          cluPSF2%mySPSF(iima)%ny0 = allStackPSF2%mySPSF(iima)%ny0
          cluPSF2%mySPSF(iima)%ny1 = allStackPSF2%mySPSF(iima)%ny1
          cluPSF2%mySPSF(iima)%ny2 = allStackPSF2%mySPSF(iima)%ny2
        enddo 
      endif
      do i=isrc+1,nsrc
        if (nclu == par%nmaxfit) exit
        rad2=(boxlist%data(0,isrc)%x_ima - boxlist%data(0,i)%x_ima)**2 + &
             (boxlist%data(0,isrc)%y_ima - boxlist%data(0,i)%y_ima)**2
        ! GL 20/11/2017: ignore sources with identical position (r^2 < epsrad2):
        if (rad2 < epsrad2) boxlist%data(0,i)%cluflag=.true.
        if (.not. boxlist%data(0,i)%cluflag .and. rad2 < scut2) then
          nclu=nclu+1
          clulist(:,nclu)=boxlist%data(:,i)
          boxlist%data(0,i)%cluflag=.true.
          if (par%shapelet_flag) then
            do iima=1,nim
             ! change 19/11/2017:
             ! cluPSF%mySPSF(iima)%c1(:,nclu)  = allStackPSF%mySPSF(iima)%c1(:,isrc)
             ! cluPSF%mySPSF(iima)%c2(:,nclu)  = allStackPSF%mySPSF(iima)%c2(:,isrc)
              cluPSF%mySPSF(iima)%c0(:,nclu)  = allStackPSF%mySPSF(iima)%c0(:,i)
              cluPSF%mySPSF(iima)%c1(:,nclu)  = allStackPSF%mySPSF(iima)%c1(:,i)
              cluPSF%mySPSF(iima)%c2(:,nclu)  = allStackPSF%mySPSF(iima)%c2(:,i)

              cluPSF%mySPSF(iima)%nx0 = allStackPSF%mySPSF(iima)%nx0
              cluPSF%mySPSF(iima)%nx1 = allStackPSF%mySPSF(iima)%nx1
              cluPSF%mySPSF(iima)%nx2 = allStackPSF%mySPSF(iima)%nx2
              cluPSF%mySPSF(iima)%ny0 = allStackPSF%mySPSF(iima)%ny0
              cluPSF%mySPSF(iima)%ny1 = allStackPSF%mySPSF(iima)%ny1
              cluPSF%mySPSF(iima)%ny2 = allStackPSF%mySPSF(iima)%ny2

             ! cluPSF2%mySPSF(iima)%c1(:,nclu) = allStackPSF2%mySPSF(iima)%c1(:,isrc)
             ! cluPSF2%mySPSF(iima)%c2(:,nclu) = allStackPSF2%mySPSF(iima)%c2(:,isrc)
              cluPSF2%mySPSF(iima)%c0(:,nclu) = allStackPSF2%mySPSF(iima)%c0(:,i)
              cluPSF2%mySPSF(iima)%c1(:,nclu) = allStackPSF2%mySPSF(iima)%c1(:,i)
              cluPSF2%mySPSF(iima)%c2(:,nclu) = allStackPSF2%mySPSF(iima)%c2(:,i)

              cluPSF2%mySPSF(iima)%nx0 = allStackPSF2%mySPSF(iima)%nx0
              cluPSF2%mySPSF(iima)%nx1 = allStackPSF2%mySPSF(iima)%nx1
              cluPSF2%mySPSF(iima)%nx2 = allStackPSF2%mySPSF(iima)%nx2

              cluPSF2%mySPSF(iima)%ny0 = allStackPSF2%mySPSF(iima)%ny0
              cluPSF2%mySPSF(iima)%ny1 = allStackPSF2%mySPSF(iima)%ny1
              cluPSF2%mySPSF(iima)%ny2 = allStackPSF2%mySPSF(iima)%ny2
            enddo 
          endif         
        endif
      end do
      !endif
    endif

    if (nmul > 1) then
      ! calulate offset pattern for split sources
      ! (circular pattern with r = 2 pix around original position)
      r_off=min(boxlist%data(0,isrc)%ecutrad/2.0,2.0)
      ! try 4 pix :
      ! r_off=min(boxlist%data(0,isrc)%ecutrad/2.0,4.0)

      do j=1,nmul


        xoff_nmul(j)=r_off*cos(real(j-1)*2.*pi/real(nmul))
        yoff_nmul(j)=r_off*sin(real(j-1)*2.*pi/real(nmul))
!        yoff_nmul(j)=r_off*cos(real(j-1)*2.*pi/real(nmul))
!        xoff_nmul(j)=r_off*sin(real(j-1)*2.*pi/real(nmul))

      enddo

      iclu=nclu
      do i=1,nclu
        splitflag = .true.
        if (par%thres_flag) then
!          write(*,*) rtname,', thresh: ', par%thres_flag,' ', par%thres_col,' ',clulist(0,nclu)%like, par%thres_val
          if ((par%thres_col == 'LIKE' .or. par%thres_col == 'like') &
               .and.  (clulist(0,nclu)%like < par%thres_val) ) splitflag = .false.
          if ((par%thres_col == 'RATE' .or. par%thres_col == 'rate') &
               .and.  (clulist(0,nclu)%rate < par%thres_val) ) splitflag = .false.
          if ((par%thres_col == 'SCTS' .or. par%thres_col == 'scts') &
               .and.  (clulist(0,nclu)%scts < par%thres_val) ) splitflag = .false.
        endif
!        write(*,*) rtname,', splitflag:',splitflag       
        if (splitflag) then 
!          write(*,*) rtname,' split'
          ! copy parameters to new sources, apply modifications to counts and positions
          nmul_act=1
          do j=2,nmul
            iclu=iclu+1
            if (iclu .le. max_fit) then
              nmul_act=nmul
              clulist(:,iclu)=clulist(:,i) 
              clulist(:,iclu)%scts=clulist(:,iclu)%scts/real(nmul)
              clulist(:,iclu)%x_ima=clulist(:,iclu)%x_ima+xoff_nmul(j)
              clulist(:,iclu)%y_ima=clulist(:,iclu)%y_ima+yoff_nmul(j)  
              ! for 2nd stage, check whether position is within mask:
              if (par%twostage_flag .and. .not.firststage) then
                dxima=clulist(1,iclu)%x_ima
                dyima=clulist(1,iclu)%y_ima
                call cagemask(dxima,dyima,mask,status) 
                clulist(:,iclu)%x_ima=real(dxima)
                clulist(:,iclu)%y_ima=real(dyima)
              endif

              
              if (par%shapelet_flag) then
                do iima=1,nim
                  cluPSF%mySPSF(iima)%c0(:,iclu) =  allStackPSF%mySPSF(iima)%c0(:,i)
                  cluPSF%mySPSF(iima)%c1(:,iclu) =  allStackPSF%mySPSF(iima)%c1(:,i)
                  cluPSF%mySPSF(iima)%c2(:,iclu) =  allStackPSF%mySPSF(iima)%c2(:,i)
                  cluPSF2%mySPSF(iima)%c0(:,iclu) = allStackPSF2%mySPSF(iima)%c0(:,i)
                  cluPSF2%mySPSF(iima)%c1(:,iclu) = allStackPSF2%mySPSF(iima)%c1(:,i)
                  cluPSF2%mySPSF(iima)%c2(:,iclu) = allStackPSF2%mySPSF(iima)%c2(:,i)
                enddo 
              endif
            else
              ierr=1
              call wrerr(rtname,ierr,'Cannot split source, MAX_FIT exceeded')
            endif                          
          enddo
          ! modify also original source:
          clulist(:,i)%scts=clulist(:,i)%scts/real(nmul_act)
          !clulist(:,i)%x_ima=clulist(:,i)%x_ima+xoff_nmul(1)
          !clulist(:,i)%y_ima=clulist(:,i)%y_ima+yoff_nmul(1)                     
        endif
      enddo 
    nclu=iclu   
    endif 

    if (par%shapelet_flag) then
      do iima=1,nim
        cluPSF%mySPSF(iima)%num_src=nclu
        cluPSF2%mySPSF(iima)%num_src=nclu
      enddo
    endif 

  end subroutine sou_clus

  !------------------------------------------------------------------------------

 


  subroutine add_srcmap(par,clulist,cluPSF,image,eventlist,pixlist,detmask,expimage,nclu,events,mask,bkgimage,srcmodrad,modelima,ierr)

  !---------------------------------------------------------------------------------
  ! Add source model to background map.
  ! Call fill_arrs with new cut-out radii
  ! and call subroutine model 
  ! Also calculate PSF-weighted exposure values here.
  !---------------------------------------------------------------------------------
   

    type(ermldet_param_type),intent(in)      :: par
!    type(srcrow_type),dimension(:,:),pointer :: clulist
    type(srcrow_type),dimension(:,:),allocatable :: clulist
    
    type(FPSF)                               :: cluPSF
    type(i4cube_type)                        :: image
    type(tEventFile), dimension(:)           :: eventlist 
    type(pixlist_type),dimension(:)          :: pixlist
    type(l1cube_type)                        :: detmask
    type(r4cube_type)                        :: expimage,bkgimage
    integer(kind=int32)                      :: srcmodrad
    integer(kind=int32)                      :: nclu
    type(events_type),pointer,dimension(:)   :: events
    type(mask_type)                          :: mask  
    real(kind=real64),dimension(:,:,:),pointer  :: modelima
    integer(kind=int32)                      :: ierr
    logical(kind=log32)                      :: newcal
    integer(kind=int32)                      :: status!, isrc,isize
    real(kind=real64)                        :: yfit0
    real(kind=real64),dimension(:),allocatable  :: yfit1
    real(kind=real64),dimension(max_ima)        :: yfit2 
    logical(kind=log32)                         :: last = .FALSE.
    real(kind=real64), dimension(max_fit)         :: x0,y0,sig         
    real(kind=real64), dimension(max_ima,max_fit) :: rate
    integer(kind=int32)                           :: ipoint,ix,iy,mx,my

    ! test output 
!    integer(kind=int32)                           :: iix,iiy,iii
!    integer(kind=int32),save                      :: imodelima=0
!    character(len=5)                              :: imstr


    character(len=*), parameter              :: rtname='ADD_SRCMAP'

    
    ierr=0
    
    ! read new cal PSF
    newcal=.true.
    last=.false.
  
    ! set cut-out radii for source model
    clulist(:,1:nclu)%ecutrad = real(srcmodrad)
 

    call fill_arrs(par,.false.,clulist,image,eventlist,pixlist,cluPSF,detmask,expimage,modelima,nclu,events,mask,status)

    

    ! allocate yfit1 , only needed for compatibility with subroutine 'model':


     if (allocated(yfit1)) deallocate(yfit1)
     allocate(yfit1(mask%nev))  

!*     if (allocated(yfit3)) deallocate(yfit3)
!*      allocate(yfit3(mask%npix))  


    ! copy source parameters from source list:

    x0  = clulist(0,:)%x_ima
    y0  = clulist(0,:)%y_ima
    sig = clulist(0,:)%ext
    rate(1:par%n_imagesets,:) = clulist(1:par%n_imagesets,:)%rate
   
   
    call model(par,image,expimage,bkgimage,mask,events,clulist,cluPSF,x0,y0,sig,rate,nclu,last,newcal,yfit0,yfit1,yfit2,modelima,status)


    do ipoint=1,par%n_imagesets
      my=0
      do iy = mask%iy_min, mask%iy_max
         mx=0
         my=my+1
         do ix = mask%ix_min, mask%ix_max
           mx=mx+1
           if (mask%mask(mx,my,ipoint) > 0) then
              bkgimage%arr(ix,iy,ipoint) = bkgimage%arr(ix,iy,ipoint)  &
                                         + real(modelima(mx,my,ipoint),kind=real32) * expimage%arr(ix,iy,ipoint)

           endif
         enddo
      enddo
    end do
     

    deallocate(yfit1)
!*    deallocate(yfit3)

 


  end subroutine add_srcmap


  !-----------------------------------------------------------------------------------------------------

  subroutine init_pixlist(eventlist,header,pixlist,ierr)

    !------------------------------------------------------------------------------
    ! Initializes pixlist and
    ! calculates image pixel coordinates from sky X and Y
    ! pixel coordinates using a SASS WCS header.
    ! 
    ! Input variables:
    !  eventlist : event list of erolib_events_mod::tEvents type
    !  header    : WCS header of corresponing image
    ! Output variables:
    !  pixlist ,containing %pix%x, %pix%y, %pix%id
    !  ierr      : error status
    ! 
    !-------------------------------------------------------------------------------

   type(tEvents)                                :: eventlist
   type(wcs_imahdr_type),intent(in)             :: header
   type(pixlist_type)                           :: pixlist
   integer(kind=int32),intent(out)              :: ierr

   integer(kind=int32)                          :: narr,i
   character(len=*),parameter                   :: rtname = 'INIT_PIXLIST'

   real(REAL64), dimension(:), pointer :: ptr_x=>null(),ptr_y=>null()

   ierr=0

   narr=eventlist%getNumRows()

   if (allocated(pixlist%pix)) then
     deallocate(pixlist%pix)
   endif

   allocate(pixlist%pix(narr))


   call eventlist%get_col_ptr("X", ptr_x)
   call eventlist%get_col_ptr("Y", ptr_y)

   ! calculate image pixels:
   do i=1,narr
     pixlist%pix(i)%id=i
     pixlist%pix(i)%x=nint(header%crpix1p+(ptr_x(i)-header%crval1p)/header%cdelt1p,kind=int16)
     pixlist%pix(i)%y=nint(header%crpix2p+(ptr_y(i)-header%crval2p)/header%cdelt2p,kind=int16) 
   enddo
   pixlist%n=narr

   nullify(ptr_x,ptr_y)
  
  end subroutine init_pixlist

  !-----------------------------------------------------------------------------------------------------


  
  subroutine mldetect(par,image,eventlist,pixlist,detmask,expimage,bkgimage,allStackPSF,cluPSF,allStackPSF2,cluPSF2,boxlist,srclist,srcmodrad,nsrc,ierr)


    type(ermldet_param_type)                 :: par
    type(i4cube_type)                        :: image
    type(tEventFile),dimension(:)            :: eventlist
    type(pixlist_type),dimension(:)          :: pixlist
    type(l1cube_type)                        :: detmask
    type(r4cube_type)                        :: expimage
    type(r4cube_type)                        :: bkgimage
!    type(r4cube_type)                        :: srcimage
    type(boxlist2d_type)                     :: boxlist
    type(srcrow_type),dimension(:,:),pointer :: srclist
    integer(kind=int32)                      :: srcmodrad  
!    type(mllist_type)                       :: mllist
    integer(kind=int32)                      :: ierr

    type(events_type),pointer,dimension(:)   :: events => NULL()
!    type(srcrow_type),dimension(:,:),pointer :: clulist  => NULL()
    type(srcrow_type),dimension(:,:),allocatable :: clulist
    type(srcrow_type),dimension(:,:),allocatable :: clulist_save
    
    type(FPSF)                               :: allStackPSF,allStackPSF2   
    type(FPSF)                               :: cluPSF,cluPSF2   
!    type(SRC),dimension(1)                   :: dummySRC                                               
    type(mask_type),save                     :: mask
!    real(kind=real32), dimension(max_mask,max_mask,max_ima) :: modelima
    real(kind=real64), dimension(:,:,:), pointer   :: modelima  => NULL() 
    logical(kind=log32)                      :: firststage,fitdone
    integer(kind=int32)                      :: nsrc_in,nsrc,isrc,status
    integer(kind=int32)                      :: nclu,nclu_save,maxclu,iclu,i,j!,k,npts
    character(len=80)                        :: outstr
    character(len=12)                        :: istr,rstr
!    real(kind=real64)                        :: IMG_PIXEL_MM
!    integer(kind=int32)                      :: MAX_PSF_IMG_SIZE
    real(kind=real32)                        :: t1,t2!,t0,t3    

    character(len=*), parameter              :: rtname='MLDETECT'

    ierr = 0
    status=0
    nclu=0
    
    nsrc_in = boxlist%shapehdr%naxis2
    maxclu=par%nmaxfit*par%nmulsou




    ! allocate clulist with summary element (0:n_imagesets,maxclu)
    
!    if (.not. associated(clulist)) allocate(clulist(0:par%n_imagesets,max_fit))
    if (.not. allocated(clulist)) allocate(clulist(0:par%n_imagesets,max_fit))
    if (.not. allocated(clulist_save)) allocate(clulist_save(0:par%n_imagesets,max_fit))

    
    call init_mllist2d(clulist)


    if (.not. associated(srclist)) allocate(srclist(0:par%n_imagesets,max_sou))

    call init_mllist2d(srclist)   



    boxlist%data(:,:)%cluflag=.false.

    nsrc = 0
    iclu = 0 


    call cut_radius(par,image,boxlist,status)

    
    do isrc=1,nsrc_in
       
       ! clean extent values in clulist array (12/12/2017)
       if (.not. par%fixed_ext_flag) then
         clulist(:,:)%ext=0.0      
       endif 

       verbose_flag = .false.
      ! if (isrc == 23 .or. isrc == 49) verbose_flag = .true.

       fcounter=0
       pcounter=0
       mcounter=0
       lcounter=0
       kcounter=0
       cclock0=0
       cclock1=0
       cclock2=0      

       call init_mllist2d(clulist)

      call cpu_time(t1) 
      fitdone=.false.
      
      if ( .not. boxlist%data(0,isrc)%cluflag ) then 

     ! moved outside source loop (19/03/2018):
       !! call cut_radius(par,image,boxlist,status)

          
         if (par%twostage_flag) then
            firststage=.true.
            
          call sou_clus(par,boxlist,allStackPSF,allStackPSF2,isrc,firststage,clulist,cluPSF,cluPSF2,nclu,mask,status)
          call fill_arrs(par,par%photon_flag,clulist,image,eventlist,pixlist,cluPSF,detmask,expimage,modelima,nclu,events,mask,status)
           if (mask%nev > 1) then
            call mulsou(par,image,eventlist,pixlist,expimage,bkgimage,detmask,clulist,cluPSF,events,mask,firststage,nclu,modelima,status)
            fitdone=.true.
         endif
        endif

          

        
        ! only if ext > extmin (GL 07/12/2016)
        ! fixed access of unitialised values ( clulist(0,:)%ext ), (GL 10/12/2020):
        
        if (.not. par%twostage_flag .or. (maxval(clulist(0,1:nclu)%ext) >= par%extmin)) then
           firststage=.false.
           clulist_save=clulist
           nclu_save=nclu
            call sou_clus(par,boxlist,allStackPSF,allStackPSF2,isrc,firststage,clulist,cluPSF,cluPSF2,nclu,mask,status)
           ! in case of 2nd stage leave fitting sub-regions unchanged  (15/12/2017):
           if (.not. par%twostage_flag) then  ! TEST
              call fill_arrs(par,par%photon_flag,clulist,image,eventlist,pixlist,cluPSF,detmask,expimage,modelima,nclu,events,mask,status)
            endif
           if (((mask%nev > (nim+3)*nclu+1 ) .or. (.not. par%fitpos_flag)) .and. (nclu > 0)) then
              call mulsou(par,image,eventlist,pixlist,expimage,bkgimage,detmask,clulist,cluPSF,events,mask,firststage,nclu,modelima,status)  
              fitdone=.true.
             !         if events limit is not met, re-use old clulist for 2nd stage:
          else if ( par%twostage_flag .and. (nclu > 0) .and.  (mask%nev <=  (nim+3)*nclu+1 )) then
              clulist=clulist_save
              nclu=nclu_save
              call mulsou(par,image,eventlist,pixlist,expimage,bkgimage,detmask,clulist,cluPSF,events,mask,firststage,nclu,modelima,status)
           endif      
        endif

        
        ! clean source list:
        i=1
        do while (i <= nclu)
           
           if (clulist(0,i)%like < par%likemin) then
             do j=i+1,nclu
              
              clulist(:,j-1)=clulist(:,j)
            enddo
            nclu=nclu-1
         else
            i=i+1
          endif
        enddo

        if (nclu > 0 .and. fitdone) then
          iclu=iclu+1
          ! copy source cluster to source list: 
          do i=1,nclu
             nsrc=nsrc+1
            srclist(:,nsrc) =  clulist(:,i)
            srclist(:,nsrc)%id_cluster = iclu
          enddo
          call add_srcmap(par,clulist,cluPSF2,image,eventlist,pixlist,detmask,expimage,nclu,events,mask,bkgimage,srcmodrad,modelima,status)
        endif
       
        ! now increase cut-out radii and calculate source models for 
        ! background maps:



!        call add_srcmap(par,clulist,image,detmask,expimage,nclu,events,mask,bkgimage,modelima,status)
         call cpu_time(t2)

         write(istr,'(I6)')  nsrc
         write(rstr,'(F12.6)')  t2-t1
         outstr = ' source no: '//istr//', CPU time: '//rstr//' s'
         call wrchatty(rtname,outstr)
      endif



    enddo

    call wrchatty(rtname, '*********** DONE input list ***********')

  end subroutine mldetect

end module  ermldet_mod
