! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.

module fftpsf

!---------------------------------------------------------------------------------
! Module containing PSF handling for ermldet.
! Author: Georg Lamer , AIP         date:  19-MAY-2010  
! update: GLamer                    date:  02-08-2013  :  correct error in fpsf 
!                                   (wrongly returning fima for largest index ipoint) 
! ermldet v 1.11
!---------------------------------------------------------------------------------


  use param
  use numkind  
  use fitsimage 
  use detect_io
  use cal 
  use psflib
  use slib_constants
  use slib_image

  
  implicit none
  public :: fpsf


  logical(kind=log32)  :: verbose_flag2,verbose_flag3,verbose_flag4

  contains

!---------------------------------------------------------------------------------------

  subroutine ftpsf(par,image,newcal,iinst,iima,ns,ix0,iy0,x0,y0,sig,ecut,cluPSF,fima,isize,ierr)

  !--------------------------------------------------------------------------------
  ! On first call or if newcal=.true., read PSF from CALDB,
  ! otherwise use PSF image saved in array sima.
  ! If sig > 0 and sig has changed wrt previous call, calculate 
  ! extent model.
  ! Resample PSF image to pixels of science image and fold with 
  ! extent model.
  ! Argument newcal is changed to .false.
  !---------------------------------------------------------------------------------


    type(ermldet_param_type)                    :: par
    type(i4cube_type)                           :: image
    real(kind=real32)                           :: ecut
    type(FPSF)                                  :: cluPSF   
    logical(kind=log32)                         :: newcal
    integer(kind=int32),intent(in)              :: ix0,iy0 
    real(kind=real64), intent(in)               :: x0, y0, sig
    
    ! TEST 01/04/2021:
    !real(kind=real64)                    :: x0, y0, sig
    
    integer(kind=int32), intent(in)             :: iinst,iima,ns
    !    real(kind=real32), dimension(:,:), pointer  :: fima
     real(kind=real64), dimension(:,:),allocatable    :: fima
       
    integer(kind=int32)                         :: ierr

    integer(kind=int32)                         :: status
!    type(cal_par_type)                          :: cal_par
!    type(psf2d_type)                            :: psf
!    character(len=clong)                        :: psf_file,psf_file_old
!    integer(kind=int32)                         :: ext,nbands,ipoint,iband,nsold=-1
    integer(kind=int32)                         :: nbands,ipoint,iband
    
!    real(kind=real32)                           :: energy
    integer(kind=int32)                                     :: isize
    integer(kind=int32)                                     :: num_src
!    real(kind=real32)                                       :: modval
    real(kind=real64)                                       :: modval
    
!    real(kind=real64), dimension(max_fit)                   :: ra,dec
    real(kind=real64)                                       :: xremain, yremain
    integer(kind=int32), dimension(:,:), allocatable       :: xsort,ysort
    real(kind=real64), dimension(:,:,:), allocatable        :: sima
    real(kind=real64), dimension(:,:,:,:), allocatable      :: bima0
    real(kind=real64), dimension(:,:),  allocatable         :: bima,mima
    real(kind=real64), dimension(:,:,:),  allocatable       :: mima0
!    real(kind=real32), dimension(3,3)                       :: sm_kernel
!    real(kind=real32), dimension(:,:), allocatable         :: pima
    complex(kind=real64), dimension(:,:),  allocatable           :: ffmodel,ffpsf,prodima,phase_arr
    complex(kind=real64), dimension(:,:,:),  allocatable         :: phase_arr0,ffmodel0
    complex(kind=real64), dimension(:,:,:,:),  allocatable       :: ffpsf0
  

    integer(kind=int64)                                     :: pplan,mplan,fplan
    logical(kind=log32)                                     :: havePlan=.false.

!    real(kind=real32), dimension(max_ima) :: psfcenx,psfceny,pfac

!    real(kind=real64) :: x,y,x1,y1
    real(kind=real64) :: x,y
    
!    real(kind=real64) :: dx,dy,ddx,ddy,xd0,yd0,xx,yy
    real(kind=real64) , dimension(1:max_fit) :: x0old,y0old,sigold,x0init,y0init
    
!    real(kind=real32) :: ssum,xyoff2,ecutold,fluxfac,fftscale,sig4sq
    real(kind=real64) :: fftscale,sig4sq
    
    real(kind=real64) :: sigmin=0.1,norm
!    real(kind=real32) :: pi=3.141592654,twopi=6.283185308

!    integer(kind=int32)             :: i,j,ix,iy,ixx,iyy,lx,ly,kx,ky,kk,np,np2,isig3,isig3sq,isrc
   integer(kind=int32)             :: i,j,isrc
     
!   integer(kind=int32)             :: isx,isy,isizeold=-1,n,iecut,ixmin,ixmax,iymin,iymax,ninst
   integer(kind=int32)             :: isizeold=-1,iecut,ninst
    
!    integer(kind=int32), dimension(max_ima) :: nx, ny
!    real(kind=real32)                       :: offax
!    real(kind=real64)               :: pxmin,pxmax,pymin,pymax,dx1,dx2,dy1,dy2
    integer(kind=int32)             :: hsize
!    integer(kind=int32)             :: pxmin0,pymin0,pxmax0,pymax0,pxmin1,pymin1,hsize
    
!    real(kind=real64)               :: img_pixel_mm
!    integer(kind=int32) :: ixcen,iycen,imaxrad,ipsfrad
!    real(kind=real32)   :: E, theta, phi, val
    logical(kind=log32) :: first = .true., imallocated=.false.
    logical(kind=log32) :: havePSF, haveModel 
    logical(kind=log32) :: survmode
!    real(kind=real32)   :: t1,t2
!    integer(kind=int32) :: clock1,clock2,clock3
    logical(kind=log32) :: event_flag
 !   integer(kind=int32)                             :: imcounter
 !   character(len=5)                                :: imstr

    character(len=*), parameter  :: rtname='FTPSF '
   
! from FFTW include file:
      INTEGER, PARAMETER :: FFTW_MEASURE=0
      INTEGER, PARAMETER :: FFTW_EXHAUSTIVE=8
      INTEGER, PARAMETER :: FFTW_PATIENT=32
      INTEGER, PARAMETER :: FFTW_ESTIMATE=64
      INTEGER, PARAMETER :: FFTW_ESTIMATE_PATIENT=128


      save :: mima,mima0,ffmodel,ffpsf,ffpsf0,prodima,xsort,ysort,phase_arr,phase_arr0,ffmodel0
      save :: bima0,bima,imallocated,first,havePSF,haveModel
      save :: sigold,x0old,y0old,isizeold,fftscale      
      save :: pplan,mplan,fplan


    if (image%stdhdr(iima)%obs_mode == 'SURVEY' .or. image%stdhdr(iima)%obs_mode =='survey') then
      survmode = .true.
    else if  (image%stdhdr(1)%obs_mode == 'SCAN' .or. image%stdhdr(iima)%obs_mode == 'scan') then
      survmode = .true. 
    else if  (image%stdhdr(1)%obs_mode == 'SLEW' .or. image%stdhdr(iima)%obs_mode == 'slew') then
      survmode = .true. 
    else if  (image%stdhdr(iima)%obs_mode == 'POINTING' .or. image%stdhdr(iima)%obs_mode =='pointing') then
      survmode = .false. 

    else 
      survmode = .false. 
      ierr=3
      call wrerr(rtname,ierr,'Invalid observing mode')
    endif

! Test new method to calculate pos. offset:
!    if (newcal) then 
!       x0init(:)=-1.D9
!       y0init(:)=-1.D9
!    endif
!    
!    if (x0init(ns) < 0.D0 ) then
!       x0init(ns)=x0     
!    endif
!     
!    if (y0init(ns) < 0.D0 ) then
!       y0init(ns)=y0     
!    endif
    
!    xremain=x0-dble(nint(x0init(ns)))
!    yremain=y0-dble(nint(y0init(ns)))
 

!--------------------------    
    !xremain=x0-dble(nint(x0))
    !yremain=y0-dble(nint(y0))
    xremain=x0-dble(ix0)
    yremain=y0-dble(iy0)



    
    nbands=num_image(iinst)

!    count_psf=count_psf+1

    if (first) then
      sigold(:)=-1.0
      y0old(:)=-1.E9
      x0old(:)=-1E9
      havePSF=.false.
      isizeold=0
!      write(*,*) 'FFTPSF legacy fortran'
     end if 

     
!    determine subimage array sizes (use 2**n sizes for FFT)  
    



! Test performance with non- 2**n sizes:

!    isize=nint(2.*ecut)+2
     !new (07/12/2017):

     
     !isize=2*nint(ecut)+2

     !test larger isize (01/04/2021):
     !isize=max(2*nint(ecut)+2,64)
     !isize=2*nint(ecut)+2
     
     if (par%shapelet_flag) isize = cluPSF%myCPSF%MAX_PSF_IMG_SIZE
     hsize = isize/2

     iecut=int(ecut)+1



    if (isizeold /= isize) then
      !*********
      havePSF=.false.
      haveModel=.false.
      fftscale=1/float((isize*isize))
      if (.not. first) then
        if (imallocated) then
          deallocate(mima)
          deallocate(mima0)
          deallocate(ffmodel)
          deallocate(ffpsf)
          deallocate(ffpsf0)
          deallocate(prodima)
          deallocate(xsort)
          deallocate(ysort) 
          deallocate(phase_arr)
          deallocate(phase_arr0)
          deallocate(ffmodel0)
          imallocated = .false.
        endif
!        write(*,*) ' deallocate bima0'
        if (allocated(bima0)) deallocate(bima0)
        if (allocated(bima)) deallocate(bima)
!        if (associated(fima)) deallocate(fima)        
        if (allocated(fima)) deallocate(fima)
      endif
      if (.not. imallocated) then     
        allocate( mima0(1:isize,1:isize,max_fit) )
        allocate( mima(1:isize,1:isize))
        allocate( ffmodel(1:hsize+1,1:isize) )
        allocate( ffpsf(1:hsize+1,1:isize) )
        allocate( ffpsf0(1:hsize+1,1:isize,nim,max_fit) )
        allocate( prodima(1:hsize+1,1:isize) )
        allocate( xsort(1:isize,1:isize) )
        allocate( ysort(1:isize,1:isize) )
        allocate(phase_arr(hsize+1,isize))
        allocate(phase_arr0(hsize+1,isize,max_fit))
        allocate(ffmodel0(hsize+1,isize,max_fit))
      
        imallocated = .true.

       ! initialize:
       mima0(:,:,:)       = 0.0D0
       mima(:,:)          = 0.0D0
       phase_arr(:,:)     = cmplx(0.D0,0.D0,kind=real64)
       phase_arr0(:,:,:)  = cmplx(0.D0,0.D0,kind=real64)
       ffmodel(:,:)       = cmplx(0.D0,0.D0,kind=real64)
       ffmodel0(:,:,:)    = cmplx(0.D0,0.D0,kind=real64)
       ffpsf(:,:)         = cmplx(0.D0,0.D0,kind=real64)
       ffpsf0(:,:,:,:)    = cmplx(0.0D0,0.0D0,kind=real64)
       prodima(:,:)       = cmplx(0.D0,0.D0,kind=real64)
       xsort(:,:)         = 0
       ysort(:,:)         = 0
      endif

      allocate( bima(1:isize,1:isize) )
      allocate( bima0(1:isize,1:isize,nim,max_fit) )
      allocate( fima(1:isize,1:isize) )
      fima(:,:)      =  0.0D0
      bima(:,:)      =  0.0D0
      bima0(:,:,:,:) =  0.0D0

!      create sorting arrays to rearrange quadrants of kernel image for FFT
!
!    1|2         4|3
!    ---         ---
!    3|4    =>   2|1 
!
       do i=1,isize
        do j=1,isize
          if (i .le. isize/2) then
            xsort(i,j) = isize/2+i
          else
            xsort(i,j) = i-isize/2
          endif
          if (j .le. isize/2) then
            ysort(i,j) = isize/2+j
          else
            ysort(i,j) = j-isize/2
          endif             
        enddo
      enddo 
      isizeold=isize 




      ! create FFTW plans:

      ! Always create plans (GL 29/11/2016)
      !if (par%fitext_flag) then
        if (havePlan) then
          call dfftw_destroy_plan(pplan)
          call dfftw_destroy_plan(mplan)
          call dfftw_destroy_plan(fplan)
       endif
        call dfftw_plan_dft_r2c_2d(pplan,isize,isize,bima,ffpsf,FFTW_MEASURE)
        call dfftw_plan_dft_r2c_2d(mplan,isize,isize,mima,ffmodel,FFTW_MEASURE)
        call dfftw_plan_dft_c2r_2d(fplan,isize,isize,prodima,fima,FFTW_MEASURE)
        havePlan=.true.
      !endif
      
!      call dfftw_print_plan(pplan)

    endif ! isize /= isizeold

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !! Test performance gain with newcal=.false.
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !newcal=.false.
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    if (newcal) then 
       havePSF=.false.
       x0init(:)=-1.D9
       y0init(:)=-1.D9
    endif
    
 
  if (.not. havePSF) then 

    if (allocated(sima)) deallocate(sima)

    if (par%shapelet_flag) then
      do ipoint=1,nim
        num_src=cluPSF%mySPSF(ipoint)%num_src
        do isrc=1,num_src
!          write(*,*) rtname, ' get_shapePSF ',par%shapelet_flag, ' ',ipoint,isrc
!          write(*,*) 'cluPSF image1, C1,C2:', cluPSF(1)%C1(1,1)
!          write(*,*) 'cluPSF image2, C1,C2:', cluPSF%mySPSF(ipoint)%C1(1:4,isrc)
!          write(*,*) 'ISNAN:',isnan(cluPSF%mySPSF(ipoint)%C1(1,isrc)) 
          if (abs(cluPSF%mySPSF(ipoint)%C1(1,isrc)) < 9999.0) then
             event_flag=.false.
            call get_shapePSF(event_flag,cluPSF,ipoint,isrc,bima,ierr)
            !!!!!!!!!  TEST2.4168  normalize to ~ 1.0
            bima=max(1.0D-7,bima)
!            write(*,*) "bima1:",sum(bima)
            !if (maxval(bima) > 1.0) write(*,*) 'max(bima),sh=',maxval(bima)
            ! TEST: extra smoothing (GL 03/06/2019):
            if (par%sixte_flag) call sixte_smoothing(bima,status) 
            !!!!!
            !write(*,*) 'SUM(BIMA): (no rescaling) ',sum(bima),ipoint,isrc,isize
           
          else 
            ! C1 is NAN:
!            write(*,*) isrc, num_src, "C1 is NaN ** :", cluPSF%mySPSF(ipoint)%C1(1,isrc)
            bima(:,:) = 1.D0/dble(cluPSF%myCPSF%MAX_PSF_IMG_SIZE*cluPSF%myCPSF%MAX_PSF_IMG_SIZE)   
         endif
          call dfftw_execute(pplan)
          ffpsf0(:,:,ipoint,isrc) = ffpsf(:,:)
        enddo 
      enddo

    else
!      write(*,*) '***** get_calPSF'
      call get_calPSF(par,image,iima,ecut,x0,y0,bima,isize,ierr)

      ! Currently use identical PSFs for all sources in cluster:
      do ninst=1,num_exp
        do iband=1,num_image(ninst)
          ipoint=ip(ninst,iband) 
          do isrc=1,max_fit 
            ! bima0 obsolete, store psf in  ffpsf0
             !bima0(:,:,ipoint,isrc) =   bima(:,:)
            call dfftw_execute(pplan)
            ffpsf0(:,:,ipoint,isrc) = ffpsf(:,:)
          enddo
        enddo
      enddo
    endif

   
       
   
    havePSF=.true.

  endif !.not. havePSF 


  if (x0 .ne. x0old(ns) .or. y0 .ne. y0old(ns) .or. newcal) then

!     write(*,*) "shift:",ns,x0,y0,ix0,iy0,xremain,yremain
!    kcounter=kcounter+1


    call phase2d(isize,xremain,yremain,phase_arr,status)

    phase_arr0(:,:,ns) = phase_arr(:,:)
    x0old(ns) = x0
    y0old(ns) = y0


  else

    phase_arr(:,:) = phase_arr0(:,:,ns)
  endif



! calculate extent model
! extent model is centered in isize*isize subimage
! needs only be calculated, if parameter 'sig' changes 

  if (sig /= sigold(ns)) haveModel = .false.

  if (sig > sigmin .and. .not. haveModel) then
  !!    write(*,*) rtname, ' calc model: '
  

!    call cpu_time(t1)
!    do iloop=1,10000
    mima(:,:)=0.0
!    mima0(:,:,ns)=0.0
    if (lowcase(par%extentmodel) == "gaussian") then
     sig4sq=max(sig*sig*16.D0,4.D0)
      do j=1,isize/2+1
         y=dble(j-isize/2-1)
        do i=1,j
          x=dble(i-isize/2-1)
          if ((x*x+y*y) .lt. sig4sq) then
          if (i/= j) then     
            modval = fgen(0.D0,x-0.5D0,x+0.5D0,sig)* &
                     fgen(0.D0,y-0.5D0,y+0.5D0,sig)
          else
            modval = fgen(0.D0,x-0.5D0,x+0.5D0,sig)
            modval = modval*modval
          endif
            mima(i,j)=modval
            if (j /= 1) mima(i,isize-j+2)=modval
            if (j /= 1 .and. i /= 1)  mima(isize-i+2,isize-j+2)=modval 
            if (i /= 1) then 
              mima(i,isize-j+2)=modval
              mima(isize-i+2,j)=modval
            endif
            if (i /= j .and. i /= 1 .and. j /= 1) then
              mima(j,i)=modval
              if (i /= 1) then
                mima(j,isize-i+2)=modval
                mima(isize-j+2,isize-i+2)=modval
              endif
              if (j /= 1) mima(isize-j+2,i)=modval
            endif
            endif
        end do
      end do    

    else
      ! beta model


      sig4sq=max(sig*sig*100.D0,5.D0)
      do j=1,isize/2+1
        y=dble(j-isize/2-1)
        do i=1,j
          x=dble(i-isize/2-1)
          if ((x*x+y*y) .lt. sig4sq) then
            modval=fking(x,y,sig)
            mima(i,j)=modval
            if (j /= 1) mima(i,isize-j+2)=modval
            if (j /= 1 .and. i /= 1)  mima(isize-i+2,isize-j+2)=modval 
            if (i /= 1) then 
              mima(i,isize-j+2)=modval
              mima(isize-i+2,j)=modval
            endif
            if (i /= j .and. i /= 1 .and. j /= 1) then
              mima(j,i)=modval
              if (i /= 1) then
                mima(j,isize-i+2)=modval
                mima(isize-j+2,isize-i+2)=modval
              endif
              if (j /= 1) mima(isize-j+2,i)=modval
            endif
          endif  
        end do
      end do 
      norm=sum(mima(:,:))
     if (norm .gt. 1.0) mima(:,:)=mima(:,:)/norm

    endif

    do j=1,isize 
      do i=1,isize
        mima0(xsort(i,j),ysort(i,j),ns)=mima(i,j)
      enddo
    enddo    


    mima(:,:)=mima0(:,:,ns)

    call dfftw_execute(mplan)
    ffmodel0(:,:,ns) = ffmodel

    haveModel=.true.
  endif  

  ! test additional initialization:
  fima(:,:)=0.0

  if (sig .gt. sigmin) then


    ffmodel= ffmodel0(:,:,ns)


    ffpsf = ffpsf0(:,:,iima,ns)

    
    prodima = ffpsf*ffmodel*fftscale*phase_arr0(:,:,ns)

    call dfftw_execute(fplan)

   ! write(*,*) 'bima=',minval(bima),maxval(bima),sum(bima),x0,y0
   ! write(*,*) 'fima=',minval(fima),maxval(fima),sum(fima),sig
 
  else  
     
      prodima = ffpsf0(:,:,iima,ns)*phase_arr0(:,:,ns)*fftscale

      call dfftw_execute(fplan)
 
      ! CPU-time ~ 4E-6 s
!     write(*,*)  "all real ffpsf0:", sum(real(ffpsf0(:,:,1,ns))), sum(real(ffpsf0(:,:,2,ns))),sum(real(ffpsf0(:,:,3,ns)))
!     write(*,*)  "all imag ffpsf0:", sum(imag(ffpsf0(:,:,1,ns))), sum(imag(ffpsf0(:,:,2,ns))),sum(imag(ffpsf0(:,:,3,ns)))
!     
!     write(*,*)  "prodima:",sum(real(prodima)),sum(imag(prodima))
!
!     write(*,*)  "phase_arr0:",sum(real(phase_arr0(:,:,ns))),sum(imag(phase_arr0(:,:,ns)))
!     write(*,*)  "bima2:",sum(bima)
!     write(*,*)  "fima:",minval(fima),maxval(fima),sum(fima)
!
!     stop
  
  endif

 
  

  newcal = .false.
  first  = .false.

  sigold(ns)=sig

  end subroutine ftpsf

!------------------------------------------------------------------------------------------------------

!  subroutine phpsf(par,image,events,newcal,ievent,ns,ix0,iy0,x0,y0,sig,ecut,cluPSF,fima,isize,ierr)
  subroutine phpsf(par,events,newcal,ievent,ns,ix0,iy0,x0,y0,sig,ecut,cluPSF,fima,isize,ierr)
    
  !--------------------------------------------------------------------------------
  ! **** Version for photon mode **** 
  ! On first call or if newcal=.true., generate shapelet PSFs,
  ! otherwise use Fourier transformed PSF images saved in ffpsf0.
  ! If sig > 0 and sig has changed wrt previous call, calculate 
  ! extent model and fold with PSF.
  !---------------------------------------------------------------------------------


    type(ermldet_param_type)                    :: par
!    type(i4cube_type)                           :: image
    type(events_type),dimension(:)              :: events
    real(kind=real32)                           :: ecut
    type(FPSF)                                  :: cluPSF   
    logical(kind=log32)                         :: newcal
     integer(kind=int32), intent(in)            :: ix0,iy0   
    real(kind=real64), intent(in)               :: x0, y0, sig
    integer(kind=int32), intent(in)             :: ievent,ns
!    real(kind=real32), dimension(:,:), pointer  :: fima   
    real(kind=real64), dimension(:,:),allocatable :: fima
    integer(kind=int32)                         :: isize
    integer(kind=int32)                         :: ierr

    integer(kind=int32)                         :: status
!    type(cal_par_type)                          :: cal_par
!    type(psf2d_type)                            :: psf
!    character(len=clong)                        :: psf_file,psf_file_old
!    integer(kind=int32)                         :: ext,ipoint,nsold=-1
    integer(kind=int32)                         :: ipoint
    
!    real(kind=real32)                           :: energy
    integer(kind=int32)                                     :: num_src
    real(kind=real64)                                       :: modval
!    real(kind=real64), dimension(max_fit)                   :: ra,dec
!    real(kind=real64)                                       :: xcen, ycen, xremain, yremain
    real(kind=real64)                                       :: xremain, yremain    

    
    integer(kind=int32), dimension(:,:), allocatable       :: xsort,ysort
!    real(kind=real32), dimension(:,:,:), allocatable        :: sima
    real(kind=real64), dimension(:,:,:,:), allocatable      :: bima0
    real(kind=real64), dimension(:,:),  allocatable         :: bima,mima
    real(kind=real64), dimension(:,:,:),  allocatable       :: mima0
!    real(kind=real32), dimension(:,:), allocatable         :: pima
    complex(kind=real64), dimension(:,:),  allocatable           :: ffmodel,ffpsf,prodima,phase_arr
    complex(kind=real64), dimension(:,:,:),  allocatable         :: phase_arr0,ffmodel0

    ! for photon mode change to 3 dims:
    complex(kind=real64), dimension(:,:,:),  allocatable         :: ffpsf0

    !test output
!    type(r4image_type)                       :: psfout
!    character(len=80)                        :: psfname
!    integer(kind=int32)                      :: ipsf,icoeff,ncoeff
!    character(len=8)                         :: istr8



    integer(kind=int64)                                     :: pplan,mplan,fplan
    logical(kind=log32)                                     :: havePlan=.false.


!    real(kind=real32), dimension(max_ima) :: psfcenx,psfceny,pfac

!    real(kind=real64) :: x,y,x1,y1
    real(kind=real64) :: x,y
    !    real(kind=real64) :: dx,dy,ddx,ddy,xd0,yd0,xx,yy
    real(kind=real64) , dimension(1:max_fit) :: x0old,y0old,sigold
!    real(kind=real32) :: ssum,xyoff2,ecutold,fluxfac,fftscale,sig4sq
    real(kind=real64) :: fftscale,sig4sq

    
    real(kind=real64) :: sigmin=0.1,norm

!    integer(kind=int32)             :: i,j,ix,iy,ixx,iyy,lx,ly,kx,ky,kk,np,np2,isig3,isig3sq,isrc
    integer(kind=int32)             :: i,j,isrc
    
!    integer(kind=int32)             :: isx,isy,isizeold=-1,n,iecut,ixmin,ixmax,iymin,iymax,ninst
    integer(kind=int32)             :: isizeold=-1,iecut
    
!    integer(kind=int32), dimension(max_ima) :: nx, ny
!    real(kind=real32)                       :: offax

!    real(kind=real64)               :: pxmin,pxmax,pymin,pymax,dx1,dx2,dy1,dy2
!    integer(kind=int32)             :: pxmin0,pymin0,pxmax0,pymax0,pxmin1,pymin1,hsize
    integer(kind=int32)             :: hsize
!    real(kind=real64)               :: img_pixel_mm
!    integer(kind=int32) :: ixcen,iycen,imaxrad,ipsfrad
!    real(kind=real32)   :: E, theta, phi, val
    logical(kind=log32) :: first = .true., imallocated=.false.
    logical(kind=log32) :: havePSF, haveModel 
!    logical(kind=log32) :: survmode
!    real(kind=real32)   :: t1,t2
!    integer(kind=int32) :: clock1,clock2,clock3,iloop
    integer(kind=int32),dimension(:,:),allocatable  :: evmap
    integer(kind=int32)                             :: nev
    logical(kind=log32)                             :: event_flag
!    integer(kind=int32)                             :: imcounter,imcounter2
!    character(len=5)                                :: imstr

    character(len=*), parameter  :: rtname='FTPSF '
    
! from FFTW include file:
      INTEGER, PARAMETER :: FFTW_MEASURE=0
      INTEGER, PARAMETER :: FFTW_EXHAUSTIVE=8
      INTEGER, PARAMETER :: FFTW_PATIENT=32
      INTEGER, PARAMETER :: FFTW_ESTIMATE=64
      INTEGER, PARAMETER :: FFTW_ESTIMATE_PATIENT=128

    save :: mima,mima0,ffmodel,ffpsf,ffpsf0,prodima,xsort,ysort,phase_arr,phase_arr0,ffmodel0
    save :: bima0,bima,imallocated,first,havePSF,haveModel
    save :: sigold,x0old,y0old,isizeold,evmap,fftscale
    save :: pplan,mplan,fplan

    ! index for test output images
    !if (first) imcounter=0
    !if (first) imcounter2=0


!    xremain=x0-dble(nint(x0))
!    yremain=y0-dble(nint(y0))
    xremain=x0-dble(ix0)
    yremain=y0-dble(iy0)


    if (first .or. newcal) then
      if (allocated(evmap)) deallocate(evmap)
      nev=size(events)  
      allocate(evmap(nim,nev))  ! actually only max. number of events per image needed, change later!
      evmap(:,:)=0
      do i=1,nev
        evmap(events(i)%ipt,events(i)%id)=i
      enddo  
    endif

!    count_psf=count_psf+1

    if (first) then
      sigold(:)=-1.0
      y0old(:)=-1.E9
      x0old(:)=-1E9
      havePSF=.false.
      isizeold=0
!      write(*,*) 'FFTPSF legacy fortran'
     end if 
     
!    determine subimage array sizes:  

   ! isize=nint(2.*ecut)+2
     !new (07/12/2017):
     !isize=2*nint(ecut)+2
     ! test large PSF (GL 15/04/2021):
     !isize=max(2*nint(ecut)+2,64)
     !isize=2*nint(ecut)+2
     ! isize is defined in subroutine model: 
     !isize = cluPSF%myCPSF%MAX_PSF_IMG_SIZE
    isize = cluPSF%myCPSF%MAX_PSF_IMG_SIZE
    hsize = isize/2
    iecut=int(ecut)+1

    if (first .or. newcal) then
      if (allocated(ffpsf0)) deallocate(ffpsf0)
      allocate(ffpsf0(1:hsize+1,1:isize,nev))
      ffpsf0(:,:,:)=cmplx(0.0,0.0,kind=real64)
    endif    


    if (isizeold /= isize) then
      havePSF=.false.
      haveModel=.false.
      fftscale=1/float((isize*isize))
      if (.not. first) then
        if (imallocated) then
          deallocate(mima)
          deallocate(mima0)
          deallocate(ffmodel)
          deallocate(ffpsf)
          deallocate(prodima)
          deallocate(xsort)
          deallocate(ysort) 
          deallocate(phase_arr)
          deallocate(phase_arr0)
          deallocate(ffmodel0)
          imallocated = .false.
        endif
!        write(*,*) ' deallocate bima0'
        deallocate(bima0)
        deallocate(bima)
        deallocate(fima)
      endif
      if (.not. imallocated) then     
        allocate(mima0(1:isize,1:isize,max_fit) )
        allocate(mima(1:isize,1:isize))
        allocate(ffmodel(1:hsize+1,1:isize) )
        allocate(ffpsf(1:hsize+1,1:isize) )
        allocate(prodima(1:hsize+1,1:isize) )
        allocate(xsort(1:isize,1:isize) )
        allocate(ysort(1:isize,1:isize) )
        allocate(phase_arr(hsize+1,isize))
        allocate(phase_arr0(hsize+1,isize,max_fit))
        allocate(ffmodel0(hsize+1,isize,max_fit))
        imallocated = .true.

       ! initialize:
       mima0(:,:,:)       = 0.0D0
       mima(:,:)          = 0.0D0
       phase_arr(:,:)     = cmplx(0.D0,0.D0,kind=real64)
       phase_arr0(:,:,:)  = cmplx(0.D0,0.D0,kind=real64)
       ffmodel(:,:)       = cmplx(0.D0,0.D0,kind=real64)
       ffmodel0(:,:,:)    = cmplx(0.D0,0.D0,kind=real64)
       ffpsf(:,:)         = cmplx(0.D0,0.D0,kind=real64)
       prodima(:,:)       = cmplx(0.D0,0.D0,kind=real64)
       xsort(:,:)         = 0
       ysort(:,:)         = 0


      endif
!      write(*,*) ' allocate bima0 ',isize,  nim, max_fit
      allocate( bima(1:isize,1:isize) )
      allocate( bima0(1:isize,1:isize,nim,max_fit) )
      allocate( fima(1:isize,1:isize) )
      fima(:,:) =  0.0D0
      bima(:,:)      =  0.0D0
      bima0(:,:,:,:) =  0.0D0

!      create sorting arrays to rearrange quadrants of kernel image for FFT
!
!    1|2         4|3
!    ---         ---
!    3|4    =>   2|1 
!
       do i=1,isize
        do j=1,isize
          if (i .le. isize/2) then
            xsort(i,j) = isize/2+i
          else
            xsort(i,j) = i-isize/2
          endif
          if (j .le. isize/2) then
            ysort(i,j) = isize/2+j
          else
            ysort(i,j) = j-isize/2
          endif             
        enddo
      enddo 
      isizeold=isize 

      ! create FFTW plans:
      !if (par%fitext_flag) then
        if (havePlan) then
          call dfftw_destroy_plan(pplan)
          call dfftw_destroy_plan(mplan)
          call dfftw_destroy_plan(fplan)
        endif
        call dfftw_plan_dft_r2c_2d(pplan,isize,isize,bima,ffpsf,FFTW_MEASURE)
        call dfftw_plan_dft_r2c_2d(mplan,isize,isize,mima,ffmodel,FFTW_MEASURE)
        call dfftw_plan_dft_c2r_2d(fplan,isize,isize,prodima,fima,FFTW_MEASURE)
      !endif
    endif ! isize /= isizeold

    if (newcal) then 
      havePSF=.false.
    endif

  !! TESTOUTPUT PSF:
 !! allocate(psfout%arr(isize,isize))

  if (.not. havePSF) then 
    ! create a PSF for each photon and store fft in ffpsf0:
!    imcounter=1
    do ipoint=1,nim
      num_src=size(cluPSF%evSPSF(ipoint)%C1,2)
      ! here isrc is actually the ID of the photon in evlist(ipoint):
      do isrc=1,num_src

         event_flag=.true.
!          call cpu_time(t1) 
           call get_shapePSF(event_flag,cluPSF,ipoint,isrc,bima,ierr)
!          call cpu_time(t2)
            !!!!!!!!!  TEST 
            bima=max(1.0E-7,bima)
            !if (maxval(bima) > 1.0) write(*,*) 'max(bima),ph=',maxval(bima)
         ! TEST: extra smoothing(GL 03/06/2019) :
         if (par%sixte_flag) call sixte_smoothing(bima,status) 

         
        call dfftw_execute(pplan) ! transform bima -> ffpsf
        
        ffpsf0(:,:,evmap(ipoint,isrc)) = ffpsf(:,:)
      enddo
    enddo
   
    havePSF=.true.

  endif !.not. havePSF 
  
 


  if (x0 .ne. x0old(ns) .or. y0 .ne. y0old(ns) .or. newcal) then
     
!    write(*,*) "shift:",ns, x0old(ns),y0old(ns),x0,y0,ix0,iy0,xremain,yremain
    call phase2d(isize,xremain,yremain,phase_arr,status)
    phase_arr0(:,:,ns) = phase_arr(:,:)
    x0old(ns) = x0
    y0old(ns) = y0
  else
    phase_arr(:,:) = phase_arr0(:,:,ns)
  endif

! calculate extent model
! extent model is centered in isize*isize subimage
! needs only be calculated, if parameter 'sig' changes 

  if (sig /= sigold(ns)) haveModel = .false.

  if (sig > sigmin .and. .not. haveModel) then
  
    mcounter=mcounter+1
    mima(:,:)=0.0
    if (lowcase(par%extentmodel) == "gaussian") then
     sig4sq=max(sig*sig*16.D0,4.D0)
      do j=1,isize/2+1
         y=dble(j-isize/2-1)
        do i=1,j
          x=dble(i-isize/2-1)
          if ((x*x+y*y) .lt. sig4sq) then
          if (i/= j) then     
            modval = fgen(0.D0,x-0.5D0,x+0.5D0,sig)* &
                     fgen(0.D0,y-0.5D0,y+0.5D0,sig)
          else
            modval = fgen(0.D0,x-0.5D0,x+0.5D0,sig)
            modval = modval*modval
          endif
            mima(i,j)=modval
            if (j /= 1) mima(i,isize-j+2)=modval
            if (j /= 1 .and. i /= 1)  mima(isize-i+2,isize-j+2)=modval 
            if (i /= 1) then 
              mima(i,isize-j+2)=modval
              mima(isize-i+2,j)=modval
            endif
            if (i /= j .and. i /= 1 .and. j /= 1) then
              mima(j,i)=modval
              if (i /= 1) then
                mima(j,isize-i+2)=modval
                mima(isize-j+2,isize-i+2)=modval
              endif
              if (j /= 1) mima(isize-j+2,i)=modval
            endif
            endif
        end do
      end do    
    else
      ! beta model
      sig4sq=max(sig*sig*100.D0,5.D0)
      do j=1,isize/2+1
        y=dble(j-isize/2-1)
        do i=1,j
          x=dble(i-isize/2-1)
          if ((x*x+y*y) .lt. sig4sq) then
            modval=fking(x,y,sig)
            mima(i,j)=modval
            if (j /= 1) mima(i,isize-j+2)=modval
            if (j /= 1 .and. i /= 1)  mima(isize-i+2,isize-j+2)=modval 
            if (i /= 1) then 
              mima(i,isize-j+2)=modval
              mima(isize-i+2,j)=modval
            endif
            if (i /= j .and. i /= 1 .and. j /= 1) then
              mima(j,i)=modval
              if (i /= 1) then
                mima(j,isize-i+2)=modval
                mima(isize-j+2,isize-i+2)=modval
              endif
              if (j /= 1) mima(isize-j+2,i)=modval
            endif
          endif  
        end do
      end do 
      norm=sum(mima(:,:))
     if (norm .gt. 1.0) mima(:,:)=mima(:,:)/norm
    endif

    do j=1,isize 
      do i=1,isize
        mima0(xsort(i,j),ysort(i,j),ns)=mima(i,j)
      enddo
    enddo    

    mima(:,:)=mima0(:,:,ns)
    call dfftw_execute(mplan)
    ffmodel0(:,:,ns) = ffmodel
    haveModel=.true.
  endif  

  if (sig .gt. sigmin) then
    ffmodel= ffmodel0(:,:,ns)
    ffpsf = ffpsf0(:,:,ievent)
    prodima = ffpsf*ffmodel*fftscale*phase_arr0(:,:,ns)
    !    write(*,*) 'fima: fold with extent model'
    call dfftw_execute(fplan) ! output -> fima
!    if (maxval(fima) > 1.0) write(*,*) 'max(fima)=',maxval(fima)
   ! write(*,*) 'ph bima=',minval(bima),maxval(bima),sum(bima),x0,y0
   ! write(*,*) 'ph fima=',minval(fima),maxval(fima),sum(fima),sig
  else  
     !write(*,*) 'fima: shift position'
     prodima = ffpsf0(:,:,ievent)*phase_arr0(:,:,ns)*fftscale
    call dfftw_execute(fplan) ! output -> fima
  endif
  !------------------------------------------
  !imcounter=imcounter+1
  !if (imcounter < 1000) then
  ! write(*,*) 'first, imcounter:',first, imcounter
  !  write(imstr,'(I5.5)') imcounter
  !  write(*,*) 'file: psf_images/fima'//imstr//'.dat'
  !  open(33,file='psf_images/fima'//imstr//'.dat',status='UNKNOWN')
  !     do i=1,isize
  !       write(33,'(32F9.5)') fima(1:isize,i)
  !     enddo 
  !  close(33)
  !endif
  !------------------------------------------
  newcal = .false.
  first  = .false.


  sigold(ns)=sig

!  write(*,*), "phpsf sum(fima): ",sum(fima)

  end subroutine phpsf

!---------------------------------------------------------------------------------------




!---------------------------------------------------------------------------------------

  real(kind=real64) function fgen(x0,x1,x2,sig)

    real(kind=real64) :: x0,x1,x2,sig
    real(kind=real64) :: fsig,eps=0.00001D0,sqrt2=1.414213562373D0
                                                                       
    fgen=0.
    fsig=sig                                                           
    if (fsig.le.eps) fsig=eps
    fsig = 1.0/(sqrt2*fsig)
    fgen = (ferrf((x2-x0)*fsig)-ferrf((x1-x0)*fsig))/2.0D0

  end function fgen     

!---------------------------------------------------------------------------------------

  real(kind=real64) function ferrf(xx)
 
    real(kind=real64) xx
    real(kind=real64) t,x

    x=abs(xx)                                                        
    ferrf=1.0D0
                                                         
    if (x<=5.0D0) then                                            
       t=1.0D0/(1.0D0+0.47047*x)                                            
       ferrf=1.0D0-(0.3480242D0-(0.0958798D0-0.7478556D0*t)*t)*t*exp(-(x*x))
     end if   
                                                                    
     if (xx<0.0D0) ferrf=-ferrf                                        

   end function ferrf  

!---------------------------------------------------------------------------------------

  real(kind=real64) function fking(x,y,rcore)
  !-----------------------------------------------------------------
  !  returns value of the King profile (1+r**2/rcore**2)**(-1.5), 
  !  averaged over the pixel with center (x,y), normalized to 1  
  !  the averageing is approximated by the radial integral between 
  !  r1 and r2, r1=r_pix-0.5, r2=r_pix+0.5
  !------------------------------------------------------------------
  real(kind=real64) :: x,y,rcore
  real(kind=real64) :: r1,r2,r1sq,r2sq,rc2,r
  real(kind=real64) :: norm,twopi=6.283185307180

    rc2=max(1.D-2,rcore*rcore)
  
    r=(x*x+y*y)**0.5
    r1=max(0.D0,r-0.5D0)
    r2=r+0.5    
    r1sq=r1*r1
    r2sq=r2*r2 

!-----------------------------------------------------
!     kingint=twopi*(rc2/(1+r1sq/rc2)-rc2/(1+r2sq/rc2))
!     area=pi*(r2sq-r1sq)
!     norm=twopi*rc2
!     fking=kingint/area/norm 
!----------------------------------------------------
!     Correct integration of king model (26/06/05 GL):

    norm=twopi*rc2
    fking=(r2/sqrt(1.+r2sq/rc2)-(r1/sqrt(1.+r1sq/rc2)))/(r2-r1)/norm
      
 
   end function fking

!---------------------------------------------------------------------------------------


subroutine phase2d(isize,xshift,yshift,phase_arr,ierr)

      integer(kind=real32)                      :: isize
      real(kind=real64)                         :: xshift,yshift
      complex(kind=real64), dimension(:,:)           :: phase_arr
      integer(kind=int32)                      :: ierr

      integer(kind=int32)                          :: hsize,isize2,i
      real(kind=real64) , dimension(:),  allocatable     ::frequency,arg, arg_u, arg_v
      complex(kind=real64), dimension(:,:),  allocatable :: uphase, vphase  
      real(kind=real64), parameter                       :: pi=3.1415927410D0

      character(len=*), parameter  :: rtname='PHASE2D'

      ierr=0

      if ( mod(isize,2) .ne. 0 ) then 
        ierr=3
        call wrerr(rtname,ierr,'isize must be even number')
      else

        hsize=isize/2+1

        allocate(frequency(isize))
        allocate(arg(isize))
        allocate(arg_u(isize))
        allocate(arg_v(isize)) 
        allocate(uphase(hsize,1)) 
        allocate(vphase(1,isize))


        
        isize2=isize/2
        do i=1,isize2+1
          frequency(i)=real(i-1)
        enddo
        do i=2,isize2
          frequency(i+isize2)=i-isize2-1
        enddo      


        arg=frequency*2.0*pi/real(isize)
        arg_u = arg * xshift
        arg_v = arg * yshift
        uphase(:,1) = cmplx(cos(arg_u(1:hsize)),-sin(arg_u(1:hsize)),kind=real64)
        vphase(1,:) = cmplx(cos(arg_v),-sin(arg_v),kind=real64)

        phase_arr =  matmul(uphase,vphase)

        deallocate(frequency)
        deallocate(arg)
        deallocate(arg_u)
        deallocate(arg_v) 
        deallocate(uphase) 
        deallocate(vphase)

      endif


end subroutine phase2d

!--------------------------------------------------------------------------  

  subroutine get_calPSF(par,image,iima,ecut,x0,y0,bima,isize,ierr)

    type(ermldet_param_type)                                :: par
    type(i4cube_type)                                       :: image
    integer(kind=int32)                                     :: iima
    real(kind=real32)                                       :: ecut
    real(kind=real64), intent(in)                           :: x0, y0
    real(kind=real64), dimension(:,:)                       :: bima
    integer(kind=int32)                                     :: isize
    integer(kind=int32)                                     :: ierr

    real(kind=real64), dimension(:,:,:), allocatable        :: sima
    integer(kind=int32)                         :: status
    integer(kind=int32)                         :: ninst,iband,ipoint,ext
    integer(kind=int32)                         :: i,j,hsize
    real(kind=real32)                           :: energy,offax,phi
    real(kind=real32),parameter                 :: offax_max=29.9    
    real(kind=real64)                           :: xcen, ycen
    real(kind=real32)                           :: deltax,deltay
    real(kind=real64)                           :: ssum
    integer(kind=int32)                         :: ixcen,iycen,imaxrad,ipsfrad
    integer(kind=int32), dimension(max_ima)     :: nx, ny
    real(kind=real64)  , dimension(max_ima)     :: psfcenx,psfceny,pfac
    type(cal_par_type)                          :: cal_par
    type(psf2d_type)                            :: psf
    character(len=clong)                        :: psf_file
    real(kind=real64)                           :: xx,yy
    integer(kind=int32)                         :: pxmin0,pymin0,pxmax0,pymax0,pxmin1,pymin1
    real(kind=real64)                           :: pxmin,pxmax,pymin,pymax,dx1,dx2,dy1,dy2
    integer(kind=int32)                         :: iecut,ixmin,ixmax,iymin,iymax,np
    logical(kind=log32) :: survmode

    character(len=*), parameter  :: rtname='get_calPSF'

    ierr=0
    status=0

    iecut=int(ecut)+1
    hsize=isize/2

    
    if (image%stdhdr(iima)%obs_mode == 'SURVEY' .or. image%stdhdr(iima)%obs_mode =='survey') then
      survmode = .true.
    else if  (image%stdhdr(iima)%obs_mode == 'POINTING' .or. image%stdhdr(iima)%obs_mode =='pointing') then
      survmode = .false. 
    else 
      survmode = .false. 
      ierr=3
      call wrerr(rtname,ierr,'Invalid observing mode')
    endif


    do ninst=1,num_exp
       do iband=1,num_image(ninst)
        ipoint=ip(ninst,iband) 

        energy=0.5*(par%emin(ipoint)+par%emax(ipoint))
        call cal_setpar(image%stdhdr(ipoint),cal_par,status) 
        if (survmode) then
          phi=0.0
          cal_par%codenam='2D_PSF'
          cal_par%expression=''
          call cal_r4toex('ENERG',energy,cal_par%expression,status)
          call cal_stringtoex('OBSMODE','SURVEY',cal_par%expression,status)
          call cal_filename(cal_par,psf_file,ext,status)
!          write(*,*) "read PSF:",psf_file,ext
          
          call get_2dpsf(psf_file,ext,phi,psf,status)          
       else ! pointed mode

          call equ2pix(image%wcshdr(ipoint),image%stdhdr(ipoint)%ra_pnt,&
               image%stdhdr(ipoint)%dec_pnt,xcen,ycen,status)

          deltax=real(abs(x0-xcen)*image%wcshdr(ipoint)%cdelt1,kind=real32)
          deltay=real(abs(y0-ycen)*image%wcshdr(ipoint)%cdelt2,kind=real32)
          if (deltax == 0.0 .and. deltay == 0.0) then
            phi = 0.0
          else
            phi = atan2(deltax,deltay)
          endif
          ! off-axis angle in arcmin:
          !offax = (deltax*deltax+deltay+deltay)**0.5/60.0
          offax = min(offax_max, 60.*(deltax*deltax+deltay+deltay)**0.5)
          write(*,*) "fftps,offax=",offax
          cal_par%expression=''
          cal_par%codenam='2D_PSF'
          call cal_r4toex('ENERG',energy,cal_par%expression,status)
          call cal_r4toex('THETA',offax,cal_par%expression,status)
          call cal_stringtoex('OBSMODE','POINTING',cal_par%expression,status)

          call cal_filename(cal_par,psf_file,ext,status)

          call get_2dpsf(psf_file,ext,phi,psf,status)   
        endif
        ixcen=nint(psf%psfhdr%crpix1)
        iycen=nint(psf%psfhdr%crpix2)

        pfac(ipoint) = abs(image%wcshdr(1)%cdelt1 / psf%psfhdr%cdelt1)
        psfcenx(ipoint) = psf%psfhdr%crpix1 
        psfceny(ipoint) = psf%psfhdr%crpix2
        imaxrad=min(min((ixcen-1),iycen-1),min((psf%psfhdr%naxis1-ixcen),psf%psfhdr%naxis1-iycen))
        ipsfrad=min(nint((ecut+2.0)*pfac(ipoint)),imaxrad)

        

        nx(iband) = 2*ipsfrad+1
        ny(iband) = 2*ipsfrad+1

!!!     Changed allocation of sima (3/04/2014):
!!!     Note: assuming constant pfac(ipoint)       
        if (allocated(sima)) deallocate(sima)
        allocate(sima(0:(isize+2)*ceiling(pfac(ipoint)),0:(isize+2)*ceiling(pfac(ipoint)),nim))
        sima(:,:,:)=0.0
        

        sima(0:nx(iband)-1,0:ny(iband)-1,ipoint) = psf%arr(ixcen-ipsfrad:ixcen+ipsfrad,iycen-ipsfrad:iycen+ipsfrad)
        ! re-normalize PSF if necessary  
        if   (psf%psfhdr%e_sumrcts .and. psf%psfhdr%sumrcts > 0.0 ) then
          sima(:,:,ipoint) = sima(:,:,ipoint)/psf%psfhdr%sumrcts
        endif
        ! apply offset to center coordinates:
        psfcenx(ipoint) = psf%psfhdr%crpix1 - ixcen + ipsfrad
        psfceny(ipoint) = psf%psfhdr%crpix2 - iycen + ipsfrad


!------------------------------
!rebin PSF to image pixel size:
!------------------------------


        bima(:,:)=0.0D0
        np = max(0,nint((pfac(ipoint)-1.0)/2.0)) 

        xx = psfcenx(ipoint) - 0.5D0*pfac(ipoint)
        yy = psfcenx(ipoint) - 0.5D0*pfac(ipoint)
    



        iymin=max(1,hsize-iecut+1,hsize-int((ny(iband)/2-np)/pfac(ipoint))+1)
        iymax=min(isize,hsize+iecut+1,hsize+int((ny(iband)/2-np)/pfac(ipoint))+1)
        ixmin=max(1,hsize-iecut+1,hsize-int((nx(iband)/2-np)/pfac(ipoint))+1)
        ixmax=min(isize,hsize+iecut+1,hsize+int((nx(iband)/2-np)/pfac(ipoint))+1)

!        pcounter=pcounter+1
  

        do j=iymin,iymax
          pymin = yy + dble(j-hsize-1)*pfac(ipoint)  ! changed 09/08/2012 
          pymin0 = int(pymin)
          pymin1 = pymin0 + 1
          pymax = pymin + pfac(ipoint)
          pymax0 = int(pymax)
          !! cut to usable area
          !!!(TO DO)!!!
          !! coefficients used in the re-binning term
          dy1 = pymin1 - pymin
          dy2 = pymax - pymax0

        
            do i=ixmin,ixmax
              
              !! loop variables describing image pixel edges in PSF coordinates
!!              pxmin = xx + dble(i-hsize)*pfac(ipoint)                !! formerly: x
              pxmin = xx + dble(i-hsize-1)*pfac(ipoint)    ! changed 09/08/2012     
              pxmin0 = int(pxmin)
!              if (pxmin0 < 0) write(*,*) 'pxmin0: ',xx,psfcenx(ipoint),xremain,pfac(ipoint),x0,y0
              pxmin1 = pxmin0 + 1
              pxmax = pxmin + pfac(ipoint)
              pxmax0 = int(pxmax)
               !! cut to usable area
               !!!TO DO!!!
               !! coefficients used in the re-binning term
              dx1 = pxmin1 - pxmin
              dx2 = pxmax - pxmax0
               !! only, if complete pixels in x and / or y are covered:
              ssum=0.0
              if (pxmax0 > pxmin1 ) then
                !! edges: complete in x, fraction in y
                ssum = ssum +                                             &
                       dy1 * sum(sima(pxmin1:pxmax0-1, pymin0, ipoint)) + &
                       dy2 * sum(sima(pxmin1:pxmax0-1, pymax0, ipoint))
              endif
              if (pymax0 > pymin1 ) then
                !! edges: complete in y, fraction in x
                ssum = ssum +                                             &
                       dx1 * sum(sima(pxmin0, pymin1:pymax0-1, ipoint)) + &
                       dx2 * sum(sima(pxmax0, pymin1:pymax0-1, ipoint))
                if (pxmax0 > pxmin1 ) then
                  !! whole (PSF) pixels
                  ssum = ssum +    sum(sima(pxmin1:pxmax0-1, pymin1:pymax0-1, ipoint))                                    
                endif
              endif

               !! always: corners, subpixel contributions
               ! i+1, j+1:
               ! indices i,j are centered around 0; bima indices start with 1
      

               bima(i,j) =  ssum +                                        &
                    dx1 * dy1 * sima(pxmin0, pymin0, ipoint) +            &
                    dx1 * dy2 * sima(pxmin0, pymax0, ipoint) +            &
                    dx2 * dy1 * sima(pxmax0, pymin0, ipoint) +            &
                    dx2 * dy2 * sima(pxmax0, pymax0, ipoint)
       

            enddo
          enddo
       

       enddo
    enddo

    !close memory leak (GL 05/12/2016)
    deallocate(psf%arr)

  end subroutine get_calPSF


!--------------------------------------------------------------------------  

  subroutine sixte_smoothing(bima,ierr)

    real(kind=real64), dimension(:,:)               :: bima
    integer(kind=int32)                             :: ierr
    real(kind=real64), dimension(3,3)               :: sm_kernel
    real(kind=real64)                               :: pixval
    integer(kind=int32)                             :: i,j,k,l    
    integer(kind=int32),  dimension(2)               :: arrshape 
    real(kind=real64), dimension(:,:),allocatable   :: sm_ima
    character(len=*), parameter                     :: rtname='SIXTE_SMOOTHING'


    ierr=0
    !sm_kernel = reshape((/ 0.09,0.12,0.09,0.12,0.16,0.12,0.09,0.12,0.09 /),shape(sm_kernel))
    !test:
    sm_kernel = reshape((/ 0.11,0.11,0.11,0.11,0.12,0.11,0.11,0.11,0.11 /),shape(sm_kernel))
    
    arrshape=shape(bima)
    allocate(sm_ima(arrshape(1),arrshape(2)))
    
    sm_ima=bima
    do j=2,arrshape(2)-1 
     do i=2,arrshape(1)-1
       pixval=0 
       do l=1,3
         do k=1,3
           pixval=pixval+sm_kernel(k,l)*bima(i+k-2,j+l-2)
         enddo
       enddo 
       sm_ima(i,j)=pixval
     enddo
    enddo
   
    bima=sm_ima

    deallocate(sm_ima)
 
  end subroutine sixte_smoothing
 


!--------------------------------------------------------------------------  

  subroutine get_shapePSF(event_flag,cluPSF,iima,ns,bima,ierr)

    logical                                     :: event_flag
    type(FPSF)                                  :: cluPSF   
    integer(kind=int32)                         :: iima
    integer(kind=int32)                         :: ns
    real(kind=real64),dimension(:,:)            :: bima
    integer(kind=int32)                         :: ierr

    real(kind=real64)                           :: sum_bima
    integer(kind=int32)                         :: isize 
    integer(kind=int32)                         :: i,n_coeff0,n_coeff1,n_coeff2 
!    type(r4image_type)                                      :: testout
!    integer(kind=int32)                         :: status   
 
!    character(len=20)                          :: testphi
!    character(len=3)                           :: istr
    character(len=*), parameter  :: rtname='GET_SHAPEPSF'

    ierr=0



    
    isize = cluPSF%myCPSF%MAX_PSF_IMG_SIZE
    n_coeff0=size(cluPSF%myCPSF%NX0)
    n_coeff1=size(cluPSF%myCPSF%NX1)
    n_coeff2=size(cluPSF%myCPSF%NX2)


    bima(:,:)=0.0D0
    if (event_flag) then


      do i = 1, n_coeff0
        bima(:,:) = bima(:,:) +  cluPSF%evSPSF(iima)%C0(i,ns) * cluPSF%myCPSF%IMG_SHAPE0(:,:, i) 
      enddo
      do i = 1, n_coeff1
        bima(:,:) = bima(:,:) +  cluPSF%evSPSF(iima)%C1(i,ns) * cluPSF%myCPSF%IMG_SHAPE1(:,:, i) 
      enddo
      do i = 1, n_coeff2
        bima(:,:) = bima(:,:) +  cluPSF%evSPSF(iima)%C2(i,ns) * cluPSF%myCPSF%IMG_SHAPE2(:,:, i)
     enddo
     sum_bima=sum(bima(:,:))
     if (sum_bima > 1.0) bima=bima/sum_bima
    else
      ! slib 1.8:

      call slib_run_write2arrays_fast(cluPSF%mySPSF(iima),cluPSF%myCPSF, ns)
       
      
      if (size(bima,1) == isize .and. size(bima,2) == isize) then
         bima(:,:) = cluPSF%myCPSF%img
         sum_bima=sum(bima(:,:))
         if (sum_bima > 1.0D0) bima=bima/sum_bima
       else
        ierr=3
        call wrerr(rtname,ierr,'PSF image size mismatch')

      endif

    endif
    
   

   

  end subroutine get_shapePSF


!--------------------------------------------------------------------------  


!--------------------------------------------------------------------------  



end module fftpsf
