! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.

module mulfit


  use param
!  use numkind
  use fitsimage  
  use detect_io
  use cal
  use psflib
  use fftpsf
!   use evtool_eventfile_mod
!    use evtool_events_mod
  use erolib_eventfile_mod
  use erolib_events_mod
  use slib_constants
  use slib_engine
  use slib_basic_calcs
  use slib_data_evt
  use slib_data_psf
  use slib_calc_space

  
!  use slib_constants
!  use slib_engine


  implicit none

  logical(kind=log32)  :: verbose_flag


contains


!-------------------------------------------------------------------------------------------------------------

  subroutine mulsou(par,image,eventlist,pixlist,expimage,bkgimage,detmask,clulist,cluPSF,events,mask,first,nsrc,modelima,ierr)

   !--------------------------------------------------------------------------
   ! Fit cluster of sources contained in clulist
   ! 
   !--------------------------------------------------------------------------  

   type(ermldet_param_type), intent(in)     :: par
   integer(kind=int32)                      :: nsrc!,nev                         
!   type(srcrow_type),dimension(:,:),pointer :: clulist
   type(srcrow_type),dimension(:,:),allocatable :: clulist
   type(FPSF)                               :: cluPSF   
   type(i4cube_type)                        :: image
   type(tEventFile),dimension(:)            :: eventlist
   type(pixlist_type),dimension(:)          :: pixlist
   type(r4cube_type)                        :: expimage,bkgimage
   type(l1cube_type)                        :: detmask
   type(events_type),pointer,dimension(:)   :: events
   type(mask_type)                          :: mask 
   real(kind=real64),dimension(:,:,:),pointer :: modelima
   logical(kind=log32)                      :: first    
 
   integer(kind=int32)                      :: ierr 

   integer(kind=int32)                      :: ntermf0,ntfix0,ntermf
   integer(kind=int32)                      :: ntfix,nffix,np, npsou,nzero
   integer(kind=int32)                      :: ipoint,iimage,nima,nn
   integer(kind=int32)                      :: nband,nspar,ddof
!   integer(kind=int32)                      :: ix,iy,x1,y1
   integer(kind=int32)                      :: ioff0,ioff1,ioff,noff,bestfit!,ifel
   real(kind=real64)                        :: expj,sc!,felhi,fello,exp,expn,ecut2
   real(kind=real64)                        :: flamda
   real(kind=real32),dimension(max_fit)     :: dml(max_fit)
!   real(kind=real64),dimension(max_par)     :: fe
   real(kind=real64)                        :: yfit0
   real(kind=real64),dimension(:),allocatable  :: yfit1
   real(kind=real64),dimension(max_ima)        :: yfit2 
!   real(kind=real32)                        :: yfit3
!   real(kind=real64),dimension(nsrc)        :: xms,yms
   real(kind=real64),dimension(max_par)     :: a,deltaa,deltae,sigmaa,atemp!,amerk
   real(kind=real64),dimension(max_par)     :: afit!,sigfit,delfit
   real(kind=real64),dimension(max_par,0:nsrc) ::   asave
   real(kind=real64)                        :: like00,like0,like1,likold,likediff
   
   real(kind=real32)                        :: dmlthresh , likecorr 
   real(kind=real64),dimension(max_ima)     :: lik_arr,like00arr
   real(kind=real64),dimension(max_ima,0:nsrc) :: lik_arr_save
   real(kind=real64),dimension(0:nsrc)      :: likesave, like2dof
   integer(kind=int32),dimension(0:nsrc)    :: fit_dof   

   integer(kind=int32),dimension(max_par)   :: iordr,iordr0 
   integer(kind=int32)                      :: iter,i,j,l,m,n,status
   logical(kind=log32)                      :: multflag, maskbad, newcal!done , 
   real(kind=real64)                        :: distsq,extdiff 
   real(kind=real64),parameter              :: flamda0=1.D-2
!   real(kind=real64),parameter              :: delt = 1.D-6
   real(kind=real64),parameter              :: delt = 1.D-4   
   real(kind=real64),parameter              :: mindist=1.5, minextdiff=1.0
   integer(kind=int32) , parameter          :: irad = 120, max_it=100

   !test:
!   integer(kind=int32)                      :: iirad,iiflux
!   real(kind=real64)                        :: steprad(max_fit),stepx(max_fit),stepy(max_fit),stepflux(max_ima,max_fit),steplike,minlike
!   real(kind=real64)                        :: t1,t2
   
 
  character(len=*), parameter              :: rtname='MULSOU'
 

   

   ierr=0          ! error status

   ntfix = 0       ! number of fixed parameters
   nspar = 3       ! fit paramaters per source: x,y,extent 
   npsou=nim+nspar ! + number of flux parameters per source 
   np=npsou*nsrc   ! total number of fit parameters
   likold=0.0


   ! read new CAL PSF for new source:
   newcal =.true.

   if (allocated(yfit1)) then
       deallocate(yfit1)
   endif
   allocate(yfit1(mask%nev))


   !********************************
   verbose_flag2=.false.
   verbose_flag3=.false.
   verbose_flag4=.false.

   ! select all:
   !verbose_flag=.true.


   ! fill parameter arrays:
   do i=1,nsrc
     ioff=npsou*(i-1)
     sc=0.0D0
     do j=1,nim
       expj=expimage%arr(ixm(nint(clulist(0,i)%x_ima),j),iym(nint(clulist(0,i)%y_ima),j),j) 
       expj=max(expj,1.0D0)
       if (clulist(j,i)%rate <= 0.) then 
           clulist(j,i)%rate = real(10.D0/expj,kind=real32)
       else if (.not. clulist(j,i)%rate > 0.) then
         !   catch NULL values from input list           
         clulist(j,i)%rate = 0.0
       endif
         a(ioff+3+j)=dble(clulist(j,i)%rate)
         deltaa(ioff+3+j)=max(sqrt(clulist(j,i)%rate*expj)/expj,1.D-5)
         deltae(ioff+3+j)=deltaa(ioff+3+j)
         sc=sc+a(ioff+3+j)*expj
     end do
     ! assume 10 counts, if all input values were 0 or NULL:  
     sc=max(sc,10.D0)
     a(ioff+1)=dble(clulist(0,i)%x_ima)
     deltaa(ioff+1)=2.0D0/sqrt(sc)
     deltae(ioff+1)=deltaa(ioff+1)
     a(ioff+2)=dble(clulist(0,i)%y_ima)
     deltaa(ioff+2)=2.0D0/sqrt(sc)
     deltae(ioff+2)=deltaa(ioff+2)
     ! use larger delta for nmulsou>1:
     ! remove (GL 05/07/2021):
     !if (nsrc>1) then
     !  deltaa(ioff+1)=1.0
     !  deltaa(ioff+2)=1.0
    ! endif
    !--------
     !  different options to initialize extent parameter
     ! ****test for single telescope fits:
     if (par%fixed_ext_flag) then
       ! use extent from input list (assume arcsec as unit):
       a(ioff+3)=dble(clulist(0,i)%ext)/(abs(image%wcshdr(1)%cdelt1)*3600.D0)
     else
     !****
       if (par%fitext_flag) then 
         a(ioff+3)=5.D0
       else
         a(ioff+3)=0.D0
       endif
     endif

     !change delta for extent parameter to 0.5 (GL 09/03/2012)
     deltaa(ioff+3)=0.5

     !   (use counts dependent stepsizes for felsu, GL 29/06/06):
     deltae(ioff+3)=10.0/sqrt(sc)
 
 
    end do
  
   !end loop over nsrc

   !main loop over sources:


   fit_loop: do
      iordr(1)=0
      call ordr(np,np,iordr) 
      iordr0=iordr                                          

      nffix=0
      nspar=3
      ntermf=np
      ! optionally fix sigmas in fit
      if (.not. par%fitext_flag) then
        do i=1,nsrc
          ioff=npsou*(i-1)
          call ordr(np,ioff+3,iordr)
        end do
        ntermf=ntermf-nsrc
        ntfix=nsrc
        nspar=nspar-1
      end if



      ! optionally fix positions in fit
      if (.not. par%fitpos_flag) then
        do n=1,nsrc
          ioff=npsou*(n-1)
          call ordr(ntermf,ioff+1,iordr)
          call ordr(ntermf,ioff+2,iordr)
        enddo
        ntermf=ntermf-2*nsrc
        ntfix=ntfix+2*nsrc
        nspar=nspar-2
      endif

      clulist(:,1:nsrc)%ndof=nspar

      !  set number of input images nband (06/11/05) 
      nband=0
      do j=1,num_exp
        nband=nband+num_image(j) 
      enddo

      ! fix count rate at zero if mask in entire source region is zero:
      !    set flux to zero, if maskzero=1    
      do n=1,nsrc
        ioff=npsou*(n-1)
        do j=1,nim         
          if (clulist(j,n)%maskzero == 1) then
            a(ioff+3+j)=0.0
            deltaa(ioff+3+j)=0.0
            call ordr(ntermf,ioff+3+j,iordr)
            ntermf=ntermf-1
            ntfix=ntfix+1
          endif
        end do
      end do
      iordr0=iordr
      ntermf0=ntermf
      ntfix0=ntfix

      ! calculate likelihood value for null hypothesis (all fluxes zero)
      atemp=a
      iordr(1)=0
      call ordr(np,np,iordr)
      do i=1,nsrc
        ioff=npsou*(i-1)
        atemp(ioff+3)=0.0 ! extent
        do j=1,nim
          atemp(ioff+3+j)=0.0 ! rates
        enddo
      end do
!*      call funct(par,image,expimage,bkgimage,mask,clulist,cluPSF,atemp,iordr,nsrc,np,newcal,yfit0,yfit1,yfit2,yfit3,modelima,status)
      call funct(par,image,events,expimage,bkgimage,mask,clulist,cluPSF,atemp,iordr,nsrc,np,newcal,yfit0,yfit1,yfit2,modelima,status)

      call like(events,mask,yfit1,yfit2,like00,like00arr,status)
      iordr=iordr0
      !!!********************************
      
     if (verbose_flag) write(*,*) 'NULL hyp. like00=',like00
     if (verbose_flag) write(*,*) 'yfit2=',yfit2(1:4)
     if (verbose_flag) write(*,*) 'yfit1=',yfit1
     if (verbose_flag) write(*,*) 'modelima:',minval(modelima(:,:,1)),maxval(modelima(:,:,1)),minval(modelima(:,:,2)),maxval(modelima(:,:,2)),minval(modelima(:,:,3)),maxval(modelima(:,:,3))
     if (verbose_flag) write(*,*) 'atemp(1:14)=',atemp(1:14)
     if (verbose_flag) write(*,*) '--1---'    


      bestfit=0
      ! first point source fit: fix all sigmas at zero:
      do i=1,nsrc
         ioff=npsou*(i-1)
         !*** GL 15/10/2018, add if-condition here: 
         if (.not. par%fixed_ext_flag) then 
           a(ioff+3)=0.0
         endif
         call ordr(np,ioff+3,iordr)
      end do
      if (par%fitext_flag) then
        ntermf=ntermf-nsrc
        ntfix=ntfix+nsrc 
        nspar=nspar-1
      endif

      call pordr3(np,a,deltaa,sigmaa,iordr,1)
      flamda= flamda0
      do iter=1,max_it


        if (verbose_flag) verbose_flag2=.true.
        call mlfit(par,image,expimage,bkgimage,clulist,cluPSF,events,mask,ntermf,a,deltaa,sigmaa, &
                   iordr,flamda,newcal,yfit1,yfit2,like0,lik_arr,nsrc,ntfix,modelima,status)
        verbose_flag2=.false.


        if(abs(likold-like0) <= delt) exit
        likold=like0
      end do
      !********************************
      
      if (verbose_flag) write(*,*) 'point sources, like0, likediff',like0,like00-like0
      if (verbose_flag) write(*,*) 'a(1:14)=',a(1:14)
      if (verbose_flag) write(*,*) 'yfit2=',yfit2(1:4)
      if (verbose_flag) write(*,*) 'yfit1=',yfit1
      if (verbose_flag) write(*,*) 'modelima:',minval(modelima(:,:,1)),maxval(modelima(:,:,1)),minval(modelima(:,:,2)),maxval(modelima(:,:,2)),minval(modelima(:,:,3)),maxval(modelima(:,:,3))
      if (verbose_flag) write(*,*) '--2---'   
 
      call pordr3(np,a,deltaa,sigmaa,iordr,-1)

      likesave(0)=like0
      lik_arr_save(:,0)=lik_arr
      nzero=0
      do n=1,nsrc
        do iimage=1,nim
          nzero=nzero+clulist(iimage,n)%maskzero
        end do
      enddo
      
      likediff=like00-like0
!      write(*,*) 'gammq1 ndof: ', nsrc*(nspar+nband)-nzero,nsrc
!       write(*,*) 'like2dof input:',like00,like0,likediff
      like2dof(0)=gammq(real(nsrc*(nspar+nband)-nzero)/2.,real(likediff,kind=real32) )
      fit_dof(0)=nsrc*(nspar+nband)-nzero
      asave(:,0)=a
      

!      write(*,*) 'START extent fit: --------------------------------------------------------------------------'
      ! ************* EXTENT FIT ***************** 
      if (par%fitext_flag) then
        ! fit only one free sigma in each run
        ntermf=ntermf+1 
        ! ntfix=nsrc-1
        ! (GL 07/05/2007) changed to:
        ntfix=ntfix-1
        nspar=nspar+1
        do n=1,nsrc
          noff=npsou*(n-1)
          ! iordr(1)=0
          ! call ordr(np,np,iordr)
          iordr=iordr0
          !fix all ext values to 0, only fit ext(n)
          do nn=1,nsrc
            if (nn .ne. n) then
              ioff=npsou*(nn-1)
              a(ioff+3)=0.0
              call ordr(np,ioff+3,iordr)
            endif
          enddo  
          call pordr3(np,a,deltaa,sigmaa,iordr,1)
          flamda= flamda0
          !******************************************
          if (verbose_flag) write(*,*) 'before ext mlfit a(1:4):',a(1:14)

          do iter=1,max_it
            !flamda= flamda0
              if (verbose_flag) write(*,*) 'before ext mlfit a(1:4):',a(1:14)
!*            call mlfit(par,image,expimage,bkgimage,clulist,cluPSF,events,mask,ntermf,a,deltaa,sigmaa, &
!*                       iordr,flamda,newcal,yfit1,yfit2,yfit3,like0,lik_arr,nsrc,ntfix,modelima,ierr)

              !**************************
!           write(*,*) "extent fit: a,deltaa:",a(1:4),deltaa(1:4)  
           if (verbose_flag) verbose_flag3=.true. 
           call mlfit(par,image,expimage,bkgimage,clulist,cluPSF,events,mask,ntermf,a,deltaa,sigmaa, &
                iordr,flamda,newcal,yfit1,yfit2,like0,lik_arr,nsrc,ntfix,modelima,ierr)
           if (verbose_flag) verbose_flag3=.false. 
           if (verbose_flag) write(*,*) '88888 after mlfit:',a(1:4),iter,like0
            if(abs(likold-like0).le.delt) exit
            flamda= flamda/10.D0
            likold=like0 
          end do
          
          call pordr3(np,a,deltaa,sigmaa,iordr,-1)

          
          if (verbose_flag) write(*,*) 'after ext mlfit a(1:14):',a(1:14)
      

          asave(:,n)=a
          likesave(n)=like0
          lik_arr_save(:,n)=lik_arr

         

 !         clulist(:,n)%ext_like = likesave(0)-like0
!          clulist(:,n)%ext_like = gammq(1.5,clulist(0,n)%ext_like)
          !-----------------------------------------------------------------------------------
          !!         New computation of ext_like and its correction for N_dof (GL 17/06/05):
          !-----------------------------------------------------------------------------------
          likediff=like00-like0

!          write(*,*) 'gammq2 ndof: ', ntermf,nsrc

          likecorr = gammq(real(ntermf)/2.,real(likediff,kind=real32))
!         clulist(:,n)%ext_like = likecorr-like2dof(0)
          !
          !-----------------------------------------------------------------------------------

          !-----------------------------------------------------------------------------------
          ! New 27/06/2014:
          ! Extent likelihood is calculated as difference between likesave(0) (N point sources)
          ! and like0 (1 extended source and N-1 point sources).
          ! The number of additional dof for extent fit is 1 
          !-----------------------------------------------------------------------------------

!          write(*,*) 'EXT_LIKE :like0,likesave(0)= ',like0,likesave(0)
          likediff=likesave(0)-like0
          clulist(:,n)%ext_like = gammq(0.5,real(likediff))

          !**********************************
          if (verbose_flag) write(*,*) 'EXT_LIKE :n,like0,likesave(0),ext_like,ext= ',n,like0,likesave(0),  clulist(:,n)%ext_like,a(noff+3)
          if (verbose_flag) write(*,*) 'yfit2=',yfit2(1:4)
          if (verbose_flag) write(*,*) 'a=(1:14)',a(1:14)
          if (verbose_flag) write(*,*) 'yfit1=',yfit1
          if (verbose_flag) write(*,*) 'modelima:',minval(modelima(:,:,1)),maxval(modelima(:,:,1)),minval(modelima(:,:,2)),maxval(modelima(:,:,2)),minval(modelima(:,:,3)),maxval(modelima(:,:,3))
          if (verbose_flag) write(*,*) '--3---' 

!          write(*,*) 'ext, OLD/NEW ext_like: ', a(noff+3), likecorr-like2dof(0), clulist(:,n)%ext_like


          do nn=1,nsrc
            do iimage=1,nim
              nzero=nzero+clulist(iimage,nn)%maskzero
            end do
          enddo                


          like2dof(n) = likecorr
          fit_dof(n)=ntermf 
          clulist(:,n)%ndof=nspar
 
!          write(*,*) 'extent fit: ',like00, like0, likecorr

!          if (clulist(0,n)%ext_like < par%extlikemin .or. a(noff+3) < par%extmin   ) then
           !  (GL 07/05/2019) add  extlike_slope:
           if (clulist(0,n)%ext_like < par%extlikemin .or. a(noff+3) < par%extmin   .or. clulist(0,n)%ext_like < ((par%extlike_slope(1) + a(noff+3)*par%extlike_slope(2)))) then
           ! write(*,*) 'extent not significant: ', n, clulist(0,n)%ext_like, a(noff+3)
            clulist(:,n)%ext_like = 0.0
            clulist(:,n)%ndof=nspar-1
           endif


          
          !!!!!!!!!  test: 09/06/2005 (GL): update  point source fit, if likelihood is better and  ext=0:
          if (like0 < likesave(0)  .and. a(noff+3) == 0.0) then
            asave(:,0)=a
            likesave(0)=like0
            !write(*,*) "update likesave(0)=",likesave(0)
            lik_arr_save(:,0)=lik_arr
            nzero=0
            do nn=1,nsrc
            do iimage=1,nim
              nzero=nzero+clulist(iimage,nn)%maskzero
            end do
            enddo           
            likediff=like00-like0
!            write(*,*) 'gammq3 ndof: ', nsrc*(nspar-1+nband)-nzero,nsrc
            like2dof(0)=gammq(real(nsrc*(nspar-1+nband)-nzero)/2.,real(likediff,kind=real32))
          endif 
         
          

!          if (clulist(0,n)%ext_like < par%extlikemin .or. a(noff+3) < par%extmin) then
           !  (GL 07/05/2019) add  extlike_slope:
           if (clulist(0,n)%ext_like < par%extlikemin .or. a(noff+3) < par%extmin  .or. clulist(0,n)%ext_like < ((par%extlike_slope(1) + a(noff+3)*par%extlike_slope(2)) )) then
!            write(*,*) 'extent not significant (2): ', n, clulist(0,n)%ext_like, a(noff+3)
            clulist(:,n)%ext_like = 0.0
            !(05/04/2005 GL:)
            a(noff+3)=0.0
            !!!!!
          else
!            write(*,*) 'extent : ', n, clulist(0,n)%ext_like, a(noff+3)
            if (n == 1) then 
              bestfit=1
             else if (like0 < likesave(bestfit)) then
              bestfit=n
             endif
          endif
       enddo  !  (nsrc)
       
        !      (05/04/2005 GL:)
        do n=1,nsrc
          noff=npsou*(n-1)
          if (asave(noff+3,bestfit) < par%extmin) then
            clulist(:,n)%ext_like = 0.0
          endif
        enddo

      endif   ! par%fitext_flag
      !*********** END EXTENT FIT ************** 
 

      a=asave(:,bestfit)     
      ntermf=ntermf0
      ntfix=ntfix0


      ! now calculate the likelihood of each of the sources
      do n=nsrc,1,-1
        iordr=iordr0
        ioff=npsou*(n-1)
        multflag=.false.
        if(nsrc.eq.1)then
          ! simply calculate likelihood for one source only
          do j=1,nim
            a(3+j)=0.
          end do

!*          call funct(par,image,expimage,bkgimage,mask,clulist,cluPSF,a(iordr),iordr,nsrc,np,newcal,yfit0,yfit1,yfit2,yfit3,modelima,status)
          call funct(par,image,events,expimage,bkgimage,mask,clulist,cluPSF,a(iordr),iordr,nsrc,np,newcal,yfit0,yfit1,yfit2,modelima,status)

          call like(events,mask,yfit1,yfit2,like0,lik_arr,status)
        else
          ! check, if 2 sources have same position and extent:
          if (par%fitpos_flag) then
            do l=n-1,1,-1
              ioff1=npsou*(l-1)
              distsq=((a(ioff+1)-a(ioff1+1))**2+(a(ioff+2)-a(ioff1+2))**2)
              extdiff=abs(a(ioff+1)-a(ioff1+1))
              if (distsq < mindist .and. extdiff < minextdiff) then
                multflag=.true.
                exit 
              endif
            enddo
          endif 
           
!          write(*,*) 'zero flux source :', n

          ! zero flux of relevant component
          nffix=0
          if (par%fitpos_flag) then
            call ordr(ntermf,ioff+1,iordr)
            call ordr(ntermf,ioff+2,iordr)
              nffix=nffix+2
          end if
          ! if (extfla .and. ext_ml(n) >= par%extlikemin) then
          ! changed 15/11/04 GL
          if (par%fitext_flag) then
            call ordr(ntermf,ioff+3,iordr)
            nffix=nffix+1
          end if
          do j=1,nim
            a(ioff+3+j)=0.
            !  only call ordr, if maskzero=0 (changed 09/07/04)
            if (clulist(j,n)%maskzero .eq. 0) then  
              call ordr(ntermf,ioff+3+j,iordr)
              nffix=nffix+1
            endif
          end do
          ! fix parameters 

          call pordr3(np,a,deltaa,sigmaa,iordr,1)
          
          flamda=flamda0
!         write(*,*) 'ntermf,nffix ,a(1:10) input: ',ntermf,nffix
!         write(*,'(10F10.4)') a(1:10)
!         write(*,*) 'iordr: ',iordr(1:10)

          do iter=1,max_it
!             write(*,*) 'ITER=',iter
!            call mlfit(par,image,expimage,bkgimage,clulist,events,mask,ntermf-nffix,a(iordr),deltaa(iordr),sigmaa(iordr), &
!                       iordr,flamda,newcal,yfit1,yfit2,yfit3,like0,lik_arr,nsrc,ntfix+nffix,modelima,ierr)
!*            call mlfit(par,image,expimage,bkgimage,clulist,cluPSF,events,mask,ntermf-nffix,a,deltaa,sigmaa, &
!*                       iordr,flamda,newcal,yfit1,yfit2,yfit3,like0,lik_arr,nsrc,ntfix+nffix,modelima,ierr)
          if (verbose_flag) verbose_flag4=.true.
            call mlfit(par,image,expimage,bkgimage,clulist,cluPSF,events,mask,ntermf-nffix,a,deltaa,sigmaa, &
                       iordr,flamda,newcal,yfit1,yfit2,like0,lik_arr,nsrc,ntfix+nffix,modelima,ierr)
          verbose_flag4=.false.
           

            if(abs(likold-like0).le.delt) exit                   
            likold=like0                                                    
          end do
          call pordr3(np,a,deltaa,sigmaa,iordr,-1)
        endif    ! if (nsrc == 1) else    

        do j=1,nim 
          clulist(j,n)%like=real( lik_arr(j)-lik_arr_save(j,bestfit),kind=real32)
!          write(*,*) 'j,n, clulist(j,n)%like: ', j,n,clulist(j,n)%like 
        end do

        nband=0
        dml(n)=0.         
        do j=1,num_exp
          nima=num_image(j)
!          if (xid_arr(j)>0) nima=xid_arr(j)-1
          do i=1,nima
            ipoint=ip(j,i)
            dml(n)=dml(n)+clulist(ipoint,n)%like
            nband=nband+1
          end do
        end do

        nzero=0
        do iimage=1,nim
          nzero=nzero+clulist(iimage,n)%maskzero
        end do


!       calculate difference wrt. best fit, not null hypothesis (10/06/2014)
!        likediff=like00-like0


        !write(*,*) 'likesave(bestfit) : ',bestfit,likesave(bestfit), like0 

         likediff = like0 - likesave(bestfit)

         if (verbose_flag) write(*,*) 'bestfit,likesave(bestfit), like0 : ',bestfit,likesave(bestfit), like0
         if (verbose_flag) write(*,*) 'yfit2=',yfit2(1:4)
         if (verbose_flag) write(*,*) 'a=',a(1:14)
         if (verbose_flag) write(*,*) 'yfit1=',yfit1
         if (verbose_flag) write(*,*) 'modelima:',minval(modelima(:,:,1)),maxval(modelima(:,:,1)),minval(modelima(:,:,2)),maxval(modelima(:,:,2)),minval(modelima(:,:,3)),maxval(modelima(:,:,3))
         if (verbose_flag) write(*,*) '--4---'   


         !! like(NULL) - like(best) GL 25/06/2014
         !!likediff = like0 - likesave(bestfit)
!        write(*,*) 'likediff 1: ',bestfit,likediff, real(ntermf-nffix)/2.,like2dof(bestfit),gammq(real(ntermf-nffix)/2.,likediff)
!         write(*,*) 'likediff 1: ', likesave(bestfit),like0,likediff

         ! Calculate dml(n) from Ndof corrected likelihoods compared to like00:
        !dml(n)=like2dof(bestfit)-gammq(real(ntermf-nffix)/2.,likediff) 

         !normalize with additional ndof wrt. dof=2 
!         write(*,*) 'fit_dof(bestfit), ntermf,nffix: ',fit_dof(bestfit), ntermf,nffix
         ddof=fit_dof(bestfit)-ntermf+nffix
!         write(*,*) 'ddof: ', ddof
         if (ddof > 0) then
           dml(n)=gammq(real(ddof)/2.,real(likediff,kind=real32))
         else
           dml(n)=real(likediff,kind=real32)
         endif

         if (verbose_flag)    write(*,*) 'dml(n):',dml(n)
!         write(*,*) 'a(1:10) output:'
!         write(*,'(10F10.4)') a(1:10)
        
        ! 20/06/05
        ! dml(n)=gammq(real(nspar+nband-nzero)/2.,dml(n))

        noff=npsou*(n-1)

        ! special case: decision between 2 pnt sources and one extended source:
        ! do not use par%likemin threshold in order to avoid bias towards extended source:
        dmlthresh=par%likemin
!        write(*,*)  'dmlthresh: ',dmlthresh
!        write(*,*)  'raw like: ',like00-like0

        if (par%fitext_flag .and. nsrc == 2 .and.  bestfit == 0 .and. n == 2 .and. a(3) > par%extmin) then
           !!!********************************
              if (verbose_flag) write(*,*) 'special case 1'
          likediff=like00-like0
          likecorr=gammq(real(nspar+nband-nzero)/2.,real(likediff,kind=real32))
          if (likecorr < like2dof(0)) then
            dmlthresh=0.0
          endif
        endif
        if (par%fitext_flag .and. nsrc == 2 .and.  bestfit == 0 .and. n == 1 .and. a(npsou+3) > par%extmin) then
          if (verbose_flag) write(*,*) 'special case 2'
          likediff=like00-like0
          likecorr=gammq(real(nspar+nband-nzero)/2.,real(likediff,kind=real32))
          if (likecorr < like2dof(0)) then
            dmlthresh=0.0
          endif
        endif


        if (verbose_flag) write(*,*) 'n,  likediff, dml(n): , ndof', n, likediff, dml(n),(nspar+nband-nzero)/2.

        if ((dml(n).lt. dmlthresh .or. multflag)) then

!          write(*,*) 'delete source ',n,dml(n), clulist(1,1)%x_ima,clulist(1,1)%y_ima , clulist(1,2)%x_ima,clulist(1,2)%x_ima,clulist(1,2)%y_ima
!          write(*,*) 'delete source a(1:11)=', a(1:11)
          ! throw out component
          do l=n+1,nsrc
            ioff0=npsou*(l-2)
            ioff1=npsou*(l-1)
            do j=1,3
              a(ioff0+j)=a(ioff1+j)
            end do          
            do j=1,nim
              a(ioff0+3+j)=a(ioff1+3+j)
            enddo
            clulist(:,l-1) = clulist(:,l)
            dml(l-1)=dml(l)
          enddo
          nsrc=nsrc-1
!           write(*,*) 'done deleted, new  a(1:11)=', a(1:11)         
          if(nsrc.eq.0)then
            !!!!!!!!  TEST output
!            write(*,*) 'exit from ' , rtname,' ', dml(n)
            return
         endif

         ! re-calculate mask for remaining source,re-size yfit1 (GL 15/04/2021):
         newcal=.true.

         
     do i=1,nsrc
  
       ioff=npsou*(i-1)
       clulist(:,i)%x_ima = real( a(ioff+1), kind=real32)
       clulist(:,i)%y_ima = real( a(ioff+2), kind=real32)
       clulist(:,i)%ext   = real( a(ioff+3), kind=real32)
       do j=1,nim
         clulist(j,i)%rate= real( a(ioff+3+j), kind=real32)
       end do
       ! (15/03/2019 GL) copy dof corrected likelihood to summary value:
       !clulist(0,i)%like=dml(i)
      enddo
    
         call fill_arrs(par,par%photon_flag,clulist,image,eventlist,pixlist,cluPSF,detmask,expimage,modelima,nsrc,events,mask,status)
          deallocate(yfit1)
          allocate(yfit1(mask%nev))
          np=nsrc*npsou
          exit
        else  


          ! get back best-fit parameters
          do m=1,np
             a(m)=asave(m,bestfit)
          enddo
           like0=likesave(bestfit)
           ! calculate final detection likelihoods for source (02/12/04, GL):
           ! zero fluxes of source n:
           ioff=npsou*(n-1)
           do j=1,nim
             a(ioff+3+j)=0.
           end do

!*           call funct(par,image,expimage,bkgimage,mask,clulist,cluPSF,a(iordr),iordr,nsrc,np,newcal,yfit0,yfit1,yfit2,yfit3,modelima,status)
           call funct(par,image,events,expimage,bkgimage,mask,clulist,cluPSF,a(iordr),iordr,nsrc,np,newcal,yfit0,yfit1,yfit2,modelima,status)

           call like(events,mask,yfit1,yfit2,like1,lik_arr,status)

           if (verbose_flag) write(*,*) 'zero source#, like1,yfit2=',n,like1,yfit2(1:4)
           if (verbose_flag) write(*,*) 'a=',a(1:14)
           if (verbose_flag) write(*,*) 'yfit1=',yfit1
           if (verbose_flag) write(*,*) 'modelima:',minval(modelima(:,:,1)),maxval(modelima(:,:,1)),minval(modelima(:,:,2)),maxval(modelima(:,:,2)),minval(modelima(:,:,3)),maxval(modelima(:,:,3))
           if (verbose_flag) write(*,*) '--5---' 

           do j=1,nim 
             clulist(j,n)%like =  real( lik_arr(j)-lik_arr_save(j,bestfit) , kind=real32)
           end do
           nband=0
           dml(n)=0.         
           do j=1,num_exp
             nima=num_image(j)
!             if (xid_arr(j)>0) nima=xid_arr(j)-1
             do i=1,nima
               ipoint=ip(j,i)
               dml(n)=dml(n)+clulist(ipoint,n)%like
               nband=nband+1
             end do
           end do
           if (par%fitext_flag .and. clulist(0,n)%ext_like == 0.0) then
!             write(*,*) 'dml 00: ',dml(n)

             dml(n)=gammq(real(nspar-1+nband-nzero)/2.,dml(n))
!             write(*,*) 'dml, ndof 0: ',dml(n),real(nspar-1+nband-nzero)/2.
           else
!             write(*,*) 'dml 1: ',dml(n)
!             write(*,*) 'gammq8 ndof: ', nspar+nband-nzero,nsrc
             dml(n)=gammq(real(nspar+nband-nzero)/2.,dml(n))
!             write(*,*) 'dml, ndof 2: ',dml(n), real(nspar+nband-nzero)/2.
           endif
           do m=1,np
             a(m)=asave(m,bestfit)
           enddo    
          if (n == 1) exit fit_loop
        endif
 
      enddo
    end do fit_loop

   !  assign output parameter values before error loop (GL 12/11/2004):
 
   do i=1,nsrc
  
    ioff=npsou*(i-1)
    clulist(:,i)%x_ima =  real( a(ioff+1), kind=real32)
    clulist(:,i)%y_ima =  real( a(ioff+2), kind=real32)
    clulist(:,i)%ext   =  real( a(ioff+3), kind=real32)
    do j=1,nim
      clulist(j,i)%rate=real( a(ioff+3+j), kind=real32)
    end do
    ! (15/03/2019 GL) copy dof corrected likelihood to summary value:
    clulist(0,i)%like=dml(i)
   enddo  

  
  ! second stage will be run, return to main program:

  if (first .and. maxval(clulist(0,1:nsrc)%ext_like) .gt. 0.0) then
 !   write(*,*) 'return for 2nd stage'
    return
  endif

  

  ! call "model" to calculate PSF weighted mask fraction:

  iordr(1)=0
  call ordr(np,np,iordr)
  afit=a
  ! call funct with scts set to 1.0:
  do i=1,nsrc
   ioff=npsou*(i-1)
    do j=1,nim
     afit(ioff+3+j)=1.0
    end do
  enddo   
  !!!!!!!!!!


  if (par%photon_flag) then
     ! ***** THIS IS A TEMPORARY FIX!****'
     clulist(:,:)%maskfrac=1.0
  else
    call funct(par,image,events,expimage,bkgimage,mask,clulist,cluPSF,afit,iordr,nsrc,np,newcal,yfit0,yfit1,yfit2,modelima,status)
    call model(par,image,expimage,bkgimage,mask,events,clulist,cluPSF,dble(clulist(0,:)%x_ima),dble(clulist(0,:)%y_ima), &
             dble(clulist(0,:)%ext),dble(clulist(1:nim,:)%scts),nsrc,.true.,newcal,yfit0,yfit1,yfit2,modelima,status) 
  endif

  ! throw out sources with PSF weighted mask fraction psffrac < psfthreshold
  do n=nsrc,1,-1
    nn=0
    ioff=npsou*(i-1)
    do i=1,num_exp
      ! check for psffrac values below the threshold:
      maskbad=.false.
      do j=1,num_image(i)
        ipoint=ip(i,j)
        if (clulist(ipoint,n)%maskfrac < psfthreshold ) maskbad=.TRUE.
      enddo
      do j=1,num_image(i)   
        ipoint=ip(i,j) 
        if (.not. maskbad) then
          nn=nn+1  
        else
          ! throw out source, if psffrac < psfthreshold in all energy bands:
          clulist(ipoint,n)%scts = 0.0
          clulist(ipoint,n)%like = 0.0
          if   (clulist(ipoint,n)%maskzero == 0) then 
            nzero=nzero+1 
            clulist(ipoint,n)%maskzero = 1
          endif
        endif
      end do
    end do
    ! check likelihood once more: 
    dml(n)=0.0
    do ipoint=1,nim
      if (clulist(ipoint,n)%maskzero == 0) dml(n)=dml(n)+clulist(ipoint,n)%like
    enddo
    if (par%fitext_flag .and. clulist(0,n)%ext_like == 0.0) then
!      write(*,*) 'dml  3) ',dml(n)

      dml(n)=gammq(real(nspar-1+nband-nzero)/2.,dml(n))
!      write(*,*) 'dml  4) ',dml(n)
    else
!      write(*,*) 'dml 5) ',dml(n)
      dml(n)=gammq(real(nspar+nband-nzero)/2.,dml(n))
!      write(*,*) 'dml 6) ',dml(n)
    endif
!    write(*,*) 'final likelihood: ',dml(n)
    if (nn == 0 .or. dml(n) .lt. par%likemin) then
    ! throw out component:
       do l=n+1,nsrc
         ioff0=npsou*(l-2)
         ioff1=npsou*(l-1)
         do j=1,3
           a(ioff0+j)=a(ioff1+j)
         end do
         do j=1,nim
           a(ioff0+3+j)=a(ioff1+3+j)
         enddo
         clulist(:,l-1) = clulist(:,l)
         dml(l-1)=dml(l)
       enddo
       nsrc=nsrc-1
       if (nsrc.eq.0) then
         ! write(*,*) 'EXIT 2 ',  dml(n)
          return   
       endif    
    endif
  enddo
    
  ! determine errors:

  iordr(1)=0

  call ordr(np,np,iordr)                                           
 

!  write(*,*) 'end mulsou', clulist(1,1:nsrc)%x_ima, clulist(1,1:2)%y_ima,clulist(1,1:nsrc)%ecutrad

!*  call paramerr(par,image,expimage,bkgimage,mask,events,clulist,cluPSF,a,deltae,nsrc,npsou,yfit1,yfit3,modelima,status)
  call paramerr(par,image,expimage,bkgimage,mask,events,clulist,cluPSF,a,deltae,nsrc,npsou,yfit1,modelima,status)





  deallocate(yfit1)


!  write(*,*) 'DONE mulsou, nsrc=',nsrc,a(1),a(2),a(3), dml(1), dml(2)
  
  end subroutine mulsou


!--------------------------------------------------------------------------------
!*
  subroutine funct(par,image,events,expimage,bkgimage,mask,clulist,cluPSF,afit,iordr,nsrc,np,newcal,yfit0,yfit1,yfit2,modelima,ierr)

  !---------------------------------------------------------------------------
  ! Calculate fit function for parameter set afit 
  ! Call routines cage and cagexy to keep parameters within
  ! given limits, then call subroutine model to calculate 
  ! model values yfit, yfit2.
  ! Change wrt. XMM-SAS : Removed argument yfit3 (total model flux, obsolete).
  !---------------------------------------------------------------------------
      
    type(ermldet_param_type), intent(in)     :: par
    type(i4cube_type)                        :: image
    type(events_type),dimension(:)           :: events
    type(r4cube_type)                        :: expimage,bkgimage
    type(mask_type)                          :: mask
!    type(srcrow_type),dimension(:,:),pointer :: clulist
    type(srcrow_type),dimension(:,:),allocatable :: clulist

    type(FPSF)                               :: cluPSF 
    integer(kind=int32), dimension(:)        :: iordr
    integer(kind=int32)                      :: nsrc,np 
    real(kind=real64), dimension(:)          :: afit  
    real(kind=real64), dimension(:)          :: yfit1,yfit2
    real(kind=real64)                        :: yfit0
    real(kind=real64), dimension(:,:,:),pointer  :: modelima
    logical(kind=log32)                      :: newcal
    integer(kind=int32)                      :: ierr  

    real(kind=real64), dimension(max_par)    :: a
    real(kind=real64), dimension(max_fit)    :: x0,y0,sig
    real(kind=real64), dimension(max_ima,max_fit)  :: scts
    real(kind=real32)                        :: ecut2
    integer(kind=int32)                      :: ndat,npars,i,n,status!,iev,nev
    character(len=*), parameter              :: rtname='FUNCT' 
 
!   call system_clock(cclock1)


    ierr = 0

    ! re-sort parameter array
    a(iordr) = afit
    ecut2 = clulist(0,1)%ecutrad*clulist(0,1)%ecutrad

    ! cage parameters
    npars=nim+3
    do n=1,nsrc 
      ndat=(n-1)*npars+1
!      write(*,*) 'before cagexy: ',n,a(ndat) ,a(ndat+1),clulist(0,1)%ecutrad, ecut2 
!      call cagexy (a(ndat),a(ndat+1),clulist(0,n)%x_ima,clulist(0,n)%y_ima,ecut2)
 ! use cagemask:
      call cagemask(a(ndat),a(ndat+1),mask,status)     
      x0(n)=a(ndat)
      y0(n)=a(ndat+1)
!      write(*,*) 'cagexy: ',n, x0(n),y0(n)
      ndat=(n-1)*npars+3
!****** New boundary routine (6/12/2017):
!      old:      
      sig(n)=cage(a,ndat,0.,par%extmax)
    !  write(*,*) "sig: ", sig(1),':',a(1:5),':',ndat,':',a(ndat)
!**********
!       !new:
!      call setparlimits(a,ndat,0.,par%extmax,ierr)
!      sig(n)=a(ndat)
!******      
   
      do i=1,nim
        ndat=(n-1)*npars+i+3
        ! wrt. XMMSAS, removed  negfla option
        scts(i,n)=cage(a,ndat,0.0,1e10) 
      end do
    end do 

    ! calculate expectation value of model + background


!*    call model(par,image,expimage,bkgimage,mask,clulist,cluPSF,x0,y0,sig,scts,nsrc,.false.,newcal,yfit0,yfit1,yfit2,yfit3,modelima,status)

    if (par%photon_flag) then
      call phmodel(par,image,expimage,bkgimage,mask,events,clulist,cluPSF,x0,y0,sig,scts,nsrc,.false.,newcal,yfit0,yfit1,yfit2,modelima,status)
    else
      call model(par,image,expimage,bkgimage,mask,events,clulist,cluPSF,x0,y0,sig,scts,nsrc,.false.,newcal,yfit0,yfit1,yfit2,modelima,status)
    endif

    call pordr(np,a,afit,iordr,1)                                


!    call system_clock(cclock2)
!    cclock0=cclock2-cclock1+cclock0



  end subroutine funct
  
           
!--------------------------------------------------------------------------------
!*
  subroutine model(par,image,expimage,bkgimage,mask,events,clulist,cluPSF,x0,y0,sig,scts,nsrc,last,newcal,yfit0,yfit1,yfit2,modelima,ierr)

  !----------------------------------------------------------------------
  ! Fill arrays yfit1, yfit2, yfit3 with model values.    
  ! Models are calculated using routine ftpsf
  ! Model images are only re-calculated, if relevant fit  
  ! parameters change. 
  ! If only source counts change, images are rescaled.
  ! Change wrt. XMM-SAS:
  ! Elements of yfit1 cover only pixels where mask%mask = 2 
  ! (i.e. where pixel value of science image > 0) 
  !----------------------------------------------------------------------

    type(ermldet_param_type), intent(in)                      :: par
    type(i4cube_type), intent(in)                             :: image
    type(r4cube_type), intent(in)                             :: expimage,bkgimage
    type(mask_type), intent(in)                               :: mask 
    type(events_type),dimension(:)                            :: events
    ! argument events is obsolete, but leave interface as is for apetool 
    real(kind=real64), dimension(:,:,:),pointer               :: modelima 
!    type(srcrow_type),dimension(:,:),pointer                  :: clulist
    type(srcrow_type),dimension(:,:),allocatable              :: clulist        
    type(FPSF)                                                :: cluPSF 
    real(kind=real64), dimension(max_fit)                     :: x0, y0, sig
    real(kind=real64), dimension(max_ima,max_fit), intent(in) :: scts
    integer(kind=int32), intent(in)                           :: nsrc 
    logical(kind=log32), intent(in)                           :: last
    real(kind=real64)                                         :: yfit0  
    real(kind=real64), dimension(:)                           :: yfit1                     
    real(kind=real64), dimension(:)                           :: yfit2
    logical(kind=log32)                                       :: newcal
    integer(kind=int32)                                       :: ierr
    
    real(kind=real64), dimension(:,:,:,:), allocatable     :: ff_save
    real(kind=real64), dimension(:,:),  allocatable        :: fima
    real(kind=real64), dimension(max_ima,max_fit)          :: scts_save
    real(kind=real64), dimension(max_ima,max_fit)          :: scts_scale
    real(kind=real64), dimension(max_fit)                  :: x0_save, y0_save, sig_save
    real(kind=real64), dimension(:,:,:), allocatable       :: exp_sel,bkg_sel,countsima  
    logical(kind=1), dimension(:,:,:), allocatable         :: mask1,mask2
    integer(kind=int32), dimension(max_fit)                :: ix0,iy0             
    real(kind=real64)                                      :: totmodel,maskmodel,totexp
!    integer(kind=int32)                                    :: iev,nev  
    real(kind=real32),parameter                            :: small=1.E-10
!    real(kind=real64)                                      :: xremain,yremain
    integer(kind=int32)                                    :: i,j,n,ix,iy,ipoint,mx,my
    integer(kind=int32)                                    ::  isize    
    integer(kind=int32)                                    :: ixs,iys,ix_mask,iy_mask,lf,ipix,status
    integer(kind=int32)                                    :: ixf1,ixf2,iyf1,iyf2,ixm1,ixm2,iym1,iym2
    logical(kind=log32), dimension(max_ima,max_fit)        :: scalepsf
    logical(kind=log32)                                    :: first = .true.
    integer(kind=int32)                                    :: nsrc_old
    character(len=*), parameter                            :: rtname='MODEL'

    save :: ff_save,fima,exp_sel,bkg_sel,countsima
    save :: isize,scts_save,scts_scale,x0_save,y0_save,sig_save,mask1,mask2,first,nsrc_old
    save :: ix0,iy0

    
    ierr = 0 

    !set isize for 2D PSF:
    if (par%shapelet_flag) then
     isize = cluPSF%myCPSF%MAX_PSF_IMG_SIZE
    else
       isize = nint(2.*clulist(0,1)%ecutrad)+2
    endif
   
    ! set ix0,iy0 to nearest integer pixel position
    ! this reference pixel remains constant during the fitting process
    if (first .or. newcal) then
      do n=1,nsrc
         ix0(n)=nint(x0(n))
        iy0(n)=nint(y0(n))
      enddo
    endif 

    
!   write(*,*) '**START MODEL**', newcal, first,x0(1),y0(1),sig(1),scts(1,1)

   if (last) then
!    calculate fraction of source flux on valid mask pixels:
     do n=1,nsrc
       do i=1,num_exp
         do j=1,num_image(i)
           ipoint=ip(i,j)
           totmodel=0.0
           maskmodel=0.0
           totexp=0.0
           my=0
           do iy = mask%iy_min, mask%iy_max
             my=my+1
             mx=0
             do ix = mask%ix_min, mask%ix_max
               mx=mx+1
               totmodel=totmodel+ff_save(mx,my,n,ipoint)
               if (mask%mask(mx,my,ipoint) > 0) then
                 maskmodel = maskmodel+ff_save(mx,my,n,ipoint)
                 totexp = totexp+ff_save(mx,my,n,ipoint)*exp_sel(mx,my,ipoint)
               endif 
             enddo
           enddo
           if (totmodel > 0.0) then
             clulist(ipoint,n)%maskfrac = real( maskmodel/totmodel, kind=real32 )
             clulist(ipoint,n)%exp_map  = real( totexp/totmodel, kind=real32 )
           endif
         enddo
       enddo
     enddo
    ! stop here
     return
   endif

    if (first .or. newcal) then
      if (allocated(ff_save)) deallocate(ff_save)
      allocate(ff_save(mask%nx,mask%ny,nsrc,nim))
      if (allocated(countsima)) deallocate(countsima)
      allocate(countsima(mask%nx,mask%ny,nim))
      if (allocated(exp_sel)) deallocate(exp_sel)
      allocate(exp_sel(mask%nx,mask%ny,nim))
      if (allocated(bkg_sel)) deallocate(bkg_sel)
      allocate(bkg_sel(mask%nx,mask%ny,nim))
      if (allocated(mask1)) deallocate(mask1)
      allocate(mask1(mask%nx,mask%ny,nim))
      if (allocated(mask2)) deallocate(mask2)
      allocate(mask2(mask%nx,mask%ny,nim))



      exp_sel = expimage%arr(mask%ix_min:mask%ix_max,mask%iy_min:mask%iy_max,:)
      bkg_sel = bkgimage%arr(mask%ix_min:mask%ix_max,mask%iy_min:mask%iy_max,:)
!      write(*,*) "selbkg:",mask%ix_min,mask%ix_max,mask%iy_min,mask%iy_max
      mask1(:,:,:) = .false.
      mask2(:,:,:) = .false. 

      where (mask%mask > 0)  mask1=.true.
      where (mask%mask == 2) mask2=.true.       

      ! initialize ff_save (13/06/2014), fixes bogus likelihood values

      ff_save(:,:,:,:)=0.0
      nsrc_old=nsrc
    else if (nsrc .ne. nsrc_old) then
       if (allocated(ff_save)) deallocate(ff_save)
      allocate(ff_save(mask%nx,mask%ny,nsrc,nim))
      ff_save(:,:,:,:)=0.0
      nsrc_old=nsrc
    endif
 

!    call system_clock(cclock2)

 


! change 22/05/2012:
    if (first .or. newcal) then
!    if (first .and. .not. newcal) then
      scalepsf(1:max_ima,1:max_fit) = .false.
    else
!   check which parameters changed since last call: 
      do n=1,nsrc
        if ((x0(n) .ne. x0_save(n)) .or. (y0(n) .ne. y0_save(n)) .or. (sig(n) .ne. sig_save(n))) then 
          scalepsf(1:max_ima,n) = .false.
!          count_model=count_model+1
        else
          do i=1,num_exp
            do j=1,num_image(i)
              ipoint=ip(i,j)
              if (scts(ipoint,n) .eq. 0.0) then
                scts_scale(ipoint,n) = 0.0
                scalepsf(ipoint,n) = .true.
              else if (scts(ipoint,n) .eq. scts_save(ipoint,n)) then
                scts_scale(ipoint,n) = 1.0
                scalepsf(ipoint,n) = .true.
              else if (abs(scts_save(ipoint,n)) .gt. small) then
                scts_scale(ipoint,n) =  scts(ipoint,n) / scts_save(ipoint,n)
                scalepsf(ipoint,n) = .true.
              else
                scts_scale(ipoint,n) = 0.0
                scalepsf(ipoint,n) = .false.
!                count_model=count_model+1
              endif
            enddo
          enddo
        endif
      enddo
    endif

    lf=1
    ipix=1
    yfit1(:)=0.
    yfit2(:)=0.
!*    yfit3(:)=0.
    yfit0=0
 


    do i=1,num_exp



        do j=1,num_image(i)
          ipoint=ip(i,j)

          !correct 31-jul-2013:
!          modelima(:,:,:)=0.0
          modelima(:,:,ipoint)=0.D0

          

          do n=1,nsrc

            ixs=ix0(n)
            iys=iy0(n)
            ix_mask=ixs-mask%ix_min
            iy_mask=iys-mask%iy_min
            ! source offset from pixel center         
            ! xremain=x0(n)-dble(ixs)
            ! yremain=y0(n)-dble(iys)
             !write(*,*) "n,x0,ixs,xremain",n,x0(n),ixs,xremain,y0(n),iys,yremain

            if (scalepsf(ipoint,n) .and. (scts_scale(ipoint,n) .ne. 0.0)) then

              !re-scale ff_save:
              if (scts_scale(ipoint,n) .ne. 1.0) then
                ff_save(:,:,n,ipoint)=ff_save(:,:,n,ipoint)*scts_scale(ipoint,n)
             endif
        !      write(*,*) "ff_save:",sum(ff_save(:,:,n,ipoint))
             modelima(:,:,ipoint)=modelima(:,:,ipoint)+ff_save(:,:,n,ipoint)


        !      write(*,*) "rescale:",maxval(modelima(:,:,ipoint)),minval(modelima(:,:,ipoint)),sum((modelima(:,:,ipoint)))

            else

               ! calculate new fima:
               call ftpsf(par,image,newcal,i,ipoint,n,ix0(n),iy0(n),x0(n),y0(n),sig(n),clulist(0,n)%ecutrad,cluPSF,fima,isize,status)

               ! set negative values to 0.D0:
               fima=max(fima,0.D0)
               
!               write(*,*) "param:",n,x0(n),y0(n),sig(n)
!               write(*,'(A5,I4,I4,3D22.14)') "fima:",n,ipoint,maxval(fima),minval(fima),sum(fima)
              ! copy fima into modelima:

              ixf1=max(1,isize/2-ix_mask+1)
              iyf1=max(1,isize/2-iy_mask+1)
              ixf2=min(isize,isize/2-ix_mask+mask%nx)
              iyf2=min(isize,isize/2-iy_mask+mask%ny)
              ixm1=max(1,ix_mask-isize/2+1)
              iym1=max(1,iy_mask-isize/2+1)
              ixm2=min(mask%nx,ix_mask+isize/2)              
              iym2=min(mask%ny,iy_mask+isize/2)

!              write(*,*) "fima:",x0(n),y0(n),ixs,iys,ixf1,ixf2,ixm1,ixm2,isize,isize/2,ix_mask,mask%nx


              ff_save(:,:,n,ipoint)=0.0

              ff_save(ixm1:ixm2,iym1:iym2,n,ipoint)=scts(ipoint,n)*fima(ixf1:ixf2,iyf1:iyf2)
!              modelima(:,:,ipoint)=modelima(:,:,ipoint)+ff_save(:,:,n,ipoint)
              ! this is slighty faster:
              modelima(ixm1:ixm2,iym1:iym2,ipoint)=modelima(ixm1:ixm2,iym1:iym2,ipoint)+ff_save(ixm1:ixm2,iym1:iym2,n,ipoint)


            endif

          end do

      end do
        ! end loop over num_image      
    end do  



    ! New code to reshape modelima to yfit arrays:

    if (par%expima_flag) then
!      write(*,*) "size=",size(modelima,1), size(exp_sel,1), size(bkg_sel,1)
      countsima = modelima*exp_sel+bkg_sel
    else
      countsima = modelima+bkg_sel
    endif

!    call system_clock(cclock1)


    ! faster method
    yfit1 = pack(countsima,mask2,yfit1)




! call system_clock(cclock2)
! cclock0 = cclock2-cclock1+cclock0
        

    do ipoint=1,nim
      yfit2(ipoint) = sum(countsima(:,:,ipoint),mask%mask(:,:,ipoint) > 0)
    enddo
    yfit0=sum(yfit2)



    first = .false.
    x0_save=x0
    y0_save=y0
    sig_save=sig
    scts_save=scts




  end subroutine model



!--------------------------------------------------------------------------------
!*
  subroutine phmodel(par,image,expimage,bkgimage,mask,events,clulist,cluPSF,x0,y0,sig,scts,nsrc,last,newcal,yfit0,yfit1,yfit2,modelima,ierr)

  !----------------------------------------------------------------------
  ! **** Version of subroutine 'model' for photon mode, preliminary ****
  ! Fill arrays yfit1, yfit2, yfit3 with model values.    
  ! Models are calculated using routine ftpsf
  ! Model images are only re-calculated, if relevant fit  
  ! parameters change. 
  ! If only source counts change, images are rescaled.
  ! Change wrt. XMM-SAS:
  ! Elements of yfit1 cover only pixels where mask%mask = 2 
  ! (i.e. where pixel value of science image > 0) 
  !----------------------------------------------------------------------

    type(ermldet_param_type), intent(in)     :: par
    type(i4cube_type)                        :: image
    type(r4cube_type)                        :: expimage,bkgimage
    type(mask_type)                          :: mask 
    type(events_type),dimension(:)            :: events
    real(kind=real64), dimension(:,:,:),pointer  :: modelima
!    type(srcrow_type),dimension(:,:),pointer :: clulist
    type(srcrow_type),dimension(:,:),allocatable :: clulist    
    type(FPSF)                               :: cluPSF 
    real(kind=real64), dimension(max_fit)    :: x0, y0, sig
    real(kind=real64), dimension(max_ima,max_fit), intent(in) :: scts
    integer(kind=int32), intent(in)          :: nsrc 
    logical(kind=log32), intent(in)          :: last
    real(kind=real64)                        :: yfit0  
    real(kind=real64), dimension(:)                       :: yfit1                     
    real(kind=real64), dimension(:)                       :: yfit2
!*    real(kind=real32), dimension(:)                     :: yfit3 

    logical(kind=log32)                                   :: newcal
    integer(kind=int32)                                   :: ierr

    integer(kind=int32), dimension(max_fit)                   :: ix0,iy0             
    
!    real(kind=real32), dimension(:,:,:,:), allocatable, SAVE :: psf_save
    real(kind=real64), dimension(:,:,:,:), allocatable :: ff_save
!    real(kind=real32), dimension(:,:), pointer, save       :: fima => NULL()    
    real(kind=real64), dimension(:,:), allocatable         :: fima 
    real(kind=real64), dimension(:,:), allocatable         :: phima 

    real(kind=real64), dimension(max_ima,max_fit) :: scts_save
    real(kind=real64), dimension(max_ima,max_fit) :: scts_scale
    real(kind=real64), dimension(max_fit)         :: x0_save, y0_save, sig_save
    real(kind=real64), dimension(:,:,:), allocatable :: exp_sel,bkg_sel,countsima  
    real(kind=real64), dimension(:), allocatable     :: yexp1,ybkg1
    real(kind=real64), dimension(:,:), allocatable   :: ysave1
    logical(kind=log32)                              :: fnewcal!,phnewcal
    logical(kind=1), dimension(:,:,:), allocatable   :: mask1,mask2
    real(kind=real64)                                      :: sumexp,sumbkg  
    integer(kind=int32)                                    :: evcount
    real(kind=real64) ::    totmodel,maskmodel,totexp
    integer(kind=int32) ::  iev,nev  ,isrc
    real(kind=real32),parameter :: small=1.E-10
    integer(kind=int32) :: i,j,n,ix,iy,ipoint,mx,my,xoff,yoff
    integer(kind=int32) :: isize
    
    integer(kind=int32) :: ixs,iys,ix_mask,iy_mask,lf,ipix,status
    integer(kind=int32) :: ixf1,ixf2,iyf1,iyf2,ixm1,ixm2,iym1,iym2
    logical(kind=log32), dimension(max_ima,max_fit) :: scalepsf
    logical(kind=log32) :: first = .true.
    integer(kind=int32) :: nsrc_old
    character(len=*), parameter :: rtname='MODEL'



    
    save :: ff_save,fima,phima,exp_sel,bkg_sel,countsima
    save :: isize,scts_save,scts_scale,x0_save,y0_save,sig_save,yexp1,ybkg1,ysave1,mask1,mask2,first,nsrc_old
    save :: ix0,iy0
    

    ierr = 0 

    nev=mask%nev


    ! set ix0,iy0 to nearest integer pixel position
    ! this reference pixel remains constant during the fitting process
    if (first .or. newcal) then
      do isrc=1,nsrc
        ix0(isrc)=nint(x0(isrc))
        iy0(isrc)=nint(y0(isrc))
      enddo
    endif 

   ! **** This is currently not used: 
   if (last) then
!    calculate fraction of source flux on valid mask pixels:
     do n=1,nsrc
       do i=1,num_exp
         do j=1,num_image(i)
           ipoint=ip(i,j)
           totmodel=0.0
           maskmodel=0.0
           totexp=0.0
           my=0
           do iy = mask%iy_min, mask%iy_max
             my=my+1
             mx=0
             do ix = mask%ix_min, mask%ix_max
               mx=mx+1
               totmodel=totmodel+ff_save(mx,my,n,ipoint)
               if (mask%mask(mx,my,ipoint) > 0) then
                 maskmodel=maskmodel+ff_save(mx,my,n,ipoint)
                 totexp = totexp+ff_save(mx,my,n,ipoint)*exp_sel(mx,my,ipoint)
               endif 
             enddo
           enddo
           if (totmodel > 0.0)  then 
             clulist(ipoint,n)%maskfrac = real( maskmodel/totmodel, kind=real32)
             clulist(ipoint,n)%exp_map  = real( totexp/totmodel, kind=real32)
             
           endif
         enddo
       enddo
     enddo
    ! stop here
     return
   endif


    
    ! ff_save is only used for global PSF ->  4th dimension=nim:
    if (first .or. newcal) then
      if (allocated(ff_save)) deallocate(ff_save)
      allocate(ff_save(mask%nx,mask%ny,nsrc,nim))

      if (allocated(countsima)) deallocate(countsima)
      allocate(countsima(mask%nx,mask%ny,nim))

      if (allocated(exp_sel)) deallocate(exp_sel)
      allocate(exp_sel(mask%nx,mask%ny,nim))

      if (allocated(bkg_sel)) deallocate(bkg_sel)
      allocate(bkg_sel(mask%nx,mask%ny,nim))

      if (allocated(mask1)) deallocate(mask1)
      allocate(mask1(mask%nx,mask%ny,nim))

      if (allocated(mask2)) deallocate(mask2)
      allocate(mask2(mask%nx,mask%ny,nim))


      if (allocated(yexp1)) deallocate(yexp1)
      allocate(yexp1(nev))

      if (allocated(ybkg1)) deallocate(ybkg1)
      allocate(ybkg1(nev))

      if (allocated(ysave1)) deallocate(ysave1)
      allocate(ysave1(nev,max_fit))
      ysave1(:,:)=0.0



      exp_sel = expimage%arr(mask%ix_min:mask%ix_max,mask%iy_min:mask%iy_max,:)
      bkg_sel = bkgimage%arr(mask%ix_min:mask%ix_max,mask%iy_min:mask%iy_max,:)      
      mask1(:,:,:) = .false.
      mask2(:,:,:) = .false. 

      !** store exposure and background values for each photon:
      do iev=1,nev
        yexp1(iev) = exp_sel( events(iev)%ix, events(iev)%iy, events(iev)%ipt )
        ybkg1(iev) = bkg_sel( events(iev)%ix, events(iev)%iy, events(iev)%ipt )
      enddo
      ! calculate exposure and background values for output list,
      ! average over photons => same for all sources of cluster 
      do i=1,num_exp
        do j=1,num_image(i)
          ipoint=ip(i,j)
          sumbkg=0.0D0
          sumexp=0.0D0
          evcount=0
          do iev=1,nev
            if (events(iev)%ipt == ipoint) then
              evcount=evcount+1
              sumbkg=sumbkg + ybkg1(iev)
              sumexp=sumexp + yexp1(iev)
            endif
           enddo 
          if (evcount > 0) then
            clulist(ipoint,:)%exp_map = real(sumexp/real(evcount, kind=real64),real32)
            clulist(ipoint,:)%bg_map =  real(sumbkg/real(evcount, kind=real64),real32)

          else
            ! this is relevant for forced photometry case only:
             if (mask%npix > 0) then
               sumexp=sum(exp_sel(:,:,ipoint)*min(mask%mask(:,:,ipoint),1))
               sumbkg=sum(bkg_sel(:,:,ipoint)*min(mask%mask(:,:,ipoint),1))
               clulist(ipoint,:)%exp_map = real(sumexp/real(sum(min(mask%mask(:,:,ipoint),1)), kind=real64), kind=real32)
               clulist(ipoint,:)%bg_map =  real(sumbkg/real(sum(min(mask%mask(:,:,ipoint),1)), kind=real64), kind=real32)

            else

               clulist(ipoint,:)%exp_map = 0.0
               clulist(ipoint,:)%bg_map =  0.0
            endif
          endif
        enddo
      enddo
      

     !**** PHMODEL

      where (mask%mask > 0)  mask1=.true.
      where (mask%mask == 2) mask2=.true.       


      ff_save(:,:,:,:)=0.0
      nsrc_old=nsrc
    else if (nsrc .ne. nsrc_old) then
      if (allocated(ff_save)) deallocate(ff_save)
      ! 4th dim = nim (GL 20/09/2017)
      allocate(ff_save(mask%nx,mask%ny,nsrc,nim ))
      ff_save(:,:,:,:)=0.0
      nsrc_old=nsrc
    endif ! first
 

    if (first .or. newcal) then
      scalepsf(1:max_ima,1:max_fit) = .false.
    else
!   check which parameters changed since last call: 
      do n=1,nsrc
        if ((x0(n) .ne. x0_save(n)) .or. (y0(n) .ne. y0_save(n)) .or. (sig(n) .ne. sig_save(n))) then 
           scalepsf(1:max_ima,n) = .false.
           !test output 24/09/2021:
           !if (sig(n) > 0.0 ) write(*,*) "nsrc,n,x,y,sig:", nsrc,n,x0(n),y0(n),sig(n)
!          count_model=count_model+1
        else
          do i=1,num_exp
            do j=1,num_image(i)
              ipoint=ip(i,j)
              if (scts(ipoint,n) .eq. 0.0) then
                scts_scale(ipoint,n) = 0.0
                scalepsf(ipoint,n) = .true.
              else if (scts(ipoint,n) .eq. scts_save(ipoint,n)) then
                scts_scale(ipoint,n) = 1.0
                scalepsf(ipoint,n) = .true.
              else if (abs(scts_save(ipoint,n)) .gt. small) then
                scts_scale(ipoint,n) =  scts(ipoint,n) / scts_save(ipoint,n)
                scalepsf(ipoint,n) = .true.
              else
                scts_scale(ipoint,n) = 0.0
                scalepsf(ipoint,n) = .false.
!                count_model=count_model+1
              endif
            enddo
          enddo
        endif
      enddo
    endif

    lf=1
    ipix=1
    yfit1(:)=0.
    yfit2(:)=0.
!*    yfit3(:)=0.
    yfit0=0.0


   ! ***PHMODEL
   ! calculate yfit1 for each photon 
    fnewcal=newcal
  

    
   do isrc=1,nsrc
      ! source offsets
     ixs=ix0(isrc)
     iys=iy0(isrc)
     ix_mask=ixs-mask%ix_min
     iy_mask=iys-mask%iy_min
     ! offset between PSF pixels and mask pixels:
     xoff=ix_mask-isize/2
     yoff=iy_mask-isize/2
     do iev=1,nev
       ipoint=events(iev)%ipt
       if (scalepsf(ipoint,isrc) .and. (scts_scale(ipoint,isrc) .ne. 0.0)) then
         if (scts_scale(ipoint,isrc) .ne. 1.0) then
           ysave1(iev,isrc)=ysave1(iev,isrc)*scts_scale(ipoint,isrc)
         endif
         ! add source(isrc):
           yfit1(iev) = yfit1(iev)+ysave1(iev,isrc)
         !if (verbose_flag2) then
          ! write(*,*) 'scalepsf:',isrc,ipoint,scts(ipoint,isrc),x0(isrc),y0(isrc),ysave1(iev,isrc),yfit1(iev)
         !endif
       else
         ! get new PSF image for event(iev)
          !  set to zero (12/12/2017)
          !  ysave1(:,isrc)=0.0

!         call phpsf(par,image,events,newcal,iev,isrc,ix0(isrc),iy0(isrc),x0(isrc),y0(isrc),sig(isrc),clulist(0,isrc)%ecutrad,cluPSF,phima,isize,status) 
         call phpsf(par,events,newcal,iev,isrc,ix0(isrc),iy0(isrc),x0(isrc),y0(isrc),sig(isrc),clulist(0,isrc)%ecutrad,cluPSF,phima,isize,status) 

         ! set negative PSF values to 0.D0 (GL 17/06/2021):
         phima=max(phima,0.D0)
 
         
          ! applying the offset is wrong:
!         ix=events(iev)%ix+xoff
!         iy=events(iev)%iy+yoff
         !GL 24/11/2017
         ix=events(iev)%ix-xoff
         iy=events(iev)%iy-yoff
         !**** if (verbose_flag2) write(*,*) 'ix,iy=',ix,iy
         if (ix > 0 .and. ix <= isize .and. iy > 0 .and. iy <= isize) then
           ysave1(iev,isrc)=scts(ipoint,isrc)*phima(ix,iy)
           yfit1(iev) = yfit1(iev)+ysave1(iev,isrc)
         else
           !add this to set ysave,yfit to 0.0, if ix,iy are outside of frame (GL 24/11/2017)
           ysave1(iev,isrc)=0.0
           !yfit1(iev) = yfit1(iev)+ysave1(iev,isrc)           
         endif

       endif
     enddo
   enddo
   ! apply exposure and background:
   yfit1 = yfit1*yexp1+ybkg1

   ! ***PHMODEL
   ! now calculate yfit2 from source averaged PSF
    do i=1,num_exp
        do j=1,num_image(i)
          ipoint=ip(i,j)

          !correct 31-jul-2013:
!         modelima(:,:,:)=0.0
          modelima(:,:,ipoint)=0.D0
          do n=1,nsrc

            !xs=nint(x0(n))
            !ys=nint(y0(n))
            ! (L 26/04/2021:)
            ixs=ix0(n)
            iys=iy0(n)            
            ix_mask=ixs-mask%ix_min
            iy_mask=iys-mask%iy_min

            if (scalepsf(ipoint,n) .and. (scts_scale(ipoint,n) .ne. 0.D0)) then

              if (scts_scale(ipoint,n) .ne. 1.D0) then
                ff_save(:,:,n,ipoint)=ff_save(:,:,n,ipoint)*scts_scale(ipoint,n)
              endif
              modelima(:,:,ipoint)=modelima(:,:,ipoint)+ff_save(:,:,n,ipoint)
!               write(*,*) 'done scale: ',scts_scale(ipoint,n),maxval(ff_save(:,:,n,ipoint))
            else
               call ftpsf(par,image,fnewcal,i,ipoint,n,ix0(n),iy0(n),x0(n),y0(n),sig(n),clulist(0,n)%ecutrad,cluPSF,fima,isize,status)
               ! copy fima into modelima:
               ! 17/06/2021  set  negative values in fima to 0.D0:
               fima=max(fima,0.D0)
               
              ixf1=max(1,isize/2-ix_mask+1)
              iyf1=max(1,isize/2-iy_mask+1)
              ixf2=min(isize,isize/2-ix_mask+mask%nx)
              iyf2=min(isize,isize/2-iy_mask+mask%ny)
              ixm1=max(1,ix_mask-isize/2+1)
              iym1=max(1,iy_mask-isize/2+1)
              ixm2=min(mask%nx,ix_mask+isize/2)
              iym2=min(mask%ny,iy_mask+isize/2)
              ff_save(:,:,n,ipoint)=0.0
!              write(*,*) 'done ftpsf: ',n,size(fima,1),size(fima,2),maxval(fima(ixf1:ixf2,iyf1:iyf2)),ixf1,ixf2,iyf1,iyf2
              ff_save(ixm1:ixm2,iym1:iym2,n,ipoint)=scts(ipoint,n)*fima(ixf1:ixf2,iyf1:iyf2)
!              modelima(:,:,ipoint)=modelima(:,:,ipoint)+ff_save(:,:,n,ipoint)
              ! this is slighty faster:
              modelima(ixm1:ixm2,iym1:iym2,ipoint)=modelima(ixm1:ixm2,iym1:iym2,ipoint)+ff_save(ixm1:ixm2,iym1:iym2,n,ipoint)

            endif

          end do

      end do
        ! end loop over num_image      
    end do  
 
    ! *** PHMODEL
    ! New code to reshape modelima to yfit arrays:

    if (par%expima_flag) then
      countsima = modelima*exp_sel+bkg_sel
    else
      countsima = modelima+bkg_sel
    endif


    do ipoint=1,nim
      yfit2(ipoint) = sum(countsima(:,:,ipoint),mask%mask(:,:,ipoint) > 0)
!    write(*,*) 'calculate yfit2:',yfit2(ipoint), minval(countsima(:,:,1)),maxval(countsima(:,:,1))
    enddo
    yfit0=sum(yfit2)
!    write(*,*) 'done calculate yfit2:',minval(yfit2),maxval(yfit2),sum(yfit2)

        



    !--------

    first = .false.
    x0_save=x0
    y0_save=y0
    sig_save=sig
    scts_save=scts


  end subroutine phmodel

!------------------------------------------------------------------------------




  !----------------------------------------------------------------------
  !* removed yfit3
  subroutine mlfit(par,image,expimage,bkgimage,clulist,cluPSF,events,mask,nterms,a,deltaa,sigmaa,iordr,flamda, &
                   newcal,yfit1,yfit2,like0,lik_arr,nsrc,nfel,modelima,ierr)
  !----------------------------------------------------------------------

    type(ermldet_param_type), intent(in)     :: par
    type(i4cube_type)                        :: image
    type(r4cube_type)                        :: expimage,bkgimage
!    type(srcrow_type),dimension(:,:),pointer :: clulist
    type(srcrow_type),dimension(:,:),allocatable :: clulist   
    type(FPSF)                               :: cluPSF 
    type(events_type),pointer,dimension(:)    :: events
    type(mask_type), intent(in)              :: mask   
    integer(kind=int32)                      :: nterms
    real(kind=real64), dimension(:)          :: a,deltaa,sigmaa   
    integer(kind=int32), dimension(:)        :: iordr
    real(kind=real64)                        :: flamda
    real(kind=real64)                        :: like0
    real(kind=real64) , dimension(:)         :: yfit1,yfit2
    real(kind=real64), dimension(:)          :: lik_arr
    integer(kind=int32)                      :: nsrc
    real(kind=real64), dimension(:,:,:),pointer  :: modelima
    logical(kind=log32)                      :: newcal
    integer(kind=int32)                      :: ierr

    integer(kind=int32)                             :: nelem
    real(kind=real64), dimension(:), allocatable    ::  b, beta
    real(kind=real64), dimension(:,:), allocatable  ::  array,alpha
    integer                        :: i,j,k,status
    integer                        :: nfel!ntermf,
    real(kind=real64)              :: det
    real(kind=real64), dimension(:),allocatable :: ytemp1
    real(kind=real64)              :: like1,liketemp!,like2,like3, like12
    real(kind=real64)              :: derv0(max_par)
    real(kind=real64)              :: yfit0,ytemp0,ecut2
    real(kind=real64), allocatable :: derv1(:,:)
    real(kind=real64), allocatable :: derv2(:),ytemp2(:)
    ! change to real64 (04/04/2014):
    real(kind=real64) :: aj,delta,delta2!,ak,deltak
!    real(kind=real64)              :: ct1,ct2
!    integer(kind=int32)            :: iloop
    logical, dimension(:), allocatable   :: limzero
    
    real, parameter :: small = 1e-6
!    real :: test1,test2

    character(len=*), parameter :: rtname='MLFIT'

    ierr=0

 ! deallocate, in case previous call was ended by 'exit' :
    if (allocated(derv1)) then
      deallocate(derv2)
!*      deallocate(derv3)
      deallocate(b)
      deallocate(beta)
      deallocate(array)
      deallocate(alpha)
      deallocate(ytemp1)
      deallocate(ytemp2)
    endif
    allocate(derv1(nterms,mask%nev))

    allocate(derv2(nterms))
!*    allocate(derv3(nterms,mask%npix))

    nelem=size(a)
    allocate(b(nelem))
    allocate(beta(nelem))
    allocate(array(nelem,nelem))
    allocate(alpha(nelem,nelem))

    nelem=size(yfit1)
    allocate(ytemp1(nelem)) 

    nelem=size(yfit2)
    allocate(ytemp2(nelem))

    allocate(limzero(nterms))

    limzero(1:nterms)=.true.
    if (par%fitpos_flag) limzero(1:2)=.false.

    ecut2 = clulist(0,1)%ecutrad * clulist(0,1)%ecutrad

!    ntermf=nterms+nfel
    if (mask%nev-nterms < 1)  then
      ! Change from ierr=3 to Warning (ierr=1, GL 09/12/2016)
      !** Need to review how to treat this 
      ierr=1
      call wrchatty(rtname,'Warning: degrees of freedom less or equal 0')
    endif


    do j=1,nterms
      aj=a(j)

      delta=deltaa(j)
!      delta2=2.0*delta
      !test:

      a(j)=aj+delta     


      !*   call funct(par,image,expimage,bkgimage,mask,clulist,cluPSF,a,iordr,nsrc,nterms+nfel,newcal,yfit0,yfit1,yfit2,yfit3,modelima,status)
      ! calculate y(a+delta):
!      write(*,*) "call funct (+):",a(1:nterms)
      call funct(par,image,events,expimage,bkgimage,mask,clulist,cluPSF,a,iordr,nsrc,nterms+nfel,newcal,yfit0,yfit1,yfit2,modelima,status)
      call like(events,mask,yfit1,yfit2,like1,lik_arr,status)
      if (verbose_flag3 .and. j < 4) then
        write(*,*) 'j,a(j),yfit0: ',j,a(j),yfit0
        write(*,*) 'j,a(j),yfit1: ',j,a(j),yfit1
      endif
      ! check bounds and set lower value for derivative calculation:
      if (limzero(j)) then
         a(j)=max(0.D0,aj-delta)
         delta2=delta+aj-a(j)
      else   
         a(j)=aj-delta
         delta2=2.D0*delta
      endif
      


      !*      call funct(par,image,expimage,bkgimage,mask,clulist,cluPSF,a,iordr,nsrc,nterms+nfel,newcal,ytemp0,ytemp1,ytemp2,ytemp3,modelima,status)
      ! calculate y(a-delta):
!      write(*,*) "call funct (-):",a(1:nterms)
      call funct(par,image,events,expimage,bkgimage,mask,clulist,cluPSF,a,iordr,nsrc,nterms+nfel,newcal,ytemp0,ytemp1,ytemp2,modelima,status)
      call like(events,mask,ytemp1,ytemp2,liketemp,lik_arr,status)     
      if (verbose_flag3 .and. j < 4) then
        write(*,*) 'j,a(j),ytemp0: ',j,a(j),ytemp0
        write(*,*) 'j,a(j),ytemp1: ',j,a(j),ytemp1
      endif

      
!      derv1(j,1:mask%nev)=(yfit1(1:mask%nev)-ytemp1(1:mask%nev))/delta2

!      if (j == 1) write(*,*) 'yfit1: ' ,yfit1(1:mask%nev)
!      if (j == 1) write(*,*) 'ytemp1: ',ytemp1(1:mask%nev)
!      if (j == 1) write(*,*) 'diff: ',yfit1(1:mask%nev)-ytemp1(1:mask%nev)

      derv1(j,:)=(yfit1-ytemp1)/delta2
!*      derv3(j,:)=(yfit3-ytemp3)/delta2
      derv0(j)=(yfit0-ytemp0)/delta2
      ! set a(j) back to original value:
      a(j)=aj
      if (verbose_flag3) write(*,*) 'mlfit:derv: ',j,derv1(j,1:mask%nev)
    end do



  
    beta(1:nterms)=0.
    alpha(1:nterms,1:nterms)=0.          


    


!*     call funct(par,image,expimage,bkgimage,mask,clulist,cluPSF,a,iordr,nsrc,nterms+nfel,newcal,yfit0,yfit1,yfit2,yfit3,modelima,status)
    call funct(par,image,events,expimage,bkgimage,mask,clulist,cluPSF,a,iordr,nsrc,nterms+nfel,newcal,yfit0,yfit1,yfit2,modelima,status)


     do i=1,mask%nev
       if (yfit1(i) .ne.  0.0) then
         do j=1,nterms
           ! GL 24/11/2017 convert %n to real
            !
            beta(j) = beta(j)+real(events(i)%n)/yfit1(i)*derv1(j,i)

           ! OLD : use yfit1 (from non-zero pixels)
            do k=1,j
              ! GL 29/09/2021: Test, go back to v1.46 equation: 
              !v1.46
              alpha(j,k)=alpha(j,k)+derv1(j,i)*derv1(k,i)/yfit1(i)
              !v1.47:
              !alpha(j,k)=alpha(j,k)+derv1(j,i)*derv1(k,i)/(yfit1(i)*yfit1(i))
           end do
         end do
       endif
     end do




     do j=1,nterms
       beta(j)=beta(j)-derv0(j)
     end do
    do j=1,nterms
      do k=1,j
        alpha(k,j)=alpha(j,k)
      end do
    end do

    call funct(par,image,events,expimage,bkgimage,mask,clulist,cluPSF,a,iordr,nsrc,nterms+nfel,newcal,yfit0,yfit1,yfit2,modelima,status)

    call like(events,mask,yfit1,yfit2,like1,lik_arr,status)


    do
      do j=1,nterms
        do k=1,nterms
          if(j.eq.k)then
            array(j,k)=1.0d0
          else 
            array(j,k)=alpha(j,k)/dsqrt(dabs(alpha(j,j))+small) &
                                 /dsqrt(dabs(alpha(k,k))+small)
          endif
        end do
        array(j,j)=1.0d0+flamda
      end do

      
      call invert_matrix(array,nterms,det,status)


      b = a
     

      do j=1,nterms
        do k=1,nterms
          b(j)=b(j)+beta(k)*array(j,k)/dsqrt(dabs(alpha(j,j))+small) &
                                      /dsqrt(dabs(alpha(k,k))+small)
        if (verbose_flag3)  write(*,*) 'beta: ',j,k, beta(k), array(j,k),alpha(j,j),alpha(k,k),beta(k)*array(j,k)/dsqrt(dabs(alpha(j,j))+small)/dsqrt(dabs(alpha(k,k))+small),b(j)
        end do 
      end do

      
      if (verbose_flag3) then
         write(*,*) 'mlfit: a=',like0,like1,a(1:nterms)
         write(*,*) 'mlfit: b=',like0,like1,b(1:nterms)
         write(*,*)
      endif

      call funct(par,image,events,expimage,bkgimage,mask,clulist,cluPSF,b,iordr,nsrc,nterms+nfel,newcal,yfit0,yfit1,yfit2,modelima,status)

      call like(events,mask,yfit1,yfit2,like0,lik_arr,status)


      if (like0 < like1) then

         flamda=max(flamda/5.D0,small)
 
        a(1:nterms)=b(1:nterms)
        if (verbose_flag2) write(*,*) '----new a: a=',like0,a(1:nterms)
        do j=1,nterms
          sigmaa(j)=dsqrt(dabs(array(j,j))/(dabs(alpha(j,j))+small))
        end do       
        exit
      else
        ! GL 08/12/2016:
        ! exit before increasing flamda to avoid overflow in flamda (mlfit is called in loop iter=1,max_it) 
        if (flamda.ge.1e10) exit
        flamda=5.*flamda
!         write(*,*) "inc. lamda:",flamda       
!        if (flamda.ge.1e10) exit
      end if
    end do



    deallocate(derv1)
    deallocate(derv2)
!*    deallocate(derv3)

    deallocate(b)
    deallocate(beta)
    deallocate(array)
    deallocate(alpha)
    deallocate(ytemp1)
    deallocate(ytemp2)
    deallocate(limzero)
!*    deallocate(ytemp3)

 

  end subroutine mlfit


  subroutine  setparlimits(array,n,vallo,valhi,ierr)

  !---------------------------------------------------------------------------------
  ! New routine to keep fit parameters within their given boundaries.
  ! Checks paramter boundaries and sets paramter array to values within boundaries.
  !--------------------------------------------------------------------------------


    real(kind=real64),dimension(:) ::        array
    integer(kind=int32)            ::        n    
!    real(kind=real64)              ::        cage,value
    real(kind=real32)              ::        vallo,valhi
    integer(kind=int32)            ::        ierr

    ierr=0
    
    if (array(n) < vallo) then
      array(n) = vallo
    else if (array(n) > valhi) then
      array(n) = valhi
    endif


  end subroutine  setparlimits

  

!--------------------------------------------------------------------------------

  function cage(array,n,vallo,valhi)

  !---------------------------------------------------------------
  ! cage fit parameters within allowed ranges
  !---------------------------------------------------------------

    real(kind=real64),dimension(:) ::        array
    real(kind=real64)              ::        cage,value
    real(kind=real32)              ::        vallo,valhi
    integer(kind=int32)            ::        ierr,n 

    ierr=0

    value=array(n)

    if (value.lt.vallo) then
      array(n)=vallo
    else if (value.gt.valhi) then
      array(n)=valhi
    endif
    cage=array(n)

  end function cage    
      
!--------------------------------------------------------------------------------

   subroutine cagexy (x1,y1,x0,y0,rcut2)

  !---------------------------------------------------------------
  ! cage positional parameters within radius^2 rcut2
  !---------------------------------------------------------------


    real(kind=real64) :: x1,y1
    real(kind=real32) :: x0,y0,rcut2  
    real(kind=real64) :: ds

    ds=sqrt(((x1-x0)**2+(y1-y0)**2)/rcut2)
    if (ds.gt.1.D0) then
      x1 = x0+(x1-x0)/ds
      y1 = y0+(y1-y0)/ds
    end if
 
  end subroutine cagexy

!----------------------------------------------------------------------
     
!--------------------------------------------------------------------------------

  subroutine cagemask (x1,y1,mask,ierr)

  !---------------------------------------------------------------
  ! Cage positional parameters within region of multiple sources defined
  ! in a structure of type mask_type regions.  If source is outside 
  ! of any source region, move it into the closest source region.
  ! 
  !---------------------------------------------------------------


    real(kind=real64)                      :: x1,y1
    type(mask_type), intent(in)            :: mask  
    integer(kind=int32),  intent(out)      :: ierr
!    real(kind=real32)                      :: x0,y0,rcut2  
    real(kind=real64),dimension(max_fit)   :: ds    
    integer(kind=int32)                    :: i,j

    character(len=*), parameter             :: rtname='CAGEMASK'


    ierr=0
    
    do i=1,mask%nsrc 
      ds(i)=sqrt(((x1-mask%srcreg_x(i))**2+(y1-mask%srcreg_y(i))**2)/(mask%srcreg_r(i)**2))
    enddo


    if (minval(ds(1:mask%nsrc)).gt.1.D0) then
      j=minloc(ds(1:mask%nsrc),dim=1)
      x1 = mask%srcreg_x(j)+(x1-mask%srcreg_x(j))/ds(j)
      y1 = mask%srcreg_y(j)+(y1-mask%srcreg_y(j))/ds(j)
    end if
 
  end subroutine cagemask

!----------------------------------------------------------------------


  subroutine like(events,mask,yfit1,yfit2,lik,lik_arr,ierr)

  !-------------------------------------------------------
  ! Calculate  likelihood per image lik_arr and total
  ! likelihood lik.
  ! Only values >= 0 are permitted in  yfit1 and yfit2.
  ! For performance reasons event list 'event' should only
  ! contain events/pixels with events%n > 0  .
  !------------------------------------------------------

    type(events_type),pointer,dimension(:), intent(in) :: events
    type(mask_type), intent(in)                       :: mask   
    real(kind=real64), dimension(:), intent(in)                    :: yfit1
    real(kind=real64), dimension(:), intent(in)                    :: yfit2

    real(kind=real64), dimension(:), intent(out)       :: lik_arr
    real(kind=real64), intent(out)                     :: lik
    integer(kind=int32),  intent(out)                  :: ierr
    integer(kind=int32)                                :: k,ipoint

!    integer(kind=int32)                                ::  tempsum
    ! for test output:
!    real(kind=real64), dimension(:),allocatable        :: liketemp   
    
    character(len=*), parameter                        :: rtname='LIKE'

    ierr = 0


    lik_arr(1:nim)=0.D0
    pcounter=pcounter+1
    do k=1,mask%nev
      ipoint=events(k)%ipt
       ! temporary fix to negative shapelib PSF values:'
      ! lik_arr(ipoint)=lik_arr(ipoint)+events(k)%n*log(max(yfit1(k),1.D-30))
      ! change lower limit to 1.D-8:
      lik_arr(ipoint)=lik_arr(ipoint)+events(k)%n*log(max(yfit1(k),1.D-8))
!      liketemp(k)=events(k)%n*log(max(yfit1(k),1.D-4))
    end do

    lik_arr(1:nim)=yfit2(1:nim)-lik_arr(1:nim)
    lik=0.D0 
    do ipoint=1,nim
      lik=lik+lik_arr(ipoint)
    end do 

    
   end subroutine like




!--------------------------------------------------------------------------------

  subroutine invert_matrix(arr,norder,det,ierr)   

  !----------------------------------------------------------
  ! Matrix inversion.
  ! Based on subroutine mati used by XMMSAS emldetect.
  ! Adapted to Fortran 90 (replaced labels, arithmetic-if, 
  ! and goto statements. (GL 12/09/2011)
  ! Replaces norder*norder elements of matrix arr by
  ! inverse matrix. 
  !----------------------------------------------------------
  

    integer(kind=int32)                :: ierr,k,norder,i,j,l
    real(kind=real64), dimension(:,:)  :: arr
    real(kind=real64)                  :: amax,save,det
    integer(kind=int32)                :: ik(max_par),jk(max_par),imax,jmax
    logical                            :: matzero
    character(len=*), parameter :: rtname='INVERT_MATRIX'

!    real(kind=real32)                  :: t1,t2
!    integer(kind=int32)                :: icount
!    real(kind=real64), dimension(:,:),allocatable  :: arrsave

    ierr=0
    det=1.D0


    matzero=.false.



    kloop: do k=1,norder


    ! find largest element arr(i,j) in rest of matrix.   
      amax=0.D0
      jmaxloop: do 
        imaxloop: do
          do i=k,norder
            do j=k,norder
              if (abs(arr(i,j)) >= abs(amax)) then
                amax=arr(i,j)
                ik(k)=i
                jk(k)=j
              endif
            enddo
          enddo
          if(amax == 0.D0) then
            det = 0.D0
            matzero = .true.
            exit kloop
          endif


         ! interchange rows and columns to put amax in arr(k,k)
          imax=ik(k)
          if (imax >= k) exit imaxloop
        end do imaxloop 
        if(imax > k) then
          do j=1,norder
            save=arr(k,j)
            arr(k,j)=arr(imax,j)
            arr(imax,j)=-save
          enddo
        endif
        jmax=jk(k)
        if (jmax >= k) exit jmaxloop
      end do jmaxloop
      if(jmax > k) then
        do i=1,norder
          save=arr(i,k)
          arr(i,k)=arr(i,jmax)
          arr(i,jmax)=-save
        enddo
      endif

      ! accumulate elements of inverse matrix
      do i=1,norder
        if (i /= k) arr(i,k) = -arr(i,k)/amax
      enddo
      do i=1,norder
        do j=1,norder
          if (i /=k ) then
            if(j /= k) then
              arr(i,j)=arr(i,j)+arr(i,k)*arr(k,j)
            endif
          endif
        enddo
      enddo
      do j=1,norder
        if(j /= k) arr(k,j)=arr(k,j)/amax
      enddo
      arr(k,k)=1./amax
      det=det*amax
    end do kloop

    if (.not. matzero) then
      ! restore ordering of matrix
      do l=1,norder
        k=norder-l+1
        j=ik(k)
        if (j > k) then
          do i=1,norder
            save=arr(i,k)
            arr(i,k)=-arr(i,j)
            arr(i,j)=save
          enddo     
        endif     
        i=jk(k)
        if (i > k) then
          do j=1,norder
            save=arr(k,j)
            arr(k,j)=-arr(i,j)
            arr(i,j)=save
          enddo
        endif
      enddo
    endif


  end subroutine invert_matrix



 !-----------------------------------------------------------------------------------
  subroutine ordr(nterms,nfix,iordr)
  !----------------------------------------------------------------------

    integer(kind=int32)                        :: i,nfix,nnfix,nterms
    integer(kind=int32), dimension(max_par)    :: iordr,io



    if(iordr(1).eq.0)then
      ! reset iordr
      do i=1,max_par
        iordr(i)=i
      end do
    endif

    do i=1,max_par
      if(iordr(i).eq.nfix) exit
      io(i)=iordr(i)
    end do

    nnfix=i

    do i=nnfix,nterms-1
      io(i)=iordr(i+1)
    end do

    io(nterms)=nfix

    if(nterms.ne.max_par) then
      do i=nterms+1,max_par
        io(i)=iordr(i)
      end do
    end if 

    iordr(1:max_par)=io(1:max_par)

  end subroutine ordr

!-----------------------------------------------------------------------------------


  subroutine pordr(nterms,a,b,iordr,mode)

  !----------------------------------------------------------------------

    real(kind=real64),  dimension(max_par) ::  a,b
    integer(kind=int32)                    ::  i,mode,nterms
    integer(kind=int32), dimension(max_par)::  iordr

    if(mode.eq.1) then
      do i=1,nterms
        b(i)=a(iordr(i))
      enddo
    else

! andere richtung
      do i=1,nterms
        a(iordr(i))=b(i)
      enddo
    end if

  end subroutine pordr

!-----------------------------------------------------------------------------------


  subroutine pordr3(nterms,arr1,arr2,arr3,iordr,mode)

  !----------------------------------------------------------------------
  ! Sort 3 arrays of dimension(max_par) according to index array iordr.
  ! Sorting is performed in place,  
  ! Only first (1:nterms) array elements are re-ordered.
  ! mode=1    : arr(1:nterms) = arr(iordr(1:nterms))
  ! mode != 1 : arr(iordr(1:nterms)) = arr(1:nterms)
  !----------------------------------------------------------------------

    real(kind=real64),  dimension(max_par)  ::  arr1,arr2,arr3
    integer(kind=int32)                     ::  mode,nterms
    integer(kind=int32), dimension(max_par) ::  iordr


    if(mode.eq.1) then
      arr1(1:nterms) = arr1(iordr(1:nterms))
      arr2(1:nterms) = arr2(iordr(1:nterms))
      arr3(1:nterms) = arr3(iordr(1:nterms))
    else
!     opposite direction:
      arr1(iordr(1:nterms)) = arr1(1:nterms)
      arr2(iordr(1:nterms)) = arr2(1:nterms)
      arr3(iordr(1:nterms)) = arr3(1:nterms)
    end if

  end subroutine pordr3

!-----------------------------------------------------------------------------------


!*
  subroutine paramerr(par,image,expimage,bkgimage,mask,events,clulist,cluPSF,a,deltae,nsrc,npsou,yfit1,modelima,ierr)

  !-------------------------------------------------------------------------------------
  ! Determine parameter errors based on Cash statistic.
  ! The 68% error margins are calculated by finding the parameter values, 
  ! where L(p_best-err_lo)=L(p_best+1.0),  L(p_best+err_hi)=L(p_best+1.0). 
  ! This is valid only for the assumption of 1 interesting fit parameter.  
  ! No re-fitting of the remaining parameters is performed. 
  ! -----------------------------------------------------------------------------------

    type(ermldet_param_type), intent(in)     :: par
    type(i4cube_type)                        :: image
    type(r4cube_type)                        :: expimage,bkgimage  
    type(mask_type)                          :: mask   
!    type(srcrow_type),dimension(:,:),pointer :: clulist
    type(srcrow_type),dimension(:,:),allocatable :: clulist    
    type(FPSF)                               :: cluPSF 
    integer(kind=int32)                      :: nsrc, npsou
    type(events_type),pointer,dimension(:), intent(in) :: events
    real(kind=real64), dimension(:)                   :: yfit1
    real(kind=real64),dimension(:)                    :: a,deltae
    real(kind=real64),dimension(:,:,:),pointer        :: modelima
    integer(kind=int32),  intent(out)                 :: ierr

    real(kind=real64),dimension(:),allocatable        :: asave    
    integer(kind=int32)                               :: np
    logical(kind=log32)                               :: newcal
    real(kind=real64), dimension(2)                   :: par_err   
    real(kind=real64), dimension(max_ima)             :: yfit2
    real(kind=real64)                                 :: yfit0
    real(kind=real64), dimension(max_ima)             :: lik_arr
    real(kind=real64)                                 :: lik, like0, likelim,liketest
    integer(kind=int32), dimension(max_par)           :: iordr   
    integer(kind=int32)                               :: isrc,ispar,ipar,i,idir!,ipoint
    integer(kind=int32),save                          :: icount=0
    logical                                           :: zerobound
    real(kind=real64)                                 :: pix2arcsec    

    
    real(kind=real64)                                 :: dx,fh,fl,swap,xh,xl,ddir!,del, f
    real(kind=real64)                                 :: xerr, mean_err, facc=0.02
    integer(kind=int32)                               :: nit,iima, status, maxit=100!j, 
    
    character(len=*), parameter                       :: rtname='PARAMERR'


!     write(*,*) "START paramerr:"

    
! count sources for test output
    icount=icount+1    
    ierr=0
    newcal=.false.
    np=nsrc*npsou

    pix2arcsec=(abs(image%wcshdr(1)%cdelt1)+abs(image%wcshdr(1)%cdelt2))*1800.0
    
    par_err(1:2)=0.D0
    if (allocated(asave)) deallocate(asave) 
    allocate(asave(np))
    ! regular order of a: 
    iordr(1)=0
    call ordr(np,np,iordr)

    ! save input parameter set and calculate likelihood with initial parameter set
    asave(1:np)=a(1:np)

!* rm yfit3
    call funct(par,image,events,expimage,bkgimage,mask,clulist,cluPSF,a,iordr,nsrc,np,newcal,yfit0,yfit1,yfit2,modelima,status)
    
    call like(events,mask,yfit1,yfit2,like0,lik_arr,status)


    ! apply factor 0.5, since like0 = C-stat/2.0

    likelim=like0+0.5D0

    fh=0.D0
    fl=0.D0

    ipar=0
    do isrc=1,nsrc
      iima=0
      do ispar=1,npsou
        ipar=ipar+1
        if ((par%fitpos_flag .or. ispar > 2)  .and. &
            ((par%fitext_flag .and. clulist(0,isrc)%ext_like > 0.0)  .or. ispar .ne. 3 )) then 
          do idir=-1,1,2
            ! for extent or rate parameter, check whether param=0.0 is within the error boundary:   
            zerobound = .false.  
            if (ispar .ge. 3 .and. idir .eq. -1) then
              a(ipar)=0.0
              call funct(par,image,events,expimage,bkgimage,mask,clulist,cluPSF,a,iordr,nsrc,np,newcal,yfit0,yfit1,yfit2,modelima,status)
              call like(events,mask,yfit1,yfit2,liketest,lik_arr,status)
              if (liketest < likelim) then
                zerobound  = .true.
                par_err(1) = asave(ipar)
              endif   
            endif 
            if (.not. zerobound) then
              a(ipar)=asave(ipar)
              fl=like0
              fh=like0
              ddir=real(idir,kind(0.D0))    
              ! find  xl, xh bracketing root of likelihood function,
              ! search with increasing step sizes:
              xl=a(ipar)
              xh=a(ipar)
              nit=0
              do
                if (nit >= maxit .or. fh >= likelim) exit
                nit=nit+1
                xl=xh
                xh=xh+ddir*real(nit,kind(0.D0))*deltae(ipar)
                 fl=fh
                a(ipar)=xh     
                call funct(par,image,events,expimage,bkgimage,mask,clulist,cluPSF,a,iordr,nsrc,np,newcal,yfit0,yfit1,yfit2,modelima,status)
                call like(events,mask,yfit1,yfit2,fh,lik_arr,status)
               enddo
              if (nit < maxit) then
                ! false position method (num. recipes),
                ! use convergence criterion in likelihood (facc):
                dx=xh-xl  
                do i=1,maxit
                  xerr = xl+dx*(fl-likelim)/(fl-fh)
                  a(ipar)=xerr 
                  call funct(par,image,events,expimage,bkgimage,mask,clulist,cluPSF,a,iordr,nsrc,np,newcal,yfit0,yfit1,yfit2,modelima,status)
                  call like(events,mask,yfit1,yfit2,lik,lik_arr,status)                              
                  
                  if (lik <  likelim) then
                    xl=xerr
                    fl=lik
                  else
                    xh=xerr
                    fh=lik
                  endif    
                  dx=xh-xl  
                  if (abs(lik-likelim) < facc) then 
                    if (idir == -1 .and. xerr > 0.D0) par_err(1) = abs((asave(ipar)-xerr))
                    if (idir  == 1) par_err(2) = abs((xerr-asave(ipar)))
                     exit
                  endif
                  if (i == maxit) then
                    ierr = 2
                    if (idir == -1) par_err(1) = -9999.9
                    if (idir  == 1) par_err(2) = -9999.9
                  endif
                enddo ! maxit
              else
                ierr = 2
                if (idir == -1) par_err(1) = -9999.9
                if (idir  == 1) par_err(2) = -9999.9
              endif
            endif ! (.not. zerobound) 
          enddo ! idir
         
          if (par_err(1) > 0.0) then
             if (par_err(2) > 0.0) then
               mean_err =  (par_err(1) +  par_err(2))/2.D0
             else
                mean_err = par_err(1)
             endif
          else
             if (par_err(2) > 0.0) then
               mean_err = par_err(2)
            else
               mean_err = -9999.9 
            endif   
          endif
         
!          write(*,*) "mean_err:",ipar,a(ipar),mean_err,par_err(1), par_err(2)
          if (ispar == 1 .and. par%fitpos_flag) then
             clulist(:,isrc)%x_ima_err    = real(mean_err, kind=real32)
             clulist(:,isrc)%x_ima_lowerr = real(par_err(1), kind=real32)
             clulist(:,isrc)%x_ima_uperr  = real(par_err(2), kind=real32)             
          endif
          
          if (ispar == 2 .and. par%fitpos_flag) then
             clulist(:,isrc)%y_ima_err =    real(mean_err, kind=real32)
             clulist(:,isrc)%y_ima_lowerr = real(par_err(1), kind=real32)
             clulist(:,isrc)%y_ima_uperr  = real(par_err(2), kind=real32)                  
          endif
          
          if (ispar == 3 .and. par%fitext_flag)  then
             clulist(:,isrc)%ext_err     = real(mean_err*pix2arcsec, kind=real32)
             clulist(:,isrc)%ext_lowerr  = real(par_err(1)*pix2arcsec, kind=real32)
             clulist(:,isrc)%ext_uperr   = real(par_err(2)*pix2arcsec, kind=real32)
             
          endif
          
          if (ispar > 3) then
            iima=iima+1
            clulist(iima,isrc)%rate_err      = real(mean_err, kind=real32)
            clulist(iima,isrc)%rate_lowerr   = real(par_err(1), kind=real32)
            clulist(iima,isrc)%rate_uperr    = real(par_err(2), kind=real32)
            
          endif 
        endif
        a(ipar)=asave(ipar)
      enddo !npsou
    enddo !nsrc
     
   
    deallocate(asave)

  end subroutine paramerr
  
  !--------------------------------------------------------------------------------------------------------------------------


  subroutine fill_arrs(par,photon_mode,clulist,image,eventlist,pixlist,cluPSF,detmask,expimage,modelima,nsrc,events,mask,ierr)

  !---------------------------------------------------------------------
  ! Extract sub-images and prepare detection masks for fitting.
  ! Structure 'events' will only contain image pixels with values > 0.
  ! mask%mask = 1 for valid pixels, = 2 for pixels containing photon events.
  ! mask%nev holds total number of non-zero image pixels.
  ! Allocate memory for  'modelima' 
  ! Reading of event files not yet implemented 
  !---------------------------------------------------------------------

    type(ermldet_param_type),intent(in)      :: par
    logical                                  :: photon_mode
!    type(srcrow_type),dimension(:,:),pointer :: clulist
    type(srcrow_type),dimension(:,:),allocatable :: clulist    
    type(tEventFile),  dimension(:)          :: eventlist 
    type(pixlist_type),dimension(:)          :: pixlist
    type(FPSF)                               :: cluPSF
    type(i4cube_type)                        :: image
    type(l1cube_type)                        :: detmask
    type(r4cube_type)                        :: expimage
!    type(r4cube_type)                        :: bkgimage
    real(kind=real64),dimension(:,:,:),pointer :: modelima
    integer(kind=int32)                      :: nsrc
    type(events_type),pointer,dimension(:)   :: events
    type(mask_type)                          :: mask                         
    integer(kind=int32)                      :: ierr


    type(tEventFile),dimension(:),allocatable :: clu_eventfile
    type(EVENTFILE)                           :: shEventfile
    type(pixlist_type),dimension(:),allocatable    :: clu_pixlist
    type(tEvents),dimension(:),allocatable         :: clu_eventlist
    integer(kind=int32)                      :: status  
    integer(kind=int32)                      :: masksum,cutsum,nev
    real(kind=real32)                        :: maskfrac
    integer(kind=int32)                      :: iev,itot  
    integer(kind=int32)                      :: xoff,yoff  
    integer(kind=int32)                      :: ipoint,i,j,k,ix,iy,imask
    integer(kind=int32)                      :: ixmin,ixmax,iymin,iymax,imx,imy
    real(kind=real32)                        :: ecut2
    integer(kind=int32)                      :: n_sel_events
    
    character(len=*), parameter              :: rtname='FILL_ARRS'

    ! for test output:
    real(kind=real32)   :: t1,t2!,t3,t4
    
    
    ierr=0

    
    ecut2 = clulist(0,1)%ecutrad * clulist(0,1)%ecutrad

    ixmin=int(minval(clulist(0,1:nsrc)%x_ima-clulist(0,1:nsrc)%ecutrad))
    ixmax=int(maxval(clulist(0,1:nsrc)%x_ima+clulist(0,1:nsrc)%ecutrad))   
    iymin=int(minval(clulist(0,1:nsrc)%y_ima-clulist(0,1:nsrc)%ecutrad))  
    iymax=int(maxval(clulist(0,1:nsrc)%y_ima+clulist(0,1:nsrc)%ecutrad))

    ! first check whether maskfrac criterion is met for each source:
    ! (for each source work out fraction of valid pixels within ecut,
    !  
    mask%ix_min=max(1,ixmin)
    mask%ix_max=min(ixmax,detmask%shapehdr(1)%naxis1)
    mask%iy_min=max(1,iymin)
    mask%iy_max=min(iymax,detmask%shapehdr(1)%naxis2)
    mask%nx=mask%ix_max-mask%ix_min+1
    mask%ny=mask%iy_max-mask%iy_min+1    

    xoff=mask%ix_min-1
    yoff=mask%iy_min-1

    mask%nsrc=nsrc
    
    do k=1,nsrc
      mask%srcreg_x(k)=clulist(0,k)%x_ima
      mask%srcreg_y(k)=clulist(0,k)%y_ima
      ! GL (21/04/2021) allow only 0.75 of ecut radius as source region:
 !     mask%srcreg_r(k)=clulist(0,k)%ecutrad
      mask%srcreg_r(k)=clulist(0,k)%ecutrad*0.75
      
      clulist(:,k)%maskzero=0
      do i=1,num_exp
        do j=1,num_image(i)
          ipoint=ip(i,j)
          imask=mask_ip(ipoint)
          masksum=0
          cutsum=0
          maskfrac=0         
          do iy=mask%iy_min,mask%iy_max
            imy=iy-mask%iy_min+1
            do ix=mask%ix_min,mask%ix_max
              imx=ix-mask%ix_min+1
              if ( ((ix-clulist(0,k)%x_ima)**2 + (iy-clulist(0,k)%y_ima)**2) <= ecut2) then
                cutsum=cutsum+1
                if (par%detmask_flag) then
                  if ( ix>=detmask%dethdr(imask)%xdatamin .and. ix<=detmask%dethdr(imask)%xdatamax & 
                      .and. iy>=detmask%dethdr(imask)%ydatamin .and. iy<=detmask%dethdr(imask)%ydatamax ) then
                      if (detmask%arr(ix,iy,imask) ) then
                        masksum=masksum+1
                      endif
                  endif
                else if (.not. par%expima_flag .or. expimage%arr(ix,iy,ipoint) > 0.0) then
                  masksum=masksum+1
                endif
              endif  ! <= ecut2  
            enddo ! ix
          enddo !iy
          maskfrac=max(maskfrac,real(masksum)/real(cutsum))
        enddo ! j
      enddo ! i
      if (maskfrac < maskthreshold) clulist(:,k)%maskzero=1
    enddo ! k
   
    ! remove sources with maskzero=1: 
    j=1
    i=0
    do k=nsrc,j,-1 
      if (clulist(1,k)%maskzero == 0) then
        j=j+1
      else
          do i=k,k+j-1
            clulist(:,i)=clulist(:,i+1)
          enddo
          nsrc=nsrc-1
      endif  
   enddo

   
    ! sizes needed for masks and events:

    if (associated(mask%mask)) deallocate(mask%mask)
   
!    if (associated(events))    deallocate(events) 
!    if (associated(modelima))  deallocate(modelima) 
    ! nullify rather than deallocate (GL/AGE 12/05/2021):
    ! also deallocate to avoid memory leak (GL 04/06/2021):
     if (associated(events)) then
       deallocate(events)
       nullify(events)
     endif   
     if (associated(modelima))   then
       deallocate(modelima)
       nullify(modelima) 
     endif    
    allocate(mask%mask(mask%nx,mask%ny,nim))
    allocate(events(nim*mask%nx*mask%ny))
    allocate(modelima(mask%nx,mask%ny,nim))   

    modelima(:,:,:)=0.0
    mask%mask(:,:,:)=0

!    if (maxval(modelima) > 100.) write(*,*) 'allocated modelima: ',maxval(modelima)

    ! fill mask and events array 
    mask%mask(:,:,:)=0
    nev=0
    mask%npix=0
    do i=1,num_exp
      do j=1,num_image(i)
        ipoint=ip(i,j)
        imask=mask_ip(ipoint)
        do iy=mask%iy_min,mask%iy_max
          imy=iy-mask%iy_min+1
          do ix=mask%ix_min,mask%ix_max
            imx=ix-mask%ix_min+1
            do k=1,nsrc
              if ( ((ix-clulist(0,k)%x_ima)**2 + (iy-clulist(0,k)%y_ima)**2)  .le. ecut2) then
                if (par%detmask_flag)  then
                  if ( ix>=detmask%dethdr(imask)%xdatamin .and. ix<=detmask%dethdr(imask)%xdatamax & 
                      .and. iy>=detmask%dethdr(imask)%ydatamin .and. iy<=detmask%dethdr(imask)%ydatamax ) then
!                   if (.not. par%expima_flag .or. expimage%arr(ix,iy,ipoint) > 0.0) then
!                     mask%mask(imx,imy,ipoint) = 1
!                   endif 
                    if (detmask%arr(ix,iy,imask)) then
                      mask%mask(imx,imy,ipoint) = 1
                    endif
                  endif ! ix>=detmask%dethdr(i)%xdatamin ...
                else if (.not. par%expima_flag .or. expimage%arr(ix,iy,ipoint) > 0.0) then
                  mask%mask(imx,imy,ipoint) = 1
                endif !(par%detmask_flag
              endif ! ecut2
            enddo ! k  
            if (mask%mask(imx,imy,ipoint) > 0) then
              mask%npix=mask%npix+1
              if (.not. photon_mode) then
                if (image%arr(ix,iy,ipoint) > 0) then
                  nev=nev+1
                  events(nev)%n=image%arr(ix,iy,ipoint)
                  events(nev)%ipt=ipoint
                  mask%mask(imx,imy,ipoint) = 2
                endif
              endif !photon_flag
            endif
          enddo ! ix
        enddo  ! iy
      enddo ! j
    enddo ! i
 
    mask%nev=nev
    
    ! **** PHOTON MODE *****
    nev=0
    if (photon_mode) then
      if (.not. allocated(clu_eventlist)) allocate(clu_eventlist(nim))
      if (.not. allocated(clu_pixlist))   allocate(clu_pixlist(nim))
      if (.not. allocated(cluPSF%evSPSF)) allocate(cluPSF%evSPSF(nim))
      if (.not. allocated(clu_eventfile)) allocate(clu_eventfile(nim))

      do ipoint=1,nim
!          imask=mask_ip(ipoint)
          call select_pixlist(pixlist(ipoint),mask,ipoint,clu_pixlist(ipoint),status)
!          write(*,*) 'done select_pixlist'
!          call select_eventlist(eventlist(ipoint)%events,pixlist(ipoint),clu_eventlist(ipoint),status)
          call select_eventlist(eventlist(ipoint)%events,clu_pixlist(ipoint),clu_eventlist(ipoint),status)
!          write(*,*) "done select_eventlist:",size(eventlist(ipoint)%events%ra),size(clu_eventlist(ipoint)%ra)
!          stop
!          write(*,*) 'done  select_eventlist'
          clu_eventfile(ipoint)=eventlist(ipoint)
          clu_eventfile(ipoint)%events = clu_eventlist(ipoint)
          
          t1=0.0
          t2=0.0
          call slib_evttool_2_myevent(shEventfile,clu_eventfile(ipoint))

           
          if (allocated(cluPSF%evSPSF(ipoint)%c0)) deallocate(cluPSF%evSPSF(ipoint)%c0)
          if (allocated(cluPSF%evSPSF(ipoint)%c1)) deallocate(cluPSF%evSPSF(ipoint)%c1)
          if (allocated(cluPSF%evSPSF(ipoint)%c2)) deallocate(cluPSF%evSPSF(ipoint)%c2)
          n_sel_events = clu_eventlist(ipoint)%getNumRows()
          allocate(cluPSF%evSPSF(ipoint)%c0(size(cluPSF%mySPSF(nim)%NX0),n_sel_events))
          allocate(cluPSF%evSPSF(ipoint)%c1(size(cluPSF%mySPSF(nim)%NX1),n_sel_events))
          allocate(cluPSF%evSPSF(ipoint)%c2(size(cluPSF%mySPSF(nim)%NX2),n_sel_events))


          do iev=1,n_sel_events
            call get_single_event(shEventfile, cluPSF%myModelPSF, cluPSF%myCPSF,    &
                     iev,cluPSF%evSPSF(ipoint)%c0(:,iev),cluPSF%evSPSF(ipoint)%c1(:,iev),cluPSF%evSPSF(ipoint)%c2(:,iev))
          enddo

          nev=nev+n_sel_events
          call slib_deallocate_event(shEventfile)
      enddo  
      if (associated(events)) deallocate(events)
      allocate(events(nev))
      itot=0
      
      do ipoint=1,nim
        n_sel_events = clu_eventlist(ipoint)%getNumRows()
        do iev=1,n_sel_events
          itot=itot+1
          events(itot)%n=1
          events(itot)%ix=clu_pixlist(ipoint)%pix(iev)%x-xoff
          events(itot)%iy=clu_pixlist(ipoint)%pix(iev)%y-yoff
          events(itot)%ipt=ipoint
          ! ID is the order of the events within clu_eventlist(ipoint):
          events(itot)%id=iev
        end do
      enddo  
      mask%nev=nev   


    endif  ! photon mode
  
    

  end subroutine fill_arrs

 !-----------------------------------------------------------------------------------------------------
  
 subroutine select_pixlist(pixlist_in,mask,iim,pixlist_out,ierr)
 
  !------------------------------------------------------------------------------------------
  ! Select pixel list pixlist_in according to mask (where mask == .true.) and write selected  
  ! entries to pixlist_out.
  !  
  !------------------------------------------------------------------------------------------

    type(pixlist_type)                                  ::  pixlist_in
    type(mask_type)                                     ::  mask
    integer(kind=int32)                                 ::  iim
    type(pixlist_type)                                  ::  pixlist_out
    integer(kind=int32)                                 ::  ierr
  
    integer(kind=int32)                                 ::  xoff,yoff,imx,imy!nx,ny,   
    integer(kind=int32)                                 ::  ipix, npix 
    type(pixlist_type)                                  ::  pixlist_tmp

    character(len=*), parameter                         :: rtname='SELECT_PIXLIST'

    ierr=0 

    xoff=mask%ix_min-1
    yoff=mask%iy_min-1


    allocate(pixlist_tmp%pix(pixlist_in%n))   
 
    npix=0




    do ipix = 1,pixlist_in%n
      if    (pixlist_in%pix(ipix)%x >= mask%ix_min .and. pixlist_in%pix(ipix)%x <= mask%ix_max  &
      .and.  pixlist_in%pix(ipix)%y >= mask%iy_min .and. pixlist_in%pix(ipix)%y <= mask%iy_max) then
        imx=pixlist_in%pix(ipix)%x-xoff
        imy=pixlist_in%pix(ipix)%y-yoff 
        if (mask%mask(imx,imy,iim) > 0) then
          npix = npix+1
          pixlist_tmp%pix(npix)  = pixlist_in%pix(ipix)
        endif
      endif
    enddo
    

    if (allocated(pixlist_out%pix)) deallocate(pixlist_out%pix)
    allocate(pixlist_out%pix(npix))


    pixlist_out%pix(1:npix) = pixlist_tmp%pix(1:npix)
    pixlist_out%n           = npix

    deallocate(pixlist_tmp%pix)


  end subroutine select_pixlist


 !-----------------------------------------------------------------------------------------------------

  subroutine select_eventlist(eventlist_in,pixlist,eventlist_out,ierr)

   type(tEvents)                                       :: eventlist_in 
   type(pixlist_type)                                  :: pixlist
   type(tEvents)                                       :: eventlist_out 
   integer(kind=int32)                                 :: ierr
   integer(kind=int32)                                 :: n_ev_in
   integer(kind=int32)                                 :: status
    character(len=*), parameter                        :: rtname='SELECT_EVENTLIST'    
   
   
    ierr=0
    status=0

    n_ev_in = eventlist_in%getNumRows()

   if (n_ev_in >= maxval(pixlist%pix%id)) then
     eventlist_out = eventlist_in ! IMS: this direct copy of the struct does not preserve pointer association, but in ermldet context this does not matter.
     !call  eventlist_out%apply_index_cols(pixlist%pix%id,status)
     call  eventlist_out%reorder(pixlist%pix%id,status)  
   else
     ierr=3
     call wrerr(rtname,ierr,'pixel list ID out of range')
   endif
   

   if (status > 0) then
     ierr=3
     call wrerr(rtname,ierr,'error in APPLY_INDEX_COLS')
   endif

   
 

  end subroutine select_eventlist


 !----------------------------------------------------------------------------------------------------- 

  
end module mulfit

