! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
! Not methods of tEvents_pp:
!  integer(i2) function calc_binned_amp(value_adu) result(binned_amp)
!  function calc_binned_amps(values_adu) result(binned_amps)
!  integer(i2) function calc_adu(binned_amp) result(value_adu)
!  function calc_adus(binned_amps) result(values_adu)
!  make_frame_histo()
module erolib_events_pp_mod

  use ero_general ! need for flag masks.
!  use erolib_timsort
  use erolib_bintable_mod
  use erolib_gtitab_mod

  implicit none

  type, extends(tBinTable) :: tEvents_pp
  contains
    procedure :: init => init_tEvents_pp
    procedure :: initHeader => initHeader_tEvents_pp
  end type tEvents_pp

  type, extends(tBinTable) :: tFrames_pp
  contains
    procedure :: init => init_tFrames_pp
  end type tFrames_pp

  type, extends(tFitsFile) :: tEvents_pp_file
    type(tEvents_pp) :: events_pp
    type(tFrames_pp) :: frames_pp
    type(tGTI_tab) :: rGTI
  contains
    procedure :: destroy => destroy_tEvents_pp_file
  end type tEvents_pp_file

  integer(i2), dimension(5), private, parameter :: adu_offsets=[3500_i2,10084_i2,11134_i2,12534_i2,14284_i2]
  integer(i2), dimension(5), private, parameter :: bna_offsets=[3500_i2,6792_i2,7142_i2,7492_i2,7842_i2]
  integer(i2), dimension(5), private, parameter :: scales=[2_i2,3_i2,4_i2,5_i2,6_i2]

contains

  !.....................................................................
  subroutine init_tEvents_pp(this, status, fileInfo)
    implicit none

    class(tEvents_pp) :: this
    integer, intent(inout) :: status
    type(tFitsInfo), intent(in), optional, target :: fileInfo

    character(*), parameter :: functionName = "init_tEvents_pp"

    if(present(fileInfo)) this%info%fileInfo => fileInfo
    this%info%tableName = "EVENTS"

    call this%initColumn("J", "FRAME")
    call this%initColumn("D", "FRAMETIME")
    call this%initColumn("J", "EXT_OTS")
    call this%initColumn("J", "INT_OTS")
    call this%initColumn("I", "TimeSRC")
    call this%initColumn("I", "DataSRC")
    call this%initColumn("J", "SCTime")
    call this%initColumn("I", "SubSec")
    call this%initColumn("D", "RecordTime")
    call this%initColumn("I", "TMFrameCounter")
    call this%initColumn("I", "RAWX", "pixel")
    call this%initColumn("I", "RAWY", "pixel")
    call this%initColumn("I", "PHA", "ADU")
    call this%initColumn("J", "env")
    call this%initColumn("I", "BinnedAmp")
    call this%initColumn("I", "Quality")
    call this%initColumn("I", "DIR1")
    call this%initColumn("I", "PHA1", "ADU")
    call this%initColumn("I", "DIR2")
    call this%initColumn("I", "PHA2", "ADU")
    call this%initColumn("J", "FLAG")      ! dummy column
    call this%initColumn("B", "PROCMODES") ! dummy column
    call this%initColumn("D", "TIME")      ! dummy column
    call this%initColumn("J", "RowNum")    ! dummy column
    call this%initColumn("J", "temp_fr_ctr") ! dummy column

    call this%initColumns(status)

    call this%initHeader(status)

  end subroutine init_tEvents_pp

  !.....................................................................
  subroutine initHeader_tEvents_pp(this, status)
    implicit none

    class(tEvents_pp), target :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "initHeader_tEvents_pp"

    this%header%tabInfo => this%info

    call this%header%initKeyword("D", "TSTART")
    call this%header%initKeyword("A", "DATE-OBS")
    call this%header%initKeyword("A", "DATE-END")
    call this%header%initKeyword("A", "TELESCOP")
    call this%header%initKeyword("A", "INSTRUME")
    call this%header%initKeyword("D", "FRAMETIM")
    call this%header%initKeyword("D", "MJDREF")
    call this%header%initKeyword("J", "TEL_ID")
    call this%header%initKeyword("J", "NINST")
    call this%header%initKeyword("A", "APP_ID")
    call this%header%initKeyword("J", "ERODAY")
    call this%header%initKeywords(status)

    call this%header%check_keywords(status)

  end subroutine initHeader_tEvents_pp

  !.....................................................................
  subroutine init_tFrames_pp(this, status, fileInfo)
    implicit none

    class(tFrames_pp) :: this
    integer,         intent(inout) :: status
    type(tFitsInfo), intent(in), optional, target :: fileInfo

    character(*), parameter :: functionName = "init_tFrames_pp"

    if(present(fileInfo)) this%info%fileInfo => fileInfo
    this%info%tableName = "FRAMES"

    call this%initColumn("J", "FRAME")
    call this%initColumn("I", "TimeSRC")
    call this%initColumn("I", "DataSRC")
    call this%initColumn("J", "SCTime")
    call this%initColumn("I", "subSec")
    call this%initColumn("D", "FRAMETIME")

    call this%initColumn("J", "FIRST_RecordTime")
    call this%initColumn("I", "FIRST_TMFrameCounter")
    call this%initColumn("J", "LAST_RecordTime")
    call this%initColumn("I", "LAST_TMFrameCounter")

    ! Temporary columns, not present in the file:
    call this%initColumn("D", "TIME")
    call this%initColumn("B", "FLAG")
    call this%initColumn("I", "N_EVENTS")

    call this%initColumns(status)

  end subroutine init_tFrames_pp

  !.....................................................................
  subroutine destroy_tEvents_pp_file(this, status)
    implicit none

    class(tEvents_pp_file) :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "destroy_tEvents_pp_file"

    call this%events_pp%destroy(status)
    call this%frames_pp%destroy(status)
    call this%rGTI%destroy(status)

  end subroutine destroy_tEvents_pp_file

  !=====================================================================
  ! The remaining routines are not methods of tEvents_pp, they're just
  ! utility routines to translate between BinnedAmp and ADU units,
  ! following the recipe set out in the CE manual, section 4.1.1.2.

  !.....................................................................
  integer(i2) function calc_binned_amp(value_adu) result(binned_amp)
    implicit none
    integer(i2), intent(in) :: value_adu

    if(     value_adu<adu_offsets(1)) then
      binned_amp =  value_adu
    else if(value_adu<adu_offsets(2)) then
      binned_amp = (value_adu - adu_offsets(1))/scales(1) + bna_offsets(1)
    else if(value_adu<adu_offsets(3)) then
      binned_amp = (value_adu - adu_offsets(2))/scales(2) + bna_offsets(2)
    else if(value_adu<adu_offsets(4)) then
      binned_amp = (value_adu - adu_offsets(3))/scales(3) + bna_offsets(3)
    else if(value_adu<adu_offsets(5)) then
      binned_amp = (value_adu - adu_offsets(4))/scales(4) + bna_offsets(4)
    else
      binned_amp = (value_adu - adu_offsets(5))/scales(5) + bna_offsets(5)
    end if

  end function calc_binned_amp

  !.....................................................................
  function calc_binned_amps(values_adu) result(binned_amps)
    implicit none
    integer(i2), dimension(:), intent(in) :: values_adu
    integer(i2), dimension(size(values_adu)) :: binned_amps

    where(values_adu<adu_offsets(1))
      binned_amps =  values_adu
    elsewhere
      where (values_adu<adu_offsets(2))
        binned_amps = (values_adu - adu_offsets(1))/scales(1) + bna_offsets(1)
      elsewhere
        where(values_adu<adu_offsets(3))
          binned_amps = (values_adu - adu_offsets(2))/scales(2) + bna_offsets(2)
        elsewhere
          where(values_adu<adu_offsets(4))
            binned_amps = (values_adu - adu_offsets(3))/scales(3) + bna_offsets(3)
          elsewhere
            where(values_adu<adu_offsets(5))
              binned_amps = (values_adu - adu_offsets(4))/scales(4) + bna_offsets(4)
            elsewhere
              binned_amps = (values_adu - adu_offsets(5))/scales(5) + bna_offsets(5)
            end where
          end where
        end where
      end where
    end where
  end function calc_binned_amps

  !.....................................................................
  integer(i2) function calc_adu(binned_amp) result(value_adu)
    implicit none
    integer(i2), intent(in) :: binned_amp

    if(     binned_amp<bna_offsets(1)) then
      value_adu = binned_amp
    else if(binned_amp<bna_offsets(2)) then
      value_adu = (binned_amp - bna_offsets(1))*scales(1) + adu_offsets(1)
    else if(binned_amp<bna_offsets(3)) then
      value_adu = (binned_amp - bna_offsets(2))*scales(2) + adu_offsets(2)
    else if(binned_amp<bna_offsets(4)) then
      value_adu = (binned_amp - bna_offsets(3))*scales(3) + adu_offsets(3)
    else if(binned_amp<bna_offsets(5)) then
      value_adu = (binned_amp - bna_offsets(4))*scales(4) + adu_offsets(4)
    else
      value_adu = (binned_amp - bna_offsets(5))*scales(5) + adu_offsets(5)
    end if

  end function calc_adu

  !.....................................................................
  function calc_adus(binned_amps) result(values_adu)
    implicit none
    integer(i2), dimension(:), intent(in) :: binned_amps
    integer(i2), dimension(size(binned_amps)) :: values_adu

    where(binned_amps<bna_offsets(1))
      values_adu = binned_amps
    elsewhere
      where(binned_amps<bna_offsets(2))
        values_adu = (binned_amps - bna_offsets(1))*scales(1) + adu_offsets(1)
      elsewhere
        where(binned_amps<bna_offsets(3))
          values_adu = (binned_amps - bna_offsets(2))*scales(2) + adu_offsets(2)
        elsewhere
          where(binned_amps<bna_offsets(4))
            values_adu = (binned_amps - bna_offsets(3))*scales(3) + adu_offsets(3)
          elsewhere
            where(binned_amps<bna_offsets(5))
              values_adu = (binned_amps - bna_offsets(4))*scales(4) + adu_offsets(4)
            elsewhere
              values_adu = (binned_amps - bna_offsets(5))*scales(5) + adu_offsets(5)
            end where
          end where
        end where
      end where
    end where

  end function calc_adus

  !.....................................................................
  subroutine get_n_events_per_frame(events_pp_file)
    ! Note that the code here is similar to the frame-matching code in evprep_mod:make_good_frame_GTIs(). However I have not been able to find a neat way to merge the two wodges of code into one.
    !
!*** note also that we have some slight improvements, fixes and comments here which could usefully be imported into the evprep routine.
    !
    implicit none

    type(tEvents_pp_file), intent(inout) :: events_pp_file

    character(*), parameter :: functionName = "get_n_events_per_frame"
    integer :: ierr,n_frames,n_good_frames,n_events
    integer :: ei_first_non_null,ei,fi_first_non_null,fi
    integer :: n_EVENTS_fids_not_in_FRAMES,n_nonmatching_FRAMETIMEs,n_nonmatching_FRAMEs
    real(r8), dimension(:), allocatable :: FRAMES_times,EVENTS_times
    integer, dimension(:), allocatable :: FRAMES_order,EVENTS_order
    real(r8) :: EVENTS_fti,FRAMES_fti
    integer(i4) :: EVENTS_fid,old_EVENTS_fid,FRAMES_fid
    character(buffsize) :: intStr

    integer(i2), dimension(:), pointer ::&
       fr_n_events   =>null()
    real(r8), dimension(:), pointer ::&
       ev_frametimes =>null()&
      ,fr_frametimes =>null()
    integer(i4), dimension(:), pointer ::&
       ev_frames     =>null()&
      ,fr_frames     =>null()
    logical, dimension(:), pointer ::&
       ev_ft_isnull  =>null()&
      ,fr_ft_isnull  =>null()&
      ,ev_fid_isnull =>null()&
      ,fr_fid_isnull =>null()

    ierr = 0

!    col_num = events_pp_file%frames_pp%getColNum("N_EVENTS", ierr)
!    if(.not.events_pp_file%frames_pp%cols(col_num)%isAllocated())&
    if(.not.events_pp_file%frames_pp%btest_col_useage("N_EVENTS", TCOL_INTRN_BIT))&
call wrstop(functionName, "You must set up the N_EVENTS column of FRAMES for internal use before calling this subroutine.")
!      call events_pp_file%frames_pp%ibset_col_useage("N_EVENTS", TCOL_INTRN_BIT)

    n_frames = events_pp_file%frames_pp%getNumRows()

!    if(.not.events_pp_file%frames_pp%cols(col_num)%isAllocated())&
!      call events_pp_file%frames_pp%cols(col_num)%alloc(n_frames, ierr)

    call events_pp_file%frames_pp%get_col_ptr("FRAMETIME", fr_frametimes, fr_ft_isnull)
    ! To cope with the stupid -255 values:
    where(.not.fr_ft_isnull .and.fr_frametimes<0.0_r8) fr_ft_isnull = .true.
    n_good_frames = n_frames - count(fr_ft_isnull)

    if(n_good_frames<1) then
      call wrerr(functionName, warn, "No non-NULL rows in FRAMES extension!")
      nullify(fr_frametimes, fr_ft_isnull)
  return
    end if

    call events_pp_file%events_pp%get_col_ptr("FRAMETIME", ev_frametimes, ev_ft_isnull)
    ! To cope with the stupid -255 values:
    where(.not.ev_ft_isnull .and.ev_frametimes<0.0_r8) ev_ft_isnull = .true.

    if(all(ev_ft_isnull)) then
      call wrerr(functionName, warn, "No good event times.")
      nullify(ev_frametimes, ev_ft_isnull)
      nullify(fr_frametimes, fr_ft_isnull)
  return
    end if

    call events_pp_file%events_pp%get_col_ptr("FRAME",  ev_frames, ev_fid_isnull)
    call events_pp_file%frames_pp%get_col_ptr("FRAME",  fr_frames, fr_fid_isnull)

    ! To cope with the stupid -255 values:
    where(.not.ev_fid_isnull .and.ev_frames<1) ev_fid_isnull = .true.
    where(.not.fr_fid_isnull .and.fr_frames<1) fr_fid_isnull = .true.

    if(any(ev_fid_isnull.and..not.ev_ft_isnull) .or. any(fr_fid_isnull.and..not.fr_ft_isnull))&
call wrstop(functionName, "NULL-valued FRAME when FRAMETIME is non-NULL.")

    call events_pp_file%frames_pp%get_col_ptr("N_EVENTS", fr_n_events)

    ! The first thing I am going to do is sort copies of the FRAMETIME
    ! column from both extensions. Mostly this is to avoid the need for
    ! nested search loops and so to speed things up.
    !
    n_events = events_pp_file%events_pp%getNumRows()
    call alloc(FRAMES_times, n_frames, ierr)
!    call alloc(FRAMES_good,  numFrames, ierr)
    call alloc(EVENTS_times, n_events, ierr)
    FRAMES_times = fr_frametimes
    EVENTS_times = ev_frametimes
    call sort(FRAMES_times, ierr, FRAMES_order)
    call sort(EVENTS_times, ierr, EVENTS_order)
    ! This should leave the NULLs grouped at the start.

    ! Find the first event and first frame with non-NULL FRAMETIME.
    !
    ei_first_non_null = 0 ! Since we have already excluded the situation in which all the FRAMETIMEs are NULL, this index should always receive a value in the loop below; still, it better to give it some value rather than rely on that assumption.
    do ei=1,n_events
      if(.not.ev_ft_isnull(EVENTS_order(ei))) then
        ei_first_non_null = ei
    exit
      end if
    end do

    fi_first_non_null = 0 ! Since we have already excluded the situation in which all the FRAMETIMEs are NULL, this index should always receive a value in the loop below; still, it better to give it some value rather than rely on that assumption.
    do fi=1,n_frames
      if(.not.fr_ft_isnull(FRAMES_order(fi))) then
        fi_first_non_null = fi
    exit
!      else
!        FRAMES_good(fi) = .false.
      end if
    end do

    if(EVENTS_times(ei_first_non_null)>FRAMES_times(n_frames))&
call wrstop(functionName, "All EVENTS frame times are larger than any FRAMES ones!")

    if(EVENTS_times(n_events)<FRAMES_times(fi_first_non_null))&
call wrstop(functionName, "All EVENTS frame times are smaller than any FRAMES ones!")

    n_EVENTS_fids_not_in_FRAMES = 0
    n_nonmatching_FRAMETIMEs = 0
    n_nonmatching_FRAMEs = 0
    ei = ei_first_non_null
    EVENTS_fid = ev_frames(EVENTS_order(ei))
!    EVENTS_fti = ev_frametimes(EVENTS_order(ei))
    EVENTS_fti = EVENTS_times(ei)
    old_EVENTS_fid = minval(ev_frames, .not.ev_ft_isnull) - 1_i4 ! just to get a starting value which cannot be equal to any valid FRAME value in the EVENTS table.
    frameloop: do fi=fi_first_non_null,n_frames
      FRAMES_fid = fr_frames(FRAMES_order(fi))
!      FRAMES_fti = fr_frametimes(FRAMES_order(fi))
      FRAMES_fti = FRAMES_times(fi)

      ! Basically here we want to find all the EVENTS rows which
      ! correspond to the FRAME row. Ideally we could just compare
      ! FRAMETIME values, but comparing floating-point values
      ! is bad karma and I don't trust it. So I have added a few more
      ! bells and whistles in an attempt to close off any loopholes.
      !
      eventloop: do
        if(EVENTS_fid==FRAMES_fid) then
          if(EVENTS_fti /= FRAMES_fti) then
            n_nonmatching_FRAMETIMEs = n_nonmatching_FRAMETIMEs + 1
          end if

!          if(btest(ev_flags(EVENTS_order(ei)), BIT_CORRUPT_FRAME)) then
!            FRAMES_good(fi) = .false.
!          end if
          fr_n_events(FRAMES_order(fi)) = fr_n_events(FRAMES_order(fi)) + 1_i2

        else ! frame IDs don't match.
          ! Now we have to compare FRAMETIME values, since they were the
          ! sort target. Frame IDs are not guaranteed to obey the same
          ! order at every point.
          !
!          if(EVENTS_times(ei) > FRAMES_times(fi)) then ! need to increment fi, the FRAMES row.
          if(EVENTS_fti > FRAMES_fti) then ! need to increment fi, the FRAMES row.
            if(ei>ei_first_non_null) old_EVENTS_fid = ev_frames(EVENTS_order(ei-1))
      exit eventloop
          end if

          ! Otherwise we stay in eventloop and (at the end of the loop
          ! pass) increment ei, the EVENTS row.
          !
!          if(EVENTS_times(ei) < FRAMES_times(fi)) then
          if(EVENTS_fti < FRAMES_fti) then
            ! If we have got to here then we have found an event row
            ! for which the frame ID is not in the sequence of IDs
            ! obtained from the FRAMES table. However we don't want to
            ! count event rows, we want to count the frame IDs in such
            ! event rows. We may have several events per such frame
            ! ID, so we keep track of the changes in frame ID and only
            ! increment our counter when we detect a change.
            !
            if(EVENTS_fid/=old_EVENTS_fid) then
              old_EVENTS_fid = EVENTS_fid
              n_EVENTS_fids_not_in_FRAMES = n_EVENTS_fids_not_in_FRAMES + 1

              intStr = int2str(EVENTS_fid)
              call wrerr(functionName, warn, "FRAME "//trim(intStr)&
                //" found in event list but not in FRAMES extension.")
            end if

          else ! implies EVENTS_fti == FRAMES_fti. This is weird if the FRAME values don't match. Log it, then get the next event.
            n_nonmatching_FRAMEs = n_nonmatching_FRAMEs + 1
          end if
        end if

        ei = ei + 1
        if(ei>n_events)&
    exit frameloop

        EVENTS_fid = ev_frames(EVENTS_order(ei))
!        EVENTS_fti = ev_frametimes(EVENTS_order(ei))
        EVENTS_fti = EVENTS_times(ei)
      end do eventloop
    end do frameloop

    if(n_EVENTS_fids_not_in_FRAMES>0) then
      intStr = int2str(n_EVENTS_fids_not_in_FRAMES)
      call wrerr(functionName, warn, trim(intStr)&
        //" frame IDs found in event list but not in FRAMES extension.")
    end if
    if(n_nonmatching_FRAMETIMEs>0) then
      intStr = int2str(n_nonmatching_FRAMETIMEs)
      call wrerr(functionName, warn, trim(intStr)&
        //" frames have different FRAMETIME values in EVENTS and FRAMES extensions.")
    end if
    if(n_nonmatching_FRAMEs>0) then
      intStr = int2str(n_nonmatching_FRAMEs)
      call wrerr(functionName, warn, trim(intStr)&
        //" frames have same FRAMETIME, but different FRAME, in EVENTS and FRAMES extensions.")
    end if

    call dealloc(FRAMES_times, ierr)
    call dealloc(EVENTS_times, ierr)
!    call dealloc(FRAMES_good, ierr)
    call dealloc(FRAMES_order, ierr)
    call dealloc(EVENTS_order, ierr)

    nullify(ev_frametimes, ev_ft_isnull)
    nullify(fr_frametimes, fr_ft_isnull)
    nullify(ev_frames, ev_fid_isnull)
    nullify(fr_frames, fr_fid_isnull)
    nullify(fr_n_events)

  end subroutine get_n_events_per_frame

  !.....................................................................
  subroutine make_frame_histo(events_pp_file, histo, n_overflows, n_null_frames)
    ! This opens the EVENTS and FRAMES extensions of the events_pp_file object and compiles a histogram of the number of events per frame.
    !
    implicit none

    type(tEvents_pp_file), intent(inout) :: events_pp_file
    integer, dimension(:), intent(out)   :: histo
    integer,               intent(out)   :: n_overflows
    integer,               intent(out)   :: n_null_frames

    character(*), parameter :: functionName = "make_frame_histo"
    integer :: n_histo_bins,n_frames,fi,n_events

    integer(i2), dimension(:), pointer ::&
       fr_n_events   =>null()
    real(r8), dimension(:), pointer ::&
       fr_frametimes =>null()
    logical, dimension(:), pointer ::&
       fr_ft_isnull  =>null()

    n_histo_bins = size(histo)
    if(n_histo_bins<1)&
  return

    call get_n_events_per_frame(events_pp_file)

    histo(:) = 0
    n_overflows = 0

    n_frames = events_pp_file%frames_pp%getNumRows()
    call events_pp_file%frames_pp%get_col_ptr("FRAMETIME", fr_frametimes, fr_ft_isnull)
    n_null_frames = count(fr_ft_isnull)

    if(n_null_frames >= n_frames) then
      nullify(fr_frametimes, fr_ft_isnull)
  return
    end if

    call events_pp_file%frames_pp%get_col_ptr("N_EVENTS", fr_n_events)

    do fi=1,n_frames
      if(fr_ft_isnull(fi))&
    cycle

      n_events = fr_n_events(fi)
      if(n_events < n_histo_bins) then
        histo(n_events+1) = histo(n_events+1) + 1
      else
        n_overflows = n_overflows + 1
      end if

if(n_events<1) print *, "No events in frame", fi
    end do

    nullify(fr_frametimes, fr_ft_isnull)
    nullify(fr_n_events)

  end subroutine make_frame_histo

end module erolib_events_pp_mod
