!-----------------------------------------------------------------------
! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------

program evatt

!------------------------------------------------------------------------------
! eROSITA TEL chain, apply corrected attitude to events.
! Author : GLamer,  AIP 
! 
!
!
!------------------------------------------------------------------------------
  

  use param
  use numkind
  use fitsimage
  use cal
  use tel_io
  use astrolib
  use evatt_mod

  implicit none

  
  integer(kind=int32)                     :: ierr,status
  type(evatt_param_type)                  :: par
  type(instpar_type)                      :: instpar 
!  type(cal_par_type)                      :: cal_par
  type(evatt_list_type)                   :: eventlist
  type(corratt_type)                      :: attitude
  type(gti_type)                          :: gti

!  character(len=64) :: version  
!  character(len=20) :: versstring

  character(len=*), parameter :: rtname='MAIN'

  status = 0
  ierr   = 0

 
  ! open parameter and log files:
  call evatt_ini(par,status)
  ierr=max(ierr,status)  

!  versstring=version()
!  write(*,*) 'version: ', versstring

  ! get task parameters:
  if (ierr < 3) then
    call evatt_par(par,status)
    ierr=max(ierr,status)
  endif


  if (ierr < 3) then
    call evatt_in(par,eventlist,attitude,gti,instpar,status)
    ierr=max(ierr,status)
  endif

  if (ierr < 3) then
    call evatt_calc(par,eventlist,attitude,gti,instpar,status)
    ierr=max(ierr,status)
  endif

  if (ierr < 3) then
    call evatt_wrcol(eventlist,status)
    ierr=max(ierr,status)
  endif

  ! close parameter and log files:   
  call  cpast(rtname,par%lun,ierr,status)

  if(status /= 0) then
    call wrerr(rtname,status,'Error closing parameter/log files')
    ierr=max(ierr,status)
  endif    



end program evatt

