! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.

module data_evt
  
  use constants
  
  implicit none
  
  


contains
  
  subroutine deallocate_event(myEVENTFILE)

    implicit none
    
    TYPE(EVENTFILE) :: myEVENTFILE

    integer ii

    IF (ALLOCATED (myEVENTFILE%EVENT%XD)) DEALLOCATE (myEVENTFILE%EVENT%XD)
    IF (ALLOCATED (myEVENTFILE%EVENT%YD)) DEALLOCATE (myEVENTFILE%EVENT%YD)
    IF (ALLOCATED (myEVENTFILE%EVENT%ZD)) DEALLOCATE (myEVENTFILE%EVENT%ZD)
 
    IF (ALLOCATED (myEVENTFILE%EVENT%RA)) DEALLOCATE (myEVENTFILE%EVENT%RA)
    IF (ALLOCATED (myEVENTFILE%EVENT%DEC)) DEALLOCATE (myEVENTFILE%EVENT%DEC)
    IF (ALLOCATED (myEVENTFILE%EVENT%RAWX)) DEALLOCATE (myEVENTFILE%EVENT%RAWX)
    IF (ALLOCATED (myEVENTFILE%EVENT%RAWY)) DEALLOCATE (myEVENTFILE%EVENT%RAWY)
    IF (ALLOCATED (myEVENTFILE%EVENT%ENERGY)) DEALLOCATE (myEVENTFILE%EVENT%ENERGY)
    IF (ALLOCATED (myEVENTFILE%EVENT%TIME)) DEALLOCATE (myEVENTFILE%EVENT%TIME)
    IF (ALLOCATED (myEVENTFILE%EVENT%ROLL)) DEALLOCATE (myEVENTFILE%EVENT%ROLL)
    IF (ALLOCATED (myEVENTFILE%EVENT%TIME)) DEALLOCATE (myEVENTFILE%EVENT%TIME)
    IF (ALLOCATED (myEVENTFILE%EVENT%EV2SRC)) DEALLOCATE (myEVENTFILE%EVENT%EV2SRC)
    IF (ALLOCATED (myEVENTFILE%EVENT%CCDNR)) DEALLOCATE (myEVENTFILE%EVENT%CCDNR)


    IF (ALLOCATED (myEVENTFILE%WHICH_CCDNR)) DEALLOCATE  (myEVENTFILE%WHICH_CCDNR)

    do ii = 1 , myEVENTFILE%NUM_ATT  
       
       IF (ALLOCATED (myEVENTFILE%ATT(ii)%RA)) DEALLOCATE (myEVENTFILE%ATT(ii)%RA)
       IF (ALLOCATED (myEVENTFILE%ATT(ii)%DEC)) DEALLOCATE (myEVENTFILE%ATT(ii)%DEC)
       IF (ALLOCATED (myEVENTFILE%ATT(ii)%ROLL)) DEALLOCATE (myEVENTFILE%ATT(ii)%ROLL)
       IF (ALLOCATED (myEVENTFILE%ATT(ii)%TIME)) DEALLOCATE (myEVENTFILE%ATT(ii)%TIME)
    enddo
    
     IF (ALLOCATED ( myEVENTFILE%ATT ) )  DEALLOCATE (myEVENTFILE%ATT) 

  end subroutine deallocate_event



  subroutine evttool_2_myevent(myEVENTFILE, evttoolEVENT)

    use basic_calcs   
    !    use evtool_mod
    !    use evtool_eventfile_mod 
    use erolib_eventfile_mod
    use erolib_events_mod
    
    implicit none
    
    type(tEventFile) evttoolEVENT
    TYPE(EVENTFILE) myEVENTFILE

    integer nrows_ev, ii, nrows, max_nrows
    double precision ra, dec, xd, yd, zd


    ! allocate EVENT
    nrows_ev = evttoolEVENT%events%nelements

    myEVENTFILE%EVENT%SIZE = nrows_ev

    allocate  ( myEVENTFILE%EVENT%XD(nrows_ev), myEVENTFILE%EVENT%YD(nrows_ev) )
    allocate  ( myEVENTFILE%EVENT%ZD(nrows_ev) )    
    allocate  ( myEVENTFILE%EVENT%RA(nrows_ev), myEVENTFILE%EVENT%DEC(nrows_ev) )
    allocate  ( myEVENTFILE%EVENT%RAWX(nrows_ev), myEVENTFILE%EVENT%RAWY(nrows_ev) )
    allocate  ( myEVENTFILE%EVENT%ENERGY(nrows_ev), myEVENTFILE%EVENT%TIME(nrows_ev) )
    allocate  ( myEVENTFILE%EVENT%ROLL(nrows_ev), myEVENTFILE%EVENT%CCDNR(nrows_ev) )
    allocate  ( myEVENTFILE%EVENT%EV2SRC(nrows_ev) )
    
    
    myEVENTFILE%EVENT%RA = evttoolEVENT%events%ra
    myEVENTFILE%EVENT%DEC = evttoolEVENT%events%dec
    myEVENTFILE%EVENT%RAWX = evttoolEVENT%events%rawx
    myEVENTFILE%EVENT%RAWY = evttoolEVENT%events%rawy
    myEVENTFILE%EVENT%ENERGY = evttoolEVENT%events%energy
    myEVENTFILE%EVENT%TIME = evttoolEVENT%events%time
    myEVENTFILE%EVENT%CCDNR = evttoolEVENT%events%ccdnr


    myEVENTFILE%EVENT%ROLL=0.D0
    myEVENTFILE%EVENT%EV2SRC=0
    
    !write(*,*) "IN:",nrows_ev, size(myEVENTFILE%EVENT%XD)

    do ii=1, nrows_ev
       
       ra= myEVENTFILE%EVENT%RA(ii) / myRAD2DEG
       dec= myEVENTFILE%EVENT%DEC(ii) / myRAD2DEG


       call  src_pos2xyz ( ra, dec, xd, yd, zd)
       
       myEVENTFILE%EVENT%XD(ii)= xd
       myEVENTFILE%EVENT%YD(ii)= yd
       myEVENTFILE%EVENT%ZD(ii)= zd

    enddo
    ! allocate event end


    ! ATTITUDE FILE
    ! first allocate the %ATT array of evt with the number of 
    ! atttitude extensions / number of active instruments   
    myEVENTFILE%NUM_ATT = evttoolEVENT%ninst
    allocate  ( myEVENTFILE%ATT(myEVENTFILE%NUM_ATT) )

    ! this step assigns to the array myEVENTFILE%WHICH_CCDNR
    ! the dimensions and the contents of the array  evttoolEVENT%ccdnr_infile
    ! which lists which CCD correaponds to what  CORRAT (ATT) extention
    ! the contents of the  evttoolEVENT%ccdnr_infile are inverse, 
    ! starting from higher and going to smaller. 
    ! CCDNR == CORRAT_EXTENTION_NUMBER BUT ATT INDEX IS INVERSE   
    ! i.e. CCDNR=6 -> CORRAR EXTENTION 6, BUT THIS IS STORED IN 
    ! ELEMENT 2 OF THE ARRAY ATT(2)
    allocate  ( myEVENTFILE%WHICH_CCDNR(evttoolEVENT%ninst  ) )
    myEVENTFILE%WHICH_CCDNR = evttoolEVENT%ccdnr_infile


    max_nrows = -9

    do ii = 1 , myEVENTFILE%NUM_ATT  
       
       nrows = evttoolEVENT%corrat(ii)%nelements
       
       if(nrows.gt.max_nrows) max_nrows = nrows

       myEVENTFILE%ATT(ii)%ACTIVE_TEL = evttoolEVENT%ccdnr_infile(ii)
       
       myEVENTFILE%ATT(ii)%SIZE = nrows
       
       allocate  ( myEVENTFILE%ATT(ii)%RA(myEVENTFILE%ATT(ii)%SIZE), &
            myEVENTFILE%ATT(ii)%DEC(myEVENTFILE%ATT(ii)%SIZE) )
       allocate  ( myEVENTFILE%ATT(ii)%ROLL(myEVENTFILE%ATT(ii)%SIZE), &
            myEVENTFILE%ATT(ii)%TIME(myEVENTFILE%ATT(ii)%SIZE) )
       
       
       myEVENTFILE%ATT(ii)%RA = evttoolEVENT%corrat(ii)%ra
       myEVENTFILE%ATT(ii)%DEC = evttoolEVENT%corrat(ii)%dec
       myEVENTFILE%ATT(ii)%ROLL = evttoolEVENT%corrat(ii)%roll
       myEVENTFILE%ATT(ii)%TIME = evttoolEVENT%corrat(ii)%time

       !write(*,*) "ATT:",evttoolEVENT%ccdnr_infile(ii), ii,evttoolEVENT%corrat(ii)%roll(1), evttoolEVENT%corrat(ii)%ccdnr, &
       !     evttoolEVENT%corrat(ii)%extname
    enddo


    call init_ccd(myEVENTFILE)
    call interpolate_roll(myEVENTFILE)

  end subroutine evttool_2_myevent









!  subroutine read_event(file, myEVENTFILE)
!
!    implicit none
!
!    CHARACTER(len=*) :: file
!
!    TYPE(EVENTFILE) :: myEVENTFILE
!
!    integer :: unit, status, rwmode, blocksize, nrows_ev
!
!
!    status=0
!    rwmode=0 ! open as readonly 
!
!    call ftgiou ( unit, status )
!
!    !write(*,*)  file
!
!    call ftopen( unit, file, rwmode, blocksize, status)  
!
!    !call ftmnhd(unit, 2 , 'EVENTS', 0, status)
!    !call ftgnrw(unit, nrows_ev, status)  
!
!    call read_event_extension( unit,  myEVENTFILE)
!
!    call read_attitude_extension( unit,  myEVENTFILE)
!
!    call init_ccd(myEVENTFILE)
!    call interpolate_roll(myEVENTFILE)
!
!  end subroutine read_event




  subroutine read_event_extension( unit,  myEVENTFILE)
    !this reads the event extensionm of an event file. 
    use basic_calcs 

    implicit none


    integer unit

    TYPE(EVENTFILE) :: myEVENTFILE

    integer nrows_ev, irow, status
    
    logical :: casesn=.false., anyf

    integer :: colra, coldec, colenergy, coltime, colrawx, colrawy, colccd

    double precision dnullval, dvalues

    double precision ra, dec, xd, yd, zd

    integer inullval, ivalues

    status=0
    dnullval=0.d0 
    inullval=0 

    call ftmnhd(unit, 2 , 'EVENTS', 0, status)
    call ftgnrw(unit, nrows_ev, status)  
    
    myEVENTFILE%EVENT%SIZE = nrows_ev


    call ftgcno(unit,casesn,'RA_DEG',colra,status)   
    !call FTGTCL(unit,colra, datacode,repeat,width,status)
    call ftgcno(unit,casesn,'DEC_DEG',coldec,status)  
    call ftgcno(unit,casesn,'ENERGY',colenergy,status)  
    call ftgcno(unit,casesn,'TIME',coltime,status)     
    call ftgcno(unit,casesn,'RAWX',colrawx,status)   
    call ftgcno(unit,casesn,'RAWY',colrawy,status)   
    call ftgcno(unit,casesn,'CCDNR',colccd,status) 
    
    allocate  ( myEVENTFILE%EVENT%XD(nrows_ev), myEVENTFILE%EVENT%YD(nrows_ev) )
    allocate  ( myEVENTFILE%EVENT%ZD(nrows_ev) )
    
    allocate  ( myEVENTFILE%EVENT%RA(nrows_ev), myEVENTFILE%EVENT%DEC(nrows_ev) )
    allocate  ( myEVENTFILE%EVENT%RAWX(nrows_ev), myEVENTFILE%EVENT%RAWY(nrows_ev) )
    allocate  ( myEVENTFILE%EVENT%ENERGY(nrows_ev), myEVENTFILE%EVENT%TIME(nrows_ev) )
    allocate  ( myEVENTFILE%EVENT%ROLL(nrows_ev), myEVENTFILE%EVENT%CCDNR(nrows_ev) )
    allocate  ( myEVENTFILE%EVENT%EV2SRC(nrows_ev) )

    myEVENTFILE%EVENT%ROLL=0.D0
    myEVENTFILE%EVENT%EV2SRC=0
    do irow=1, nrows_ev

       call FTGCVD(unit,colra,irow,1,1,dnullval,dvalues,anyf,status)
       myEVENTFILE%EVENT%RA(irow)= dvalues

       call FTGCVD(unit,coldec,irow,1,1,dnullval,dvalues,anyf,status)
       myEVENTFILE%EVENT%DEC(irow)= dvalues

       call FTGCVD(unit,colenergy,irow,1,1,dnullval,dvalues,anyf,status)
       myEVENTFILE%EVENT%ENERGY(irow)= dvalues

       call FTGCVD(unit,coltime,irow,1,1,dnullval,dvalues,anyf,status)
       myEVENTFILE%EVENT%TIME(irow)= dvalues

       call FTGCVI(unit,colrawx,irow,1,1,inullval,ivalues,anyf,status)
       myEVENTFILE%EVENT%RAWX(irow)= ivalues

       call FTGCVI(unit,colrawy,irow,1,1,inullval,ivalues,anyf,status)
       myEVENTFILE%EVENT%RAWY(irow)= ivalues

       call FTGCVI(unit,colccd,irow,1,1,inullval,ivalues,anyf,status)
       myEVENTFILE%EVENT%CCDNR(irow)= ivalues

       
       ra= myEVENTFILE%EVENT%RA(irow) / myRAD2DEG
       dec= myEVENTFILE%EVENT%DEC(irow) / myRAD2DEG

       call  src_pos2xyz ( ra, dec, xd, yd, zd)

       myEVENTFILE%EVENT%XD(irow)= xd
       myEVENTFILE%EVENT%YD(irow)= yd
       myEVENTFILE%EVENT%ZD(irow)= zd



    end do

  end subroutine read_event_extension




!  subroutine read_attitude_extension( unit,  myEVENTFILE)
!    ! this reads the attitude extenstions of the EVENT file
!    ! the assumption is that they are called CORRAT1-7
!    ! if an attitude extension is missing the routine
!    ! will not crash but it is assumed that the 
!    ! telescope is not there
!    ! the roll, time and ra, dec are read. This
!    ! will be used to interpolate the roll angle
!    ! of a given photon event
!    
!
!    implicit none
!
!    integer unit
!
!    TYPE(EVENTFILE) :: myEVENTFILE
!
!    integer max_nrows, nrows, iactive, itelescope, irow, status
!    
!    logical :: casesn=.false., anyf
!
!    integer :: colra, coldec, coltime, colroll
!
!    double precision dnullval, dvalues
!
!    character*20 cornam
!
!    dnullval=0.d0 
!    max_nrows=0
!    iactive=0
!
!
!    ! this loop checks if the att extensions for all 
!    ! telescopes are present and upates the relevant 
!    ! parameters, i.e. array that holds the index
!    ! of each telescope (ACTIVE_TEL) 
!    ! the number of colums in the att file of each telescope (SIZE)
!    ! and the toal number of active telescopes NUM_ACTIVE_TEL
!    ! For the array assignment later on the maximum number of
!    ! rows in a given ATT extenstion is used
!    do itelescope=1, MAX_NUM_TELESCOPES
!       write (cornam, '(''CORRATT'', I1)') itelescope
!
!       status=0
!       call ftmnhd(unit, 2 ,cornam , 0, status)
!       write(*,*) cornam, status
!       if(status.eq.0)then
!          
!          iactive=iactive+1
!          
!          call ftgnrw(unit, nrows, status)  
!       
!          if(nrows.gt.max_nrows)max_nrows=nrows
!    
!          myEVENTFILE%ATT%ACTIVE_TEL(iactive)= itelescope
!
!          myEVENTFILE%ATT%REV_ACTIVE_TEL(itelescope) = iactive
!
!          myEVENTFILE%ATT%SIZE(iactive)= nrows
!
!       endif
!    end do
!
!    myEVENTFILE%ATT%NUM_ACTIVE_TEL = iactive 
!
!    allocate  ( myEVENTFILE%ATT%RA( iactive, max_nrows), &
!         myEVENTFILE%ATT%DEC(iactive, max_nrows) )
!    allocate  ( myEVENTFILE%ATT%ROLL(iactive, max_nrows), &
!         myEVENTFILE%ATT%TIME(iactive, max_nrows) )
!   
! 
!    
!    do iactive=1, myEVENTFILE%ATT%NUM_ACTIVE_TEL 
!       write (cornam, '(''CORRATT'', I1)') myEVENTFILE%ATT%ACTIVE_TEL(iactive)
!       
!       status=0
!       call ftmnhd(unit, 2 ,cornam , 0, status)
!       
!       call ftgcno(unit,casesn,'RA',colra,status)   
!       call ftgcno(unit,casesn,'DEC',coldec,status)  
!       call ftgcno(unit,casesn,'TIME',coltime,status)     
!       call ftgcno(unit,casesn,'ROLL',colroll,status)   
!       
!       do irow=1, myEVENTFILE%ATT%SIZE(iactive)
!
!          call FTGCVD(unit,colra,irow,1,1,dnullval,dvalues,anyf,status)
!          myEVENTFILE%ATT%RA(iactive, irow )= dvalues
!          
!          call FTGCVD(unit,coldec,irow,1,1,dnullval,dvalues,anyf,status)
!          myEVENTFILE%ATT%DEC(iactive, irow)= dvalues
!          
!          call FTGCVD(unit,coltime,irow,1,1,dnullval,dvalues,anyf,status)
!          myEVENTFILE%ATT%TIME(iactive, irow)= dvalues
!          
!          call FTGCVD(unit,colroll,irow,1,1,dnullval,dvalues,anyf,status)
!          myEVENTFILE%ATT%ROLL(iactive, irow)= dvalues
!          
!       end do
!    end do
!    
!
!   
!  end subroutine read_attitude_extension




  subroutine init_ccd(myEVENTFILE)    
    ! this simply initialises the centre of each CCD
    ! in the future we may want to do more complex things
    ! i.e. read from header etc
    implicit none

    TYPE(EVENTFILE) :: myEVENTFILE

    integer irow, iactive


    !do iactive=1, myEVENTFILE%ATT%NUM_ACTIVE_TEL 
    do iactive=1, MAX_NUM_TELESCOPES

    
       myEVENTFILE%CCD%RAW_X0(iactive)=192.0D0
       myEVENTFILE%CCD%RAW_Y0(iactive)=192.0D0

    enddo

    
  end subroutine init_ccd



  subroutine interpolate_roll(myEVENTFILE)    
    
    implicit none

    TYPE(EVENTFILE) :: myEVENTFILE

    integer isrc, ii, iccdnr, left, right, jj
    
    double precision atime,x0, x1, y0, y1, y
 

    left = 0

    do isrc = 1, myEVENTFILE%EVENT%SIZE
       
       atime = myEVENTFILE%EVENT%TIME(isrc)
       iccdnr = myEVENTFILE%EVENT%CCDNR(isrc)
       
       ! find which element of array myEVENTFILE%WHICH_CCDNR
       ! the ccd number is
       do jj = 1, SIZE(myEVENTFILE%WHICH_CCDNR)
          if( myEVENTFILE%WHICH_CCDNR(jj) == iccdnr ) ii=jj
       enddo

       !write(*,*) "1",isrc, atime, iccdnr, ii
       !write(*,*)  myEVENTFILE%WHICH_CCDNR
       !write(*,*) "1",isrc, atime, iccdnr, ii, myEVENTFILE%WHICH_CCDNR(ii)
       !where ( (myEVENTFILE%ATT(ii)%TIME - atime) .ge. 0 ) 
       !   left=minloc((myEVENTFILE%ATT(ii)%TIME - atime),1)
       !ENDWHERE
       !left =  minloc(abs(myEVENTFILE%ATT(ii)%TIME - atime),1)
       !if(left<1)left=1
       !if(left>=myEVENTFILE%ATT(ii)%SIZE)left=myEVENTFILE%ATT(ii)%SIZE-1
       !right =left+1
       
       call find_bracket (myEVENTFILE, atime, ii, left, right )
       
       !write(*,*) "2", left, right

       y0=myEVENTFILE%ATT(ii)%ROLL(left) 
       y1=myEVENTFILE%ATT(ii)%ROLL(right)
       x0=myEVENTFILE%ATT(ii)%TIME(left) 
       x1=myEVENTFILE%ATT(ii)%TIME(right)

       !write(*,*) "3", x0, atime, x1, y0, y1


       y = y0 + (y1 - y0) * (atime - x0) / (x1 - x0)

       myEVENTFILE%EVENT%ROLL(isrc)=y !* myRAD2DEG

       !write(*,*) "ASSIGN:",isrc, iccdnr, ii, x0, atime, x1, "   ", y0, y, y1

    enddo

    
  end subroutine interpolate_roll



  subroutine find_bracket (myEVENTFILE, atime, itelesc, left, right )
    
    implicit none

    TYPE(EVENTFILE) :: myEVENTFILE

    integer irow, itelesc, left, right
    
    double precision atime
    
    do irow = left + 1, myEVENTFILE%ATT(itelesc)%SIZE - 1
       
       if ( atime < myEVENTFILE%ATT(itelesc)%TIME(irow) ) then
          left = irow - 1
          right = irow
          return
       end if

    end do
    
!    left = myEVENTFILE%ATT%SIZE(itelesc) - 1
!    right = myEVENTFILE%ATT%SIZE(itelesc)
    left = myEVENTFILE%ATT(itelesc)%SIZE - 1
    right = myEVENTFILE%ATT(itelesc)%SIZE
   
  end subroutine find_bracket

  



end  module data_evt
