! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.

program main
  


  !use iso_fortran_env ! a more standard replacement for numkind
  !use param          ! standard eROSITA module handling parameter/logging/error interface
  !use ero_general    ! standard eROSITA module handling flag definitions etc
  
  use slib_constants
  use slib_engine
  use slib_image
  use erolib_bintable_mod, only : TCOL_READ_BIT
  use erolib_eventfile_mod
  use slib_data_psf
  use slib_data_src
  use detect_io
  use slib_heal

  
  use slib_calc_space, only : slib_fluxR,  slib_fluxEEF
 
  
  implicit none

  !TYPE(STACKPSF) myStackPSF(2)
  

  character(len=256), dimension(1) :: eventlist !, namelist
  character(len=256) srcfile

  integer status

  INTEGER NUM_EVENTS, IEVT, ci, kk, corratt_i
  !INTEGER I, J
  type(tEventFile), dimension(:), allocatable :: EventFile_in
  type(SRC)  ::  srclist
  TYPE(FPSF) myFPSF

  !type(mllist_type) :: mllist
  !INTEGER :: ierr
  
  INTEGER DIMS(2), NUM_SRC, ISRC, INAME, MAX_PSF_IMG_SIZE 
  character ( len = 256) filename, name
  DOUBLE PRECISION IMG_PIXEL_MM
  real               :: t1,t2
  INTEGER, PARAMETER :: MR=100
  !DOUBLE PRECISION RR(MR), FLUXR(MR)

  !these are for data/time, benchmarking
  !character(8)  :: date
  !character(10) :: time
  !character(5)  :: zone
  !integer,dimension(8) :: values
  INTEGER :: &
  !      nb_ticks_initial, & ! initial value of the clock tick counter
  !     nb_ticks_final,   & ! final value of the clock tick counter
       nb_ticks_max,     & ! maximum value of the clock counter
       nb_ticks_sec !,     & ! number of clock ticks per second
  !     nb_ticks           ! number of clock ticks of the code
  !REAL :: elapsed_time  ! real time in seconds
  !these are for data/time, benchmarking

  logical :: HEALFLAG
  
  !REAL SLAMCH

  !integer IPIX(7), IND(7)
  integer, dimension(:), allocatable :: ccdnrs
  
  ! THESE PARAMS ARE HARD-CORED EEF
  INTEGER, parameter :: MEEF=12
  real(kind=real64), DIMENSION(MEEF) :: EEF 
  !real(kind=real64), DIMENSION(MEEF) :: REEF
  ! THESE PARAMS ARE HARD-CORED EEF
  EEF = (/0.4,0.45,0.5,0.55,0.6,0.65,0.7,0.75,0.8,0.85,0.9,0.95/)

  !IPIX=(/10,4,3,5,10,20,1/)
  !call sortPIX(ipix, ind)
  !write(*,*) IPIX
  !write(*,*) IND
  !stop
  
  CALL SYSTEM_CLOCK(COUNT_RATE=nb_ticks_sec, COUNT_MAX=nb_ticks_max)

  !DO I=1,MR
  !   RR(I)=I * 1.0
  !END DO
  IMG_PIXEL_MM = 0.03102808 ! 4 arcsec from Georg
  MAX_PSF_IMG_SIZE  = 100

  !eventlist(1)="/local/age/CLERC/erositasimul-synthetic/ccd1_new_events.fits"
  !eventlist(2)="/local/age/CLERC/erositasimul-synthetic/ccd2_new_events.fits"
  !eventlist(3)="/local/age/CLERC/erositasimul-synthetic/ccd3_new_events.fits"
  !eventlist(4)="/local/age/CLERC/erositasimul-synthetic/ccd4_new_events.fits"
  !eventlist(5)="/local/age/CLERC/erositasimul-synthetic/ccd5_new_events.fits"
  !eventlist(6)="/local/age/CLERC/erositasimul-synthetic/ccd6_new_events.fits"
  !eventlist(7)="/local/age/CLERC/erositasimul-synthetic/ccd7_new_events.fits"
  !eventlist(1)="/local/age/CLERC/erositasimul-synthetic/RES/cal_pnt_sft_f1em13_NOB_022_EventList_0.fits"
  eventlist(1)="/local/age/CLERC/erositasimul-synthetic/calccd1_021_EventList_0.fits"
  srcfile="/local/age/CLERC/erositasimul-synthetic/table3.fits"
  name="psfold"
  call slib_read_src(srcfile, srclist)

  !srcfile="/local/age/CLERC/erositasimul-synthetic/RES/cal_pnt_sft_f1em13_NOB_020_MaxLikSourceList_0.fits"
  !call slib_read_mllist(srcfile, mllist, ierr)
  !call slib_ml2src(mllist, srclist, ierr) 

  
  NUM_EVENTS=SIZE(eventlist)

  ! this erolib call
  status = 0
  call alloc_eventfiles(EventFile_in, NUM_EVENTS, status)


  !open events with erolib and read information
   do IEVT=1, NUM_EVENTS
   !do IEVT=1, 1

      ! these are erolib calls
      !
      ! Set up the extensions expected in the file, plus the columns expected for each extension.
      !
      call EventFile_in(IEVT)%read_ccd_nrs(eventlist(IEVT), ccdnrs, status)
      call EventFile_in(IEVT)%init(ccdnrs, status)
      deallocate(ccdnrs)

      ! We only want to read this subset of the EVENT-extension columns:
      call EventFile_in(IEVT)%events%ibset_col_useage("TIME",  TCOL_READ_BIT)
      call EventFile_in(IEVT)%events%ibset_col_useage("RA",    TCOL_READ_BIT)
      call EventFile_in(IEVT)%events%ibset_col_useage("DEC",   TCOL_READ_BIT)
      call EventFile_in(IEVT)%events%ibset_col_useage("PI",    TCOL_READ_BIT)
      call EventFile_in(IEVT)%events%ibset_col_useage("RAWX",  TCOL_READ_BIT)
      call EventFile_in(IEVT)%events%ibset_col_useage("RAWY",  TCOL_READ_BIT)
      call EventFile_in(IEVT)%events%ibset_col_useage("TM_NR", TCOL_READ_BIT)
      EventFile_in(IEVT)%events%useageIsSet = .true.

      corratt_i = EventFile_in(IEVT)%get_ext_index("ext", extCORRATT)
      do kk=1,size(EventFile_in(IEVT)%ccdnr_infile)
        ci = EventFile_in(IEVT)%ccdnr_infile(kk) ! TM number
        call EventFile_in(IEVT)%extList(corratt_i)%extPerTM(ci)%ibset_all_useage(TCOL_READ_BIT) ! read all cols.
        EventFile_in(IEVT)%extList(corratt_i)%extPerTM(ci)%useageIsSet = .true.
      end do

      ! Now we can open the file and read the required data.
      call EventFile_in(IEVT)%open(readonly, status, filename=eventlist(IEVT))
      call EventFile_in(IEVT)%read(status)
      call EventFile_in(IEVT)%close(status)


   enddo

   call cpu_time(t1)   
   HEALFLAG =.False.   
   call slib_run_wrap(EventFile_in, srclist, myFPSF, MAX_PSF_IMG_SIZE, IMG_PIXEL_MM, HEALFLAG)
   !call slib_run(EventFile_in, srcfile, myFPSF, MAX_PSF_IMG_SIZE, IMG_PIXEL_MM)

   call cpu_time(t2)
   write(*,*) 'slib_run_array : cpu_time =  ',t2-t1

   !deallocate events via erolib
   call dealloc_eventfiles(EventFile_in, status) ! IMS: this does a complete free of each eventfile.

   !IEVT=1
   !!ISRC=1
   !!write(*,*) "ERITE"
   !call slib_recon_inimage(eventlist(1), myFPSF%myCPSF, myFPSF%mySPSF(IEVT), srclist)
   !stop

   
   iname = index(name,' ') - 1  
   do IEVT=1, NUM_EVENTS

      DIMS = SHAPE(myFPSF%mySPSF(IEVT)%C1)
      
      NUM_SRC = DIMS(2)
      
      do ISRC = 1, NUM_SRC 
         !call date_and_time(date,time,zone,values)         
         !write(*,*)"starting on Date: ", date, " Time: ",time
         !CALL SYSTEM_CLOCK(COUNT=nb_ticks_initial)

         !call slib_fluxR(myFPSF%mySPSF(IEVT),myFPSF%myCPSF, ISRC, MR, RR, FLUXR)
         !do J = 1, MR
         !   write(20,*) RR(J), FLUXR(J), ISRC
         !enddo
         !write(*,*) SIZE(myStackPSF(IEVT)%NX1)
         !write(*,*) "DIMS",DIMS
         call cpu_time(t1)
         call  slib_run_write2arrays_fast(myFPSF%mySPSF(IEVT), myFPSF%myCPSF, ISRC)
         !call slib_fluxR_image(myFPSF%myCPSF, MR, RR, FLUXR)
         !do J = 1, MR
         !   write(30,*) RR(J), FLUXR(J), ISRC
         !enddo

         !call slib_fluxEEF(myFPSF%mySPSF(IEVT),myFPSF%myModelPSF(1), ISRC, MEEF, EEF, REEF)

         
         call cpu_time(t2)
         write(*,*) 'cpu_time, isrc=',t2-t1 ,isrc     
         write (filename, "(I4.4, A5, I4.4)") &
              IEVT, "_SRC_", ISRC
         filename=name(1:iname)//"_EVT_"//trim(filename)//'.img'
         write(*,*) filename         
         call slib_run_write2image(myFPSF%myCPSF%IMG, filename)
         write(*,*) "SUM IS:", sum(myFPSF%myCPSF%IMG)

      enddo
   enddo

   call slib_dealloc_FPSF(myFPSF) 


  stop
end program main
