! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.

module slib_cimage

  
  use iso_fortran_env

  implicit none

  private

  public :: slib_write2file

contains 

  subroutine slib_write2file(N_COEFF0, NX0, NY0, BETA0, &
       N_COEFF1, C_COEFF1, &       
       NX1, NY1, BETA1, &
       N_COEFF2, C_COEFF2, NX2, NY2, BETA2, &
       NIMG, ISRC, IE, IP)  
  
    integer(kind=int32), INTENT(IN) :: N_COEFF0, N_COEFF1, N_COEFF2, NIMG, ISRC, IE, IP
    
    complex(kind=real64) :: C_COEFF0(N_COEFF0), C_COEFF1(N_COEFF1), C_COEFF2(N_COEFF2)
    
    integer, INTENT(IN), DIMENSION(N_COEFF0) :: NX0, NY0
    integer, INTENT(IN), DIMENSION(N_COEFF1) :: NX1, NY1
    integer, INTENT(IN), DIMENSION(N_COEFF2) :: NX2, NY2
    
    REAL, INTENT(IN) :: BETA0, BETA1, BETA2
    
    REAL(kind=real64) :: x, y, xo, yo, xprime, yprime, pixsize=1
    DOUBLE PRECISION, DIMENSION(0:N_COEFF0) :: cx0,cy0  
    DOUBLE PRECISION, DIMENSION(0:N_COEFF1) :: cx1,cy1  
    DOUBLE PRECISION, DIMENSION(0:N_COEFF2) :: cx2,cy2  
    REAL, DIMENSION(NIMG, NIMG) :: array
    INTEGER i, ix, iy, ix_hermite, iy_hermite
    !REAL normx, normy, A1
    DOUBLE PRECISION :: normx, normy, A1
    real(kind=real64) :: PI=3.141592653589793D0 !DEG2RAD=3.141592653589793D0/180.D0

    
    !Fitsio needs these
    integer :: status,unit,blocksize,bitpix,naxis
    integer, DIMENSION(2) :: naxes(2)
    integer ::  group,fpixel,nelements
    logical simple,extend
    
    character ( len = 80 ) filename


    ! center of PSF image
    xo=dble(NIMG)/2.0D0+1.0D0
    yo=dble(NIMG)/2.0D0+1.0D0
    array=0.0
    
    !write(*,*) LOOKUP_N1
    !write(*,*) LOOKUP_N2
    !write(*,*) C_COEFF
 
    i=1
    do ix=1,NIMG
       x=dble(ix)*pixsize-pixsize/2.D0
       
       do iy=1,NIMG
          y=dble(iy)*pixsize-pixsize/2.D0
          


          xprime=(x-xo)
          yprime=(y-yo)
          xprime=xprime/beta0
          yprime=yprime/beta0
          
          !write(*,*) x,y,ix,iy,xprime,yprime,beta, NIMG
          
          call hermite_poly ( N_COEFF0, dble(xprime), cx0 )
          call hermite_poly ( N_COEFF0, dble(yprime), cy0 )

          do i = 1, N_COEFF0
             ix_hermite=NX0(I)
             iy_hermite=NY0(I)

             !write(*,*) ix_hermite, iy_hermite, C_COEFF(i)
             
             normx=1.D0 / sqrt( 2.D0**ix_hermite * sqrt(PI) * GAMMA(real(ix_hermite+1,kind=real64)) ) !factrl(ix_hermite) )
             normx=normx*1./sqrt(BETA0)*exp(-xprime**2/2.D0)  
             normy=1.D0 / sqrt( 2.D0**iy_hermite * sqrt(PI) * GAMMA(real(iy_hermite+1,kind=real64)) ) !factrl(iy_hermite) )
             normy=normy*1.D0/sqrt(BETA0)*exp(-yprime**2/2.D0)  
             
             !if(ix==10 && iy==4)then
             !   write(*,*) normx, 2**ix_hermite, sqrt(PI), GAMMA(real(ix_hermite+1)), exp(-xprime**2/2) , xprime, sqrt(BETA)
             !enddo
             A1=REALPART(C_COEFF0(i))
             array(ix,iy) = array(ix,iy) + REAL(A1 * normx *normy * real(cx0(ix_hermite),kind=real64) * real(cy0(iy_hermite),kind=real64))
             
             !write(*,*) ix_hermite, iy_hermite, normx, normy,  A1 ,  real(cx(ix_hermite)),  real(cy(iy_hermite)),  array(ix,iy)

          enddo


          
          xprime=(x-xo)
          yprime=(y-yo)
          xprime=xprime/beta1
          yprime=yprime/beta1
          
          !write(*,*) x,y,ix,iy,xprime,yprime,beta, NIMG
          
          call hermite_poly ( N_COEFF1, dble(xprime), cx1 )
          call hermite_poly ( N_COEFF1, dble(yprime), cy1 )

          do i = 1, N_COEFF1
             ix_hermite=NX1(I)
             iy_hermite=NY1(I)

             !write(*,*) ix_hermite, iy_hermite, C_COEFF(i)
             
             normx=1.D0 / sqrt( 2**ix_hermite * sqrt(PI) * GAMMA(real(ix_hermite+1,kind=real64)) ) !factrl(ix_hermite) )
             normx=normx*1./sqrt(BETA1)*exp(-xprime**2/2.D0)  
             normy=1.D0 / sqrt( 2**iy_hermite * sqrt(PI) * GAMMA(real(iy_hermite+1,kind=real64)) ) !factrl(iy_hermite) )
             normy=normy*1.D0/sqrt(BETA1)*exp(-yprime**2/2.D0)  
             
             !if(ix==10 && iy==4)then
             !   write(*,*) normx, 2**ix_hermite, sqrt(PI), GAMMA(real(ix_hermite+1)), exp(-xprime**2/2) , xprime, sqrt(BETA)
             !enddo
             A1=REALPART(C_COEFF1(i))
             array(ix,iy) = array(ix,iy) + REAL(A1 * normx *normy * real(cx1(ix_hermite),kind=real64) * real(cy1(iy_hermite),kind=real64))
             
             !write(*,*) ix_hermite, iy_hermite, normx, normy,  A1 ,  real(cx(ix_hermite)),  real(cy(iy_hermite)),  array(ix,iy)

          enddo
          

          xprime=(x-xo)
          yprime=(y-yo)
          xprime=xprime/beta2
          yprime=yprime/beta2
          
          !write(*,*) x,y,ix,iy,xprime,yprime,beta, NIMG
          
          call hermite_poly ( N_COEFF2, dble(xprime), cx2 )
          call hermite_poly ( N_COEFF2, dble(yprime), cy2 )

          do i = 1, N_COEFF2
             ix_hermite=NX2(I)
             iy_hermite=NY2(I)

             !write(*,*) ix_hermite, iy_hermite, C_COEFF(i)
             
             normx=1.D0 / sqrt( 2.D0**ix_hermite * sqrt(PI) * GAMMA(real(ix_hermite+1,kind=real64)) ) !factrl(ix_hermite) )
             normx=normx*1./sqrt(BETA2)*exp(-xprime**2/2.D0)  
             normy=1.D0 / sqrt( 2.D0**iy_hermite * sqrt(PI) * GAMMA(real(iy_hermite+1,kind=real64)) ) !factrl(iy_hermite) )
             normy=normy*1.D0/sqrt(BETA2)*exp(-yprime**2/2.D0)  
             
             !if(ix==10 && iy==4)then
             !   write(*,*) normx, 2**ix_hermite, sqrt(PI), GAMMA(real(ix_hermite+1)), exp(-xprime**2/2) , xprime, sqrt(BETA)
             !enddo
             A1=REALPART(C_COEFF2(i))
             array(ix,iy) = array(ix,iy) + REAL(A1 * normx *normy * real(cx2(ix_hermite),kind=real64) * real(cy2(iy_hermite),kind=real64))
             
             !write(*,*) ix_hermite, iy_hermite, normx, normy,  A1 ,  real(cx(ix_hermite)),  real(cy(iy_hermite)),  array(ix,iy)

          enddo
          
          !write(*,*) ix, iy, array(ix,iy)

       enddo
    enddo


 
    ! WRITE_IMAGE
    
    status=0
!
    !  Name of the FITS file to be created:
    !
    !  Name of the FITS file to be created:
    write (filename, "(I4.4, A2, I4.4, A2, I4.4)") &
         ISRC, "_E", IE, "_P", IP
    filename="comp_SRC"//trim(filename)//'.img'

    write(*,*) filename

    !filename="psf.fits"  
    !
    !  Delete the file if it already exists, so we can then recreate it.
    !
    call delete_file ( filename, status )
    !
    !  Get an unused Logical Unit Number to use to open the FITS file.
    !
    call ftgiou ( unit, status )
    !
    !  Create the new empty FITS file.
    !
    blocksize=1
    call ftinit(unit,filename,blocksize,status)
    !
    !  Initialize parameters about the FITS image (300 x 200 16-bit integers).
    !
    simple=.true. 
    bitpix=-32
    naxis=2
    naxes(1)=NIMG
    naxes(2)=NIMG
    extend=.true.
    
    
    !  Write the required header keywords.
    call ftphpr(unit,simple,bitpix,naxis,naxes,0,1,extend,status)
    
    !  Write the array to the FITS file.
    
    group=1
    fpixel=1
    nelements=naxes(1)*naxes(2)
    !call ftpprd(unit,group,fpixel,nelements,array_gauss,status)
    call ftppre(unit,group,fpixel,nelements,array,status)
    !
    !  Write another optional keyword to the header.
    !
    !  call ftpkyj(unit,'EXPOSURE',1500,'Total Exposure Time',status)
    !
    !  Close the file and free the unit number.
    !
    call ftclos(unit, status)
    call ftfiou(unit, status)
    !
    !  Check for any error, and if so print out error messages
    !
    if (status > 0) then
       call print_error(status)
    end if
    
  end subroutine slib_write2file




subroutine print_error(status)
!
!*******************************************************************************
!
!! PRINT_ERROR prints out the FITSIO error messages to the user.
!
  integer status
  character ( len = 30 ) errtext
  character ( len = 80 ) errmessage
!
!  Check if status is OK (no error); if so, simply return.
!
  if (status <= 0) then
    return
  end if
!
!  Get the text string which describes the error
!
  call ftgerr(status,errtext)
  print *,'FITSIO Error Status =',status,': ',errtext
!
!  Read and print out all the error messages on the FITSIO stack
!
  call ftgmsg(errmessage)
  do while (errmessage .ne. ' ')
      print *,errmessage
      call ftgmsg(errmessage)
  end do

  return
end subroutine print_error



subroutine delete_file(filename,status)
!
!*******************************************************************************
!
!! DELETE_FILE deletes a FITS file.
!
  integer status,unit,blocksize
  character ( len = * ) filename
!
!  Simply return if status is greater than zero.
!

  !write(*,*) filename, status


  if (status > 0) then
    return
  end if
!
!  Get an unused Logical Unit Number to use to open the FITS file
!
  call ftgiou ( unit, status )
!
!  Try to open the file, to see if it exists
!
  call ftopen ( unit, filename, 1, blocksize, status )

  !write(*,*) unit, filename, blocksize, status

  if ( status == 0 ) then
!
!  File was opened;  so now delete it 
!
    call ftdelt(unit,status)

  else if (status == 103)then
!
!     File doesn't exist, so just reset status to zero and clear errors
!
    status=0
    call ftcmsg

  else
!
!  There was some other error opening the file; delete the file anyway
!
      status=0
      call ftcmsg
      call ftdelt(unit,status)
  end if
!
!  Free the unit number for later reuse.
!
  call ftfiou(unit, status)

  return
end subroutine delete_file
!****************************


subroutine hermite_poly ( n, x, cx )

!*****************************************************************************80
!
!! HERMITE_POLY evaluates the physicist's Hermite polynomials at X.
!
!  Differential equation:
!
!    Y'' - 2 X Y' + 2 N Y = 0
!
!  First terms:
!
!      1
!      2 X
!      4 X^2     -  2
!      8 X^3     - 12 X
!     16 X^4     - 48 X^2     + 12
!     32 X^5    - 160 X^3    + 120 X
!     64 X^6    - 480 X^4    + 720 X^2    - 120
!    128 X^7   - 1344 X^5   + 3360 X^3   - 1680 X
!    256 X^8   - 3584 X^6  + 13440 X^4  - 13440 X^2   + 1680
!    512 X^9   - 9216 X^7  + 48384 X^5  - 80640 X^3  + 30240 X
!   1024 X^10 - 23040 X^8 + 161280 X^6 - 403200 X^4 + 302400 X^2 - 30240
!
!  Recursion:
!
!    H(0,X) = 1,
!    H(1,X) = 2*X,
!    H(N,X) = 2*X * H(N-1,X) - 2*(N-1) * H(N-2,X)
!
!  Norm:
!
!    Integral ( -oo < X < oo ) exp ( - X^2 ) * H(N,X)^2 dX
!    = sqrt ( PI ) * 2^N * N!
!
!    H(N,X) = (-1)^N * exp ( X^2 ) * dn/dXn ( exp(-X^2 ) )
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license. 
!
!  Modified:
!
!    01 October 2002
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Milton Abramowitz, Irene Stegun,
!    Handbook of Mathematical Functions,
!    National Bureau of Standards, 1964,
!    ISBN: 0-486-61272-4,
!    LC: QA47.A34.
!
!    Larry Andrews,
!    Special Functions of Mathematics for Engineers,
!    Second Edition, 
!    Oxford University Press, 1998.
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) N, the highest order polynomial to compute.
!    Note that polynomials 0 through N will be computed.
!
!    Input, real ( kind = 8 ) X, the point at which the polynomials are 
!    to be evaluated.
!
!    Output, real ( kind = 8 ) CX(0:N), the values of the first N+1 Hermite
!    polynomials at the point X.
!
  implicit none

  integer ( kind = int32 ) n

  real ( kind = real64 ) cx(0:n)
  integer ( kind = int32 ) i
  real ( kind =  real64 ) x

  if ( n < 0 ) then
    return
  end if

  cx(0) = 1.0D+00

  if ( n == 0 ) then
    return
  end if

  cx(1) = 2.0D+00 * x
   !write(*,*) "1",cx(1),x

  do i = 2, n
    cx(i) = 2.0D+00 * x * cx(i-1) - 2.0D+00 * real ( i - 1, kind = real64 ) * cx(i-2)
    !write(*,*) i,cx(i),x
  end do
  
  

  return
end subroutine hermite_poly



! APPARENTLY THIS IS NOT USED - COMMENT OUT TO SUPPRESS WARNINGS
!subroutine l_polynomial ( m, n, x, v )
!
!*****************************************************************************80
!
!! L_POLYNOMIAL evaluates the Laguerre polynomial L(n,x).
!
!  First terms:
!
!      1
!     -X     +  1
!   (  X^2 -  4 X      +  2 ) / 2
!   ( -X^3 +  9 X^2 -  18 X    +    6 ) / 6
!   (  X^4 - 16 X^3 +  72 X^2 -   96 X +      24 ) / 24
!   ( -X^5 + 25 X^4 - 200 X^3 +  600 X^2 -   600 X    +  120 ) / 120
!   (  X^6 - 36 X^5 + 450 X^4 - 2400 X^3 +  5400 X^2 -  4320 X     + 720 ) 
!     / 720
!   ( -X^7 + 49 X^6 - 882 X^5 + 7350 X^4 - 29400 X^3 + 52920 X^2 - 35280 X 
!     + 5040 ) / 5040
!
!  Recursion:
!
!    L(0,X) = 1
!    L(1,X) = 1 - X
!    L(N,X) = (2*N-1-X)/N * L(N-1,X) - (N-1)/N * L(N-2,X)
!
!  Orthogonality:
!
!    Integral ( 0 <= X < oo ) exp ( - X ) * L(N,X) * L(M,X) dX = delta ( M, N )
!
!  Relations:
!
!    L(N,X) = (-1)^N / N! * exp ( x ) * (d/dx)^n ( exp ( - x ) * x^n )  
!
!  Licensing:
!
!    This code is distributed under the GNU LGPL license. 
!
!  Modified:
!
!    07 March 2012
!
!  Author:
!
!    John Burkardt
!
!  Reference:
!
!    Milton Abramowitz, Irene Stegun,
!    Handbook of Mathematical Functions,
!    National Bureau of Standards, 1964,
!    ISBN: 0-486-61272-4,
!    LC: QA47.A34.
!
!  Parameters:
!
!    Input, integer ( kind = 4 ) M, the number of evaluation points.
!
!    Input, integer ( kind = 4 ) N, the highest order polynomial to compute.
!    Note that polynomials 0 through N will be computed.
!
!    Input, real ( kind = 8 ) X(M), the evaluation points.
!
!    Output, real ( kind = 8 ) V(M,0:N), the function values.
!
!  implicit none
!
!  integer ( kind = int32 ) m
!  integer ( kind = int32 ) n
!
!  integer ( kind = int32) j
!  real ( kind = real64 ) v(m,0:n)
!  real ( kind = real64 ) x(m)
!
!  if ( n < 0 ) then
!    return
!  end if
!
!  v(1:m,0) = 1.0D0
!
!  if ( n == 0 ) then
!    return
!  end if
!
!  v(1:m,1) = 1.0D0 - x(1:m)
! 
!  do j = 2, n
!
!    v(1:m,j) = ( ( real ( 2 * j - 1, kind = real64 ) - x(1:m) ) * v(1:m,j-1)   &
!                 - real (     j - 1, kind = real64 )            * v(1:m,j-2) ) &
!                 / real (     j,     kind = real64 )
!
!  end do
!
!  return
!end subroutine l_polynomial



end module slib_cimage
