! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.

module slib_data_psf
  
  use iso_fortran_env
  use slib_constants

  implicit none

  private 
  

  ! AGE 14/03/17: added  slib_init_SPSF as a public routine
  !               added  slib_deallocate_myStackPSF as a public routine
  public :: slib_init_FPSF, slib_dealloc_FPSF, &
            slib_init_CPSF, slib_dealloc_CPSF, &
            slib_read_model_psf,  slib_read_alltel_model_psf, &
            slib_init_SPSF, slib_deallocate_myStackPSF, &
            slib_dealloc_model_psf

contains



   !**********************************************************************
   subroutine slib_init_FPSF(myFPSF, myCPSF, myModelPSF, NUMFPSF, NUMSPSF) 

     implicit none

     TYPE(FPSF) :: myFPSF
     TYPE(CPSF) :: myCPSF
     TYPE(MODELPSF) :: myModelPSF
     INTEGER :: NUMFPSF
     INTEGER :: NUMSPSF(:)
     INTEGER I

     myFPSF%myCPSF = myCPSF
     
     allocate(myFPSF%mySPSF(NUMFPSF))

     
     do I=1,SIZE(myFPSF%mySPSF)
        call  slib_init_SPSF(myFPSF%mySPSF(I), myModelPSF, NUMSPSF(I))
     enddo
    
   end subroutine slib_init_FPSF



   !**********************************************************************
   subroutine slib_dealloc_FPSF(myFPSF) 

     implicit none

     TYPE(FPSF) :: myFPSF
     
     INTEGER I

     call slib_dealloc_CPSF(myFPSF%myCPSF)


     do I=1,SIZE(myFPSF%mySPSF)
        call  slib_deallocate_myStackPSF(myFPSF%mySPSF(I))
     enddo

     if(ALLOCATED(myFPSF%mySPSF)) DEALLOCATE (myFPSF%mySPSF)
     
   end subroutine slib_dealloc_FPSF



  !**********************************************************************
  subroutine slib_init_SPSF(myStackPSF, myModelPSF, NUMPSF)

    implicit none

    TYPE(STACKPSF), intent(out) :: myStackPSF 
    TYPE(MODELPSF) :: myModelPSF
    INTEGER NUMPSF

    INTEGER TOTAL_COEFF0, TOTAL_COEFF1, TOTAL_COEFF2

    TOTAL_COEFF0 = SIZE( myModelPSF%NX0)
    TOTAL_COEFF1 = SIZE( myModelPSF%NX1)
    TOTAL_COEFF2 = SIZE( myModelPSF%NX2)

    allocate  ( myStackPSF%NX0(TOTAL_COEFF0), myStackPSF%NY0(TOTAL_COEFF0) )
    allocate  ( myStackPSF%NX1(TOTAL_COEFF1), myStackPSF%NY1(TOTAL_COEFF1) )
    allocate  ( myStackPSF%NX2(TOTAL_COEFF2), myStackPSF%NY2(TOTAL_COEFF2) )

    allocate  ( myStackPSF%C0(TOTAL_COEFF0, NUMPSF) )
    allocate  ( myStackPSF%C1(TOTAL_COEFF1, NUMPSF) )
    allocate  ( myStackPSF%C2(TOTAL_COEFF2, NUMPSF) )

    myStackPSF%NX0=myModelPSF%NX0
    myStackPSF%NY0=myModelPSF%NY0
    myStackPSF%NX1=myModelPSF%NX1
    myStackPSF%NY1=myModelPSF%NY1
    myStackPSF%NX2=myModelPSF%NX2
    myStackPSF%NY2=myModelPSF%NY2

    ! added by AGE 18.3.2018
    myStackPSF%NUM_SRC=NUMPSF

  end subroutine slib_init_SPSF



!  !**********************************************************************
!  ! obsolete in v1.8(?)
!  subroutine slib_initialise_stack_psf(myModelPSF, mySRC, myStackPSF, &
!       MAX_PSF_IMG_SIZE, IMG_PIXEL_MM  )
!
!    implicit none
!
!    TYPE(STACKPSF), intent(out) :: myStackPSF 
!    TYPE(MODELPSF), intent(in) :: myModelPSF
!    TYPE(SRC), intent(in) :: mySRC
!    double precision, intent(in) :: IMG_PIXEL_MM   
!    integer, intent(in) :: MAX_PSF_IMG_SIZE
!
!
!    INTEGER TOTAL_COEFF0, TOTAL_COEFF1, TOTAL_COEFF2, TOTAL_SRC
!
!    TOTAL_SRC= mySRC%TOTAL
!    TOTAL_COEFF0 = SIZE( myModelPSF%NX0)
!    TOTAL_COEFF1 = SIZE( myModelPSF%NX1)
!    TOTAL_COEFF2 = SIZE( myModelPSF%NX2)
!
!    allocate  ( myStackPSF%NX0(TOTAL_COEFF0), myStackPSF%NY0(TOTAL_COEFF0) )
!    allocate  ( myStackPSF%NX1(TOTAL_COEFF1), myStackPSF%NY1(TOTAL_COEFF1) )
!    allocate  ( myStackPSF%NX2(TOTAL_COEFF2), myStackPSF%NY2(TOTAL_COEFF2) )
!
!    allocate  ( myStackPSF%C0(TOTAL_COEFF0, TOTAL_SRC) )
!    allocate  ( myStackPSF%C1(TOTAL_COEFF1, TOTAL_SRC) )
!    allocate  ( myStackPSF%C2(TOTAL_COEFF2, TOTAL_SRC) )
!
!    myStackPSF%NX0=myModelPSF%NX0
!    myStackPSF%NY0=myModelPSF%NY0
!    myStackPSF%NX1=myModelPSF%NX1
!    myStackPSF%NY1=myModelPSF%NY1
!    myStackPSF%NX2=myModelPSF%NX2
!    myStackPSF%NY2=myModelPSF%NY2
!    
!  end subroutine slib_initialise_stack_psf


  !**********************************************************************
  subroutine slib_deallocate_myStackPSF(myStackPSF)

    implicit none

    TYPE(STACKPSF) :: myStackPSF
 
    IF (ALLOCATED (myStackPSF%NX0)) DEALLOCATE (myStackPSF%NX0)
    IF (ALLOCATED (myStackPSF%NX1)) DEALLOCATE (myStackPSF%NX1)
    IF (ALLOCATED (myStackPSF%NX2)) DEALLOCATE (myStackPSF%NX2)
    IF (ALLOCATED (myStackPSF%NY0)) DEALLOCATE (myStackPSF%NY0)
    IF (ALLOCATED (myStackPSF%NY1)) DEALLOCATE (myStackPSF%NY1)
    IF (ALLOCATED (myStackPSF%NY2)) DEALLOCATE (myStackPSF%NY2)

    IF (ALLOCATED (myStackPSF%C0)) DEALLOCATE (myStackPSF%C0)
    IF (ALLOCATED (myStackPSF%C1)) DEALLOCATE (myStackPSF%C1)
    IF (ALLOCATED (myStackPSF%C2)) DEALLOCATE (myStackPSF%C2)

  end subroutine slib_deallocate_myStackPSF



!########################################################################
 subroutine slib_init_CPSF(calfile, myCPSF, MAX_PSF_IMG_SIZE, IMG_PIXEL_MM)
    
    implicit none
    
    CHARACTER(len=*) :: calfile

    TYPE(CPSF) :: myCPSF

    double precision, intent(in) :: IMG_PIXEL_MM   
    integer, intent(in) :: MAX_PSF_IMG_SIZE


    integer :: unit, status, rwmode, blocksize
    integer :: nhdu, hdutype, ie, ixy, col, inullval
    character*80 keyword, comment

    logical :: casesn=.false., anyf

    integer :: TOTAL_COEFF0, &
         TOTAL_COEFF1, TOTAL_COEFF2
    integer :: NH0, NH1, NH2
    double precision :: BETA0, BETA1, BETA2

    myCPSF%MAX_PSF_IMG_SIZE = MAX_PSF_IMG_SIZE
    myCPSF%IMG_PIXEL_MM  = IMG_PIXEL_MM  
 

    status=0
    rwmode=0 ! open as readonly 

    call ftgiou ( unit, status )

    !write(*,*)  "IN:", calfile

    call ftopen( unit, calfile, rwmode, blocksize, status)  
   

    !move to primary header extension to read keywords
    !about array sizes
    nhdu=1
    call FTMAHD(unit,nhdu, hdutype,status)
    

    !read basic shapelet prameters
    keyword="BETA0"
    call FTGKYD(unit, keyword, BETA0, comment, status)
    keyword="BETA1"
    call FTGKYD(unit, keyword, BETA1, comment, status)
    keyword="BETA2"
    call FTGKYD(unit, keyword, BETA2, comment, status)
    keyword="N0" 
    call FTGKYJ(unit, keyword, NH0, comment, status)
    keyword="N1" 
    call FTGKYJ(unit, keyword, NH1, comment, status)
    keyword="N2"
    call FTGKYJ(unit, keyword, NH2, comment, status)


    ! read sizes of various arrays
    keyword="SIZE_C0"
    call FTGKYJ(unit, keyword, TOTAL_COEFF0, comment, status)
    keyword="SIZE_C1"
    call FTGKYJ(unit, keyword, TOTAL_COEFF1, comment, status)
    keyword="SIZE_C2"
    call FTGKYJ(unit, keyword, TOTAL_COEFF2, comment, status)

    allocate  ( myCPSF%NX0(TOTAL_COEFF0), myCPSF%NY0(TOTAL_COEFF0) )
    allocate  ( myCPSF%NX1(TOTAL_COEFF1), myCPSF%NY1(TOTAL_COEFF1) )
    allocate  ( myCPSF%NX2(TOTAL_COEFF2), myCPSF%NY2(TOTAL_COEFF2) )

    allocate  ( myCPSF%PN0(TOTAL_COEFF0), myCPSF%PM0(TOTAL_COEFF0) )
    allocate  ( myCPSF%PN1(TOTAL_COEFF1), myCPSF%PM1(TOTAL_COEFF1) )
    allocate  ( myCPSF%PN2(TOTAL_COEFF2), myCPSF%PM2(TOTAL_COEFF2) )

    ALLOCATE ( myCPSF%MATRIX0 (TOTAL_COEFF0, TOTAL_COEFF0) )
    ALLOCATE ( myCPSF%MATRIX_INV0 (TOTAL_COEFF0, TOTAL_COEFF0) )
    ALLOCATE ( myCPSF%MATRIX1 (TOTAL_COEFF1, TOTAL_COEFF1) )
    ALLOCATE ( myCPSF%MATRIX_INV1 (TOTAL_COEFF1, TOTAL_COEFF1) )
    ALLOCATE ( myCPSF%MATRIX2 (TOTAL_COEFF2, TOTAL_COEFF2) )
    ALLOCATE ( myCPSF%MATRIX_INV2 (TOTAL_COEFF2, TOTAL_COEFF2) )

    myCPSF%BETA0=BETA0
    myCPSF%BETA1=BETA1
    myCPSF%BETA2=BETA2

    myCPSF%NH0=NH0
    myCPSF%NH1=NH1
    myCPSF%NH2=NH2

    !write(*,*) "TOTAL_COEFF0, TOTAL_COEFF1, TOTAL_COEFF2",TOTAL_COEFF0, TOTAL_COEFF1, TOTAL_COEFF2
    !write(*,*) SIZE(myCPSF%NX0)
    !write(*,*) SIZE(myCPSF%NX1)
    !write(*,*) SIZE(myCPSF%NX2)
    !write(*,*) "MAX_PSF_IMG_SIZE", MAX_PSF_IMG_SIZE
    !do ie=1,TOTAL_ENERGIES
    ie=1
    call FTMAHD(unit,ie+1, hdutype,status)
   
    call ftgcno(unit,casesn,'NX0',col,status)   
    ixy=1
    call FTGCVJ(unit,col,ixy,1,TOTAL_COEFF0, inullval, myCPSF%NX0,anyf,status)       
    call ftgcno(unit,casesn,'NX1',col,status)   
    call FTGCVJ(unit,col,ixy,1,TOTAL_COEFF1, inullval, myCPSF%NX1,anyf,status)       
    call ftgcno(unit,casesn,'NX2',col,status)   
    call FTGCVJ(unit,col,ixy,1,TOTAL_COEFF2, inullval, myCPSF%NX2,anyf,status)
    call ftgcno(unit,casesn,'NY0',col,status)   
    call FTGCVJ(unit,col,ixy,1,TOTAL_COEFF0, inullval, myCPSF%NY0,anyf,status)       
    call ftgcno(unit,casesn,'NY1',col,status)   
    call FTGCVJ(unit,col,ixy,1,TOTAL_COEFF1, inullval, myCPSF%NY1,anyf,status)
    call ftgcno(unit,casesn,'NY2',col,status)   
    call FTGCVJ(unit,col,ixy,1,TOTAL_COEFF2, inullval, myCPSF%NY2,anyf,status)
    
    call get_polar_nm(TOTAL_COEFF0, myCPSF%NX0, myCPSF%NY0, myCPSF%PN0, myCPSF%PM0)
    call get_polar_nm(TOTAL_COEFF1, myCPSF%NX1, myCPSF%NY1, myCPSF%PN1, myCPSF%PM1)
    call get_polar_nm(TOTAL_COEFF2, myCPSF%NX2, myCPSF%NY2, myCPSF%PN2, myCPSF%PM2)
    
    call make_matrix( TOTAL_COEFF0, myCPSF%NX0, myCPSF%NY0, myCPSF%MATRIX0)
    myCPSF%MATRIX_INV0=myCPSF%MATRIX0
    CALL invert_matrix(TOTAL_COEFF0, myCPSF%MATRIX_INV0)

    call make_matrix( TOTAL_COEFF1, myCPSF%NX1, myCPSF%NY1, myCPSF%MATRIX1)
    myCPSF%MATRIX_INV1=myCPSF%MATRIX1
    CALL invert_matrix(TOTAL_COEFF1, myCPSF%MATRIX_INV1)
    
    call make_matrix( TOTAL_COEFF2, myCPSF%NX2, myCPSF%NY2, myCPSF%MATRIX2)
    myCPSF%MATRIX_INV2=myCPSF%MATRIX2
    CALL invert_matrix(TOTAL_COEFF2, myCPSF%MATRIX_INV2)   


    !AGE:  IMG_SHAPE1, IMG_SHAPE2 are allocatatble arrays 
    ALLOCATE ( myCPSF%IMG(MAX_PSF_IMG_SIZE, MAX_PSF_IMG_SIZE) )
    ALLOCATE ( myCPSF%IMG_SHAPE0(MAX_PSF_IMG_SIZE, MAX_PSF_IMG_SIZE, TOTAL_COEFF0) )
    ALLOCATE ( myCPSF%IMG_SHAPE1(MAX_PSF_IMG_SIZE, MAX_PSF_IMG_SIZE, TOTAL_COEFF1) )
    ALLOCATE ( myCPSF%IMG_SHAPE2(MAX_PSF_IMG_SIZE, MAX_PSF_IMG_SIZE, TOTAL_COEFF2) )
    ! this funcion estimates vaues at the centre of each pixel
    !call slib_fill_IMG_arrays(myCPSF)
    !this function estimates integrals within pixels and should be prefered
    call slib_integr_IMG_arrays(myCPSF)
     
  end subroutine slib_init_CPSF




!********************************************************************
  subroutine slib_dealloc_CPSF(myCPSF)

    implicit none

    TYPE(CPSF) :: myCPSF

    if(ALLOCATED (myCPSF%NX0)) DEALLOCATE (myCPSF%NX0)
    if(ALLOCATED (myCPSF%NX1)) DEALLOCATE (myCPSF%NX1)
    if(ALLOCATED (myCPSF%NX2)) DEALLOCATE (myCPSF%NX2)

    if(ALLOCATED (myCPSF%NY0)) DEALLOCATE (myCPSF%NY0)
    if(ALLOCATED (myCPSF%NY1)) DEALLOCATE (myCPSF%NY1)
    if(ALLOCATED (myCPSF%NY2)) DEALLOCATE (myCPSF%NY2)

    if(ALLOCATED (myCPSF%PN0)) DEALLOCATE (myCPSF%PN0)
    if(ALLOCATED (myCPSF%PN1)) DEALLOCATE (myCPSF%PN1)
    if(ALLOCATED (myCPSF%PN2)) DEALLOCATE (myCPSF%PN2)

    if(ALLOCATED (myCPSF%PM0)) DEALLOCATE (myCPSF%PM0)
    if(ALLOCATED (myCPSF%PM1)) DEALLOCATE (myCPSF%PM1)
    if(ALLOCATED (myCPSF%PM2)) DEALLOCATE (myCPSF%PM2)

    if(ALLOCATED (myCPSF%MATRIX0)) DEALLOCATE (myCPSF%MATRIX0)
    if(ALLOCATED (myCPSF%MATRIX1)) DEALLOCATE (myCPSF%MATRIX1)
    if(ALLOCATED (myCPSF%MATRIX2)) DEALLOCATE (myCPSF%MATRIX2)

    if(ALLOCATED (myCPSF%MATRIX_INV0)) DEALLOCATE (myCPSF%MATRIX_INV0)
    if(ALLOCATED (myCPSF%MATRIX_INV1)) DEALLOCATE (myCPSF%MATRIX_INV1)
    if(ALLOCATED (myCPSF%MATRIX_INV2)) DEALLOCATE (myCPSF%MATRIX_INV2)


    if(ALLOCATED (myCPSF%IMG_SHAPE0)) DEALLOCATE (myCPSF%IMG_SHAPE0)
    if(ALLOCATED (myCPSF%IMG_SHAPE1)) DEALLOCATE (myCPSF%IMG_SHAPE1)
    if(ALLOCATED (myCPSF%IMG_SHAPE2)) DEALLOCATE (myCPSF%IMG_SHAPE2)
    IF (ALLOCATED (myCPSF%IMG)) DEALLOCATE (myCPSF%IMG)

  end subroutine slib_dealloc_CPSF


!#################################################################################  
  subroutine slib_read_alltel_model_psf(myModelPSF, myCPSF, MAX_PSF_IMG_SIZE, IMG_PIXEL_MM)

    use slib_constants
 
    TYPE(MODELPSF), intent(out) :: myModelPSF(:)
    TYPE(CPSF) myCPSF
    double precision, intent(in) :: IMG_PIXEL_MM   
    integer, intent(in) :: MAX_PSF_IMG_SIZE

    integer ITEL

    !gtcalf needs those to access CALDB
    !CG: chatter of gtcalf tied to prep_gtcalf (SASS_VERBOSITY)
    !integer, parameter :: chatter=1  ! caldb verbosity
    integer :: chatter
    character(len=30), parameter :: telescop = "eROSITA"
    character(len=30)            :: instrume
    character(len=30), parameter :: filter = "-", detnam = "-"
    character(len=30), parameter :: codenam = "SLET_PSF"
    character(len=30), parameter :: expression="-"
    character(len=30), parameter ::  strtdate="-", strtime="-", stpdate="-", stptime="-"
    integer, parameter :: group = 1, maxret = 10 
    integer, dimension(maxret) :: extno
    character(len=256), dimension(maxret) :: filename , online
    integer :: nret, nfound, status
    character(len=1)             :: istr
    !gtcalf needs those
 
   ! LOAD Model PSF FOR EACH INSTRUM
    ! To do:  instrume should defined in the loop
    do ITEL=1,MAX_NUM_TELESCOPES
       ! *** Changed 20/10/2016 GL
       ! Works only for FM1..FM7
       ! this to tweak
       write(istr,'(I1)') itel

       ! this is the correct
       instrume='TM'//istr

       ! this is for testing
       !II=1
       !write(istr,'(I1)') II
       !instrume='FM'//istr
       call gtcalf (chatter, telescop, instrume, detnam, filter, &
            codenam, strtdate, strtime, stpdate, stptime, &
            expression, maxret, filename, extno, online, &
            nret, nfound, status)
       !filename(1) = "/local/age/MyESASS/caldb/data/erosita/fm1/bcf/fm1_slet_psf_181217v03.fits"
       !write(*,*) filename
       if(ITEL.eq.1) call slib_init_CPSF(filename(1), myCPSF, MAX_PSF_IMG_SIZE, IMG_PIXEL_MM)
       
       call slib_read_model_psf(filename(1), myModelPSF(ITEL), MAX_PSF_IMG_SIZE, IMG_PIXEL_MM)
    enddo
   ! LOAD Model PSF FOR EACH INSTRUM
  end subroutine slib_read_alltel_model_psf


!#################################################################################  
  subroutine slib_read_model_psf(calfile, myMPSF, MAX_PSF_IMG_SIZE, IMG_PIXEL_MM)
    
    implicit none
    
    CHARACTER(len=*) :: calfile

    TYPE(MODELPSF) :: myMPSF

    double precision, intent(in) :: IMG_PIXEL_MM   
    integer, intent(in) :: MAX_PSF_IMG_SIZE

    integer :: unit, status, rwmode, blocksize
    integer :: nhdu, hdutype, ie, ixy, col, inullval
    character*80 keyword, comment
    double precision :: dd
    double precision dnullval
    logical :: casesn=.false., anyf

    integer :: TOTAL_ENERGIES, TOTAL_COEFF0, TOTAL_COEFF1, TOTAL_COEFF2, TOTAL_XY
    integer :: NH0, NH1, NH2
    double precision :: BETA0, BETA1, BETA2, XMIN, XMAX, XBIN, YMIN, YMAX, YBIN

    myMPSF%MAX_PSF_IMG_SIZE = MAX_PSF_IMG_SIZE
    myMPSF%IMG_PIXEL_MM  = IMG_PIXEL_MM  
 

    status=0
    rwmode=0 ! open as readonly 

    call ftgiou ( unit, status )

    !write(*,*)  "IN:", calfile

    call ftopen( unit, calfile, rwmode, blocksize, status)  
    
    !get the number of all hdus. this num_energies-1 (primary)
    call FTTHDU( unit, TOTAL_ENERGIES, status)

    TOTAL_ENERGIES=TOTAL_ENERGIES - 1

    !write(*,*) "TOTAL_ENERGIES:",TOTAL_ENERGIES


    !move to primary header extension to read keywords
    !about array sizes
    nhdu=1
    call FTMAHD(unit,nhdu, hdutype,status)
    

    !read basic shapelet prameters
    keyword="BETA0"
    call FTGKYD(unit, keyword, BETA0, comment, status)
    keyword="BETA1"
    call FTGKYD(unit, keyword, BETA1, comment, status)
    keyword="BETA2"
    call FTGKYD(unit, keyword, BETA2, comment, status)
    keyword="N0" 
    call FTGKYJ(unit, keyword, NH0, comment, status)
    keyword="N1" 
    call FTGKYJ(unit, keyword, NH1, comment, status)
    keyword="N2"
    call FTGKYJ(unit, keyword, NH2, comment, status)

    keyword="XOFF_MIN"
    call FTGKYD(unit, keyword, XMIN, comment, status)
    keyword="XOFF_MAX"
    call FTGKYD(unit, keyword, XMAX, comment, status)
    keyword="XOFF_BIN"
    call FTGKYD(unit, keyword, XBIN, comment, status)
    keyword="YOFF_MIN"
    call FTGKYD(unit, keyword, YMIN, comment, status)
    keyword="YOFF_MAX"
    call FTGKYD(unit, keyword, YMAX, comment, status)
    keyword="YOFF_BIN"
    call FTGKYD(unit, keyword, YBIN, comment, status)

    ! read sizes of various arrays
    keyword="SIZE_C0"
    call FTGKYJ(unit, keyword, TOTAL_COEFF0, comment, status)
    keyword="SIZE_C1"
    call FTGKYJ(unit, keyword, TOTAL_COEFF1, comment, status)
    keyword="SIZE_C2"
    call FTGKYJ(unit, keyword, TOTAL_COEFF2, comment, status)
    keyword="SIZE_XY"
    call FTGKYJ(unit, keyword, TOTAL_XY, comment, status)
    
    IF (ALLOCATED (myMPSF%ENERGY)) DEALLOCATE (myMPSF%ENERGY)
    allocate  ( myMPSF%ENERGY(TOTAL_ENERGIES) )

    IF (ALLOCATED (myMPSF%XOFF)) DEALLOCATE (myMPSF%XOFF)
    IF (ALLOCATED (myMPSF%YOFF)) DEALLOCATE (myMPSF%YOFF)
    allocate  ( myMPSF%XOFF(TOTAL_XY), myMPSF%YOFF(TOTAL_XY) )

    IF (ALLOCATED (myMPSF%NX0)) DEALLOCATE (myMPSF%NX0)
    IF (ALLOCATED (myMPSF%NX1)) DEALLOCATE (myMPSF%NX1)
    IF (ALLOCATED (myMPSF%NX2)) DEALLOCATE (myMPSF%NX2)
    IF (ALLOCATED (myMPSF%NY0)) DEALLOCATE (myMPSF%NY0)
    IF (ALLOCATED (myMPSF%NY1)) DEALLOCATE (myMPSF%NY1)
    IF (ALLOCATED (myMPSF%NY2)) DEALLOCATE (myMPSF%NY2)
    allocate  ( myMPSF%NX0(TOTAL_COEFF0), myMPSF%NY0(TOTAL_COEFF0) )
    allocate  ( myMPSF%NX1(TOTAL_COEFF1), myMPSF%NY1(TOTAL_COEFF1) )
    allocate  ( myMPSF%NX2(TOTAL_COEFF2), myMPSF%NY2(TOTAL_COEFF2) )

    IF (ALLOCATED (myMPSF%PN0)) DEALLOCATE (myMPSF%PN0)
    IF (ALLOCATED (myMPSF%PN1)) DEALLOCATE (myMPSF%PN1)
    IF (ALLOCATED (myMPSF%PN2)) DEALLOCATE (myMPSF%PN2)
    IF (ALLOCATED (myMPSF%PM0)) DEALLOCATE (myMPSF%PM0)
    IF (ALLOCATED (myMPSF%PM1)) DEALLOCATE (myMPSF%PM1)
    IF (ALLOCATED (myMPSF%PM2)) DEALLOCATE (myMPSF%PM2)
    allocate  ( myMPSF%PN0(TOTAL_COEFF0), myMPSF%PM0(TOTAL_COEFF0) )
    allocate  ( myMPSF%PN1(TOTAL_COEFF1), myMPSF%PM1(TOTAL_COEFF1) )
    allocate  ( myMPSF%PN2(TOTAL_COEFF2), myMPSF%PM2(TOTAL_COEFF2) )

    IF (ALLOCATED (myMPSF%MATRIX0)) DEALLOCATE (myMPSF%MATRIX0)
    ALLOCATE ( myMPSF%MATRIX0 (TOTAL_COEFF0, TOTAL_COEFF0) )

    IF (ALLOCATED (myMPSF%MATRIX_INV0)) DEALLOCATE (myMPSF%MATRIX_INV0)
    ALLOCATE ( myMPSF%MATRIX_INV0 (TOTAL_COEFF0, TOTAL_COEFF0) )

    IF (ALLOCATED (myMPSF%MATRIX1)) DEALLOCATE (myMPSF%MATRIX1)
    ALLOCATE ( myMPSF%MATRIX1 (TOTAL_COEFF1, TOTAL_COEFF1) )

    IF (ALLOCATED (myMPSF%MATRIX_INV1)) DEALLOCATE (myMPSF%MATRIX_INV1)
    ALLOCATE ( myMPSF%MATRIX_INV1 (TOTAL_COEFF1, TOTAL_COEFF1) )

    IF (ALLOCATED (myMPSF%MATRIX2)) DEALLOCATE (myMPSF%MATRIX2)
    ALLOCATE ( myMPSF%MATRIX2 (TOTAL_COEFF2, TOTAL_COEFF2) )

    IF (ALLOCATED (myMPSF%MATRIX_INV2)) DEALLOCATE (myMPSF%MATRIX_INV2)
    ALLOCATE ( myMPSF%MATRIX_INV2 (TOTAL_COEFF2, TOTAL_COEFF2) )

    IF (ALLOCATED (myMPSF%C0)) DEALLOCATE (myMPSF%C0)
    allocate  ( myMPSF%C0(TOTAL_COEFF0, TOTAL_XY, TOTAL_ENERGIES) )

    IF (ALLOCATED (myMPSF%C1)) DEALLOCATE (myMPSF%C1)
    allocate  ( myMPSF%C1(TOTAL_COEFF1, TOTAL_XY, TOTAL_ENERGIES) )

    IF (ALLOCATED (myMPSF%C2)) DEALLOCATE (myMPSF%C2)
    allocate  ( myMPSF%C2(TOTAL_COEFF2, TOTAL_XY, TOTAL_ENERGIES) )
    myMPSF%BETA0=BETA0
    myMPSF%BETA1=BETA1
    myMPSF%BETA2=BETA2

    myMPSF%XMIN=XMIN
    myMPSF%XMAX=XMAX
    myMPSF%XBIN=XBIN

    myMPSF%YMIN=YMIN
    myMPSF%YMAX=YMAX
    myMPSF%YBIN=YBIN

    myMPSF%NH0=NH0
    myMPSF%NH1=NH1
    myMPSF%NH2=NH2

    !write(*,*) SIZE(myMPSF%ENERGY), TOTAL_ENERGIES
    !write(*,*) "TOTAL XY",SIZE(myMPSF%XOFF), TOTAL_XY
    !write(*,*) "TOTAL_XY",SIZE(myMPSF%YOFF), TOTAL_XY
    !write(*,*) "C1:",SHAPE(myMPSF%C1)
    !write(*,*) "C2:",SHAPE(myMPSF%C2)


    
    do ie=1,TOTAL_ENERGIES

       call FTMAHD(unit,ie+1, hdutype,status)
       
       keyword="ENERGY"
       !key_no = 1
       !call FTGREC(unit,key_no, card,status)       
       call FTGKYD(unit, keyword, dd, comment, status)
       keyword="SIZE_XOF"
       call FTGKYJ(unit, keyword, TOTAL_XY, comment, status)
  

       myMPSF%ENERGY(ie)=dd
       !write(*,*) myMPSF%ENERGY(ie), ie, status, TOTAL_XY
       do ixy=1, TOTAL_XY

          call ftgcno(unit,casesn,'COEFF0',col,status)   
          call FTGCVD(unit,col,ixy,1,TOTAL_COEFF0,dnullval, myMPSF%C0(:, ixy, ie),anyf,status)
          call ftgcno(unit,casesn,'COEFF1',col,status)   
          call FTGCVD(unit,col,ixy,1,TOTAL_COEFF1,dnullval, myMPSF%C1(:, ixy, ie),anyf,status)
          call ftgcno(unit,casesn,'COEFF2',col,status)   
          call FTGCVD(unit,col,ixy,1,TOTAL_COEFF2,dnullval, myMPSF%C2(:, ixy, ie),anyf,status)

          call ftgcno(unit,casesn,'NX0',col,status)   
          call FTGCVJ(unit,col,ixy,1,TOTAL_COEFF0, inullval, myMPSF%NX0,anyf,status)
          call ftgcno(unit,casesn,'NX1',col,status)   
          call FTGCVJ(unit,col,ixy,1,TOTAL_COEFF1, inullval, myMPSF%NX1,anyf,status)

          call ftgcno(unit,casesn,'NX2',col,status)   
          call FTGCVJ(unit,col,ixy,1,TOTAL_COEFF2, inullval, myMPSF%NX2,anyf,status)

          call ftgcno(unit,casesn,'NY0',col,status)   
          call FTGCVJ(unit,col,ixy,1,TOTAL_COEFF0, inullval, myMPSF%NY0,anyf,status)
          call ftgcno(unit,casesn,'NY1',col,status)   
          call FTGCVJ(unit,col,ixy,1,TOTAL_COEFF1, inullval, myMPSF%NY1,anyf,status)
          call ftgcno(unit,casesn,'NY2',col,status)   
          call FTGCVJ(unit,col,ixy,1,TOTAL_COEFF2, inullval, myMPSF%NY2,anyf,status)

          call ftgcno(unit,casesn,'XOFFSET',col,status)   
          call FTGCVD(unit,col,ixy,1,1,dnullval, myMPSF%XOFF(ixy),anyf,status)
          call ftgcno(unit,casesn,'YOFFSET',col,status)   
          call FTGCVD(unit,col,ixy,1,1,dnullval, myMPSF%YOFF(ixy),anyf,status)

          !do i=1,TOTAL_COEFF1          
          !    write(*,*) i, ixy, ie, myMPSF%C1(i, ixy, ie), total_xy
          ! enddo

       enddo
    enddo

    call get_polar_nm(TOTAL_COEFF0, myMPSF%NX0, myMPSF%NY0, myMPSF%PN0, myMPSF%PM0)
    call get_polar_nm(TOTAL_COEFF1, myMPSF%NX1, myMPSF%NY1, myMPSF%PN1, myMPSF%PM1)
    call get_polar_nm(TOTAL_COEFF2, myMPSF%NX2, myMPSF%NY2, myMPSF%PN2, myMPSF%PM2)
    
    call make_matrix( TOTAL_COEFF0, myMPSF%NX0, myMPSF%NY0, myMPSF%MATRIX0)
    myMPSF%MATRIX_INV0=myMPSF%MATRIX0
    CALL invert_matrix(TOTAL_COEFF0, myMPSF%MATRIX_INV0)

    call make_matrix( TOTAL_COEFF1, myMPSF%NX1, myMPSF%NY1, myMPSF%MATRIX1)
    myMPSF%MATRIX_INV1=myMPSF%MATRIX1
    CALL invert_matrix(TOTAL_COEFF1, myMPSF%MATRIX_INV1)
    
    call make_matrix( TOTAL_COEFF2, myMPSF%NX2, myMPSF%NY2, myMPSF%MATRIX2)
    myMPSF%MATRIX_INV2=myMPSF%MATRIX2
    CALL invert_matrix(TOTAL_COEFF2, myMPSF%MATRIX_INV2)   

    !AGE:  IMG_SHAPE1, IMG_SHAPE2 are allocatatble arrays 
    !ALLOCATE ( myMPSF%IMG_SHAPE1(MAX_PSF_IMG_SIZE, MAX_PSF_IMG_SIZE, TOTAL_COEFF1) )
    !ALLOCATE ( myMPSF%IMG_SHAPE2(MAX_PSF_IMG_SIZE, MAX_PSF_IMG_SIZE, TOTAL_COEFF2) )
    !call slib_fill_IMG_arrays(myMPSF)
         
  end subroutine slib_read_model_psf

subroutine slib_dealloc_model_psf(myMPSF)
    
    implicit none
    
    TYPE(MODELPSF) :: myMPSF(:)

    integer ITEL
    
    do ITEL=1,MAX_NUM_TELESCOPES
       IF (ALLOCATED (myMPSF(ITEL)%ENERGY)) DEALLOCATE (myMPSF(ITEL)%ENERGY)
       IF (ALLOCATED (myMPSF(ITEL)%XOFF)) DEALLOCATE (myMPSF(ITEL)%XOFF)
       IF (ALLOCATED (myMPSF(ITEL)%YOFF)) DEALLOCATE (myMPSF(ITEL)%YOFF)
       
       IF (ALLOCATED (myMPSF(ITEL)%NX0)) DEALLOCATE (myMPSF(ITEL)%NX0)
       IF (ALLOCATED (myMPSF(ITEL)%NX1)) DEALLOCATE (myMPSF(ITEL)%NX1)
       IF (ALLOCATED (myMPSF(ITEL)%NX2)) DEALLOCATE (myMPSF(ITEL)%NX2)
       IF (ALLOCATED (myMPSF(ITEL)%NY0)) DEALLOCATE (myMPSF(ITEL)%NY0)
       IF (ALLOCATED (myMPSF(ITEL)%NY1)) DEALLOCATE (myMPSF(ITEL)%NY1)
       IF (ALLOCATED (myMPSF(ITEL)%NY2)) DEALLOCATE (myMPSF(ITEL)%NY2)
       
       IF (ALLOCATED (myMPSF(ITEL)%PN0)) DEALLOCATE (myMPSF(ITEL)%PN0)
       IF (ALLOCATED (myMPSF(ITEL)%PN1)) DEALLOCATE (myMPSF(ITEL)%PN1)
       IF (ALLOCATED (myMPSF(ITEL)%PN2)) DEALLOCATE (myMPSF(ITEL)%PN2)
       IF (ALLOCATED (myMPSF(ITEL)%PM0)) DEALLOCATE (myMPSF(ITEL)%PM0)
       IF (ALLOCATED (myMPSF(ITEL)%PM1)) DEALLOCATE (myMPSF(ITEL)%PM1)
       IF (ALLOCATED (myMPSF(ITEL)%PM2)) DEALLOCATE (myMPSF(ITEL)%PM2)
       
       IF (ALLOCATED (myMPSF(ITEL)%MATRIX0)) DEALLOCATE (myMPSF(ITEL)%MATRIX0)
       IF (ALLOCATED (myMPSF(ITEL)%MATRIX_INV0)) DEALLOCATE (myMPSF(ITEL)%MATRIX_INV0)
       IF (ALLOCATED (myMPSF(ITEL)%MATRIX1)) DEALLOCATE (myMPSF(ITEL)%MATRIX1)
       IF (ALLOCATED (myMPSF(ITEL)%MATRIX_INV1)) DEALLOCATE (myMPSF(ITEL)%MATRIX_INV1)
       IF (ALLOCATED (myMPSF(ITEL)%MATRIX2)) DEALLOCATE (myMPSF(ITEL)%MATRIX2)
       IF (ALLOCATED (myMPSF(ITEL)%MATRIX_INV2)) DEALLOCATE (myMPSF(ITEL)%MATRIX_INV2)
       IF (ALLOCATED (myMPSF(ITEL)%C0)) DEALLOCATE (myMPSF(ITEL)%C0)
       IF (ALLOCATED (myMPSF(ITEL)%C1)) DEALLOCATE (myMPSF(ITEL)%C1)
       IF (ALLOCATED (myMPSF(ITEL)%C2)) DEALLOCATE (myMPSF(ITEL)%C2)
    enddo
  end subroutine slib_dealloc_model_psf






 !**********************************************************************
  subroutine get_polar_nm(n, lookup_nr, lookup_nl, lookup_n, lookup_m)

    integer, intent(in) ::  n

    integer, DIMENSION(n), intent(in) :: lookup_nr, lookup_nl

    integer, DIMENSION(n) :: lookup_n, lookup_m
    
    integer icount

    do icount = 1, n
       lookup_n(icount) = lookup_nr(icount) + lookup_nl(icount)
       lookup_m(icount) = lookup_nr(icount) - lookup_nl(icount)
    enddo
 
  end subroutine get_polar_nm


  subroutine invert_matrix(N_COEFF, MATRIX)

    implicit none

    integer, intent(in) :: N_COEFF
 
   integer LDA, LWORK

    !PARAMETER (LDA = N_COEFF, LWORK=64*N_COEFF)

    complex(kind=real64) :: MATRIX(:, :) !, MATRIX_TEMP(N_COEFF, N_COEFF)

    INTEGER IPIV(N_COEFF)
    
    complex(kind=real64) :: WORK(64*N_COEFF)

    INTEGER INFO

!   complex*16 versions (GL 12/03/2021):    
!    external CGETRF
!    external CGETRI
    external ZGETRF
    external ZGETRI

    LDA = N_COEFF
    LWORK=64*N_COEFF
    
    

    !LU decomposition with lapack
    !MATRIX_TEMP=MATRIX
    ! use the complex*16 version of the factorization (GL 12/03/2021)
!    CALL cgetrf (N_COEFF, N_COEFF , MATRIX, LDA, IPIV, INFO)
    CALL zgetrf (N_COEFF, N_COEFF , MATRIX, LDA, IPIV, INFO)
    
 
!    !print LU 
    !write(*,*) "HERE",N_COEFF, IPIV
    !do i=1,N_COEFF
    !   WRITE (*,*) (MATRIX(i,j),j=1,N_COEFF)        
    !enddo

    !do the actual inversion
    ! use the complex*16 version of the inversion (GL 12/03/2021)
    !CALL CGETRI(N_COEFF, MATRIX, LDA, IPIV, WORK, LWORK, INFO)!
    CALL ZGETRI(N_COEFF, MATRIX, LDA, IPIV, WORK, LWORK, INFO)!

!    !print inverted
    !write(*,*) IPIV
!    do i=1,N_COEFF
!       do j =1, N_COEFF
!          WRITE (*,*) i, j, MATRIX(i,j), N_COEFF
!       enddo
!    enddo

  end subroutine invert_matrix



 !**********************************************************************
  subroutine make_matrix(N_COEFF, LOOKUP1, LOOKUP2, MATRIX)
    
    !use some_functions 
    
    integer, intent(in) :: N_COEFF
    
    integer, DIMENSION(N_COEFF), intent(in) :: LOOKUP1, LOOKUP2
    
    complex(kind=real64) :: MATRIX(N_COEFF, N_COEFF)
    
    integer ipolar, icart
    
    complex(kind=real64) :: factor
    
    integer n_r, n_l, n_1, n_2, n_rd, n_ld
    
    integer good
    
    real(kind=real64) ::  intermediate
    
    ! these are function
    !integer factorial, choose
    
    
    ! this is setting matrix elements to zero
    MATRIX=0.
    
    do ipolar=1,N_COEFF
       
       n_r = LOOKUP1(ipolar)
       n_l = LOOKUP2(ipolar)
       
       
       do icart=1,N_COEFF
          
          
          n_1 = LOOKUP1(icart)
          n_2 = LOOKUP2(icart)
          
          ! the actual calcs are done here
          ! this is identifical to Massey's IDL code
          if( (n_1+n_2)==(n_r+n_l) )then
             
             !write(*,*) n_1, n_2
             
             !             factor=complex(0.,0.)
             factor=cmplx(0.D0,0.D0,kind=real64)
             good=0
             do n_rd=0,n_r
                do n_ld=0,n_l
                   if( (n_rd+n_ld) == n_1) then 
                      good=good+1
                      factor = factor                &
                           + cmplx(0.D0,1.D0,kind=real64)**(n_ld-n_rd) &
                           * choose(n_r,n_rd)             &
                           * choose(n_l,n_ld)
                   endif
                enddo
             enddo
             
             !                    write(*,*) "FACTOR:", factor

             intermediate=real(factorial(n_1),kind=real64)    &
                  *  real(factorial(n_2),kind=real64)    &
                  /  real(factorial(n_r),kind=real64)    &
                  /  real(factorial(n_l),kind=real64)
             !
             matrix(ipolar, icart) = 2.D0**(-0.5D0*(n_r+n_l)) &
                  * cmplx(0.D0,1.D0,kind=real64)**(n_r-n_l) &
                  * sqrt( intermediate ) &
                  * factor
             
             
          endif !if(n_r + n_l == n_1 + n_2)
          
          
       enddo ! cart
    enddo ! polar    
  end subroutine make_matrix
  

 !**********************************************************************
  function factorial (n) 
    
    implicit none
    
    integer factorial
    integer, intent (in) :: n
    integer :: res
    integer :: i
    
    res = product ((/(i, i = 1, n)/))
    
    factorial = res
    
  end function factorial
  
  
  function choose (n, k) 
    
    implicit none
    integer choose
    integer, intent (in) :: n
    integer, intent (in) :: k
    integer :: res
    
    res = factorial (n) / (factorial (k) * factorial (n - k))
    
    choose=res
  
  end function choose





! this is apparently not used - comment out to suppress warnings
! !**********************************************************************
!  subroutine slib_fill_IMG_arrays(myPSF)
!
!    use slib_basic_math
!
!    implicit none
!
!
!    !TYPE(STACKPSF) :: myPSF
!    !TYPE(MODELPSF) :: myPSF
!    TYPE(CPSF) :: myPSF
!
!
!  
!    DOUBLE PRECISION, DIMENSION(0:SIZE(myPSF%NX0)) :: cx0,cy0  
!
!    DOUBLE PRECISION, DIMENSION(0:SIZE(myPSF%NX1)) :: cx1,cy1  
!    
!    DOUBLE PRECISION, DIMENSION(0:SIZE(myPSF%NX2)) :: cx2,cy2  
!    
!    INTEGER  N_COEFF0, N_COEFF1, N_COEFF2
!    
!    DOUBLE PRECISION BETA0, BETA1, BETA2
!
!    
!    DOUBLE PRECISION x, y, xo, yo, xprime, yprime, pixsize, dpix
!    
!    INTEGER i, ix, iy, ix_hermite, iy_hermite
!    
!    DOUBLE PRECISION normx, normy, SUBSAMPLE
!
!    !REAL, DIMENSION(myPSF%MAX_PSF_IMG_SIZE, myPSF%MAX_PSF_IMG_SIZE) :: array
!    
!
!    SUBSAMPLE = RAW_PIX_MICRON / myPSF%IMG_PIXEL_MM
!    pixsize= 1.0D0 / SUBSAMPLE
!    dpix=pixsize/2.0D0
!    i = int(SUBSAMPLE * 100)
!
!    if( myPSF%MAX_PSF_IMG_SIZE < i  )then
!       write(*,*) "WARNING: PSF RESOLUTION SUBSAMPLED BY FACTOR", SUBSAMPLE 
!       write(*,*) "WARNING: OUTPUT ARRAY SIZE:", myPSF%MAX_PSF_IMG_SIZE," MAY BE TOO SMALL"
!       write(*,*) "YOU MAY WANT TO CHANGE PARAMETER MAX_PSF_IMG_SIZE OR IMG_PIXEL_MM"
!    endif
!
!
!
!    ! center of PSF 
!    ! *** GL 20/10/1016: shift PSF centre:
!!    xo=myPSF%MAX_PSF_IMG_SIZE/2.0+1.0
!!    yo=myPSF%MAX_PSF_IMG_SIZE/2.0+1.0
!    xo=myPSF%MAX_PSF_IMG_SIZE/2.0D0/SUBSAMPLE+0.5D0
!    yo=myPSF%MAX_PSF_IMG_SIZE/2.0D0/SUBSAMPLE+0.5D0
!
!    !myPSF%IMG=0.0
!    
!    
!    N_COEFF0 = SIZE(myPSF%NX0)
!    N_COEFF1 = SIZE(myPSF%NX1)
!    N_COEFF2 = SIZE(myPSF%NX2)
!    BETA0  = myPSF%BETA0
!    BETA1  = myPSF%BETA1 
!    BETA2  = myPSF%BETA2
!
!
!    !write(*,*) "FILL ARRAYS",
!    !write(*,*) "FILL_ARRAY:",LOOKUP_N2
!    !write(*,*) "FILL_ARRAYS:",C_COEFF
!
!    ! *** GL (27/10/2016)
!    !i=1
!    
!    ! center of pixel is integer
!    do ix=1,myPSF%MAX_PSF_IMG_SIZE
!       x=real(ix,kind=real64)*pixsize !-pixsize/2.
!          
!       do iy=1,myPSF%MAX_PSF_IMG_SIZE
!          y=real(iy,kind=real64)*pixsize !-pixsize/2.
!
!
!             
!          ! this is because NORTH is always aligned with
!          ! ERO-X and EAST with ERO-Y. I need to create images
!          ! with NORTH==+X, EAST=-Y
!          xprime=(y-yo)
!          yprime=(x-xo)
!          xprime=xprime/BETA0
!          yprime=yprime/BETA0
!             
!
!          call hermite_poly ( N_COEFF0, dble(xprime), cx0 )
!          call hermite_poly ( N_COEFF0, dble(yprime), cy0 )
!             
!          do i = 1, N_COEFF0
!             ix_hermite=myPSF%NX0(I)
!             iy_hermite=myPSF%NY0(I)
!                
!             !write(*,*) ix_hermite, iy_hermite, C_COEFF(i)
!             ! using GAMMA(ix_hermite+1) instead of factrl(ix_hermite)
!             normx=1.D0 / sqrt( 2.D0**ix_hermite * sqrt(myPI) * GAMMA(real(ix_hermite+1,kind=real64)) ) 
!             normx=normx*1./sqrt(BETA0)*exp(-xprime**2/2.D0)  
!             normy=1.D0 / sqrt( 2.D0**iy_hermite * sqrt(myPI) * GAMMA(real(iy_hermite+1,kind=real64)) ) 
!             normy=normy*1./sqrt(BETA0)*exp(-yprime**2/2.D0)  
!
!             myPSF%IMG_SHAPE0(ix,iy, i) =  normx *normy * cx1(ix_hermite) * cy1(iy_hermite) * pixsize * pixsize
!
!
!          enddo
!
!          
!          
!          ! this is because NORTH is always aligned with
!          ! ERO-X and EAST with ERO-Y. I need to create images
!          ! with NORTH==+X, EAST=-Y
!          xprime=(y-yo)
!          yprime=(x-xo)
!          xprime=xprime/BETA1
!          yprime=yprime/BETA1
!             
!
!          call hermite_poly ( N_COEFF1, dble(xprime), cx1 )
!          call hermite_poly ( N_COEFF1, dble(yprime), cy1 )
!             
!          do i = 1, N_COEFF1
!             ix_hermite=myPSF%NX1(I)
!             iy_hermite=myPSF%NY1(I)
!                
!             !write(*,*) ix_hermite, iy_hermite, C_COEFF(i)
!             ! using GAMMA(ix_hermite+1) instead of factrl(ix_hermite)
!             normx=1.D0 / sqrt( 2.D0**ix_hermite * sqrt(myPI) * GAMMA(real(ix_hermite+1,kind=real64)) ) 
!             normx=normx*1./sqrt(BETA1)*exp(-xprime**2/2.D0)  
!             normy=1.D0 / sqrt( 2.D0**iy_hermite * sqrt(myPI) * GAMMA(real(iy_hermite+1,kind=real64)) ) 
!             normy=normy*1./sqrt(BETA1)*exp(-yprime**2/2.D0)  
!
!             myPSF%IMG_SHAPE1(ix,iy, i) =  normx *normy * cx1(ix_hermite) * cy1(iy_hermite) * pixsize * pixsize
!
!
!          enddo
!
!
!
!          ! this is because NORTH is always aligned with
!          ! ERO-X and EAST with ERO-Y. I need to create images
!          ! with NORTH==+X, EAST=-Y
!          xprime=(y-yo)
!          yprime=(x-xo)
!          xprime=xprime/BETA2
!          yprime=yprime/BETA2
!          
!          !write(*,*) x,y,ix,iy,xprime,yprime,beta, NIMG
!          !call  hermite_integrate2D(N_COEFF2, BETA2, x, x+1.0, y, y+1.0, xo, yo, cx2, cy2)
!          
!          call hermite_poly ( N_COEFF2, dble(xprime), cx2 )
!          call hermite_poly ( N_COEFF2, dble(yprime), cy2 )
!          
!          do i = 1, N_COEFF2
!             ix_hermite=myPSF%NX2(I)
!             iy_hermite=myPSF%NY2(I)
!                
!             !write(*,*) ix_hermite, iy_hermite, C_COEFF(i)
!             ! using GAMMA(ix_hermite+1) instead of factrl(ix_hermite)
!             normx=1.D0 / sqrt( 2.D0**ix_hermite * sqrt(myPI) * GAMMA(real(ix_hermite+1,kind=real64)) ) 
!             normx=normx*1./sqrt(BETA2)*exp(-xprime**2/2.D0)  
!             normy=1.D0 / sqrt( 2.D0**iy_hermite * sqrt(myPI) * GAMMA(real(iy_hermite+1,kind=real64)) ) 
!             normy=normy*1./sqrt(BETA2)*exp(-yprime**2/2.D0)  
!             
!             myPSF%IMG_SHAPE2(ix,iy, i) = normx * normy * cx2(ix_hermite) * cy2(iy_hermite) * pixsize * pixsize
!
!
!          enddo
!             
!             
!       enddo
!
!    enddo
!
!    !call slib_test_write2image(myPSF%MAX_PSF_IMG_SIZE, myPSF%IMG_SHAPE2(:,:, 9), "!junking.img")
!    !call slib_test_write2image(myPSF, myPSF%IMG_SHAPE1(:,:, 3), "test.fits")
!    !write(*,*) "I am out"
!    !stop
!
!  end subroutine slib_fill_IMG_arrays





 !**********************************************************************
  subroutine slib_integr_IMG_arrays(myPSF)

    use slib_basic_math

    implicit none
    TYPE(CPSF) :: myPSF
  
    DOUBLE PRECISION, DIMENSION(0:SIZE(myPSF%NX0)) :: cx0,cy0  

    DOUBLE PRECISION, DIMENSION(0:SIZE(myPSF%NX1)) :: cx1,cy1  
    
    DOUBLE PRECISION, DIMENSION(0:SIZE(myPSF%NX2)) :: cx2,cy2  
    
    INTEGER N_COEFF0, N_COEFF1, N_COEFF2
    
    DOUBLE PRECISION BETA0, BETA1, BETA2

    
    DOUBLE PRECISION x, y, xo, yo, pixsize, dpix
    
    INTEGER i, ix, iy, ix_hermite, iy_hermite
    
    DOUBLE PRECISION SUBSAMPLE


    SUBSAMPLE = RAW_PIX_MICRON / myPSF%IMG_PIXEL_MM
    pixsize=1.0D0 / SUBSAMPLE
    dpix =  pixsize/2.0D0

    !    i = int(SUBSAMPLE * 100)    
!    if( myPSF%MAX_PSF_IMG_SIZE < i  )then
!       write(*,*) "WARNING: PSF RESOLUTION SUBSAMPLED BY FACTOR", SUBSAMPLE 
!       write(*,*) "WARNING: OUTPUT ARRAY SIZE:", myPSF%MAX_PSF_IMG_SIZE," MAY BE TOO SMALL"
!       write(*,*) "YOU MAY WANT TO CHANGE PARAMETER MAX_PSF_IMG_SIZE OR IMG_PIXEL_MM"
!    endif



    ! center of PSF 
    ! *** GL 20/10/1016: shift PSF centre:
!    xo=myPSF%MAX_PSF_IMG_SIZE/2.0+1.0
!    yo=myPSF%MAX_PSF_IMG_SIZE/2.0+1.0
    xo=myPSF%MAX_PSF_IMG_SIZE/2.0D0/SUBSAMPLE+0.5D0
    yo=myPSF%MAX_PSF_IMG_SIZE/2.0D0/SUBSAMPLE+0.5D0
    
    N_COEFF0 = SIZE(myPSF%NX0)
    N_COEFF1 = SIZE(myPSF%NX1)
    N_COEFF2 = SIZE(myPSF%NX2)
    BETA0  = myPSF%BETA0
    BETA1  = myPSF%BETA1
    BETA2  = myPSF%BETA2

    
    ! center of pixels is integer
    do ix=1,myPSF%MAX_PSF_IMG_SIZE
       x=real(ix,kind=real64)*pixsize !-pixsize/2. 
          
       do iy=1,myPSF%MAX_PSF_IMG_SIZE
          y=real(iy,kind=real64)*pixsize !-pixsize/2.

          !write(*,*) x, y, xo, yo

          
          ! this is because NORTH is always aligned with
          ! ERO-X and EAST with ERO-Y. I need to create images
          ! with NORTH==+X, EAST=-Y
          call  hermite_integrate2D(N_COEFF0, BETA0, y-dpix, y+dpix, x-dpix, x+dpix, yo, xo, cx0, cy0)
          do i = 1, N_COEFF0
             ix_hermite=myPSF%NX0(I)
             iy_hermite=myPSF%NY0(I)
             !write(*,*) x, y, ix_hermite, iy_hermite, cx1(ix_hermite), cy1(iy_hermite), xo, yo
             myPSF%IMG_SHAPE0(ix,iy, i) =  cx0(ix_hermite) * cy0(iy_hermite)

          enddo

          

          
          ! this is because NORTH is always aligned with
          ! ERO-X and EAST with ERO-Y. I need to create images
          ! with NORTH==+X, EAST=-Y
          call  hermite_integrate2D(N_COEFF1, BETA1, y-dpix, y+dpix, x-dpix, x+dpix, yo, xo, cx1, cy1)
          do i = 1, N_COEFF1
             ix_hermite=myPSF%NX1(I)
             iy_hermite=myPSF%NY1(I)
             !write(*,*) x, y, ix_hermite, iy_hermite, cx1(ix_hermite), cy1(iy_hermite), xo, yo
             myPSF%IMG_SHAPE1(ix,iy, i) =  cx1(ix_hermite) * cy1(iy_hermite)

          enddo


          ! this is because NORTH is always aligned with
          ! ERO-X and EAST with ERO-Y. I need to create images
          ! with NORTH==+X, EAST=-Y
          call  hermite_integrate2D(N_COEFF2, BETA2, y-dpix, y+dpix, x-dpix, x+dpix, yo, xo, cx2, cy2)
          
          do i = 1, N_COEFF2
             ix_hermite=myPSF%NX2(I)
             iy_hermite=myPSF%NY2(I)
                
             myPSF%IMG_SHAPE2(ix,iy, i) = cx2(ix_hermite) * cy2(iy_hermite)


          enddo
             
             
       enddo

    enddo

    !call slib_test_write2image(myPSF%MAX_PSF_IMG_SIZE, myPSF%IMG_SHAPE2(:,:, 9), "!junking_int.img")

  end subroutine slib_integr_IMG_arrays





! this is only for testing - Comment out to suprress warnings
! !**********************************************************************
! subroutine slib_test_write2image(MAX_PSF_IMG_SIZE, ARRAY, filename)
!
!    !TYPE(STACKPSF) :: myPSF
!   ! TYPE(MODELPSF) :: myPSF
!   integer MAX_PSF_IMG_SIZE
!   character ( len = *) filename
!   DOUBLE PRECISION, DIMENSION(MAX_PSF_IMG_SIZE,MAX_PSF_IMG_SIZE) :: ARRAY
!
!    
!
!    !Fitsio needs these
!    integer :: status,unit,blocksize,bitpix,naxis
!    integer, DIMENSION(2) :: naxes(2)
!    integer ::  group,fpixel,nelements
!    logical simple,extend
!    !DOUBLE PRECISION, DIMENSION(myPSF%MAX_PSF_IMG_SIZE,myPSF%MAX_PSF_IMG_SIZE) :: ARRAY
!
!    
!    write(*,*) "LENGTH:", SHAPE(ARRAY)
!   
!    !ARRAY = IMG
!
!    ! WRITE_IMAGE
!       
!    status=0
!       
!    !  Name of the FITS file to be created:
!    !write (filename, "(I4.4, A2, I4.4)") &
!    !     ISRC, "_E", IE
!    !   filename=name(1:iname)//"_SRC"//trim(filename)//'.img'
!
!
!    !  Delete the file if it already exists, so we can then recreate it.
!    call delete_file ( filename, status )
!       
!    !  Get an unused Logical Unit Number to use to open the FITS file.
!    call ftgiou ( unit, status )
!    !
!    !  Create the new empty FITS file.
!    !
!    blocksize=1
!    call ftinit(unit,filename,blocksize,status)
!    !
!    !  Initialize parameters about the FITS image (300 x 200 16-bit integers).
!    !
!    simple=.true. 
!    bitpix=-32
!    naxis=2
!    naxes(1)=MAX_PSF_IMG_SIZE
!    naxes(2)=MAX_PSF_IMG_SIZE
!    extend=.true.
!       
!    !  Write the required header keywords.
!    call ftphpr(unit,simple,bitpix,naxis,naxes,0,1,extend,status)
!    
!    !  Write the array to the FITS file.
!       
!    group=1
!    fpixel=1
!    nelements=naxes(1)*naxes(2)
!    !call ftpprd(unit,group,fpixel,nelements,array_gauss,status)
!    call ftpprd(unit,group,fpixel,nelements,ARRAY,status)
!    !
!    !  Write another optional keyword to the header.
!    !
!    !  call ftpkyj(unit,'EXPOSURE',1500,'Total Exposure Time',status)
!    !
!    !  Close the file and free the unit number.
!    !
!    call ftclos(unit, status)
!    call ftfiou(unit, status)
!    !
!    !  Check for any error, and if so print out error messages
!    !
!    !if (status > 0) then
!    !   call print_error(status)
!    !end if
!
!    
!  end subroutine slib_test_write2image



!subroutine delete_file(filename,status)
!
!*******************************************************************************
!
!! DELETE_FILE deletes a FITS file.
!
!  integer status,unit,blocksize
!  character ( len = * ) filename
!!
!!  Simply return if status is greater than zero.
!!
!
!  !write(*,*) filename, status
!
!
!  if (status > 0) then
!    return
!  end if
!!
!!  Get an unused Logical Unit Number to use to open the FITS file
!!
!  call ftgiou ( unit, status )
!!
!!  Try to open the file, to see if it exists
!!
!  call ftopen ( unit, filename, 1, blocksize, status )
!
!  !write(*,*) unit, filename, blocksize, status
!
!  if ( status == 0 ) then
!!
!!  File was opened;  so now delete it 
!!
!    call ftdelt(unit,status)
!
!  else if (status == 103)then
!!
!!     File doesn't exist, so just reset status to zero and clear errors
!!
!    status=0
!    call ftcmsg
!
!  else
!!
!!  There was some other error opening the file; delete the file anyway
!!
!      status=0
!      call ftcmsg
!      call ftdelt(unit,status)
!  end if
!!
!!  Free the unit number for later reuse.
!!
!  call ftfiou(unit, status)
!
!  return
!end subroutine delete_file
!!****************************

end module slib_data_psf
