#!/bin/tcsh -f

#can do this all on 
#ds43

set OUT_BASE = "${HOME}/erosita/srctool_versions"
# set SRC_DIR   = "${HOME}/working/SASS_devel/task/srctool"
set SRC_DIR   = "${HOME}/gitwork/SASS_devel/task/srctool"

#set TAG = "1p11"
#set DIR = "v1.11"

if ( $#argv == 2 ) then 
  set TAG = "${argv[1]}"
  set DIR = "${argv[2]}"
else
  echo "Error: correct usage: ${0:t} TAG DIR"
  echo "e.g.: ${0:t} '1p11' 'v1.11'"
  exit 1
endif

set OUT_DIR = "${OUT_BASE}/$DIR"

echo "You requested the following:"
echo "TAG     = ${TAG}"
echo "OUT_DIR = ${OUT_DIR}"


if ( ! -d ${SRC_DIR} ) then
  echo "Failed to find source code directory: ${SRC_DIR}"
  exit 1
endif

# ################################################################################
# echo "Checking that CREATOR_STRING is correct in srctool_generic_mod.f90:"
# echo "---"
# #egrep -i -h CREATOR_STRING ${SRC_DIR}/srctool_datatypes_mod.f90
# egrep -i -h CREATOR_STRING ${SRC_DIR}/srctool_generic_mod.f90
# echo "---"
# ################################################################################
################################################################################

################################################################################
echo "Checking that Version number and date is correct in Changelog:"
echo "---"
egrep -i -h ^Version ${SRC_DIR}/ChangeLog | head -n 1
echo "---"
################################################################################


echo "##-- Press enter if happy with the above (or ^C if not)..."
set KEYSTROKE = $<                    

cd ${SRC_DIR}
echo "Here are the existing git tags:"
echo "# -----------------------------------------------"
git tag
echo "# -----------------------------------------------"
echo "Note, to delete an existing tag do something like this:"
echo "# -----------------------------------------------"
echo "#>  cd ${SRC_DIR}"
echo "#>  git tag -d ${TAG}"
echo "#>  git push origin :refs/tags/${TAG}"
echo "# -----------------------------------------------"
echo "##-- Press enter if you are still happy to add new tag=${TAG} (or ^C if not)..."
set KEYSTROKE = $<                    


# to tag the files do:
cd ${SRC_DIR}
#cvs -R tag srctool_version_${TAG}
git tag -a ${TAG} -m "srctool_version_${TAG}"
git push --tags

if ( $? > 0 ) then
  echo "Failed to add tag to git repo"
  echo "Note, to delete an existing tag do something like this:"
  echo "#>  cd ${SRC_DIR}"
  echo "#>  git tag -d ${TAG}"
  echo "#>  git push origin :refs/tags/${TAG}"
  exit 1
endif



#To export the files to a new directory do:
git archive -o ${OUT_BASE}/srctool_${DIR}.tar.gz --prefix=${DIR}/  $TAG 

# and make tar file read-only 
chmod a-w  ${OUT_BASE}/srctool_${DIR}.tar.gz

if ( $? > 0 ) then
  echo "Failed to extract archive from git"
  exit 1
endif

# now untar the files
cd ${OUT_BASE}
tar -xzvf srctool_${DIR}.tar.gz

if ( $? > 0 || ! -d ${OUT_DIR} ) then
  echo "Failed to untar repo"
  exit 1
endif

### remove build tests for now - will have to do these manually.
# #Then test the build with:
# 
# cd ${OUT_DIR}
# #source ${HOME}/erosita/scripts/sass-setup_local.csh eSASSdevel
# 
# source ${HOME}/gitwork/SASS_devel/sass-setup_local.csh eSASSdevel ${OUT_DIR}
# 
# make -f local_Makefile
# 
# if ( $? > 0 ) then
#   echo "Failed to build cleanly"
#   exit 1
# endif
# 
# 
# make -f local_Makefile test
# 
# if ( $? > 0 ) then
#   echo "Failed built-in tests"
#   exit 1
# endif
# 
# 
# # then clean and re-export the build
# cd ${OUT_BASE}
# rm -r ${DIR}
# cvs export -r srctool_version_${TAG} -d  ${OUT_DIR} SASS_devel/task/srctool
#
#if ( $? > 0 ) then
#  echo "Failed to export from CVS"
#  exit 1
#endif

## now make a tar-ball
#cd ${OUT_BASE}
#tar -cvzf  srctool_${DIR}.tar.gz ${DIR}

# and make archive directory read-only 
chmod -R a-w  ${OUT_DIR}


# now remind the user how to install into main erosita sw tree

echo "# ----------------------------------------------------------- #"
echo "# ----------------------------------------------------------- #"
echo "# Now you can install into the main eROSITA eSASSdevel tree"
echo "#> ds53"
echo "#ds53> cd /home/erosita/sw/eSASSdevel/erosita/task/"
echo "#ds53> mv srctool srctool-PREVIOUS_VERSION_NUMBER"
echo "#ds53> mkdir srctool/"
echo "#ds53> cd srctool/"
echo "#ds53> tar -xzvf /home/tdwelly/erosita/srctool_versions/srctool_${DIR}.tar.gz --strip 1"
echo "#ds53> source /home/erosita/sw/sass-setup.csh eSASSdevel"
echo "#ds53> make clean"
echo "#ds53> make"
echo "#ds53> test/srctool_test.csh"
echo "#ds53> cd ../"
echo "#ds53> chown -R ${USER}:erosita srctool/"
echo "#ds53> chmod -R g+wr srctool/"
echo "done"
echo "# ----------------------------------------------------------- #"
echo "# ----------------------------------------------------------- #"

exit
