  !---------------------------------------------------------------------------------
  ! Combined merge and insertion sort
  ! This is a very simplified form of the "timsort" algorithm
  ! This is a generic algorithm, densigned to be included in a typed-wrapper

    DATATYPE, dimension(:), intent(inout) :: data

    integer(kind=int32) :: num, chunksize, startp, endp, ssize, leftp, midp, rightp
    DATATYPE, dimension(:), allocatable :: tmp_left, tmp_right

    ! For testing
    ! DATATYPE :: temp
    ! integer(kind=int32) :: i

    num = size(data)

    if ( num < 2 ) then
       return
    end if

    allocate( tmp_left(num), tmp_right(num) )

    ! initial chunk size
    chunksize = find_chunksize(num)

    ! run insertion sort on smaller chunks
    do startp=1, num, chunksize
      endp = min(startp+chunksize, num)
      call do_insertion_sort(startp, endp)
    end do

    ! repeatedly run merge sort on the chunks, doubling the size each time
    ssize = chunksize
    do while( ssize < num )
      do leftp=1, num, 2*ssize
        midp = min(num, leftp+ssize-1)
        rightp = min(leftp+2*ssize-1, num)

        call do_merge(leftp, midp, rightp)
      end do

      ssize = 2 * ssize
    end do

    ! For testing
    ! temp = data(1)
    ! do i=2, num
    !   if(data(i) < temp ) stop
    !   temp = data(i)
    ! end do

    deallocate( tmp_left, tmp_right )

  contains

    !----------------------------------------------------------------------
    ! do an insertion sort on the array between leftp and rightp
    recursive subroutine do_insertion_sort(leftp, rightp)
      integer(kind=int32), intent(in) :: leftp, rightp

      integer(kind=int32) :: i, j
      DATATYPE :: tmp

      do i=leftp+1, rightp
        tmp = data(i)
        j = i-1
        do
          if( j<leftp ) exit
          if( tmp>=data(j) ) exit
          data(j+1) = data(j)
          j = j-1
        end do
        data(j+1) = tmp
      end do

    end subroutine do_insertion_sort

    !----------------------------------------------------------------------
    ! do a merge sort of leftp->midp and midp->rightp
    recursive subroutine do_merge(leftp, midp, rightp)

      integer(kind=int32), intent(in) :: leftp, midp, rightp
      integer(kind=int32) :: len1, len2, i, j, k

      len1 = midp - leftp + 1
      len2 = rightp - midp

      tmp_left(:len1) = data(leftp:leftp+len1-1)
      tmp_right(:len2) = data(midp+1:midp+1+len2-1)

      i = 1
      j = 1
      k = leftp

      do while ( j <= len2 .and. i <= len1 )
        if( tmp_left(i) <= tmp_right(j) ) then
          data(k) = tmp_left(i)
          i = i + 1
        else
          data(k) = tmp_right(j)
          j = j + 1
        end if
        k = k + 1
      end do

      do while ( i <= len1 )
        data(k) = tmp_left(i)
        k = k + 1
        i = i + 1
      end do

      do while ( j <= len2 )
        data(k) = tmp_right(j)
        k = k + 1
        j = j + 1
      end do
      
    end subroutine do_merge

    !----------------------------------------------------------------------
    ! find optimal chunck size so number of chunks is close to power of two
    recursive function find_chunksize(n_in) result(x)
      integer(kind=int32), intent(in) :: n_in
      integer(kind=int32) :: n, x, q
      integer(kind=int32), parameter :: MINIMUM_RUN=64

      n = n_in
      q = 0
      do while ( n >= MINIMUM_RUN )
        q = ior(q, iand(n, 1))
        n = n / 2
      end do

      x = n + q
    end function find_chunksize
