# global make include
# HBrunner 2006-06-21 (last update 2014-08-15)

include $(SASS_SETUP)/default.libs

tasklib	:= ../../lib/$(task).a

# make f90 object names (don't use .f90 and .for modules in same task)
obj_f90	:= $(patsubst %_mod.f90,%_mod.o,$(patsubst %_mod.for,,$(modules))) $(patsubst %.f90,%.o,$(patsubst %_mod.f90,,$(wildcard *.f90)))

# also consider .for, .f,  and .c files
obj_for	:= $(patsubst %_mod.for,%_mod.o,$(patsubst %_mod.f90,,$(modules))) $(patsubst %.for,%.o,$(patsubst %_mod.for,,$(patsubst %_init.for,,$(patsubst version.for,,$(wildcard *.for)))))
obj_f	:= $(patsubst %.f,%.o,$(wildcard *.f))
obj_c	:= $(patsubst %.c,%.o,$(wildcard *.c))

# look for init file
obj_ini	:= $(patsubst %.for,%.o,$(wildcard *_init.for))

# objects for compiling
obj	:= $(obj_f90) $(obj_ini) $(obj_for) $(obj_f) $(obj_c)    

# objects for library
obj_lib := $(obj_f90) $(obj_for) $(obj_f) $(obj_c)
obj_lib  := $(patsubst %_main.o,,$(obj)) 

# executables
exe	:= $(patsubst %_main.f90,$$SASS_BIN_ROOT/%,$(wildcard *_main.f90)) $(patsubst %_main.for,$$SASS_BIN_ROOT/%,$(wildcard *_main.for)) $(patsubst %_main.c,$$SASS_BIN_ROOT/%,$(wildcard *_main.c))

CFLAGS	:= -I$$E_MOD -I$(PIL) \
	-static -static-libgfortran -fdollar-ok -fall-intrinsics -std=gnu \
	-fno-automatic -fbounds-check -frange-check \
	-ffixed-line-length-none -ffree-line-length-none
#       -fbacktrace \

all: $(obj) $(tasklib) $(exe)

compile: $(obj)

lib: $(tasklib)

link: $(exe)

$(obj_f90):%.o:%.f90
	${FC} $(CFLAGS) $(TASK_CFLAGS) -c $< -o $@ || grep -q -s compile status.log || echo 'compile:err' >> status.log

$(obj_for):%.o:%.for
	${FC} $(CFLAGS) $(TASK_CFLAGS) -c $< -o $@ || grep -q -s compile status.log || echo 'compile:err' >> status.log

$(obj_f):%.o:%.f
	${FC} $(CFLAGS) $(TASK_CFLAGS) -c $< -o $@ || grep -q -s compile status.log || echo 'compile:err' >> status.log 

$(obj_c):%.o:%.c
	${CC} $(CFLAGS) $(TASK_CFLAGS) -c $< -o $@ || grep -q -s compile status.log || echo 'compile:err' >> status.log

$(tasklib):$(obj_lib)
	$$SASS_SETUP/lib.csh ../../lib/$(task).a "$?"
	$$SASS_SETUP/touch_lib.csh $(task)

$(exe):$$SASS_BIN_ROOT/%:%_main.o $(tasklib) %.libs
	$$SASS_SETUP/version.csh
	$$SASS_SETUP/depend.csh || grep -q -s depend status.log || echo 'depend:err' >> status.log
	if test -e $*.par ; then cp $*.par $$SASS_PAR ; fi
	${FC} -c version.for -o version.o
	echo '${FC} $(TASK_LFLAGS) -std=gnu -o $@ $< \' > $*.lnk
	if test -e $*_init.o ; then echo '$*_init.o \' >> $*.lnk ; fi
	echo 'version.o \' >> $*.lnk
	if test -e ../../lib/$(task).a ; then grep -q -s "$(task)\.a" $*.libs || echo '../../lib/$(task).a \' >> $*.lnk ; fi
	if test -e $*.libs ; then grep "\.a" $*.libs >> $*.lnk ; fi; test 0
	echo '$(LIB_SLATEC) $(LIB_SPLINE) $(LIB_FFTW) $(LIB_FTOOLS) $(PGP_LIB) $(PIL_LIB) $(TASK_LIBS) $(READLINE_LIB)' >> $*.lnk	
	. ./$*.lnk || grep -q -s link status.log || echo 'link:err' >> status.log       
	chmod ugo+x $$SASS_BIN_ROOT/$*
	sync

.PHONY  : clean dist
clean   :
	rm -f *.o $(tasklib) ../../mod/$(task)*.mod $(exe) *.lnk *.log version.for *~ \#*\#
	$$SASS_SETUP/rmod.csh

dist	: clean
	$$SASS_SETUP/make_dist.csh $(task)














