#
# Script to compile fftw 3.3.8 or higher for the 3 precisions, 2 of them used by eSASS tasks on October 2021.
# to be used when the fftw libraries are not available at the system level
##
# this script must be launched from the autoconf folder of the esass distribution after controling that
# a fftw-3.3.8 folder exist in  external.
# it can be customized if the fftw is stored in an other path or for other versions: the line cd external/fft-3.3.8 must then we adapted
# the fftw tarball can be downloaded from the official fftw.org web site :
# http://fftw.org/
# and should be unpacked in the external subfolder
# when this script is used, the option
# --with-fftw=/fftw/full/installation/path/
# should be added to the main configure option of the eSASS installation

# Author : Alain Gueguen agueguen@mpe.mpg.de
# last modification 23 Feb 2022

FOLDERSTART=$PWD
# to be customized
FFTINSTALLLIB=external/fftw-3.3.8
if [ -d $FFTINSTALLLIB  ] ;then
   cd $FFTINSTALLLIB
else
   echo ""
   echo "       ATTENTION: "
   echo " The Folder ${FFTINSTALLLIB} does not exist."
   echo " Please check the path of your fftw library and fix the value of the FFTINSTALLLIB variable in this script, (line19)."
   echo ""
   exit 0
fi
LIBFOLD=libprecision
mkdir $LIBFOLD
FULLpathLIB=${FFTINSTALLLIB}/${LIBFOLD}
echo "### Default precision ###"
make clean
autoreconf -fi
./configure
make
cp -r .libs $LIBFOLD/
cp libfft* $LIBFOLD/
make clean
echo "### Enable Float ###"
autoreconf -fi
./configure --enable-float
make
cp -r .libs $LIBFOLD/
cp libfft* $LIBFOLD/
make clean
echo "### enable long double ###"
autoreconf -fi
./configure --enable-long-double
make
cp -r .libs $LIBFOLD/
cp libfft* $LIBFOLD/
PATHFOROPTION=$PWD
ln -s .libs lib
cd $FOLDERSTART
#echo "library files saved in ${LIBFOLD}"
echo "library files saved in ${PATHFOROPTION}"
#echo realpath($LIBFOLD)

echo "**************************************************************************"
echo "** To use this  FFTW please use the option "
echo "** --with-fftw=$(realpath $FULLpathLIB)"
echo "** In the configure command"
echo "**************************************************************************"
