! Copyright (C) 2019-2023:
!   Max Planck Institut fuer extraterrestrische Physik (MPE)
!   Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eROSITA's Science Analysis Software System
! (eSASS) (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!-----------------------------------------------------------------------
!
module erolib_bintable_mod

  ! Non-methods in this module:
  !  subroutine merge_columns(colInA, colInB, colOut, status)!*********** obsolete?
  !  subroutine derive_diffs_column(inCol, outI2Col, status)
  !  subroutine dealloc_bintables(tables, status)
  !  subroutine alloc_bintables(tables, n, status)
  !  subroutine copy_bintable_scalars(tableIn, tableOut)!*********** obsolete?
  !  function checkMergable_bintables(tableInA, tableInB) result(theyAreMergable)!*********** obsolete?
  !  subroutine merge_two_bintables(tableInA, tableInB, tableOut, status)!*********** obsolete?
  !  subroutine merge_bintable_ptrs(listOfTablePtrs, tableOut, status)!*********** not working?

  use erolib_general_mod
  use erolib_ffits_mod ! ***only needed for open_FITS_file() and close_FITS_file().
  use erolib_str_mod
  use erolib_sort_mod
  use erolib_timsort
  use erolib_linklist_mod !*** change some time to erolib_linklist2_mod!! Also the other uses of this obsolete module.

  implicit none

!logical, private :: doTest=.false.

  private

  real(r4), parameter    :: r4defaultNull = -999.0
  real(r8), parameter    :: r8defaultNull = -999.0
  integer(i1), parameter :: i1defaultNull = -99
!*** use huge()???
  integer(i2), parameter :: i2defaultNull = -999
  integer(i4), parameter :: i4defaultNull = -999
  integer(i8), parameter :: i8defaultNull = -999

  ! Regarding tColumn%useageFlags:
  integer, parameter, public :: isOptionalBit    = 0 ! Try to read it, no error on failure.
  integer, parameter, public :: readBit          = 1
  integer, parameter, public :: writeBit         = 2
  integer, parameter, public :: internalUseBit   = 3 ! We don't intent to read or write it, but we want to use it.
  integer, parameter, public :: TCOL_ISOPT_BIT   = 0 ! Try to read it, no error on failure.
  integer, parameter, public :: TCOL_READ_BIT    = 1
  integer, parameter, public :: TCOL_WRITE_BIT   = 2
  integer, parameter, public :: TCOL_INTRN_BIT   = 3 ! We don't intent to read or write it, but we want to use it.
  integer, parameter, public :: maxBitNumUsed = 3

  integer, parameter, public :: fillWithZeroes  = 1
  integer, parameter, public :: fillWithNulls   = 2
  integer, parameter, public :: TCOL_FILL_ZEROES  = 1
  integer, parameter, public :: TCOL_FILL_NULLS   = 2

  integer, parameter, public :: ignoreTnullKwds    = 0
  integer, parameter, public :: readTnullKwds      = 1
  integer, parameter, public :: overwriteTnullKwds = 2

  type :: tScale
    real(r8) :: tzero = 0.0
    real(r8) :: tscal = 0.0
    logical :: containsData = .false.
  end type tScale

  type :: tTabInfo
    type(tFitsInfo), pointer :: fileInfo=>null()
    ! tBinTable-inheriting init routines should point this at a tFitsInfo-type object.
    character(buffsize) :: tableName=""
    logical :: haveAdvancedToHDU = .false.

    integer, private :: numRows = 0  ! the number of rows in the table.
    ! (This is typed as integer rather than i8 for example because its
    ! range is limited by the call ffits_get_num_rows() which ultimately
    ! is used to set it.)
  end type tTabInfo

  type, public :: tColumn
    type(tTabInfo), pointer :: tabInfo=>null()
    character(buffsize) :: nameStr = ""
    integer, private :: num = 0 ! the column number in a FITS file.
    character(1), private :: dataTypeFITS = "" ! from the set ['L','A','E','D','B','I','J','K'].
    character(20), private :: unitStr = ""
    logical :: writeNulls = .false.
    logical, private :: colNumFoundInFITS = .false.
    integer, private :: num_chars=1 ! for this%dataTypeFITS=='A'.
    integer, private :: num_vec_elements=0 ! for vector columns (not permitted for data type 'A').
    integer, private :: group_i=0

    ! Re tColumn%useageFlags, the intent with a read flag is that the
    ! read() method should look for that column in whatever FITS file
    ! you hand it. However, if it doesn't find it, it will only throw
    ! an error if the optional bit is NOT set.
    integer(i1), private :: useageFlags = 0 ! LS bit = optional, 2nd is
    ! R, 3rd bit = W, 4th bit indicates internal use.

    real(r4)    :: r4null = r4defaultNull
    real(r8)    :: r8null = r8defaultNull
    integer(i1) :: i1null = i1defaultNull
    integer(i2) :: i2null = i2defaultNull
    integer(i4) :: i4null = i4defaultNull
    integer(i8) :: i8null = i8defaultNull
    character(buffsize) :: strnull = "-999"

    type(tScale) :: scal
!    integer, dimension(:), allocatable, private :: felem

    character(buffsize), allocatable, dimension(:), private :: strdata ! no vector string data.
    real(r4),    allocatable, dimension(:), private :: r4data
    real(r8),    allocatable, dimension(:), private :: r8data
    integer(i1), allocatable, dimension(:), private :: i1data
    integer(i2), allocatable, dimension(:), private :: i2data
    integer(i4), allocatable, dimension(:), private :: i4data
    integer(i8), allocatable, dimension(:), private :: i8data
    logical,     allocatable, dimension(:), private :: b0data
    logical,     allocatable, dimension(:)          :: isNull
    real(r4),    allocatable, dimension(:,:), private :: r4vecdata
    real(r8),    allocatable, dimension(:,:), private :: r8vecdata
    integer(i1), allocatable, dimension(:,:), private :: i1vecdata
    integer(i2), allocatable, dimension(:,:), private :: i2vecdata
    integer(i4), allocatable, dimension(:,:), private :: i4vecdata
    integer(i8), allocatable, dimension(:,:), private :: i8vecdata
    logical,     allocatable, dimension(:,:), private :: b0vecdata
    logical,     allocatable, dimension(:,:)          :: isNullVec
  contains
    procedure          :: to_str                       => to_str_tColumn
    procedure, private :: blanks                       => blanks_tColumn
    procedure, private :: dealloc                     => dealloc_tColumn
    procedure, private :: destroy                     => destroy_tColumn
    procedure, private :: alloc                         => alloc_tColumn
    procedure          :: allocVec                   => allocVec_tColumn
    procedure, private :: realloc                     => realloc_tColumn
    procedure, private :: createFITSCol         => createFITSCol_tColumn
    procedure, private :: readRows                   => readRows_tColumn
    procedure, private :: setNullsFromMask   => setNullsFromMask_tColumn
    procedure, private :: writeRows                 => writeRows_tColumn
    procedure, private :: readTNULL                 => readTNULL_tColumn
    procedure, private :: setupFITSnulls       => setupFITSnulls_tColumn
    procedure, private :: readScaleKwds         => readScaleKwds_tColumn
    procedure, private :: writeScaleKwds       => writeScaleKwds_tColumn
    procedure, private :: updateTNULL             => updateTNULL_tColumn
    procedure, private :: pack                           => pack_tColumn
    procedure, private :: reorder                     => reorder_tColumn
    procedure, private :: unpack                       => unpack_tColumn
    procedure, private :: overwrite_section => overwrite_section_tColumn
    procedure, private :: setUseage                 => setUseage_tColumn
    procedure, private :: ibset_useage           => ibset_useage_tColumn
    procedure, private :: ibclr_useage           => ibclr_useage_tColumn
    procedure, private :: btest_useage           => btest_useage_tColumn
    procedure, private :: isAllocated             => isAllocated_tColumn
    procedure          :: isAllocatedVec       => isAllocatedVec_tColumn
    procedure, private :: isSame                       => isSame_tColumn
    procedure, private :: getNumRows               => getNumRows_tColumn
    procedure          :: getVectorSize         => getVectorSize_tColumn
    procedure          :: getDataType             => getDataType_tColumn
    procedure, private :: copy_data                 => copy_data_tColumn
    procedure, private :: copy                           => copy_tColumn
    procedure, private :: get_unit                   => get_unit_tColumn
    procedure, private :: set_unit                   => set_unit_tColumn
    procedure, private :: append                       => append_tColumn
    procedure, private :: append_zeroes         => append_zeroes_tColumn
    procedure, private :: str                         => get_str_tColumn
  end type tColumn

  type :: t_col_comparator
    character(1) :: dataTypeFITS = "" ! from the set ['L','A','E','D','B','I','J','K'].
    character(buffsize), pointer, dimension(:) :: strdata=>null()
    real(r4),    pointer, dimension(:) :: r4data=>null()
    real(r8),    pointer, dimension(:) :: r8data=>null()
    integer(i1), pointer, dimension(:) :: i1data=>null()
    integer(i2), pointer, dimension(:) :: i2data=>null()
    integer(i4), pointer, dimension(:) :: i4data=>null()
    integer(i8), pointer, dimension(:) :: i8data=>null()
    logical,     pointer, dimension(:) :: b0data=>null()
    logical,     pointer, dimension(:) :: isNull=>null()
    logical :: is_initialized=.false.
  contains
    procedure :: init => init_t_col_comparator
    procedure :: compare => compare_t_col_comparator
    procedure :: destroy => destroy_t_col_comparator
  end type t_col_comparator

  type :: tKeyword
    character(fkeylength) :: nameStr = ""
    integer(i1) :: useageFlags = 6 ! LS bit = optional, 2nd is R, 3rd bit = W. Note default is R/W, non-optional.
    character(1), private :: dataTypeFITS = '' ! Currently must be from the set ['E','D','J','A'].
    logical, private :: contains_data = .false.
    character(fcommentlength), private :: strValue=""
    integer(i4), private :: i4Value=0_i4
    real(r4), private :: r4Value=0.0_r4
    real(r8), private :: r8Value=0.0_r8
  contains
    procedure, private :: set                   => set_tKeyword
    procedure, private :: getValue         => getValue_tKeyword
    procedure, private :: setValue         => setValue_tKeyword
    procedure, private :: containsData => containsData_tKeyword
    procedure, private :: getDataType   => getDataType_tKeyword
  end type tKeyword

  type, public :: tHeader
    type(tTabInfo), pointer :: tabInfo=>null()!****** not yet doing anything with it. Note: after we do, should put machinery in place to nullify this on destuction of the binTable object.
    type(tKeyword), dimension(:), allocatable :: keywords
    class(list), private, pointer :: kwd_specs=>null()
  contains
    procedure, non_overridable :: initKeyword         =>     initKeyword_tHeader
    procedure, non_overridable :: deleteLinkedList    =>deleteLinkedList_tHeader
    procedure, non_overridable :: initKeywords        =>    initKeywords_tHeader
    procedure, private         :: dealloc_kwds        =>    dealloc_kwds_tHeader
    procedure, non_overridable :: alloc_kwds          =>      alloc_kwds_tHeader
    procedure, non_overridable :: dealloc             =>         dealloc_tHeader
    procedure, non_overridable :: alloc               =>           alloc_tHeader
    procedure, non_overridable :: check_keywords      =>  check_keywords_tHeader
    procedure, private         :: read                =>            read_tHeader
    procedure, private         :: write               =>           write_tHeader
    procedure, non_overridable :: append_kwd          =>      append_kwd_tHeader
    procedure, non_overridable :: containsData        =>    containsData_tHeader
    procedure, non_overridable :: getNumKeywords      =>  getNumKeywords_tHeader
    procedure, non_overridable :: getKeywordNum       =>   getKeywordNum_tHeader
    procedure, non_overridable :: setKeyword          =>      setKeyword_tHeader
    procedure, non_overridable :: getKeywordValue     => getKeywordValue_tHeader
    procedure, non_overridable :: setKeywordValue     => setKeywordValue_tHeader
    procedure, non_overridable :: keywordContainsData => kwdContainsData_tHeader
    procedure, non_overridable :: getKeywordDataType  =>  getKwdDataType_tHeader
    procedure, non_overridable :: set_kwd_useage      =>  set_kwd_useage_tHeader
  end type tHeader

  type, public :: tFitsInfo
    character(buffsize), private :: fileName=""
    integer :: unitNo = 0 ! fitsio unit to access the file
    logical :: fileOpenedForRead = .false.
    logical :: fileOpenedForWrite = .false.
    logical :: isRenamed=.false.
  contains
    procedure, non_overridable :: setFileName => setFileName_tFitsInfo
    procedure, non_overridable :: getFileName => getFileName_tFitsInfo
  end type tFitsInfo

  type, public :: tBinTable
    type(tTabInfo) :: info
    type(tHeader) :: header

    ! This is intended for use in method delete_duplicates(). It should be set up in this%init().
    integer, dimension(:), allocatable :: sort_sequence

    type(tColumn), dimension(:), allocatable :: cols
    logical :: isInitialized = .false.
    logical :: useageIsSet = .false.
    logical :: haveReadColNums = .false.
    logical :: foundInFile = .false.
    class(list), private, pointer :: col_specs=>null()

    real :: dummy=0.0 ! just to prevent warnings from -Wall! Better to make the class abstract.

  contains
!    procedure                  :: init => init_tBinTable ! Can't list it here because in some of the tBinTable subclasses we need to change the number and type of arguments. F90 doesn't seem to allow that sort of overloading.
    procedure                  :: initHeader             => initHeader_tBinTable
    procedure, non_overridable :: initColumn             => initColumn_tBinTable
    procedure, non_overridable :: deleteLinkedList => deleteLinkedList_tBinTable
    procedure, non_overridable :: initColumns           => initColumns_tBinTable
    procedure, non_overridable :: check                       => check_tBinTable
    procedure, non_overridable :: existsInFile         => existsInFile_tBinTable
    procedure, non_overridable :: read                         => read_tBinTable
    procedure, non_overridable :: write                       => write_tBinTable
    procedure, private         :: dealloc_cols           => dealloc_cols_tBinTable
    procedure, non_overridable :: alloc_cols               => alloc_cols_tBinTable
    procedure, non_overridable :: alloc                         => alloc_tBinTable
    procedure, non_overridable :: dealloc                     => dealloc_tBinTable
    procedure, non_overridable :: realloc                     => realloc_tBinTable
    procedure, non_overridable :: allocVec                   => allocVec_tBinTable
    procedure, non_overridable :: destroy                     => destroy_tBinTable
    procedure, non_overridable :: pack                           => pack_tBinTable
    procedure, non_overridable :: reorder                     => reorder_tBinTable
    procedure, non_overridable :: unpack                       => unpack_tBinTable
    procedure, non_overridable :: overwrite_section => overwrite_section_tBinTable
    procedure, non_overridable :: sort => sort_tBinTable
    procedure, non_overridable :: getColNum => getColNum_tBinTable
    procedure, non_overridable :: read_succeeded => read_succeeded_tBinTable
    procedure, private         :: detect_same_rows => detect_same_rows_tBinTable
    procedure, non_overridable :: flag_duplicates => flag_duplicates_tBinTable
    generic :: delete_duplicates => del_dup_seq,del_dup_char
    procedure, private         :: del_dup_seq  => del_dup_seq_tBinTable
    procedure, private         :: del_dup_char => del_dup_char_tBinTable
    procedure, private         :: getColNumsForReading => getColNumsForReading_tBinTable
    procedure, private         :: createTable => createTable_tBinTable
    procedure, private         :: getColNumsForWriting => getColNumsForWriting_tBinTable
    procedure, non_overridable :: sort_in_sequence => sort_in_sequence_tBinTable
    procedure, non_overridable :: getNumCols => getNumCols_tBinTable
    procedure, non_overridable :: containsData => containsData_tBinTable
    procedure, non_overridable :: getNumRows => getNumRows_tBinTable
    procedure, non_overridable :: set_all_useage => set_all_useage_tBinTable
    procedure, non_overridable :: ibset_all_useage => ibset_all_useage_tBinTable
    procedure, non_overridable :: set_col_useage => set_col_useage_tBinTable
    procedure, non_overridable :: ibset_col_useage => ibset_col_useage_tBinTable
    procedure, non_overridable :: ibclr_col_useage => ibclr_col_useage_tBinTable
    procedure, non_overridable :: btest_col_useage => btest_col_useage_tBinTable
    procedure, non_overridable :: isAllocated => isAllocated_tBinTable
    procedure, non_overridable :: col_is_allocated => col_is_allocated_tBinTable
    procedure, non_overridable :: copy_data_from_col => copy_col_tBinTable
    procedure, non_overridable :: get_col_unit => get_col_unit_tBinTable
    procedure, non_overridable :: set_col_unit => set_col_unit_tBinTable
    procedure, non_overridable :: can_append => can_append_tBinTable
    procedure, non_overridable :: append => append_tBinTable
    procedure, non_overridable :: set_col_blank => set_col_blank_tBinTable
    procedure, private         :: check_col_ptr => check_col_ptr_tBinTable
    procedure, private         :: get_null_col_ptr => get_null_col_ptr_tBinTable
    procedure, private         :: get_vec_null_col_ptr => get_vec_null_col_ptr_tBinTable

    procedure, private         :: gcp_str => get_col_ptr_str_tBinTable
    procedure, private         :: gcp_i1 => get_col_ptr_i1_tBinTable
    procedure, private         :: gcp_i2 => get_col_ptr_i2_tBinTable
    procedure, private         :: gcp_i4 => get_col_ptr_i4_tBinTable
    procedure, private         :: gcp_i8 => get_col_ptr_i8_tBinTable
    procedure, private         :: gcp_r4 => get_col_ptr_r4_tBinTable
    procedure, private         :: gcp_r8 => get_col_ptr_r8_tBinTable
    procedure, private         :: gcp_b0 => get_col_ptr_bool_tBinTable
    procedure, private         :: gcpv_i1 => get_col_ptr_i1_vec_tBinTable
    procedure, private         :: gcpv_i2 => get_col_ptr_i2_vec_tBinTable
    procedure, private         :: gcpv_i4 => get_col_ptr_i4_vec_tBinTable
    procedure, private         :: gcpv_i8 => get_col_ptr_i8_vec_tBinTable
    procedure, private         :: gcpv_r4 => get_col_ptr_r4_vec_tBinTable
    procedure, private         :: gcpv_r8 => get_col_ptr_r8_vec_tBinTable
    procedure, private         :: gcpv_b0 => get_col_ptr_bool_vec_tBinTable
    generic :: get_col_ptr => gcp_str,gcp_i1,gcp_i2,gcp_i4,gcp_i8,gcp_r4,gcp_r8,gcp_b0
    generic :: get_col_ptr => gcpv_i1,gcpv_i2,gcpv_i4,gcpv_i8,gcpv_r4,gcpv_r8,gcpv_b0
  end type tBinTable

  type, extends(t_sortable), public :: t_sort_bintable
    type(tBinTable), pointer, private :: bintable_obj=>null()
    type(t_col_comparator), dimension(:), allocatable, private :: col_comparators
    integer, private :: n_cols=0
    logical, private :: is_initialized=.false.
  contains
    procedure :: init => init_t_sort_bintable
    procedure :: get_n_elements => get_n_elements_t_sort_bintable
    procedure :: compare => compare_t_sort_bintable
    procedure :: check => check_t_sort_bintable
    procedure :: destroy => destroy_t_sort_bintable
  end type t_sort_bintable

  ! This will be useful when we want to construct an array of pointers to binary tables. Note that
  !	type(tBinTable), pointer, dimension(:) :: dummyPtrs
  ! is a pointer to an array, not an array of pointers.
!*** didn't work too well. Try class(tBinTable) rather than type?
  type :: tBinTablePtr
    type(tBinTable), pointer :: ptr=>null()
  end type tBinTablePtr

  type, public :: tFitsFile ! You'll probably want to extend this as well since you need to specify what extensions it contains. When I get really clever I'll pre-set this superclass up with an allocatable array of generic tBinTable objects.
    type(tFitsInfo) :: info
!    real :: dummy=0.0 ! just to prevent warnings from -Wall! Better to make the class abstract.
  contains
    procedure, non_overridable :: open => open_tFitsFile
!!!    procedure, non_overridable :: read => read_tFitsFile
!!!    procedure                  :: read_extensions => read_ext_tFitsFile
    procedure, non_overridable :: close => close_tFitsFile
  end type tFitsFile

  public :: derive_diffs_column,checkMergable_bintables
  public :: merge_bintable_ptrs,merge_two_bintables,dealloc_bintables,alloc_bintables

contains

  !=====================================================================
  ! Methods of tColumn:

  !.....................................................................
  subroutine blanks_tColumn(this, fillType, status, tabInfo)
    implicit none

    class(tColumn) :: this
    integer, intent(in)    :: fillType
    integer, intent(inout) :: status
    type(tTabInfo), optional, target, intent(in) :: tabInfo

    character(buffsize) :: msg
    character(*), parameter :: functionName = "blanks_tColumn"

    if(.not.this%isAllocated()) then
      call wrerr(functionName, s_err, "Column "//trim(this%nameStr)//" is not allocated.")
      status = s_errRep
  return
    end if

    if(.not.(fillType==TCOL_FILL_ZEROES .or. fillType==TCOL_FILL_NULLS)) then
      write (msg,'(a,i1)') "Column "//trim(this%nameStr)//" unrecognized fillType", fillType
      call wrerr(functionName, s_err, msg)
      status = s_errRep
  return
    end if

    if(present(tabInfo)) this%tabInfo => tabInfo

    select case(this%dataTypeFITS)
    case("L")
      if(this%num_vec_elements>0) then
        this%b0vecdata = .false.
      else
        this%b0data = .false.
      end if

    case("A")
      if(this%num_vec_elements>0)&
call wrstop(functionName, "Vector string columns are not yet permitted.")

      if(fillType==TCOL_FILL_ZEROES) then
        this%strdata = ''
      else
        this%strdata = this%strnull
      end if

    case("E")
      if(this%num_vec_elements>0) then
        if(fillType==TCOL_FILL_ZEROES) then
          this%r4vecdata = 0.0
        else
         this%r4vecdata = this%r4null
        end if
      else
        if(fillType==TCOL_FILL_ZEROES) then
          this%r4data = 0.0
        else
         this%r4data = this%r4null
        end if
      end if

    case("D")
      if(this%num_vec_elements>0) then
        if(fillType==TCOL_FILL_ZEROES) then
          this%r8vecdata = 0.0
        else
         this%r8vecdata = this%r8null
        end if
      else
        if(fillType==TCOL_FILL_ZEROES) then
          this%r8data = 0.0
        else
         this%r8data = this%r8null
        end if
      end if

    case("B")
      if(this%num_vec_elements>0) then
        if(fillType==TCOL_FILL_ZEROES) then
          this%i1vecdata = 0_i1
        else
         this%i1vecdata = this%i1null
        end if
      else
        if(fillType==TCOL_FILL_ZEROES) then
          this%i1data = 0_i1
        else
         this%i1data = this%i1null
        end if
      end if

    case("I")
      if(this%num_vec_elements>0) then
        if(fillType==TCOL_FILL_ZEROES) then
          this%i2vecdata = 0_i2
        else
         this%i2vecdata = this%i2null
        end if
      else
        if(fillType==TCOL_FILL_ZEROES) then
          this%i2data = 0_i2
        else
         this%i2data = this%i2null
        end if
      end if

    case("J")
      if(this%num_vec_elements>0) then
        if(fillType==TCOL_FILL_ZEROES) then
          this%i4vecdata = 0_i4
        else
         this%i4vecdata = this%i4null
        end if
      else
        if(fillType==TCOL_FILL_ZEROES) then
          this%i4data = 0_i4
        else
         this%i4data = this%i4null
        end if
      end if

    case("K")
      if(this%num_vec_elements>0) then
        if(fillType==TCOL_FILL_ZEROES) then
          this%i8vecdata = 0_i8
        else
         this%i8vecdata = this%i8null
        end if
      else
        if(fillType==TCOL_FILL_ZEROES) then
          this%i8data = 0_i8
        else
         this%i8data = this%i8null
        end if
      end if

    case default
      call wrerr(functionName, s_err, "Unrecognized data type '"//trim(this%dataTypeFITS)//"'")
      status = s_errRep
  return
    end select

    if(this%num_vec_elements>0) then
      if(fillType==TCOL_FILL_ZEROES) then
        this%isNullVec = .false.
      else
        this%isNullVec = .true.
      end if
    else
      if(fillType==TCOL_FILL_ZEROES) then
        this%isNull = .false.
      else
        this%isNull = .true.
      end if
    end if

  end subroutine blanks_tColumn

  !.....................................................................
  subroutine dealloc_tColumn(this, status)
    implicit none

    ! This deallocates everything, rather than trust the data type is correct.

    class(tColumn) :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "dealloc_tColumn"
    integer :: stat
    character(buffsize) :: errmsg

    if(allocated(this%b0data)) then
      errmsg = ""
      deallocate(this%b0data, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

    if(allocated(this%b0vecdata)) then
      errmsg = ""
      deallocate(this%b0vecdata, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

    if(allocated(this%strdata)) then
      errmsg = ""
      deallocate(this%strdata, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

    if(allocated(this%r4data)) then
      errmsg = ""
      deallocate(this%r4data, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

    if(allocated(this%r4vecdata)) then
      errmsg = ""
      deallocate(this%r4vecdata, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

    if(allocated(this%r8data)) then
      errmsg = ""
      deallocate(this%r8data, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

    if(allocated(this%r8vecdata)) then
      errmsg = ""
      deallocate(this%r8vecdata, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

    if(allocated(this%i1data)) then
      errmsg = ""
      deallocate(this%i1data, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

    if(allocated(this%i1vecdata)) then
      errmsg = ""
      deallocate(this%i1vecdata, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

    if(allocated(this%i2data)) then
      errmsg = ""
      deallocate(this%i2data, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

    if(allocated(this%i2vecdata)) then
      errmsg = ""
      deallocate(this%i2vecdata, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

    if(allocated(this%i4data)) then
      errmsg = ""
      deallocate(this%i4data, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

    if(allocated(this%i4vecdata)) then
      errmsg = ""
      deallocate(this%i4vecdata, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

    if(allocated(this%i8data)) then
      errmsg = ""
      deallocate(this%i8data, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

    if(allocated(this%i8vecdata)) then
      errmsg = ""
      deallocate(this%i8vecdata, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

    if(allocated(this%isNull)) then
      errmsg = ""
      deallocate(this%isNull, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

    if(allocated(this%isNullVec)) then
      errmsg = ""
      deallocate(this%isNullVec, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      if(status==s_errRep)&
  return
    end if

!    if(allocated(this%felem)) then
!      errmsg = ""
!      deallocate(this%felem, stat=stat, errmsg=errmsg)
!      call FError(functionName, stat, errmsg, status)
!      if(status==s_errRep)&
!  return
!    end if

  end subroutine dealloc_tColumn

  !.....................................................................
  subroutine destroy_tColumn(this, status)
    implicit none

    class(tColumn) :: this
    integer, intent(inout) :: status

    call this%dealloc(status)
    nullify(this%tabInfo)

  end subroutine destroy_tColumn

  !.....................................................................
  subroutine alloc_tColumn(this, numRows, status, fillType)
    implicit none

    class(tColumn) :: this
    integer,           intent(in)    :: numRows
    integer,           intent(inout) :: status
    integer, optional, intent(in)    :: fillType

    character(*), parameter :: functionName = "alloc_tColumn"
    integer :: stat
    character(buffsize) :: errmsg

    if(numRows < 0) then
      call wrerr(functionName, warn, "You can't allocate a column to less than zero size.")
  return
    end if

    call this%dealloc(status)
    errmsg = ""

    select case(this%dataTypeFITS)
    case("L")
      allocate(this%b0data(numRows), stat=stat, errmsg=errmsg)

    case("A")
      allocate(this%strdata(numRows), stat=stat, errmsg=errmsg)

    case("E")
      allocate(this%r4data(numRows), stat=stat, errmsg=errmsg)

    case("D")
      allocate(this%r8data(numRows), stat=stat, errmsg=errmsg)

    case("B")
      allocate(this%i1data(numRows), stat=stat, errmsg=errmsg)

    case("I")
      allocate(this%i2data(numRows), stat=stat, errmsg=errmsg)

    case("J")
      allocate(this%i4data(numRows), stat=stat, errmsg=errmsg)

    case("K")
      allocate(this%i8data(numRows), stat=stat, errmsg=errmsg)

    case default
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
      status = s_errRep
  return
    end select

    call FError(functionName, stat, errmsg, status)
    if(status==s_errRep)&
  return

    errmsg = ""
    allocate(this%isNull(numRows), stat=stat, errmsg=errmsg)
    call FError(functionName, stat, errmsg, status)
    if(status==s_errRep)&
  return

    this%num_vec_elements = 0

    if(present(fillType)) then
      call this%blanks(fillType, status)
    else
      call this%blanks(TCOL_FILL_ZEROES, status)
    end if

  end subroutine alloc_tColumn

  !.....................................................................
  subroutine allocVec_tColumn(this, num_vec_elements, status, fillType)
    implicit none

    class(tColumn) :: this
    integer,           intent(in)    :: num_vec_elements
    integer,           intent(inout) :: status
    integer, optional, intent(in)    :: fillType

    character(*), parameter :: functionName = "allocVec_tColumn"
    integer :: stat
    character(buffsize) :: errmsg
    integer :: numRows

    numRows = this%getNumRows(status)

    if(num_vec_elements<1) then
      if(this%num_vec_elements > 0) then ! the column used to be vector; we're going to reallocate it as scalar.
        call this%alloc(numRows, status, fillType)
      else ! no action, although remark on it if chatty.
        call wrchatty(functionName, "Attempting to allocate column "&
          //trim(this%nameStr)//" as vector, but with zero width. Ignored.")
      end if
  return
    end if

    this%num_vec_elements = num_vec_elements

    call this%dealloc(status)
    errmsg = ""

    select case(this%dataTypeFITS)
    case("L")
      allocate(this%b0vecdata(this%num_vec_elements,numRows), stat=stat, errmsg=errmsg)

    case("A")
call wrstop(functionName, "Vector string columns are not yet permitted.")

    case("E")
      allocate(this%r4vecdata(this%num_vec_elements,numRows), stat=stat, errmsg=errmsg)

    case("D")
      allocate(this%r8vecdata(this%num_vec_elements,numRows), stat=stat, errmsg=errmsg)

    case("B")
      allocate(this%i1vecdata(this%num_vec_elements,numRows), stat=stat, errmsg=errmsg)

    case("I")
      allocate(this%i2vecdata(this%num_vec_elements,numRows), stat=stat, errmsg=errmsg)

    case("J")
      allocate(this%i4vecdata(this%num_vec_elements,numRows), stat=stat, errmsg=errmsg)

    case("K")
      allocate(this%i8vecdata(this%num_vec_elements,numRows), stat=stat, errmsg=errmsg)

    case default
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
      status = s_errRep
  return
    end select

    call FError(functionName, stat, errmsg, status)
    if(status==s_errRep)&
  return

    errmsg = ""
    allocate(this%isNullVec(this%num_vec_elements,numRows), stat=stat, errmsg=errmsg)
    call FError(functionName, stat, errmsg, status)
    if(status==s_errRep)&
  return

!    errmsg = ""
!    allocate(this%felem(this%num_vec_elements), stat=stat, errmsg=errmsg)
!    call FError(functionName, stat, errmsg, status)
!    if(status==s_errRep)&
!  return
!    this%felem(:) = 1

    if(present(fillType)) then
      call this%blanks(fillType, status)
    else
      call this%blanks(TCOL_FILL_ZEROES, status)
    end if

  end subroutine allocVec_tColumn

  !.....................................................................
  subroutine realloc_tColumn(this, new_n_rows, status)
    implicit none

    class(tColumn)         :: this
    integer, intent(in)    :: new_n_rows
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "realloc_tColumn"
    integer :: temp_n_vec_elements

    temp_n_vec_elements = this%num_vec_elements
    call this%alloc(new_n_rows, status)
    call this%allocVec(temp_n_vec_elements, status)

  end subroutine realloc_tColumn

  !.....................................................................
  function getNumRows_tColumn(this, status) result(numRows)
    implicit none

    class(tColumn) :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "getNumRows_tColumn"
    integer :: numRows

    numRows=0

    select case(this%dataTypeFITS)
    case("L")
      if(allocated(this%b0data)) then
        numRows = size(this%b0data)
      elseif(allocated(this%b0vecdata)) then
        numRows = size(this%b0vecdata, 2)
      end if

    case("A")
      if(allocated(this%strdata)) then
        numRows = size(this%strdata)
      end if

    case("E")
      if(allocated(this%r4data)) then
        numRows = size(this%r4data)
      elseif(allocated(this%r4vecdata)) then
        numRows = size(this%r4vecdata, 2)
      end if

    case("D")
      if(allocated(this%r8data)) then
        numRows = size(this%r8data)
      elseif(allocated(this%r8vecdata)) then
        numRows = size(this%r8vecdata, 2)
      end if

    case("B")
      if(allocated(this%i1data)) then
        numRows = size(this%i1data)
      elseif(allocated(this%i1vecdata)) then
        numRows = size(this%i1vecdata, 2)
      end if

    case("I")
      if(allocated(this%i2data)) then
        numRows = size(this%i2data)
      elseif(allocated(this%i2vecdata)) then
        numRows = size(this%i2vecdata, 2)
      end if

    case("J")
      if(allocated(this%i4data)) then
        numRows = size(this%i4data)
      elseif(allocated(this%i4vecdata)) then
        numRows = size(this%i4vecdata, 2)
      end if

    case("K")
      if(allocated(this%i8data)) then
        numRows = size(this%i8data)
      elseif(allocated(this%i8vecdata)) then
        numRows = size(this%i8vecdata, 2)
      end if

    case default
      call wrerr(functionName, s_err, "Unrecognized choice '"&
        //trim(this%dataTypeFITS)//"'")
      status = s_errRep
  return
    end select

  end function getNumRows_tColumn

  !.....................................................................
  function getVectorSize_tColumn(this) result(vecSize)
    ! We want to keep tColumn%num_vec_elements private to remove the temptation to set it directly.

    implicit none

    class(tColumn) :: this

    character(*), parameter :: functionName = "getVectorSize_tColumn"
    integer :: vecSize

    vecSize=this%num_vec_elements

  end function getVectorSize_tColumn

  !.....................................................................
  function getDataType_tColumn(this) result(dataType)
    ! We want to keep tColumn%dataTypeFITS private to remove the temptation to set it directly.

    implicit none

    class(tColumn) :: this

    character(*), parameter :: functionName = "getDataType_tColumn"
    character(1) :: dataType

    dataType=this%dataTypeFITS

  end function getDataType_tColumn

  !.....................................................................
  subroutine pack_tColumn(this, mask, status)
    implicit none

    class(tColumn) :: this
    logical, dimension(:), intent(in) :: mask
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "pack_tColumn"

    if(this%num_vec_elements>0)&
call wrstop(functionName, "Cannot pack vector columns.")

    select case(this%dataTypeFITS)
    case("L")
      this%b0data = pack(this%b0data, mask)
    case("A")
      this%strdata = pack(this%strdata, mask)
    case("E")
      this%r4data = pack(this%r4data, mask)
    case("D")
      this%r8data = pack(this%r8data, mask)
    case("B")
      this%i1data = pack(this%i1data, mask)
    case("I")
      this%i2data = pack(this%i2data, mask)
    case("J")
      this%i4data = pack(this%i4data, mask)
    case("K")
      this%i8data = pack(this%i8data, mask)
    case default
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
      status = s_errRep
  return
    end select

    this%isNull = pack(this%isNull, mask)

  end subroutine pack_tColumn

  !.....................................................................
  subroutine reorder_tColumn(this, order, status)
    ! Also packs if size(order)/=this%info%numRows.

    implicit none

    class(tColumn) :: this
    integer, dimension(:), intent(in)    :: order
    integer,               intent(inout) :: status

    character(*), parameter :: functionName = "reorder_tColumn"

    if(size(order) < 1) then
      call this%alloc(0, status)
  return
    end if

    if(this%num_vec_elements>0)&
call wrstop(functionName, "Cannot order vector columns.")

    select case(this%dataTypeFITS)
    case("L")
      this%b0data = this%b0data(order)
    case("A")
      this%strdata = this%strdata(order)
    case("E")
      this%r4data = this%r4data(order)
    case("D")
      this%r8data = this%r8data(order)
    case("B")
      this%i1data = this%i1data(order)
    case("I")
      this%i2data = this%i2data(order)
    case("J")
      this%i4data = this%i4data(order)
    case("K")
      this%i8data = this%i8data(order)
    case default
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
      status = s_errRep
  return
    end select

    this%isNull = this%isNull(order)

  end subroutine reorder_tColumn

  !.....................................................................
  subroutine unpack_tColumn(this, colB, order, status, fillType)
    implicit none

!**** better without explicit colB?

    class(tColumn) :: this
    type(tColumn),     intent(in)    :: colB
    integer, dimension(:), intent(in)    :: order
    integer,               intent(inout) :: status
    integer, optional,     intent(in)    :: fillType

    character(*), parameter :: functionName = "unpack_tColumn"

    if(size(order) <= 0)&
  return

    if(this%num_vec_elements>0)&
call wrstop(functionName, "Cannot unpack vector columns.")

    if(this%dataTypeFITS /= colB%dataTypeFITS) then
      call wrerr(functionName, s_err, "Columns "//trim(this%nameStr)//" and "&
        //trim(colB%nameStr)//" have different data types "&
        //trim(this%dataTypeFITS)//" and "//trim(colB%dataTypeFITS)//".")
      status = s_errRep
  return
    end if

    if(present(fillType))&
      call this%blanks(fillType, status)

    select case(this%dataTypeFITS)
    case("L")
      this%b0data(order) = colB%b0data
    case("A")
      this%strdata(order) = colB%strdata
    case("E")
      this%r4data(order) = colB%r4data
    case("D")
      this%r8data(order) = colB%r8data
    case("B")
      this%i1data(order) = colB%i1data
    case("I")
      this%i2data(order) = colB%i2data
    case("J")
      this%i4data(order) = colB%i4data
    case("K")
      this%i8data(order) = colB%i8data
    case default
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
      status = s_errRep
  return
    end select

    this%isNull(order) = colB%isNull

  end subroutine unpack_tColumn

  !.....................................................................
  subroutine overwrite_section_tColumn(this, colB, startI, finisI, status)
    implicit none

    class(tColumn) :: this
    type(tColumn),  intent(in)    :: colB
    integer,            intent(in)    :: startI,finisI
    integer,            intent(inout) :: status

    character(*), parameter :: functionName = "overwrite_section_tColumn"

    if(this%dataTypeFITS /= colB%dataTypeFITS) then
      call wrerr(functionName, s_err, "Columns "//trim(this%nameStr)//" and "&
        //trim(colB%nameStr)//" have different data types "&
        //trim(this%dataTypeFITS)//" and "//trim(colB%dataTypeFITS)//".")
      status = s_errRep
  return
    end if

    select case(this%dataTypeFITS)
    case("L")
      if(this%num_vec_elements>0) then
        this%b0vecdata(:,startI:finisI) = colB%b0vecdata
      else
        this%b0data(startI:finisI) = colB%b0data
      end if
    case("A")
      if(this%num_vec_elements>0)&
call wrstop(functionName, "Vector string columns are not yet permitted.")

      this%strdata(startI:finisI) = colB%strdata
    case("E")
      if(this%num_vec_elements>0) then
        this%r4vecdata(:,startI:finisI) = colB%r4vecdata
      else
        this%r4data(startI:finisI) = colB%r4data
      end if
    case("D")
      if(this%num_vec_elements>0) then
        this%r8vecdata(:,startI:finisI) = colB%r8vecdata
      else
        this%r8data(startI:finisI) = colB%r8data
      end if
    case("B")
      if(this%num_vec_elements>0) then
        this%i1vecdata(:,startI:finisI) = colB%i1vecdata
      else
        this%i1data(startI:finisI) = colB%i1data
      end if
    case("I")
      if(this%num_vec_elements>0) then
        this%i2vecdata(:,startI:finisI) = colB%i2vecdata
      else
        this%i2data(startI:finisI) = colB%i2data
      end if
    case("J")
      if(this%num_vec_elements>0) then
        this%i4vecdata(:,startI:finisI) = colB%i4vecdata
      else
        this%i4data(startI:finisI) = colB%i4data
      end if
    case("K")
      if(this%num_vec_elements>0) then
        this%i8vecdata(:,startI:finisI) = colB%i8vecdata
      else
        this%i8data(startI:finisI) = colB%i8data
      end if
    case default
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
      status = s_errRep
  return
    end select

    if(this%num_vec_elements>0) then
      this%isNullVec(:,startI:finisI) = colB%isNullVec
    else
      this%isNull(startI:finisI) = colB%isNull
    end if

  end subroutine overwrite_section_tColumn

  !.....................................................................
  subroutine readRows_tColumn(this, unitNo, frow, lrow, felem, nrows, status)
    implicit none

    class(tColumn) :: this
    integer, intent(in)    :: unitNo,frow,lrow,felem,nrows
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "readRows_tColumn"
    integer :: ierr
    logical :: dummySomeNullsFound
    integer, parameter :: maxdim=1
!    integer, dimension(maxdim) :: naxes

    ierr = 0

!**** must get this at some point.
!    call FTGTDM(unitNo, this%num, maxdim, > naxis,naxes,ierr)
!    call report_FITS_err(functionName, ierr, status, " - column "//trim(this%nameStr))

    select case(this%dataTypeFITS)
    case("L")
call wrstop(functionName, "Read of boolean columns is not permitted.")
    case("A")
      if(this%num_vec_elements>0)&
call wrstop(functionName, "Vector string columns are not yet permitted.")

      call FTGCFS(unitNo, this%num, frow, felem, nrows&
        , this%strdata(frow:lrow), this%isNull(frow:lrow), dummySomeNullsFound, ierr)
    case("E")
      if(this%num_vec_elements>0) then
        call FTGCFE(unitNo, this%num, frow, felem, nrows*this%num_vec_elements&
          , this%r4vecdata(:,frow:lrow), this%isNullVec(:,frow:lrow), dummySomeNullsFound, ierr)
      else
        call FTGCFE(unitNo, this%num, frow, felem, nrows&
          , this%r4data(frow:lrow), this%isNull(frow:lrow), dummySomeNullsFound, ierr)
      end if
    case("D")
      if(this%num_vec_elements>0) then
        call FTGCFD(unitNo, this%num, frow, felem, nrows*this%num_vec_elements&
          , this%r8vecdata(:,frow:lrow), this%isNullVec(:,frow:lrow), dummySomeNullsFound, ierr)
      else
        call FTGCFD(unitNo, this%num, frow, felem, nrows&
          , this%r8data(frow:lrow), this%isNull(frow:lrow), dummySomeNullsFound, ierr)
      end if
    case("B")
      if(this%num_vec_elements>0) then
        call FTGCFB(unitNo, this%num, frow, felem, nrows*this%num_vec_elements&
          , this%i1vecdata(:,frow:lrow), this%isNullVec(:,frow:lrow), dummySomeNullsFound, ierr)
      else
        call FTGCFB(unitNo, this%num, frow, felem, nrows&
          , this%i1data(frow:lrow), this%isNull(frow:lrow), dummySomeNullsFound, ierr)
      end if
    case("I")
      if(this%num_vec_elements>0) then
        call FTGCFI(unitNo, this%num, frow, felem, nrows*this%num_vec_elements&
          , this%i2vecdata(:,frow:lrow), this%isNullVec(:,frow:lrow), dummySomeNullsFound, ierr)
      else
        call FTGCFI(unitNo, this%num, frow, felem, nrows&
          , this%i2data(frow:lrow), this%isNull(frow:lrow), dummySomeNullsFound, ierr)
      end if
    case("J")
      if(this%num_vec_elements>0) then
        call FTGCFJ(unitNo, this%num, frow, felem, nrows*this%num_vec_elements&
          , this%i4vecdata(:,frow:lrow), this%isNullVec(:,frow:lrow), dummySomeNullsFound, ierr)
      else
        call FTGCFJ(unitNo, this%num, frow, felem, nrows&
          , this%i4data(frow:lrow), this%isNull(frow:lrow), dummySomeNullsFound, ierr)
      end if
    case("K")
      if(this%num_vec_elements>0) then
        call FTGCFK(unitNo, this%num, frow, felem, nrows*this%num_vec_elements&
          , this%i8vecdata(:,frow:lrow), this%isNullVec(:,frow:lrow), dummySomeNullsFound, ierr)
      else
        call FTGCFK(unitNo, this%num, frow, felem, nrows&
          , this%i8data(frow:lrow), this%isNull(frow:lrow), dummySomeNullsFound, ierr)
      end if

    case default
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
      status = s_errRep
  return
    end select

    call report_FITS_err(functionName, ierr, status, " - column "//trim(this%nameStr))

  end subroutine readRows_tColumn

  !.....................................................................
  subroutine setNullsFromMask_tColumn(this, status)
    implicit none

    class(tColumn) :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "setNullsFromMask_tColumn"

    if(.not.this%writeNulls)&
  return

    select case(this%dataTypeFITS)
    case("L")
      if(this%num_vec_elements>0) then
        where(this%isNullVec) this%b0vecdata = .false.
      else
        where(this%isNull) this%b0data = .false.
      end if
    case("A")
      if(this%num_vec_elements>0)&
call wrstop(functionName, "Vector string columns are not yet permitted.")

      where(this%isNull) this%strdata = this%strnull
    case("E")
      if(this%num_vec_elements>0) then
        where(this%isNullVec) this%r4vecdata = this%r4null
      else
        where(this%isNull) this%r4data = this%r4null
      end if
    case("D")
      if(this%num_vec_elements>0) then
        where(this%isNullVec) this%r8vecdata = this%r8null
      else
        where(this%isNull) this%r8data = this%r8null
      end if
    case("B")
      if(this%num_vec_elements>0) then
        where(this%isNullVec) this%i1vecdata = this%i1null
      else
        where(this%isNull) this%i1data = this%i1null
      end if
    case("I")
      if(this%num_vec_elements>0) then
        where(this%isNullVec) this%i2vecdata = this%i2null
      else
        where(this%isNull) this%i2data = this%i2null
      end if
    case("J")
      if(this%num_vec_elements>0) then
        where(this%isNullVec) this%i4vecdata = this%i4null
      else
        where(this%isNull) this%i4data = this%i4null
      end if
    case("K")
      if(this%num_vec_elements>0) then
        where(this%isNullVec) this%i8vecdata = this%i8null
      else
        where(this%isNull) this%i8data = this%i8null
      end if
    case default
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
      status = s_errRep
  return
    end select

  end subroutine setNullsFromMask_tColumn

  !.....................................................................
  subroutine writeRows_tColumn(this, unitNo, frow, lrow, nrows, status)
    ! FITS nulls: 
    !	- For floating-point (types E and D) values, the call FTPCN[E,D]
    !	  will cause FITS to flag any value as NULL which is equal to
    !	  the supplied this%r[4,8]null.
    !
    !	- For integer values we assume here that an appropriate TNULL
    !	  keyword exists in the file. If so, the call FTPCL[B,I,J,K]
    !	  will cause FITS to flag any value as NULL which is equal to
    !	  the supplied this%r?null.
    !
    !	- All such values will be displayed as NULL in fv etc and, upon
    !	  reading the file again, will set .true. values in the boolean
    !	  array returned by the call FTGCF?.
    !
    implicit none

    class(tColumn) :: this
    integer, intent(in)    :: unitNo,frow,lrow,nrows
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "writeRows_tColumn"
    integer :: ierr

    ierr = 0

    select case(this%dataTypeFITS)
    case("L")
call wrstop(functionName, "Write of boolean columns is not permitted.")
    case("A")
      if(this%num_vec_elements>0)&
call wrstop(functionName, "Vector string columns are not yet permitted.")

      if(this%writeNulls) then
        call FTPCNS(unitNo, this%num, frow, 1, nrows&
          , this%strdata(frow:lrow), this%strnull, ierr)
      else
        call FTPCLS(unitNo, this%num, frow, 1, nrows&
          , this%strdata(frow:lrow), ierr)
      end if
    case("E")
      if(this%num_vec_elements>0) then
        if(this%writeNulls) then
!          call FTPCNE(unitNo, this%num, frow, this%felem, nrows*this%num_vec_elements&
          call FTPCNE(unitNo, this%num, frow, 1, nrows*this%num_vec_elements&
            , this%r4vecdata(:,frow:lrow), this%r4null, ierr)
        else
!          call FTPCLE(unitNo, this%num, frow, this%felem, nrows*this%num_vec_elements&
          call FTPCLE(unitNo, this%num, frow, 1, nrows*this%num_vec_elements&
            , this%r4vecdata(:,frow:lrow), ierr)
        end if
      else
        if(this%writeNulls) then
          call FTPCNE(unitNo, this%num, frow, 1, nrows&
            , this%r4data(frow:lrow), this%r4null, ierr)
        else
          call FTPCLE(unitNo, this%num, frow, 1, nrows&
            , this%r4data(frow:lrow), ierr)
        end if
      end if
    case("D")
      if(this%num_vec_elements>0) then
        if(this%writeNulls) then
!          call FTPCND(unitNo, this%num, frow, this%felem, nrows*this%num_vec_elements&
          call FTPCND(unitNo, this%num, frow, 1, nrows*this%num_vec_elements&
            , this%r8vecdata(:,frow:lrow), this%r8null, ierr)
        else
!          call FTPCLD(unitNo, this%num, frow, this%felem, nrows*this%num_vec_elements&
          call FTPCLD(unitNo, this%num, frow, 1, nrows*this%num_vec_elements&
            , this%r8vecdata(:,frow:lrow), ierr)
        end if
      else
        if(this%writeNulls) then
          call FTPCND(unitNo, this%num, frow, 1, nrows&
           , this%r8data(frow:lrow), this%r8null, ierr)
        else
          call FTPCLD(unitNo, this%num, frow, 1, nrows&
            , this%r8data(frow:lrow), ierr)
        end if
      end if
    case("B")
      if(this%num_vec_elements>0) then
!        call FTPCLB(unitNo, this%num, frow, this%felem, nrows*this%num_vec_elements&
        call FTPCLB(unitNo, this%num, frow, 1, nrows*this%num_vec_elements&
          , this%i1vecdata(:,frow:lrow), ierr)
      else
        call FTPCLB(unitNo, this%num, frow, 1, nrows&
          , this%i1data(frow:lrow), ierr)
      end if
    case("I")
      if(this%num_vec_elements>0) then
!        call FTPCLI(unitNo, this%num, frow, this%felem, nrows*this%num_vec_elements&
        call FTPCLI(unitNo, this%num, frow, 1, nrows*this%num_vec_elements&
          , this%i2vecdata(:,frow:lrow), ierr)
      else
        call FTPCLI(unitNo, this%num, frow, 1, nrows&
          , this%i2data(frow:lrow), ierr)
      end if
    case("J")
      if(this%num_vec_elements>0) then
!        call FTPCLJ(unitNo, this%num, frow, this%felem, nrows*this%num_vec_elements&
        call FTPCLJ(unitNo, this%num, frow, 1, nrows*this%num_vec_elements&
          , this%i4vecdata(:,frow:lrow), ierr)
      else
        call FTPCLJ(unitNo, this%num, frow, 1, nrows&
          , this%i4data(frow:lrow), ierr)
      end if
    case("K")
      if(this%num_vec_elements>0) then
!        call FTPCLK(unitNo, this%num, frow, this%felem, nrows*this%num_vec_elements&
        call FTPCLK(unitNo, this%num, frow, 1, nrows*this%num_vec_elements&
          , this%i8vecdata(:,frow:lrow), ierr)
      else
        call FTPCLK(unitNo, this%num, frow, 1, nrows&
          , this%i8data(frow:lrow), ierr)
      end if
    case default
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
      status = s_errRep
  return
    end select

    call report_FITS_err(functionName, ierr, status, " - column "//trim(this%nameStr))

  end subroutine writeRows_tColumn

  !.....................................................................
  subroutine readTNULL_tColumn(this, unitNo, setMyNull, tnullFound, status)
    implicit none

    class(tColumn) :: this
    integer, intent(in)    :: unitNo
    logical, intent(in)    :: setMyNull ! if .false., the routine just detects the presence of the TNULLn.
    logical, intent(out)   :: tnullFound
    integer, intent(inout) :: status

    integer :: ierr
    character(fkeylength) :: kwdname_tnull
    integer(i4) :: local_i4null,i4_val
    integer(i8) :: local_i8null
    character(fcommentlength) :: tmp_comment
    character(*), parameter :: functionName = "readTNULL_tColumn"

    tnullFound = .false. ! default

    kwdname_tnull = "TNULL"//trim(int2str(this%num, status))
    ierr = 0

    select case(this%dataTypeFITS)
    case("B")
      call FTGKYJ(unitNo, kwdname_tnull, i4_val, tmp_comment, ierr)
      if (ierr==0 .and. setMyNull) this%i1null = int(i4_val, kind(this%i1null))
!*** test for range?

    case("I")
      call FTGKYJ(unitNo, kwdname_tnull, i4_val, tmp_comment, ierr)
      if (ierr==0 .and. setMyNull) this%i2null = int(i4_val, kind(this%i2null))
!*** test for range?

    case("J")
      call FTGKYJ(unitNo, kwdname_tnull, local_i4null, tmp_comment, ierr)
      if (ierr==0 .and. setMyNull) this%i4null = local_i4null

    case("K")
      call FTGKYK(unitNo, kwdname_tnull, local_i8null, tmp_comment, ierr)
      if (ierr==0 .and. setMyNull) this%i8null = local_i8null

    case default
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
      status = s_errRep
  return
    end select

    if(ierr==0) then
      tnullFound = .true.
    else if(ierr/=202) then !**** 20210428 IMS: eh? According to HEASARC, 202 is KEY_NO_EXIST.
      call report_FITS_err(functionName, ierr, status, "- column "//trim(this%nameStr))
    end if

  end subroutine readTNULL_tColumn

  !.....................................................................
  subroutine setupFITSnulls_tColumn(this, unitNo, status)
    implicit none

    class(tColumn) :: this
    integer, intent(in)    :: unitNo
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "setupFITSnulls_tColumn"
    logical :: tnullKwdFound
    character(fkeylength) :: kwdname_tnull
    integer :: ierr

!*** type A??

    if(  this%dataTypeFITS=='B' .or. this%dataTypeFITS=='I'&
    .or. this%dataTypeFITS=='J' .or. this%dataTypeFITS=='K') then
      if(this%writeNulls) then
        call this%readTNULL(unitNo, .true., tnullKwdFound, status) ! sets this%??null from TNULLn kwd if found.
        if(status==0 .and. .not.tnullKwdFound) then
          call this%updateTNULL(unitNo, status)
        end if

      else
        call this%readTNULL(unitNo, .false., tnullKwdFound, status)
        if(status==0 .and. tnullKwdFound) then ! delete with a warning.
          kwdname_tnull = "TNULL"//trim(int2str(this%num, status))
          call wrerr(functionName, warn, "TNULL was found set for column "&
            //trim(this%nameStr)//", but you specified to write no nulls. Deleting keyword!")
          ierr = 0
          call FTDKEY(unitNo, kwdname_tnull, ierr)
          call report_FITS_err(functionName, ierr, status, "- column "//trim(this%nameStr))
        end if
      end if
    end if

!**** warn if any other type and col%writeNulls has been set.

  end subroutine setupFITSnulls_tColumn

  !.....................................................................
  subroutine readScaleKwds_tColumn(this, unitNo, status)
    implicit none

    class(tColumn) :: this
    integer, intent(in)    :: unitNo
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "readScaleKwds_tColumn"
    character(fkeylength) :: kwdname_tzero
    character(fkeylength) :: kwdname_tscal
    integer :: ierr
    character(fcommentlength) :: tmp_comment
    real(r8) :: local_tzero,local_tscal

    kwdname_tzero = "TZERO"//trim(int2str(this%num, status))
    ierr = 0
    call FTGKYD(unitNo, kwdname_tzero, local_tzero, tmp_comment, ierr)
    if(ierr==0) then
      kwdname_tscal = "TSCAL"//trim(int2str(this%num, status))
      ierr = 0
      call FTGKYD(unitNo, kwdname_tscal, local_tscal, tmp_comment, ierr)
      if(ierr==0) then
        this%scal%tzero = local_tzero
        this%scal%tscal = local_tscal
        this%scal%containsData = .true.
      end if
    end if

  end subroutine readScaleKwds_tColumn

  !.....................................................................
  subroutine writeScaleKwds_tColumn(this, unitNo, status)
    implicit none

    class(tColumn) :: this
    integer, intent(in)    :: unitNo
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "writeScaleKwds_tColumn"
    character(fkeylength) :: kwdname_tzero
    character(fkeylength) :: kwdname_tscal
    integer :: ierr
    integer :: decimals=15

    if(this%scal%containsData) then
      kwdname_tzero = "TZERO"//trim(int2str(this%num, status))
      ierr = 0
      call FTUKYD(unitNo, kwdname_tzero, this%scal%tzero, decimals&
        , "Offset for "//trim(this%nameStr)//" scaling", ierr)
      call report_FITS_err(functionName, ierr, status, "- column "//trim(this%nameStr))

      kwdname_tscal = "TSCAL"//trim(int2str(this%num, status))
      ierr = 0
      call FTUKYD(unitNo, kwdname_tscal, this%scal%tscal, decimals&
        , "Scaling value for "//trim(this%nameStr)//" column", ierr)
      call report_FITS_err(functionName, ierr, status, "- column "//trim(this%nameStr))
    end if

  end subroutine writeScaleKwds_tColumn

  !.....................................................................
  subroutine updateTNULL_tColumn(this, unitNo, status)
    implicit none

    class(tColumn) :: this
    integer, intent(in)    :: unitNo
    integer, intent(inout) :: status

    integer :: ierr
    character(fkeylength) :: keyname_tnull
    integer(i4) :: i4_val
    character(*), parameter :: functionName = "updateTNULL_tColumn"

    keyname_tnull = "TNULL"//trim(int2str(this%num, status))
    ierr = 0

    select case(this%dataTypeFITS)
!*** We should not write TNULL keywords for floating-point values, and we don't need to.
!    case("E")
!      call FTUKYE(unitNo, keyname_tnull, this%r4null, -14, "NULL VALUE", ierr)
!    case("D")
!      call FTUKYD(unitNo, keyname_tnull, this%r8null, -14, "NULL VALUE", ierr)
    case("B")
      i4_val = int(this%i1null, kind(i4_val))
!*** test for range?
      call FTUKYJ(unitNo, keyname_tnull, i4_val, "NULL VALUE", ierr)
    case("I")
      i4_val = int(this%i2null, kind(i4_val))
!*** test for range?
      call FTUKYJ(unitNo, keyname_tnull, i4_val, "NULL VALUE", ierr)
    case("J")
      call FTUKYJ(unitNo, keyname_tnull, this%i4null, "NULL VALUE", ierr)
    case("K")
      call FTUKYK(unitNo, keyname_tnull, this%i8null, "NULL VALUE", ierr)
    case default
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
      status = s_errRep
  return
    end select

    call report_FITS_err(functionName, ierr, status, " - column "//trim(this%nameStr))

    call wrchatty(functionName,"Update FITS null for "//trim(this%nameStr)//".")

  end subroutine updateTNULL_tColumn

  !.....................................................................
  subroutine createFITSCol_tColumn(this, unitNo, nColsInTable, status)
    implicit none

    class(tColumn) :: this
    integer, intent(in)    :: unitNo
    integer, intent(inout) :: nColsInTable
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "createFITSCol_tColumn"
    integer :: ierr
    character(100) :: tform

    if(this%dataTypeFITS=='L')&
call wrstop(functionName, "Creation of boolean columns is not permitted.")

    if(this%dataTypeFITS=='A') then
      write(tform, '(a)') trim(int2str(this%num_chars))//this%dataTypeFITS
    else if(this%num_vec_elements>0) then
      write(tform, '(a)') trim(int2str(this%num_vec_elements))//this%dataTypeFITS
    else
      write(tform, '(a)') this%dataTypeFITS
    end if

    ierr = 0
    call FTICOL(unitNo, nColsInTable+1, this%nameStr, tform, ierr)
    call report_FITS_err(functionName, ierr, status, " - column "//trim(this%nameStr))

    nColsInTable = nColsInTable + 1
    this%num = nColsInTable

    call wrlog(functionName,"Appended the column "//trim(this%nameStr)//".")

  end subroutine createFITSCol_tColumn

  !.....................................................................
  subroutine setUseage_tColumn(this, optionalFlag, readFlag&
    , writeFlag, useInternallyFlag)

    implicit none

    class(tColumn) :: this
    integer, intent(in) :: optionalFlag,readFlag,writeFlag,useInternallyFlag

    character(*), parameter :: functionName = "setUseage_tColumn"

    this%useageFlags = 0 ! default

    if(optionalFlag>0)&
      this%useageFlags = ibset(this%useageFlags, isOptionalBit)
    if(readFlag>0)&
      this%useageFlags = ibset(this%useageFlags, readBit)
    if(writeFlag>0)&
      this%useageFlags = ibset(this%useageFlags, writeBit)
    if(useInternallyFlag>0)&
      this%useageFlags = ibset(this%useageFlags, internalUseBit)

    if(this%dataTypeFITS=='L' .and.(btest(this%useageFlags,readBit) .or. btest(this%useageFlags,writeBit)))&
call wrstop(functionName, "Read and write useage bits may not be set for a boolean column.")

  end subroutine setUseage_tColumn

  !.....................................................................
  subroutine ibset_useage_tColumn(this, bitNum)
    implicit none

    class(tColumn) :: this
    integer, intent(in) :: bitNum

    character(*), parameter :: functionName = "ibset_useage_tColumn"

    if(bitNum<0 .or. bitNum>maxBitNumUsed) then
      call wrerr(functionName, warn, "Invalid bit number.")
  return
    end if

    if(this%dataTypeFITS=='L' .and.(bitNum==readBit .or. bitNum==writeBit))&
call wrstop(functionName, "Read and write useage bits may not be set for a boolean column.")

    this%useageFlags = ibset(this%useageFlags, bitNum)

  end subroutine ibset_useage_tColumn

  !.....................................................................
  subroutine ibclr_useage_tColumn(this, bitNum)
    implicit none

    class(tColumn) :: this
    integer, intent(in) :: bitNum

    character(*), parameter :: functionName = "ibclr_useage_tColumn"

    if(bitNum<0 .or. bitNum>maxBitNumUsed) then
      call wrerr(functionName, warn, "Invalid bit number.")
  return
    end if

    if(this%dataTypeFITS=='L' .and.(bitNum==readBit .or. bitNum==writeBit))&
call wrstop(functionName, "Read and write useage bits may not be set for a boolean column.")

    this%useageFlags = ibclr(this%useageFlags, bitNum)

  end subroutine ibclr_useage_tColumn

  !.....................................................................
  logical function btest_useage_tColumn(this, bitNum)
    class(tColumn) :: this
    integer, intent(in) :: bitNum

    character(*), parameter :: functionName = "btest_useage_tColumn"

    if(bitNum<0 .or. bitNum>maxBitNumUsed) then
      call wrerr(functionName, warn, "Invalid bit number.")
      btest_useage_tColumn = .false.
    else
      btest_useage_tColumn = btest(this%useageFlags, bitNum)
    end if

  end function btest_useage_tColumn

  !.....................................................................
  function isAllocatedVec_tColumn(this) result(isAllocated)

    implicit none

    class(tColumn) :: this
    logical :: isAllocated

    character(*), parameter :: functionName = "isAllocatedVec_tColumn"

    isAllocated = .false. ! default

    if(allocated(this%b0vecdata)) then
      if(isAllocated)&
        call wrerr(functionName, warn, "Column values of more than 1 data type are allocated!")

      if(this%dataTypeFITS /= "L")&
        call wrerr(functionName, warn, "Column of wrong data type is allocated! L found, should be "&
          //trim(this%dataTypeFITS)//".")

      isAllocated = .true.
    end if

    if(allocated(this%i1vecdata)) then
      if(isAllocated)&
        call wrerr(functionName, warn, "Column values of more than 1 data type are allocated!")

      if(this%dataTypeFITS /= "B")&
        call wrerr(functionName, warn, "Column of wrong data type is allocated! B found, should be "&
          //trim(this%dataTypeFITS)//".")

      isAllocated = .true.
    end if

    if(allocated(this%i2vecdata)) then
      if(isAllocated)&
        call wrerr(functionName, warn, "Column values of more than 1 data type are allocated!")

      if(this%dataTypeFITS /= "I")&
        call wrerr(functionName, warn, "Column of wrong data type is allocated! I found, should be "&
          //trim(this%dataTypeFITS)//".")

      isAllocated = .true.
    end if

    if(allocated(this%i4vecdata)) then
      if(isAllocated)&
        call wrerr(functionName, warn, "Column values of more than 1 data type are allocated!")

      if(this%dataTypeFITS /= "J")&
        call wrerr(functionName, warn, "Column of wrong data type is allocated! J found, should be "&
          //trim(this%dataTypeFITS)//".")

      isAllocated = .true.
    end if

    if(allocated(this%i8vecdata)) then
      if(isAllocated)&
        call wrerr(functionName, warn, "Column values of more than 1 data type are allocated!")

      if(this%dataTypeFITS /= "K")&
        call wrerr(functionName, warn, "Column of wrong data type is allocated! K found, should be "&
          //trim(this%dataTypeFITS)//".")

      isAllocated = .true.
    end if

    if(allocated(this%r4vecdata)) then
      if(isAllocated)&
        call wrerr(functionName, warn, "Column values of more than 1 data type are allocated!")

      if(this%dataTypeFITS /= "E")&
        call wrerr(functionName, warn, "Column of wrong data type is allocated! E found, should be "&
          //trim(this%dataTypeFITS)//".")

      isAllocated = .true.
    end if

    if(allocated(this%r8vecdata)) then
      if(isAllocated)&
        call wrerr(functionName, warn, "Column values of more than 1 data type are allocated!")

      if(this%dataTypeFITS /= "D")&
        call wrerr(functionName, warn, "Column of wrong data type is allocated! D found, should be "&
          //trim(this%dataTypeFITS)//".")

      isAllocated = .true.
    end if

  end function isAllocatedVec_tColumn

  !.....................................................................
  function isAllocated_tColumn(this) result(isAllocated)

    implicit none

    class(tColumn) :: this
    logical :: isAllocated

    character(*), parameter :: functionName = "isAllocated_tColumn"
    logical :: vectorIsAllocated

    isAllocated = .false. ! default

    if(allocated(this%b0data)) then
      if(isAllocated)&
        call wrerr(functionName, warn, "Column values of more than 1 data type are allocated!")

      if(this%dataTypeFITS /= "L")&
        call wrerr(functionName, warn, "Column of wrong data type is allocated! L found, should be "&
          //trim(this%dataTypeFITS)//".")

      isAllocated = .true.
    end if

    if(allocated(this%strdata)) then
      if(isAllocated)&
        call wrerr(functionName, warn, "Column values of more than 1 data type are allocated!")

      if(this%dataTypeFITS /= "A")&
        call wrerr(functionName, warn, "Column of wrong data type is allocated! A found, should be "&
          //trim(this%dataTypeFITS)//".")

      isAllocated = .true.
    end if

    if(allocated(this%i1data)) then
      if(isAllocated)&
        call wrerr(functionName, warn, "Column values of more than 1 data type are allocated!")

      if(this%dataTypeFITS /= "B")&
        call wrerr(functionName, warn, "Column of wrong data type is allocated! B found, should be "&
          //trim(this%dataTypeFITS)//".")

      isAllocated = .true.
    end if

    if(allocated(this%i2data)) then
      if(isAllocated)&
        call wrerr(functionName, warn, "Column values of more than 1 data type are allocated!")

      if(this%dataTypeFITS /= "I")&
        call wrerr(functionName, warn, "Column of wrong data type is allocated! I found, should be "&
          //trim(this%dataTypeFITS)//".")

      isAllocated = .true.
    end if

    if(allocated(this%i4data)) then
      if(isAllocated)&
        call wrerr(functionName, warn, "Column values of more than 1 data type are allocated!")

      if(this%dataTypeFITS /= "J")&
        call wrerr(functionName, warn, "Column of wrong data type is allocated! J found, should be "&
          //trim(this%dataTypeFITS)//".")

      isAllocated = .true.
    end if

    if(allocated(this%i8data)) then
      if(isAllocated)&
        call wrerr(functionName, warn, "Column values of more than 1 data type are allocated!")

      if(this%dataTypeFITS /= "K")&
        call wrerr(functionName, warn, "Column of wrong data type is allocated! K found, should be "&
          //trim(this%dataTypeFITS)//".")

      isAllocated = .true.
    end if

    if(allocated(this%r4data)) then
      if(isAllocated)&
        call wrerr(functionName, warn, "Column values of more than 1 data type are allocated!")

      if(this%dataTypeFITS /= "E")&
        call wrerr(functionName, warn, "Column of wrong data type is allocated! E found, should be "&
          //trim(this%dataTypeFITS)//".")

      isAllocated = .true.
    end if

    if(allocated(this%r8data)) then
      if(isAllocated)&
        call wrerr(functionName, warn, "Column values of more than 1 data type are allocated!")

      if(this%dataTypeFITS /= "D")&
        call wrerr(functionName, warn, "Column of wrong data type is allocated! D found, should be "&
          //trim(this%dataTypeFITS)//".")

      isAllocated = .true.
    end if

    vectorIsAllocated = this%isAllocatedVec()
    if(vectorIsAllocated .and. isAllocated)&
      call wrerr(functionName, warn, "Both vector and scalar column values are allocated!")

    isAllocated = isAllocated .or. vectorIsAllocated

  end function isAllocated_tColumn

  !.....................................................................
!  function isSame_tColumn(this, ri1, ri2, dt) result(isSame)
  function isSame_tColumn(this, ri1, ri2) result(isSame)

    implicit none

    class(tColumn) :: this
    integer, intent(in) :: ri1, ri2
    logical :: isSame
!logical, intent(in), optional :: dt

    character(*), parameter :: functionName = "isSame_tColumn"
!logical :: ldt

!if(present(dt)) then
!  ldt=.true.
!else
!  ldt=.false.
!end if

    isSame = .false. ! default

!!print *, "+++ isSame_tColumn(). present(dt)?", present(dt)
!if(doTest .and. ldt) print *, ">>> isSame_tColumn(). Column "//trim(this%nameStr)
!if(doTest .and. ldt) print *, "+++ isSame_tColumn(). ri1, ri2:", ri1, ri2

    select case(this%dataTypeFITS)
    case("L")
      if(this%num_vec_elements>0) then
        if(all(this%b0vecdata(:,ri1) .eqv. this%b0vecdata(:,ri2)))&
          isSame = .true.
      else
        if(this%b0data(ri1) .eqv. this%b0data(ri2))&
          isSame = .true.
      end if
    case("A")
      if(this%num_vec_elements>0)&
call wrstop(functionName, "Vector string columns are not yet permitted.")

      if(this%strdata(ri1) == this%strdata(ri2))&
        isSame = .true.
    case("E")
      if(this%num_vec_elements>0) then
        if(all(abs(this%r4vecdata(:,ri1) - this%r4vecdata(:,ri2)) < epsilon(this%r4vecdata(1,1))))&
        isSame = .true.
      else
        if(abs(this%r4data(ri1) - this%r4data(ri2)) < epsilon(this%r4data(1)))&
        isSame = .true.
      end if
    case("D")
      if(this%num_vec_elements>0) then
        if(all(abs(this%r8vecdata(:,ri1) - this%r8vecdata(:,ri2)) < epsilon(this%r8vecdata(1,1))))&
          isSame = .true.
      else
        if(abs(this%r8data(ri1) - this%r8data(ri2)) < epsilon(this%r8data(1)))&
          isSame = .true.
      end if
    case("B")
      if(this%num_vec_elements>0) then
        if(all(this%i1vecdata(:,ri1) == this%i1vecdata(:,ri2)))&
          isSame = .true.
      else
        if(this%i1data(ri1) == this%i1data(ri2))&
          isSame = .true.
      end if
    case("I")
      if(this%num_vec_elements>0) then
        if(all(this%i2vecdata(:,ri1) == this%i2vecdata(:,ri2)))&
          isSame = .true.
      else
!if(doTest .and. ldt) print *, "+++ isSame_tColumn(). row(ri1), row(ri2):", this%i2data(ri1), this%i2data(ri2)
        if(this%i2data(ri1) == this%i2data(ri2))&
          isSame = .true.
      end if
    case("J")
      if(this%num_vec_elements>0) then
        if(all(this%i4vecdata(:,ri1) == this%i4vecdata(:,ri2)))&
          isSame = .true.
      else
!if(doTest .and. ldt) print *, "+++ isSame_tColumn(). row(ri1), row(ri2):", this%i4data(ri1), this%i4data(ri2)
        if(this%i4data(ri1) == this%i4data(ri2))&
          isSame = .true.
      end if
    case("K")
      if(this%num_vec_elements>0) then
        if(all(this%i8vecdata(:,ri1) == this%i8vecdata(:,ri2)))&
          isSame = .true.
      else
        if(this%i8data(ri1) == this%i8data(ri2))&
          isSame = .true.
      end if
    case default
      call wrerr(functionName, s_err, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
    end select

  end function isSame_tColumn

  !.....................................................................
  subroutine to_str_tColumn(this, ri, str, vi)
    ! This is not used anywhere.
    implicit none

    class(tColumn) :: this
    integer,           intent(in) :: ri ! row number
    character(*),      intent(out) :: str
    integer, optional, intent(in) :: vi ! element number for vector columns

    character(*), parameter :: functionName = "to_str_tColumn"
    integer :: local_vi=1

    if(present(vi))&
      local_vi = vi

    if(this%num_vec_elements>0) then
      select case(this%dataTypeFITS)
      case("L")
        write(str, '(l)') this%b0vecdata(local_vi,ri)
      case("A")
        call wrstop(functionName, "Vector string columns are not yet permitted.")
      case("E")
        write(str, *) this%r4vecdata(local_vi,ri)
      case("D")
        write(str, *) this%r8vecdata(local_vi,ri)
      case("B")
        write(str, *) this%i1vecdata(local_vi,ri)
      case("I")
        write(str, *) this%i2vecdata(local_vi,ri)
      case("J")
        write(str, *) this%i4vecdata(local_vi,ri)
      case("K")
        write(str, *) this%i8vecdata(local_vi,ri)
      case default
        call wrerr(functionName, s_err, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
      end select
    else
      select case(this%dataTypeFITS)
      case("L")
        write(str, '(l)') this%b0data(ri)
      case("A")
        write(str, '(a)') this%strdata(ri)
      case("E")
        write(str, *) this%r4data(ri)
      case("D")
        write(str, *) this%r8data(ri)
      case("B")
        write(str, *) this%i1data(ri)
      case("I")
        write(str, *) this%i2data(ri)
      case("J")
        write(str, *) this%i4data(ri)
      case("K")
        write(str, *) this%i8data(ri)
      case default
        call wrerr(functionName, s_err, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
      end select
    end if

  end subroutine to_str_tColumn

  !.....................................................................
  subroutine copy_data_tColumn(this, source_col)
    ! Since this is a private method, I'm going to assume that the calling routine has checked that both columns are allocated and of the same size.
    implicit none

    class(tColumn) :: this,source_col

    character(*), parameter :: functionName = "copy_data_tColumn"

    if(this%dataTypeFITS /= source_col%dataTypeFITS) then
call wrstop(functionName, "Column "//trim(this%nameStr)//": data types must match. Source table has '"&
  //source_col%dataTypeFITS//"' but destination table has '"//this%dataTypeFITS//"'.")
    end if

    select case(this%dataTypeFITS)
    case("L")
      if(this%num_vec_elements>0) then
        this%b0vecdata = source_col%b0vecdata
      else
        this%b0data = source_col%b0data
      end if

    case("A")
      if(this%num_vec_elements>0)&
call wrstop(functionName, "Vector string columns are not yet permitted.")

      this%strdata = source_col%strdata

    case("E")
      if(this%num_vec_elements>0) then
        this%r4vecdata = source_col%r4vecdata
      else
        this%r4data = source_col%r4data
      end if

    case("D")
      if(this%num_vec_elements>0) then
        this%r8vecdata = source_col%r8vecdata
      else
        this%r8data = source_col%r8data
      end if

    case("B")
      if(this%num_vec_elements>0) then
        this%i1vecdata = source_col%i1vecdata
      else
        this%i1data = source_col%i1data
      end if

    case("I")
      if(this%num_vec_elements>0) then
        this%i2vecdata = source_col%i2vecdata
      else
        this%i2data = source_col%i2data
      end if

    case("J")
      if(this%num_vec_elements>0) then
        this%i4vecdata = source_col%i4vecdata
      else
        this%i4data = source_col%i4data
      end if

    case("K")
      if(this%num_vec_elements>0) then
        this%i8vecdata = source_col%i8vecdata
      else
        this%i8data = source_col%i8data
      end if

    case default
call wrstop(functionName, "Unrecognized data type '"//trim(this%dataTypeFITS)//"'")
    end select

    if(this%num_vec_elements>0) then
      this%isNullVec = source_col%isNullVec
    else
      this%isNull = source_col%isNull
    end if

  end subroutine copy_data_tColumn

  !.....................................................................
  function copy_tColumn(this) result(new_col)
    implicit none

    class(tColumn) :: this!, new_col
    type(tColumn) :: new_col

    character(*), parameter :: functionName = "copy_tColumn"

!    new_col = this
    new_col%tabInfo => this%tabInfo
    new_col%nameStr           = this%nameStr
    new_col%num               = this%num
    new_col%dataTypeFITS      = this%dataTypeFITS
    new_col%unitStr           = this%unitStr
    new_col%writeNulls        = this%writeNulls
    new_col%colNumFoundInFITS = this%colNumFoundInFITS
    new_col%num_chars         = this%num_chars
    new_col%num_vec_elements  = this%num_vec_elements
    new_col%group_i           = this%group_i
    new_col%useageFlags       = this%useageFlags
    new_col%r4null            = this%r4null
    new_col%r8null            = this%r8null
    new_col%i1null            = this%i1null
    new_col%i2null            = this%i2null
    new_col%i4null            = this%i4null
    new_col%i8null            = this%i8null
    new_col%strnull           = this%strnull
    new_col%scal              = this%scal
!    if(allocated(this%felem)) new_col%felem = this%felem

    if(allocated(this%strdata  )) new_col%strdata   = this%strdata
    if(allocated(this%r4data   )) new_col%r4data    = this%r4data
    if(allocated(this%r8data   )) new_col%r8data    = this%r8data
    if(allocated(this%i1data   )) new_col%i1data    = this%i1data
    if(allocated(this%i2data   )) new_col%i2data    = this%i2data
    if(allocated(this%i4data   )) new_col%i4data    = this%i4data
    if(allocated(this%i8data   )) new_col%i8data    = this%i8data
    if(allocated(this%b0data   )) new_col%b0data    = this%b0data
    if(allocated(this%isNull   )) new_col%isNull    = this%isNull
    if(allocated(this%r4vecdata)) new_col%r4vecdata = this%r4vecdata
    if(allocated(this%r8vecdata)) new_col%r8vecdata = this%r8vecdata
    if(allocated(this%i1vecdata)) new_col%i1vecdata = this%i1vecdata
    if(allocated(this%i2vecdata)) new_col%i2vecdata = this%i2vecdata
    if(allocated(this%i4vecdata)) new_col%i4vecdata = this%i4vecdata
    if(allocated(this%i8vecdata)) new_col%i8vecdata = this%i8vecdata
    if(allocated(this%b0vecdata)) new_col%b0vecdata = this%b0vecdata
    if(allocated(this%isNullVec)) new_col%isNullVec = this%isNullVec

  end function copy_tColumn

  !.....................................................................
  subroutine get_unit_tColumn(this, unit_str)
    implicit none

    class(tColumn) :: this
    character(*), intent(out) :: unit_str

    character(*), parameter :: functionName = "get_unit_tColumn"

    unit_str = trim(this%unitStr)

  end subroutine get_unit_tColumn

  !.....................................................................
  subroutine set_unit_tColumn(this, unit_str)
    implicit none

    class(tColumn) :: this
    character(*), intent(in) :: unit_str

    character(*), parameter :: functionName = "set_unit_tColumn"

    this%unitStr = trim(unit_str)

  end subroutine set_unit_tColumn

  !.....................................................................
  subroutine append_zeroes_tColumn(this, n_new_elements, status)
    implicit none

    class(tColumn)         :: this
    integer, intent(in)    :: n_new_elements
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "append_zeroes_tColumn"
    integer :: size_a,size_out
    type(tColumn) :: temp_col

    size_a = this%tabInfo%numRows
    size_out = size_a + n_new_elements

    if(.not.this%isAllocated()) then
      call wrerr(functionName, warn, "Base table column "//trim(this%nameStr)//" is not allocated. No append.")
  return
    end if

    temp_col = this%copy()
    call this%realloc(size_out, status)

    if(this%num_vec_elements>0) then
!      call colOut%allocVec(this%num_vec_elements, status)

      select case(this%dataTypeFITS)
      case("L")
        this%b0vecdata(:,1:size_a) = temp_col%b0vecdata

!      case("A") not allowed for vector columns.
      case("E")
        this%r4vecdata(:,1:size_a) = temp_col%r4vecdata

      case("D")
        this%r8vecdata(:,1:size_a) = temp_col%r8vecdata

      case("B")
        this%i1vecdata(:,1:size_a) = temp_col%i1vecdata

      case("I")
        this%i2vecdata(:,1:size_a) = temp_col%i2vecdata

      case("J")
        this%i4vecdata(:,1:size_a) = temp_col%i4vecdata

      case("K")
        this%i8vecdata(:,1:size_a) = temp_col%i8vecdata

      case default
        call wrerr(functionName, s_err, "Unrecognized choice '"&
          //trim(this%dataTypeFITS)//"'")
        status = s_errRep
  return
      end select

      this%isNullVec(:,1:size_a) = temp_col%isNullVec

    else
      select case(this%dataTypeFITS)
      case("L")
        this%b0data(1:size_a) = temp_col%b0data

      case("A")
        this%strdata(1:size_a) = temp_col%strdata

      case("E")
        this%r4data(1:size_a) = temp_col%r4data

      case("D")
        this%r8data(1:size_a) = temp_col%r8data

      case("B")
        this%i1data(1:size_a) = temp_col%i1data

      case("I")
        this%i2data(1:size_a) = temp_col%i2data

      case("J")
        this%i4data(1:size_a) = temp_col%i4data

      case("K")
        this%i8data(1:size_a) = temp_col%i8data

      case default
        call wrerr(functionName, s_err, "Unrecognized choice '"&
          //trim(this%dataTypeFITS)//"'")
        status = s_errRep
  return
      end select

      this%isNull(1:size_a) = temp_col%isNull
    end if

    call temp_col%destroy(status)

  end subroutine append_zeroes_tColumn

  !.....................................................................
  subroutine append_tColumn(this, new_col, status)
    ! It's a private function, hence no checks that the input columns match. It is assumed that the calling routine has done this.

    implicit none

    class(tColumn)               :: this
    type(tColumn), intent(in)    :: new_col
    integer,       intent(inout) :: status

    character(*), parameter :: functionName = "append_tColumn"
    integer :: size_a,size_b,size_out
    type(tColumn) :: temp_col

!integer :: ii

!    size_a = this%getNumRows(status)
!    size_b = colInB%getNumRows(status)
    size_a = this%tabInfo%numRows
    size_b = new_col%tabInfo%numRows
    size_out = size_a + size_b

    if(.not.this%isAllocated()) then
      if(new_col%isAllocated())&
        call wrerr(functionName, warn, "Base table column "//trim(this%nameStr)//" is not allocated. No append.")
  return
    end if

    if(.not.new_col%isAllocated()) then
      call this%append_zeroes(size_b, status)
  return
    end if

    temp_col = this%copy()
    call this%realloc(size_out, status)

    if(this%num_vec_elements>0) then
!      call colOut%allocVec(this%num_vec_elements, status)

      select case(this%dataTypeFITS)
      case("L")
        this%b0vecdata(:,1:size_a)          = temp_col%b0vecdata
        this%b0vecdata(:,size_a+1:size_out) = new_col%b0vecdata

!      case("A") not allowed for vector columns.
      case("E")
        this%r4vecdata(:,1:size_a)          = temp_col%r4vecdata
        this%r4vecdata(:,size_a+1:size_out) = new_col%r4vecdata

      case("D")
        this%r8vecdata(:,1:size_a)          = temp_col%r8vecdata
        this%r8vecdata(:,size_a+1:size_out) = new_col%r8vecdata

      case("B")
        this%i1vecdata(:,1:size_a)          = temp_col%i1vecdata
        this%i1vecdata(:,size_a+1:size_out) = new_col%i1vecdata

      case("I")
        this%i2vecdata(:,1:size_a)          = temp_col%i2vecdata
        this%i2vecdata(:,size_a+1:size_out) = new_col%i2vecdata

      case("J")
        this%i4vecdata(:,1:size_a)          = temp_col%i4vecdata
        this%i4vecdata(:,size_a+1:size_out) = new_col%i4vecdata

      case("K")
        this%i8vecdata(:,1:size_a)          = temp_col%i8vecdata
        this%i8vecdata(:,size_a+1:size_out) = new_col%i8vecdata

      case default
        call wrerr(functionName, s_err, "Unrecognized choice '"&
          //trim(this%dataTypeFITS)//"'")
        status = s_errRep
  return
      end select

      this%isNullVec(:,1:size_a)          = temp_col%isNullVec
      this%isNullVec(:,size_a+1:size_out) = new_col%isNullVec

    else ! column is scalar
      select case(this%dataTypeFITS)
      case("L")
        this%b0data = [temp_col%b0data, new_col%b0data]

      case("A")
        this%strdata = [temp_col%strdata, new_col%strdata]

      case("E")
        this%r4data = [temp_col%r4data, new_col%r4data]

      case("D")
        this%r8data = [temp_col%r8data, new_col%r8data]

      case("B")
        this%i1data = [temp_col%i1data, new_col%i1data]

      case("I")
        this%i2data = [temp_col%i2data, new_col%i2data]

      case("J")
        this%i4data = [temp_col%i4data, new_col%i4data]

      case("K")
        this%i8data = [temp_col%i8data, new_col%i8data]

      case default
        call wrerr(functionName, s_err, "Unrecognized choice '"&
          //trim(this%dataTypeFITS)//"'")
        status = s_errRep
  return
      end select

      this%isNull = [temp_col%isNull, new_col%isNull]
    end if

    call temp_col%destroy(status)

  end subroutine append_tColumn

  !.....................................................................
  function get_str_tColumn(this, row_i) result(str)
    implicit none

    class(tColumn)      :: this
    integer, intent(in) :: row_i
    character(buffsize) :: str

    character(*), parameter :: functionName = "get_str_tColumn"

    str = ""

    if(.not.this%isAllocated()) then
      call wrerr(functionName, warn, "Table "//trim(this%tabInfo%tableName)&
      //" column "//trim(this%nameStr)//" is not allocated.")
  return
    end if

    if(this%num_vec_elements>0) then
      call wrerr(functionName, warn, "Table "//trim(this%tabInfo%tableName)&
      //" column "//trim(this%nameStr)//" is vector.")
  return
    end if

    if(row_i<1 .or. row_i>this%tabInfo%numRows)&
call wrstop(functionName, "Table "//trim(this%tabInfo%tableName)//" column "&
  //trim(this%nameStr)//": index out of range.")

    if(this%isNull(row_i)) then
      str = "NULL"

    else
      select case(this%dataTypeFITS)
      case("L")
        if(this%b0data(row_i)) then
          str = "True"
        else
          str = "False"
        end if

      case("A")
        str = this%strdata(row_i)
      case("E")
        write(str, *) this%r4data(row_i)
      case("D")
        write(str, *) this%r8data(row_i)
      case("B")
        write(str, *) this%i1data(row_i)
      case("I")
        write(str, *) this%i2data(row_i)
      case("J")
        write(str, *) this%i4data(row_i)
      case("K")
        write(str, *) this%i8data(row_i)

      case default
call wrstop(functionName, "Table "//trim(this%tabInfo%tableName)//" column "&
  //trim(this%nameStr)//": unrecognized choice '"//trim(this%dataTypeFITS)//"'")
      end select
    end if

  end function get_str_tColumn

  !=====================================================================
  ! The remaining routines are not methods of tColumn, they're just
  ! utility routines that manipulate the objects.

  !.....................................................................
  subroutine merge_columns(colInA, colInB, colOut, status)
    ! It's a private function, hence no checks that the input columns match.

    implicit none

    type(tColumn), intent(in)    :: colInA,colInB
    type(tColumn), intent(out)   :: colOut
    integer,           intent(inout) :: status

    character(*), parameter :: functionName = "merge_columns"
    integer :: size_a,size_b,size_out

    size_a = colInA%getNumRows(status)
    size_b = colInB%getNumRows(status)
    size_out = size_a + size_b
    call colOut%alloc(size_out, status)

    if(colInA%num_vec_elements>0) then
      call colOut%allocVec(colInA%num_vec_elements, status)

      select case(colInA%dataTypeFITS)
      case("L")
        colOut%b0vecdata(:,1:size_a)          = colInA%b0vecdata
        colOut%b0vecdata(:,size_a+1:size_out) = colInB%b0vecdata

!      case("A") not allowed for vector columns.
      case("E")
        colOut%r4vecdata(:,1:size_a)          = colInA%r4vecdata
        colOut%r4vecdata(:,size_a+1:size_out) = colInB%r4vecdata

      case("D")
        colOut%r8vecdata(:,1:size_a)          = colInA%r8vecdata
        colOut%r8vecdata(:,size_a+1:size_out) = colInB%r8vecdata

      case("B")
        colOut%i1vecdata(:,1:size_a)          = colInA%i1vecdata
        colOut%i1vecdata(:,size_a+1:size_out) = colInB%i1vecdata

      case("I")
        colOut%i2vecdata(:,1:size_a)          = colInA%i2vecdata
        colOut%i2vecdata(:,size_a+1:size_out) = colInB%i2vecdata

      case("J")
        colOut%i4vecdata(:,1:size_a)          = colInA%i4vecdata
        colOut%i4vecdata(:,size_a+1:size_out) = colInB%i4vecdata

      case("K")
        colOut%i8vecdata(:,1:size_a)          = colInA%i8vecdata
        colOut%i8vecdata(:,size_a+1:size_out) = colInB%i8vecdata

      case default
        call wrerr(functionName, s_err, "Unrecognized choice '"&
          //trim(colInA%dataTypeFITS)//"'")
        status = s_errRep
  return
      end select

      colOut%isNullVec(:,1:size_a)          = colInA%isNullVec
      colOut%isNullVec(:,size_a+1:size_out) = colInB%isNullVec

    else
      select case(colInA%dataTypeFITS)
      case("L")
        colOut%b0data = [colInA%b0data, colInB%b0data]

      case("A")
        colOut%strdata = [colInA%strdata, colInB%strdata]

      case("E")
        colOut%r4data = [colInA%r4data, colInB%r4data]

      case("D")
        colOut%r8data = [colInA%r8data, colInB%r8data]

      case("B")
        colOut%i1data = [colInA%i1data, colInB%i1data]

      case("I")
        colOut%i2data = [colInA%i2data, colInB%i2data]

      case("J")
        colOut%i4data = [colInA%i4data, colInB%i4data]

      case("K")
        colOut%i8data = [colInA%i8data, colInB%i8data]

      case default
        call wrerr(functionName, s_err, "Unrecognized choice '"&
          //trim(colInA%dataTypeFITS)//"'")
        status = s_errRep
  return
      end select

      colOut%isNull = [colInA%isNull, colInB%isNull]
    end if

  end subroutine merge_columns

  !.....................................................................
  subroutine derive_diffs_column(inCol, outI2Col, status)
    implicit none

    type(tColumn), intent(in)    :: inCol
    type(tColumn), intent(inout) :: outI2Col
    integer,       intent(inout) :: status

    integer :: numRows,ii
    character(*), parameter :: functionName = "derive_diffs_column"

    if(status > 100 .or. status == s_errRep) return

    numRows = inCol%getNumRows(status)
    if(numRows < 1)&
  return

    if(.not.inCol%isAllocated()) then
      call wrerr(functionName, s_err, "Input column "//trim(inCol%nameStr)//" contains no data.")
      status = s_errRep
  return
    end if
    if(.not.outI2Col%isAllocated()) then
      call wrerr(functionName, s_err, "Output column "//trim(outI2Col%nameStr)//" is not allocated.")
      status = s_errRep
  return
    end if
    if(inCol%num_vec_elements>0) then
      call wrerr(functionName, s_err, "Difference of vector columns is not yet supported.")
      status = s_errRep
  return
    end if
    if(inCol%dataTypeFITS=="A") then
      call wrerr(functionName, s_err, "Difference of character columns is not supported.")
      status = s_errRep
  return
    end if
    if(inCol%dataTypeFITS=="L") then
      call wrerr(functionName, s_err, "Difference of boolean columns is not supported.")
      status = s_errRep
  return
    end if

    outI2Col%i2data(1) = 0

    if(numRows < 2)&
  return

    do ii = 2, numRows
      select case(inCol%dataTypeFITS)
      case("E")
        if(inCol%r4data(ii) == inCol%r4data(ii-1)) then
          outI2Col%i2data(ii) = 0
        else if(inCol%r4data(ii) > inCol%r4data(ii-1)) then
          outI2Col%i2data(ii) = 1
        else ! inCol%r4data(ii) < inCol%r4data(ii-1)
          outI2Col%i2data(ii) = -1
        end if

      case("D")
        if(inCol%r8data(ii) == inCol%r8data(ii-1)) then
          outI2Col%i2data(ii) = 0
        else if(inCol%r8data(ii) > inCol%r8data(ii-1)) then
          outI2Col%i2data(ii) = 1
        else ! inCol%r8data(ii) < inCol%r8data(ii-1)
          outI2Col%i2data(ii) = -1
        end if

      case("B")
        if(inCol%i1data(ii) == inCol%i1data(ii-1)) then
          outI2Col%i2data(ii) = 0
        else if(inCol%i1data(ii) > inCol%i1data(ii-1)) then
          outI2Col%i2data(ii) = 1
        else ! inCol%i1data(ii) < inCol%i1data(ii-1)
          outI2Col%i2data(ii) = -1
        end if

      case("I")
        if(inCol%i2data(ii) == inCol%i2data(ii-1)) then
          outI2Col%i2data(ii) = 0
        else if(inCol%i2data(ii) > inCol%i2data(ii-1)) then
          outI2Col%i2data(ii) = 1
        else ! inCol%i2data(ii) < inCol%i2data(ii-1)
          outI2Col%i2data(ii) = -1
        end if

      case("J")
        if(inCol%i4data(ii) == inCol%i4data(ii-1)) then
          outI2Col%i2data(ii) = 0
        else if(inCol%i4data(ii) > inCol%i4data(ii-1)) then
          outI2Col%i2data(ii) = 1
        else ! inCol%i4data(ii) < inCol%i4data(ii-1)
          outI2Col%i2data(ii) = -1
        end if

      case("K")
        if(inCol%i8data(ii) == inCol%i8data(ii-1)) then
          outI2Col%i2data(ii) = 0
        else if(inCol%i8data(ii) > inCol%i8data(ii-1)) then
          outI2Col%i2data(ii) = 1
        else ! inCol%i8data(ii) < inCol%i8data(ii-1)
          outI2Col%i2data(ii) = -1
        end if

      case default
        call wrerr(functionName, s_err, "Unrecognized choice '"&
          //trim(inCol%dataTypeFITS)//"'")
        status = s_errRep
  return
      end select
    end do
  end subroutine derive_diffs_column

  !=====================================================================
  ! Methods of t_col_comparator:

  !.....................................................................
  subroutine init_t_col_comparator(this, bintable_obj, column_name)
    implicit none
    class(t_col_comparator)      :: this
    class(tBinTable), intent(in) :: bintable_obj
    character(*),     intent(in) :: column_name

    character(*), parameter :: functionName = "init_t_col_comparator"
    integer :: ierr,col_id

    ierr = 0

    col_id = bintable_obj%getColNum(column_name, ierr)
    this%dataTypeFITS = bintable_obj%cols(col_id)%dataTypeFITS

    select case(this%dataTypeFITS)
    case("L")
      call bintable_obj%get_col_ptr(column_name, this%b0data, this%isNull)
    case("A")
      call bintable_obj%get_col_ptr(column_name, this%strdata, this%isNull)
    case("E")
      call bintable_obj%get_col_ptr(column_name, this%r4data, this%isNull)
    case("D")
      call bintable_obj%get_col_ptr(column_name, this%r8data, this%isNull)
    case("B")
      call bintable_obj%get_col_ptr(column_name, this%i1data, this%isNull)
    case("I")
      call bintable_obj%get_col_ptr(column_name, this%i2data, this%isNull)
    case("J")
      call bintable_obj%get_col_ptr(column_name, this%i4data, this%isNull)
    case("K")
      call bintable_obj%get_col_ptr(column_name, this%i8data, this%isNull)
    case default
call wrstop(functionName, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
    end select

    this%is_initialized = .true.

  end subroutine init_t_col_comparator

  !.....................................................................
  integer function compare_t_col_comparator(this, i1, i2) result(comp)
    implicit none
    class(t_col_comparator) :: this
    integer, intent(in)  :: i1, i2

    character(*), parameter :: functionName = "compare_t_col_comparator"

    ! Always list nulls before non-nulls:
    !
    if(this%isNull(i1) .or. this%isNull(i2)) then
      if(this%isNull(i1)) then
        if(this%isNull(i2)) then
          comp = COMP_EQ
        else ! 1==null .and. 2/=null
          comp = COMP_LT
        end if
      else ! 1/=null .and. 2==null
        comp = COMP_GT
      end if
  return
    end if

    select case(this%dataTypeFITS)
    case("L")
      ! True before false.
      if(.not.this%b0data(i1) .and. this%b0data(i2)) then
        comp = COMP_GT
      else if(this%b0data(i1) .and..not. this%b0data(i2)) then
        comp = COMP_LT
      else
        comp = COMP_EQ
      end if

    case("A")
      if(this%strdata(i1) > this%strdata(i2)) then
        comp = COMP_GT
      else if(this%strdata(i1) < this%strdata(i2)) then
        comp = COMP_LT
      else
        comp = COMP_EQ
      end if

    case("E")
      if(this%r4data(i1) > this%r4data(i2)) then
        comp = COMP_GT
      else if(this%r4data(i1) < this%r4data(i2)) then
        comp = COMP_LT
      else
        comp = COMP_EQ
      end if

    case("D")
      if(this%r8data(i1) > this%r8data(i2)) then
        comp = COMP_GT
      else if(this%r8data(i1) < this%r8data(i2)) then
        comp = COMP_LT
      else
        comp = COMP_EQ
      end if

    case("B")
      if(this%i1data(i1) > this%i1data(i2)) then
        comp = COMP_GT
      else if(this%i1data(i1) < this%i1data(i2)) then
        comp = COMP_LT
      else
        comp = COMP_EQ
      end if

    case("I")
      if(this%i2data(i1) > this%i2data(i2)) then
        comp = COMP_GT
      else if(this%i2data(i1) < this%i2data(i2)) then
        comp = COMP_LT
      else
        comp = COMP_EQ
      end if

    case("J")
      if(this%i4data(i1) > this%i4data(i2)) then
        comp = COMP_GT
      else if(this%i4data(i1) < this%i4data(i2)) then
        comp = COMP_LT
      else
        comp = COMP_EQ
      end if

    case("K")
      if(this%i8data(i1) > this%i8data(i2)) then
        comp = COMP_GT
      else if(this%i8data(i1) < this%i8data(i2)) then
        comp = COMP_LT
      else
        comp = COMP_EQ
      end if

    case default
call wrstop(functionName, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
    end select

  end function compare_t_col_comparator

  !.....................................................................
  subroutine destroy_t_col_comparator(this)
    implicit none
    class(t_col_comparator) :: this

    character(*), parameter :: functionName = "destroy_t_col_comparator"

    if(.not.this%is_initialized)&
call wrstop(functionName, "The object is not initialized.")

    select case(this%dataTypeFITS)
    case("L")
      nullify(this%b0data, this%isNull)
    case("A")
      nullify(this%strdata, this%isNull)
    case("E")
      nullify(this%r4data, this%isNull)
    case("D")
      nullify(this%r8data, this%isNull)
    case("B")
      nullify(this%i1data, this%isNull)
    case("I")
      nullify(this%i2data, this%isNull)
    case("J")
      nullify(this%i4data, this%isNull)
    case("K")
      nullify(this%i8data, this%isNull)
    case default
call wrstop(functionName, "Unrecognized choice '"//trim(this%dataTypeFITS)//"'")
    end select

  end subroutine destroy_t_col_comparator

  !=====================================================================
  ! Methods of tKeyword:

  !.....................................................................
  subroutine set_tKeyword(this, dataTypeFITS, nameStr, value)!, unitStr, commentStr)
    implicit none

    class(tKeyword) :: this
    character(*),       intent(in)    :: dataTypeFITS, nameStr
    class(*), optional, intent(in)    :: value
!    character(*), intent(in), optional :: unitStr, commentStr

    character(*), parameter :: functionName = "set_tKeyword"

    this%nameStr = nameStr
    this%dataTypeFITS = dataTypeFITS

    if(present(value))&
      call this%setValue(value)

  end subroutine set_tKeyword

  !.....................................................................
  subroutine getValue_tKeyword(this, value)
!!!  function getValue_tKeyword(this) result(value)
    implicit none

    class(tKeyword) :: this
    class(*)   :: value ! compiler doesn't seem to like 'intent(out)' with polymorphic arguments.

    character(1) :: empiricalDataChar
    character(*), parameter :: functionName = "getValue_tKeyword"

    select type (value) ! ['E','D','J','A']
    type is (integer(i4))
      value = this%i4Value
      empiricalDataChar = 'J'
    type is (real(r4))
      value = this%r4Value
      empiricalDataChar = 'E'
    type is (real(8))
      value = this%r8Value
      empiricalDataChar = 'D'
    type is (character(*))
      value = this%strValue
      empiricalDataChar = 'A'
    class default
call wrstop(functionName, "Unrecognized choice of data type for keyword.")
    end select

    if(trim(this%dataTypeFITS) /= empiricalDataChar) then
      call wrerr(functionName, warn, "Data type of requested value ("//empiricalDataChar&
        //") does not match the keyword data type ("//trim(this%dataTypeFITS)//").")
    end if

  end subroutine getValue_tKeyword

  !.....................................................................
  subroutine setValue_tKeyword(this, value)
    ! This is the preferred way to set a keyword, rather than setting this%*Value directly, because it checks the data type for one thing, and sets the contains_data flag for another.
    implicit none

    class(tKeyword) :: this
    class(*), intent(in) :: value

    character(1) :: empiricalDataChar
    character(*), parameter :: functionName = "setValue_tKeyword"

    select type (value) ! ['E','D','J','A']
    type is (integer(i4))
      this%i4Value = value
      empiricalDataChar = 'J'
    type is (real(r4))
      this%r4Value = value
      empiricalDataChar = 'E'
    type is (real(8))
      this%r8Value = value
      empiricalDataChar = 'D'
    type is (character(*))
      this%strValue = value
      empiricalDataChar = 'A'
    class default
call wrstop(functionName, "Unrecognized choice of data type for keyword.")
    end select

    if(trim(this%dataTypeFITS) /= empiricalDataChar) then
      call wrerr(functionName, warn, "Data type of supplied value ("//empiricalDataChar&
        //") does not match the data type you claim ("//trim(this%dataTypeFITS)//").")

      this%dataTypeFITS = empiricalDataChar
    end if

    this%contains_data = .true.

  end subroutine setValue_tKeyword

  !.....................................................................
  logical function containsData_tKeyword(this)
    implicit none

    class(tKeyword) :: this

    containsData_tKeyword = this%contains_data

  end function containsData_tKeyword

  !.....................................................................
  function getDataType_tKeyword(this) result(dataType)
    ! We want to keep tColumn%dataTypeFITS private to remove the temptation to set it directly.

    implicit none

    class(tKeyword) :: this

    character(*), parameter :: functionName = "getDataType_tKeyword"
    character(1) :: dataType

    dataType=this%dataTypeFITS

  end function getDataType_tKeyword

  !=====================================================================
  ! Methods of tHeader:

  !.....................................................................
  subroutine initKeyword_tHeader(this, dataTypeFITS, nameStr)!!!, unitStr, commentStr)
    ! This should be called by all routines wishing to set up a tKeyword object as part of a FITS binary table header. Note that, after all desired keywords have been set up by such calls, the initialization of the set of keyword specifications should be completed by calling tHeader%initKeywords(status).
    !
    implicit none

    class(tHeader), target :: this
    character(*), intent(in) :: dataTypeFITS, nameStr
!!!    character(*), intent(in), optional :: unitStr, commentStr

    character(*), parameter :: functionName = "initKeyword_tHeader"
    type(tKeyword) :: kwd

!    kwd%tabInfo => this%info

    kwd%nameStr = nameStr
    kwd%dataTypeFITS = dataTypeFITS
!!!    if(present(unitStr))&
!!!      kwd%unitStr = unitStr
!!!    if(present(commentStr))&
!!!      kwd%commentStr = commentStr

    if(.not. associated(this%kwd_specs)) allocate(this%kwd_specs)
    call this%kwd_specs%addObj(kwd)

  end subroutine initKeyword_tHeader

  !.....................................................................
  subroutine deleteLinkedList_tHeader(this, status)
    implicit none

    class(tHeader) :: this
    integer, intent(inout) ::  status

    character(*), parameter :: functionName = "deleteLinkedList_tHeader"
    integer :: stat
!    class(*), pointer :: tmp

    if(status > 100 .or. status == s_errRep) return

!    if(associated(this%kwd_specs)) then
!      call this%kwd_specs%reset()
!      if(associated(tmp)) deallocate(tmp)
!    end if

    call this%kwd_specs%reset()
    stat =  this%kwd_specs%delAllLinks()
    if(stat /= 0) then
      call wrerr(functionName, s_err, "deallocation of LinkedList failed")
      status = s_errRep
    end if

    if(associated(this%kwd_specs)) deallocate(this%kwd_specs)

  end subroutine deleteLinkedList_tHeader

  !.....................................................................
  subroutine initKeywords_tHeader(this, status)
    ! Having loaded as many keyword specifications into the linked list this%kwd_specs as we want, we convert this LL to a vector of tKeyword objects.
    !
    implicit none

    class(tHeader) :: this ! target??
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "initKeywords_tHeader"
    integer :: nKwdsUsed,ii,jj
    class(*), pointer :: tmp
    character(buffsize) :: str1,str2

    nKwdsUsed = this%kwd_specs%getNumLinks()
    call this%alloc_kwds(nKwdsUsed, status)!**********************************
    call this%kwd_specs%reset() ! resets the leaf index to the start of the list.

    do ii = 1, nKwdsUsed
      tmp => this%kwd_specs%currObj()
      select type(tmp)
        type is (tKeyword)
          this%keywords(ii) = tmp
      end select

      call this%kwd_specs%next()
    end do

    call this%deleteLinkedList(status)

    do ii = 1, nKwdsUsed-1
      do jj=ii+1,nKwdsUsed
        if(trim(this%keywords(ii)%nameStr)==trim(this%keywords(jj)%nameStr)) then
          str1 = int2str(ii)
          str2 = int2str(jj)
call wrstop(functionName, "Keywords "//trim(str1)//" and "//trim(str2)&
 //" have the same name "//trim(this%keywords(ii)%nameStr))
        end if
      end do
    end do

!    this%isInitialized = .true.

  end subroutine initKeywords_tHeader

  !.....................................................................
  subroutine dealloc_kwds_tHeader(this, status)
    implicit none

    class(tHeader) :: this
    integer, intent(inout) :: status

    integer :: stat
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "dealloc_kwds_tHeader"

    if(allocated(this%keywords)) then
      call this%dealloc(status) ! deallocates arrays in each tKeyword object individually.
      errmsg = ""
      deallocate(this%keywords, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
    end if

!    this%isInitialized = .false.
!    this%useageIsSet = .false.

  end subroutine dealloc_kwds_tHeader

  !.....................................................................
  subroutine alloc_kwds_tHeader(this, nKwdsUsed, status)
    implicit none

    class(tHeader) :: this
    integer, intent(in)    :: nKwdsUsed
    integer, intent(inout) :: status

    integer :: stat
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "alloc_kwds_tHeader"

    if(nKwdsUsed < 1) then
      status = c_err
      call wrerr(functionName, status, "Can't allocate keyword list with <1 elements.")
  return
    end if

    call this%dealloc_kwds(status)
    errmsg = ""
    allocate(this%keywords(nKwdsUsed), stat=stat, errmsg=errmsg)
    call FError(functionName, stat, errmsg, status)

  end subroutine alloc_kwds_tHeader

  !.....................................................................
  subroutine append_kwd_tHeader(this, dataTypeFITS, nameStr, status, value)!, unitStr, commentStr)
    implicit none

    class(tHeader) :: this
    character(*),       intent(in)    :: dataTypeFITS, nameStr
    integer,            intent(inout) :: status
    class(*), optional, intent(in)    :: value
!!!    character(*), intent(in), optional :: unitStr, commentStr

    integer :: oldNumKwds
    type(tKeyword), dimension(:), allocatable :: tempKwds
    integer :: stat
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "append_kwd_tHeader"

    if(allocated(this%keywords)) then
      oldNumKwds = size(this%keywords)
      tempKwds = this%keywords
      call this%alloc(oldNumKwds+1, status)
      this%keywords(:oldNumKwds) = tempKwds
      call this%keywords(oldNumKwds+1)%set(dataTypeFITS, nameStr, value)!, unitStr, commentStr)
    else
      errmsg = ""
      allocate(this%keywords(1), stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
      call this%keywords(1)%set(dataTypeFITS, nameStr, value)!, unitStr, commentStr)
    end if

  end subroutine append_kwd_tHeader

  !.....................................................................
  function containsData_tHeader(this)
    implicit none

    class(tHeader) :: this
    logical :: containsData_tHeader

    integer :: i
    character(*), parameter :: functionName = "containsData_tHeader"

    containsData_tHeader = .false. ! default

    if(.not.allocated(this%keywords))&
  return

    do i=1,size(this%keywords)
      if(this%keywords(i)%containsData()) then
        containsData_tHeader = .true.
    exit
      end if
    end do

  end function containsData_tHeader

  !.....................................................................
  function getNumKeywords_tHeader(this) result(nKwdsUsed)
    implicit none

    class(tHeader) :: this
    integer :: nKwdsUsed

    if(allocated(this%keywords)) then
      nKwdsUsed = size(this%keywords)
    else
      nKwdsUsed = 0
!*** maybe size() returns anyway if not allocated?
    end if

  end function getNumKeywords_tHeader

  !.....................................................................
  function getKeywordNum_tHeader(this, kwdName, ierr) result(kwdNum)
    ! This searches the keywords for kwdName and returns the index of that keyword in the list this%keywords.

    implicit none

    class(tHeader) :: this
    character(*),      intent(in)  :: kwdName
    integer, optional, intent(out) :: ierr

    integer :: kwdNum,ii,nKwdsUsed
    character(*), parameter :: functionName = "getKeywordNum_tHeader"

    kwdNum = 0 ! Means no column of that name found.
    if(present(ierr)) ierr=0

!!!    if(.not. this%isInitialized) then
!!!      if(present(ierr)) then
!!!        ierr = 1
!!!  return
!!!      else
!!!call wrstop(functionName, "Table "//trim(this%tabInfo%tableName)&
!!!  //": you must first initialize (set the number of columns to be used, and their names etc).")
!!!      end if
!!!    end if

    if(.not. allocated(this%keywords)) then
      if(present(ierr)) then
        ierr = 2
  return
      else
call wrstop(functionName, "Table "//trim(this%tabInfo%tableName)//": your initialization routine should allocate this%keywords.")
!call wrstop(functionName, "Table xxx: your initialization routine should allocate this%keywords.")
      end if
    end if

    nKwdsUsed = this%getNumKeywords()
    if(nKwdsUsed<1) then
      if(present(ierr)) then
        ierr = 3
  return
      else
call wrstop(functionName, "Table "//trim(this%tabInfo%tableName)//": you don't have any keywords.")
!call wrstop(functionName, "Table xxx: you don't have any keywords.")
      end if
    end if

    do ii=1,nKwdsUsed
      if(trim(this%keywords(ii)%nameStr) == trim(kwdName)) then
        kwdNum = ii
    exit
      end if
    end do

    if(present(ierr)) then
      if(kwdNum<1) then
        ierr = 4
      else
        ierr = 0
      end if
    else if(kwdNum<1) then
call wrstop(functionName, "Keyword "//trim(kwdName)&
  //" not found in header of table "//trim(this%tabInfo%tableName))
!  //" not found in header of table xxx")
    end if

  end function getKeywordNum_tHeader

  !.....................................................................
  subroutine dealloc_tHeader(this, status)
    implicit none

    class(tHeader) :: this
    integer, intent(inout) :: status

    integer :: stat
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "dealloc_tHeader"

    if(allocated(this%keywords)) then
      errmsg = ""
      deallocate(this%keywords, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
    end if

  end subroutine dealloc_tHeader

  !.....................................................................
  subroutine alloc_tHeader(this, nKwdsDesired, status)
    implicit none

    class(tHeader) :: this
    integer, intent(in)    :: nKwdsDesired
    integer, intent(inout) :: status

    integer :: stat
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "alloc_tHeader"

    call this%dealloc(status)
    errmsg = ""
    allocate(this%keywords(nKwdsDesired), stat=stat, errmsg=errmsg)
    call FError(functionName, stat, errmsg, status)

  end subroutine alloc_tHeader

  !.....................................................................
  subroutine check_keywords_tHeader(this, status)
    implicit none

    class(tHeader) :: this
    integer, intent(inout) :: status

    integer :: ii
    character(*), parameter :: functionName = "check_keywords_tHeader"

    if(.not.allocated(this%keywords)) then
      call wrerr(functionName, c_err, "Keywords array is not allocated.")
      status = c_err
  return
    end if

    do ii=1,size(this%keywords)
      if(this%keywords(ii)%nameStr=='') then
        call wrerr(functionName, c_err, "Zero-length keyword name.")
       status = c_err
  return
      end if

      if(this%keywords(ii)%dataTypeFITS=='') then
        call wrerr(functionName, c_err, "Zero-length keyword data type.")
       status = c_err
  return
      end if

      if(.not.(this%keywords(ii)%dataTypeFITS=="E"&
           .or.this%keywords(ii)%dataTypeFITS=="D"&
           .or.this%keywords(ii)%dataTypeFITS=="J"&
           .or.this%keywords(ii)%dataTypeFITS=="A"))then
        call wrerr(functionName, c_err, "Unrecognized data type "&
          //trim(this%keywords(ii)%dataTypeFITS)//" for keyword "&
          //trim(this%keywords(ii)%nameStr))
        status = c_err
    return
      end if
    end do

  end subroutine check_keywords_tHeader

  !.....................................................................
  subroutine read_tHeader(this, fileInfo, status)
    implicit none

    class(tHeader) :: this
    type(tFitsInfo), pointer, intent(in) :: fileInfo
    integer, intent(inout) :: status

    integer :: ii,ierr
    character(fcommentlength) :: tmp_comment
    character(*), parameter :: functionName = "read_tHeader"

    if(.not.associated(fileInfo)) then
      call wrerr(functionName, s_err, "You must associate the fileInfo pointer!")
      status = s_errRep
  return
    end if

    if(.not.fileInfo%fileOpenedForRead) then
      call wrerr(functionName, s_err, "You must first open the FITS file for reading.")
      status = s_errRep
  return
    end if

!*** probably ought to test for the tBintable%info%haveAdvancedToHDU, but this involves more complication.

    if(.not. allocated(this%keywords)) then
  return
    end if

    do ii=1,size(this%keywords)
      if(.not.btest(this%keywords(ii)%useageFlags, readBit))&
    cycle

      ierr = 0
      select case(this%keywords(ii)%dataTypeFITS)
      case("E")
        call FTGKYE(fileInfo%unitNo, this%keywords(ii)%nameStr, this%keywords(ii)%r4Value, tmp_comment, ierr)
      case("D")
        call FTGKYD(fileInfo%unitNo, this%keywords(ii)%nameStr, this%keywords(ii)%r8Value, tmp_comment, ierr)
      case("J")
        call FTGKYJ(fileInfo%unitNo, this%keywords(ii)%nameStr, this%keywords(ii)%i4Value, tmp_comment, ierr)
      case("A")
        call FTGKYS(fileInfo%unitNo, this%keywords(ii)%nameStr, this%keywords(ii)%strValue, tmp_comment, ierr)
      case default
        call wrerr(functionName, s_err, "Unrecognized choice '"//trim(this%keywords(ii)%dataTypeFITS)//"'")
        status = s_errRep
  return
      end select

      if(ierr==0) then
        this%keywords(ii)%contains_data = .true.
      else if(.not.btest(this%keywords(ii)%useageFlags, isOptionalBit)) then
        call report_FITS_err(functionName, ierr, status, "- keyword "//trim(this%keywords(ii)%nameStr))
      end if
    end do

  end subroutine read_tHeader

  !.....................................................................
  subroutine write_tHeader(this, fileInfo, status, comment)
    implicit none

    class(tHeader) :: this
    type(tFitsInfo), pointer, intent(in) :: fileInfo
    integer, intent(inout) :: status
    character(*), optional, intent(in) :: comment

!**** put in some machinery to deal with absent comment!!!

    integer :: ii,ierr
    integer :: decimals=-14
    character(*), parameter :: functionName = "write_tHeader"

    if(.not.this%containsData())& ! also covers .not.allocated(this%keywords)
  return

    if(.not.associated(fileInfo)) then
      call wrerr(functionName, s_err, "You must associate the fileInfo pointer!")
      status = s_errRep
  return
    end if

    if(.not.fileInfo%fileOpenedForWrite) then
      call wrerr(functionName, s_err, "You must first open the FITS file for writing.")
      status = s_errRep
  return
    end if

!*** probably ought to test for the tBintable%info%haveAdvancedToHDU, but this involves more complication.

    do ii=1,size(this%keywords)
      if(.not.this%keywords(ii)%containsData())&
    cycle

      if(.not.btest(this%keywords(ii)%useageFlags, writeBit))&
    cycle

      ierr = 0
      select case(this%keywords(ii)%dataTypeFITS)
      case("E")
        call FTUKYE(fileInfo%unitNo, this%keywords(ii)%nameStr, this%keywords(ii)%r4Value, decimals, comment, ierr)
      case("D")
        call FTUKYD(fileInfo%unitNo, this%keywords(ii)%nameStr, this%keywords(ii)%r8Value, decimals, comment, ierr)
      case("J")
        call FTUKYJ(fileInfo%unitNo, this%keywords(ii)%nameStr, this%keywords(ii)%i4Value, comment, ierr)
      case("A")
        call FTUKYS(fileInfo%unitNo, this%keywords(ii)%nameStr, this%keywords(ii)%strValue, comment, ierr)
      case default
        call wrerr(functionName, s_err, "Unrecognized choice '"//trim(this%keywords(ii)%dataTypeFITS)//"'")
        status = s_errRep
  return
      end select

      call report_FITS_err(functionName, ierr, status, "- keyword "//trim(this%keywords(ii)%nameStr))
    end do

  end subroutine write_tHeader

  !.....................................................................
  subroutine setKeyword_tHeader(this, kwd_name, dataTypeFITS, nameStr, value)!, unitStr, commentStr)
    implicit none

    class(tHeader) :: this
    character(*),       intent(in) :: kwd_name
    character(*),       intent(in) :: dataTypeFITS, nameStr
    class(*), optional, intent(in) :: value

    integer :: kwd_num

    kwd_num = this%getKeywordNum(kwd_name)
    call this%keywords(kwd_num)%set(dataTypeFITS, nameStr, value)!, unitStr, commentStr)

  end subroutine setKeyword_tHeader

  !.....................................................................
  subroutine getKeywordValue_tHeader(this, kwd_name, value)
    implicit none

    class(tHeader) :: this
    character(*), intent(in) :: kwd_name
    class(*)                 :: value ! compiler doesn't seem to like 'intent(out)' with polymorphic arguments.

    integer :: kwd_num

    kwd_num = this%getKeywordNum(kwd_name)
    call this%keywords(kwd_num)%getValue(value)

  end subroutine getKeywordValue_tHeader

  !.....................................................................
  subroutine setKeywordValue_tHeader(this, kwd_name, value)
    implicit none

    class(tHeader) :: this
    character(*), intent(in)    :: kwd_name
    class(*),     intent(in)    :: value

    integer :: kwd_num

    kwd_num = this%getKeywordNum(kwd_name)
    call this%keywords(kwd_num)%setValue(value)

  end subroutine setKeywordValue_tHeader

  !.....................................................................
  logical function kwdContainsData_tHeader(this, kwdName)
    implicit none

    class(tHeader) :: this
    character(*), intent(in) :: kwdName

    integer :: kwdI

    kwdI = this%getKeywordNum(kwdName)
    kwdContainsData_tHeader = this%keywords(kwdI)%containsData()

  end function kwdContainsData_tHeader

  !.....................................................................
  function getKwdDataType_tHeader(this, kwdName) result(dataType)
    ! We want to keep tColumn%dataTypeFITS private to remove the temptation to set it directly.

    implicit none

    class(tHeader) :: this
    character(*), intent(in) :: kwdName
    character(1) :: dataType

    character(*), parameter :: functionName = "getKwdDataType_tHeader"
    integer :: kwdI

    kwdI = this%getKeywordNum(kwdName)
    dataType=this%keywords(kwdI)%getDataType()

  end function getKwdDataType_tHeader

  !.....................................................................
  subroutine set_kwd_useage_tHeader(this, kwd_name, optionalFlag, readFlag&
    , writeFlag, useInternallyFlag)

    implicit none
    class(tHeader)           :: this
    character(*), intent(in) :: kwd_name
    integer,      intent(in) :: optionalFlag,readFlag,writeFlag,useInternallyFlag

    character(*), parameter :: functionName = "set_kwd_useage_tHeader"
    integer :: kwd_num

    kwd_num = this%getKeywordNum(kwd_name)

    this%keywords(kwd_num)%useageFlags = 0 ! default

    if(optionalFlag>0)&
      this%keywords(kwd_num)%useageFlags = ibset(this%keywords(kwd_num)%useageFlags, isOptionalBit)
    if(readFlag>0)&
      this%keywords(kwd_num)%useageFlags = ibset(this%keywords(kwd_num)%useageFlags, readBit)
    if(writeFlag>0)&
      this%keywords(kwd_num)%useageFlags = ibset(this%keywords(kwd_num)%useageFlags, writeBit)
    if(useInternallyFlag>0)&
      this%keywords(kwd_num)%useageFlags = ibset(this%keywords(kwd_num)%useageFlags, internalUseBit)

  end subroutine set_kwd_useage_tHeader

  !=====================================================================
  ! Methods of tFitsInfo.

  !.....................................................................
  subroutine setFileName_tFitsInfo(this, fileName)
    implicit none

    class(tFitsInfo) :: this
    character(*), intent(in) :: fileName

    if(this%fileName /= fileName)&
      this%isRenamed = .true.

    this%fileName = fileName

  end subroutine setFileName_tFitsInfo

  !.....................................................................
  subroutine getFileName_tFitsInfo(this, fileName)
    implicit none

    class(tFitsInfo) :: this
    character(*), intent(out) :: fileName

    fileName = this%fileName

  end subroutine getFileName_tFitsInfo

  !=====================================================================
  ! Methods of tBinTable.

  !.....................................................................
!  subroutine init_tBinTable(this, status, fileInfo)
    ! This is a virtual method that you will need to overwrite with one
    ! you have written yourself for whatever subclass of tBinTable
    ! you are constructing. Your init() method should do the following:
    !
    !	if(present(fileInfo)) this%info%fileInfo => fileInfo
    !	- Set this%info%tableName.
    !	- Choose nColsUsed.
    !	- call this%alloc_cols(nColsUsed, status).
    !	- Set values of this%cols(i)%nameStr for all i.
    !	- Set values of this%cols(i)%dataTypeFITS for all i.
    !	- Set values of this%cols(i)%unitStr for all i (at least if you
    !	  intend to write these columns to a table). However, these you
    !	  can safely leave at default empty strings if you wish.
    !	- Set this%isInitialized = .true.
    !
    ! You might also want to allocate and fill the attribute
    !
    !	this%sort_sequence
    !
    ! Things you probably won't want your method to do (because these
    ! matters are not part of the definition of your class, they depend
    ! rather on what you want to do with it at any given moment):
    !
    !	- Set this%cols(i)%useageFlags
    !
    ! Note that whatever routine you write to set %useageFlags, be sure
    ! to set this%useageIsSet = .true. when you're finished. Otherwise
    ! you won't be able to read, write or allocate any columns.

!    implicit none

!    class(tBinTable) :: this
!    integer, intent(inout) :: status
!    type(tFitsInfo), intent(in), optional, target :: fileInfo

!    character(*), parameter :: functionName = "init_tBinTable"

!call wrstop(functionName, "You must replace this method by one tailored to your particular binTable.")

    ! Template for init routine:
    !
    !    if(present(fileInfo)) this%info%fileInfo => fileInfo
    !    this%info%tableName = "blah"
    !
    !    call this%initColumn(<data type char>, <column name>, <optional unit string>)
    !    etc for all desired columns.
    !
    !    call this%initColumns(status)
    !
    !    call this%initHeader(status)


    ! A template for setting the useage flags (such should usually not
    ! be placed in the init method, but in a non-method subroutine):
    !
    !                                  - optional read - no error if col not found.
    !                                 / - read.
    !                                / / - write.
    !                               / / / - no read, write, but used internally.
    ! this%set_col_useage(col_name,x,x,x,x)
    !
    ! this%useageIsSet = .true.

    ! just to prevent warnings from -Wall:
!    if(present(fileInfo)) this%info%fileInfo => fileInfo
!    if(status==0) this%dummy=0.0

!  end subroutine init_tBinTable

  !.....................................................................
  subroutine initHeader_tBinTable(this, status)
    implicit none

    class(tBinTable), target :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "initHeader_tBinTable"

    status = 0
    call wrstop(functionName, "You should overload this with an init() routine tailored to your header requirements.")

    ! Template for initHeader routine:
    !
    !   this%header%tabInfo => this%info
    !
    !   call this%header%initKeyword(<data type char>, <name>)
    !	  ...etc for other desired keywords.
    !   call this%header%initKeywords(status)
    !
    ! Finally you should do:
    !
    !	call this%header%check_keywords(status)

    if(status==0) this%dummy=0.0 ! just to prevent warnings from -Wall!

  end subroutine initHeader_tBinTable

  !.....................................................................
  subroutine initColumn_tBinTable(this, dataTypeFITS, nameStr, unitStr, num_chars, group_i)
    ! This should be called by all routines wishing to set up a tColumn object as part of a FITS binary table. Note that, after all desired columns have been set up by such calls, the initialization of the set of column specifications should be completed by calling tBinTable%initColumns(status).
    !
    implicit none

    class(tBinTable), target :: this
    character(*), intent(in) :: dataTypeFITS, nameStr
    character(*), intent(in), optional :: unitStr
    integer,      intent(in), optional :: num_chars,group_i

    character(*), parameter :: functionName = "initColumn_tBinTable"
    type(tColumn) :: col

    col%tabInfo => this%info

    col%nameStr = nameStr
    col%dataTypeFITS = dataTypeFITS
    if(present(unitStr))&
      col%unitStr = unitStr

    col%num_chars = 1 ! default
    if(present(num_chars)) then
      if(col%dataTypeFITS=="A") then
        col%num_chars = num_chars
      else
        call wrerr(functionName, warn, "Column "//trim(col%nameStr)&
          //" has data type "//col%dataTypeFITS&
          //", but must have type A for num_chars to make any sense.")
      end if
    end if

    if(present(group_i))&
      col%group_i = group_i

    col%num_vec_elements = 0 ! default. Vector columns can only be allocated as such via allocVec_tColumn().

    if(.not. associated(this%col_specs)) allocate(this%col_specs)
    call this%col_specs%addObj(col)

  end subroutine initColumn_tBinTable

  !.....................................................................
  subroutine deleteLinkedList_tBinTable(this, status)
    implicit none

    class(tBinTable) :: this
    integer, intent(inout) ::  status

    character(*), parameter :: functionName = "deleteLinkedList_tBinTable"
    integer :: stat
    class(*), pointer :: tmp

    if(status > 100 .or. status == s_errRep) return

    if(associated(this%col_specs)) then
      call this%col_specs%reset()
      do while (associated(this%col_specs%currLink))
        tmp => this%col_specs%currLink%getObj()

        select type (tmp)
          type is (tColumn)
            call tmp%destroy(status)
            if(status == s_errRep)&
  return
        end select

        call this%col_specs%next()
      end do

      if(associated(tmp)) deallocate(tmp)
    end if

    call this%col_specs%reset()
    stat =  this%col_specs%delAllLinks()
    if(stat /= 0) then
      call wrerr(functionName, s_err, "deallocation of LinkedList failed")
      status = s_errRep
    end if

    if(associated(this%col_specs)) deallocate(this%col_specs)

  end subroutine deleteLinkedList_tBinTable

  !.....................................................................
  subroutine initColumns_tBinTable(this, status)
    ! Having loaded as many column specifications into the linked list this%col_specs as we want, we convert this LL to a vector of tColumn objects.
    !
    implicit none

    class(tBinTable) :: this ! target??
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "initColumns_tBinTable"
    integer :: nColsUsed,ii,jj
    class(*), pointer :: tmp
    character(buffsize) :: str1,str2

    nColsUsed = this%col_specs%getNumLinks()
    call this%alloc_cols(nColsUsed, status)
    call this%col_specs%reset() ! resets the leaf index to the start of the list.

    do ii = 1, nColsUsed
      tmp => this%col_specs%currObj()
      select type(tmp)
        type is (tColumn)
          this%cols(ii) = tmp
      end select

      call this%col_specs%next()
    end do

    call this%deleteLinkedList(status)

    do ii = 1, nColsUsed-1
      do jj=ii+1,nColsUsed
        if(trim(this%cols(ii)%nameStr)==trim(this%cols(jj)%nameStr)) then
          str1 = int2str(ii)
          str2 = int2str(jj)
call wrstop(functionName, "Columns "//trim(str1)//" and "//trim(str2)&
 //" have the same name "//trim(this%cols(ii)%nameStr))
        end if
      end do
    end do

    this%isInitialized = .true.

  end subroutine initColumns_tBinTable

  !.....................................................................
  function check_tBinTable(this, functionName) result(status)
    ! This does some basic sanity checks.

    implicit none

    class(tBinTable) :: this
    character(*), intent(in) :: functionName
    integer :: status

    status=0

    if(.not.this%isInitialized) then
      call wrerr(functionName, s_err, "Table "//trim(this%info%tableName)&
        //": you must first designate the column properties.")
      status = s_errRep
  return
    end if

    if(trim(this%info%tableName)=="") then
      call wrerr(functionName, warn, "You should fill the tableName attribute.")
    end if

    if(.not. allocated(this%cols)) then
      call wrerr(functionName, s_err, "Table "//trim(this%info%tableName)&
        //": your initialization routine should allocate this%cols.")
      status = s_errRep
  return
    end if

    if(size(this%cols)<1) then
      call wrerr(functionName, s_err, "Table "//trim(this%info%tableName)&
        //": you don't have any columns.")
      status = s_errRep
  return
    end if

    if(.not.this%useageIsSet) then
      call wrerr(functionName, s_err, "Table "//trim(this%info%tableName)&
        //": you must first set this%cols%useageFlags.")
      status = s_errRep
  return
    end if

    if(.not.associated(this%info%fileInfo)) then
      call wrerr(functionName, s_err, "Table "//trim(this%info%tableName)&
        //": you must associate the fileInfo pointer!")
      status = s_errRep
  return
    end if

  end function check_tBinTable

  !.....................................................................
  subroutine getColNumsForReading_tBinTable(this, nums_vec_elements, status)
    ! For each column flagged to be read, this obtains the column number
    ! in the FITS file and stores it in this%cols(ii)%num. The data type
    ! of the column is also checked by comparing
    ! this%cols(ii)%dataTypeFITS with the value of the matching TFORMx
    ! keyword. Columns for which this succeeds (i.e., which exist in the
    ! file) are flagged with this%cols(ii)%colNumFoundInFITS = .true.,
    ! .false. otherwise. Non-found columns are otherwise ignored so long
    ! as the 'optional' flag in this%cols(ii)%useageFlags is set.

    implicit none

    class(tBinTable) :: this
    integer, dimension(:), allocatable :: nums_vec_elements
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "getColNumsForReading_tBinTable"
    integer :: ii,ierr,nColsUsed
    character(buffsize) :: dummyName
    character(buffsize) :: dataTypeFITS
    character(8) :: unitStr
    integer :: repeat_val,tnull
    real(dp) :: tscal,tzero
    character(buffsize) :: tdisp

    nColsUsed = this%getNumCols()
    call alloc(nums_vec_elements, nColsUsed, status)
    nums_vec_elements(:) = 0 ! default for scalar columns.

    do ii=1, nColsUsed
      this%cols(ii)%colNumFoundInFITS = .false. ! default

      if(btest(this%cols(ii)%useageFlags, readBit)) then
        ierr = 0
        ! Try to read the column number.
        call FTGCNO(this%info%fileInfo%unitNo,.true.,this%cols(ii)%nameStr,this%cols(ii)%num,ierr)

        if (ierr.ne.0) then
          if(.not.btest(this%cols(ii)%useageFlags, isOptionalBit)) then
            call report_FITS_err(functionName, ierr, status, "- column "//trim(this%cols(ii)%nameStr))
          end if
        else
          ! Fetch the data type:
          !
          call FTGBCL(this%info%fileInfo%unitNo, this%cols(ii)%num, dummyName, unitStr&
            , dataTypeFITS, repeat_val, tscal, tzero, tnull, tdisp, ierr)

          call report_FITS_err(functionName, ierr, status, "- column "//trim(this%cols(ii)%nameStr))
          if (ierr==0) then
            if(dataTypeFITS(1:1) == this%cols(ii)%dataTypeFITS) then
              this%cols(ii)%colNumFoundInFITS = .true.
            else
              call wrerr(functionName, s_err, "Column "//trim(this%cols(ii)%nameStr)&
                //": specified data type is "//trim(this%cols(ii)%dataTypeFITS)&
                //" but file has "//trim(dataTypeFITS))
              status = s_errRep
  return
            end if

            if(repeat_val>1) then
              if(this%cols(ii)%dataTypeFITS=="A") then ! interpret as the number of chars.
                this%cols(ii)%num_chars = repeat_val
              else ! interpret as the size of a vector column.
                nums_vec_elements(ii) = repeat_val
              end if
            end if
          end if
        end if
      end if
    end do

    this%haveReadColNums = .true.

  end subroutine getColNumsForReading_tBinTable

  !.....................................................................
  function existsInFile_tBinTable(this) result(exists)
    implicit none

    class(tBinTable) :: this
    logical :: exists

    character(*), parameter :: functionName = "existsInFile_tBinTable"
    integer :: ierr,dummyStatus

    exists = .false. ! default

    if(trim(this%info%tableName)=="") then
      call wrerr(functionName, c_err, "You should fill the tableName attribute.")
  return
    end if

    if(.not.this%info%fileInfo%fileOpenedForRead) then
      call wrerr(functionName, c_err, "You should first open the FITS file for reading.")
  return
    end if

    ierr = 0
    call FTMNHD(this%info%fileInfo%unitNo, BINARY_TBL, this%info%tableName, 0, ierr)
    if(ierr==0) then
      exists = .true.
!    else if(ierr/=301) then
    else if(ierr/=BAD_HDU_NUM) then
      call report_FITS_err(functionName, ierr, dummyStatus, "- extension "//trim(this%info%tableName))
    end if

  end function existsInFile_tBinTable

  !.....................................................................
  subroutine read_tBinTable(this, status)
    implicit none

    class(tBinTable) :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "read_tBinTable"
    integer :: rowSize,felem,ii,frow,nrows,lrow,ierr,nColsUsed
    logical :: tnullKwdFound
    integer, dimension(:), allocatable :: nums_vec_elements

    status = this%check(functionName)
    if(status/=0)&
  return

    if(.not.this%info%fileInfo%fileOpenedForRead) then
      call wrerr(functionName, s_err, "You must first open the FITS file for reading.")
      status = s_errRep
  return
    end if

    if(.not. this%info%haveAdvancedToHDU) then
      ! do so.
      ierr = 0
      call FTMNHD(this%info%fileInfo%unitNo, BINARY_TBL, this%info%tableName, 0, ierr)

      if(ierr==BAD_HDU_NUM) then
        status = c_err
        call wrerr(functionName, c_err, "Extension "//trim(this%info%tableName)//" not found.")
  return
      else if(ierr /= 0) then
        status = c_err
        call report_FITS_err(functionName, ierr, status, "- extension "//trim(this%info%tableName))
  return
      end if

      this%info%haveAdvancedToHDU = .true.
    end if

    this%foundInFile = .true. !********** is this used anywhere?

    call this%header%read(this%info%fileInfo, status)

    if(.not.this%haveReadColNums) then
      ! This looks for column numbers, also checks the data types.
      call this%getColNumsForReading(nums_vec_elements, status) ! sets this%haveReadColNums = .true.
    end if

    ! Get the number of rows in the table.
    !
    ierr = 0
    call FTGNRW(this%info%fileInfo%unitNo, this%info%numRows, ierr)
    call report_FITS_err(functionName, ierr, status, "- extension "//trim(this%info%tableName))

!*** if bad status or no elements, warn and return?

    nColsUsed = this%getNumCols()

    do ii = 1, nColsUsed
      if(btest(this%cols(ii)%useageFlags,readBit) .and. this%cols(ii)%colNumFoundInFITS) then
        if(  this%cols(ii)%dataTypeFITS=='B' .or. this%cols(ii)%dataTypeFITS=='I'&
        .or. this%cols(ii)%dataTypeFITS=='J' .or. this%cols(ii)%dataTypeFITS=='K')&
          call this%cols(ii)%readTNULL(this%info%fileInfo%unitNo, .true., tnullKwdFound, status) ! sets this%cols(ii)%??null from TNULLn kwd if found.

        call this%cols(ii)%readScaleKwds(this%info%fileInfo%unitNo, status)
        call this%cols(ii)%alloc(this%info%numRows, status, fillType=TCOL_FILL_NULLS)
        if(nums_vec_elements(ii)>0)&
          call this%cols(ii)%allocVec(nums_vec_elements(ii), status) ! note that getColNumsForReading() sets num_vec_elements from the FITS table.
      end if
    end do

    if(this%info%numRows<1)&
  return

    ! Get best row size:
    !
    ierr = 0
    call FTGRSZ(this%info%fileInfo%unitNo, rowSize, ierr)
    call report_FITS_err(functionName, ierr, status, "- extension "//trim(this%info%tableName))

    felem = 1

    do frow = 1, this%info%numRows, rowSize
      nrows = min(this%info%numRows - frow + 1, rowSize)
      lrow = frow + nrows - 1
      do ii = 1, nColsUsed
        if(btest(this%cols(ii)%useageFlags,readBit) .and. this%cols(ii)%colNumFoundInFITS)&
          call this%cols(ii)%readRows(this%info%fileInfo%unitNo, frow, lrow, felem, nrows, status) ! sets containsData
      end do
    end do

    call dealloc(nums_vec_elements, status)

  end subroutine read_tBinTable

  !.....................................................................
  subroutine createTable_tBinTable(this, status)
    implicit none

    class(tBinTable) :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "createTable_tBinTable"
    integer :: ii,jj,ierr,nColsUsed
    character(fcolnamelength), dimension(:), allocatable :: ttype,tform,tunit
    integer :: numColsToWrite

    nColsUsed = this%getNumCols()

    ! Find out how many columns to write.
    !
    numColsToWrite = 0
    do ii=1,nColsUsed
      if(btest(this%cols(ii)%useageFlags, writeBit)) numColsToWrite = numColsToWrite + 1
    end do

    call alloc(ttype, numColsToWrite, status)
    call alloc(tform, numColsToWrite, status)
    call alloc(tunit, numColsToWrite, status)

    jj = 0
    do ii=1,nColsUsed
      if(btest(this%cols(ii)%useageFlags, writeBit)) then
        jj = jj + 1
        ttype(jj) = trim(this%cols(ii)%nameStr)

        if(this%cols(ii)%dataTypeFITS=='A') then
          write(tform(jj), '(a)') trim(int2str(this%cols(ii)%num_chars))//this%cols(ii)%dataTypeFITS
          ! From the description in the FITSIO docs it would seem that the number of chars should come after the 'A', but in practice this seems only to result in the first letter of each string being stored in the table (in a scalar column).
        else if(this%cols(ii)%num_vec_elements>0) then
          write(tform(jj), '(a)') trim(int2str(this%cols(ii)%num_vec_elements))//this%cols(ii)%dataTypeFITS
        else
          write(tform(jj), '(a)') this%cols(ii)%dataTypeFITS
        end if

        tunit(jj) = this%cols(ii)%unitStr
      end if
    end do

    ! Create the FITS table:
    !
    ierr = 0
    call FTIBIN(this%info%fileInfo%unitNo, 0, numColsToWrite, ttype, tform, tunit, this%info%tableName, 0, ierr)
    call report_FITS_err(functionName, ierr, status, "- extension "//trim(this%info%tableName))

    call dealloc(ttype, status)
    call dealloc(tform, status)
    call dealloc(tunit, status)

    this%haveReadColNums = .false. ! because whatever ones we may have read are no longer valid. Safest is to reset them from the FITS table itself, now that we have made it.

  end subroutine createTable_tBinTable

  !.....................................................................
  subroutine getColNumsForWriting_tBinTable(this, status)
    implicit none

    class(tBinTable) :: this
    integer, intent(inout) :: status

    integer :: ii,ierr,nColsUsed,nColsInTable
    character(*), parameter :: functionName = "getColNumsForWriting_tBinTable"

    nColsUsed = this%getNumCols()

    nColsInTable = 0
    do ii=1, nColsUsed
      if(btest(this%cols(ii)%useageFlags, writeBit)) then
        ierr = 0
        ! Try to read column number.
        call FTGCNO(this%info%fileInfo%unitNo,.true.,this%cols(ii)%nameStr,this%cols(ii)%num,ierr)
!*** are we going to assume it has the right data type and dimensions?

        if (ierr==0) then
          nColsInTable = nColsInTable + 1
        else if(ierr==COL_NOT_FOUND) then ! column doesn't exist, create it.
          call this%cols(ii)%createFITSCol(this%info%fileInfo%unitNo, nColsInTable, status) ! sets this%cols(ii)%num
        else
          call report_FITS_err(functionName, ierr, status, "- column "//trim(this%cols(ii)%nameStr))
call wrstop(functionName, "FTGCNO error.")
        end if
      end if
    end do

    this%haveReadColNums = .true.

  end subroutine getColNumsForWriting_tBinTable

  !.....................................................................
  subroutine write_tBinTable(this, status)
    implicit none

    class(tBinTable) :: this
    integer,           intent(inout) :: status

    character(*), parameter :: functionName = "write_tBinTable"
    integer :: rowSize,ii,frow,nrows,lrow,ierr
    integer :: nColsUsed
    character(buffsize) :: str0,str1

    call wrchatty(functionName, "Writing extension "//trim(this%info%tableName))

    status = this%check(functionName)
    if(status/=0)&
  return

    nColsUsed = this%getNumCols()

    do ii = 1, nColsUsed
      if(btest(this%cols(ii)%useageFlags, writeBit)) then
        if(.not.this%cols(ii)%isAllocated()) then
          call wrerr(functionName, s_err, "Column "//trim(this%cols(ii)%nameStr)&
            //" is not allocated.")
          status = s_errRep
  return
        end if
      end if
    end do

    if(.not. this%info%fileInfo%fileOpenedForWrite) then
      call wrerr(functionName, s_err, "You must first open the FITS file for writing.")
!**** better to try to open this%fileName.
      status = s_errRep
  return
    end if

    if(this%info%fileInfo%isRenamed .or..not. this%info%haveAdvancedToHDU) then
      ! do so.
      ierr = 0
      call FTMNHD(this%info%fileInfo%unitNo, BINARY_TBL, this%info%tableName, 0, ierr)

      if(ierr/=0) then
!*** should be more specific about the error.
        call this%createTable(status)
        ierr = 0
      else
        ! Get the existing number of rows in the table.
        ierr = 0
        call FTGNRW(this%info%fileInfo%unitNo, nrows, ierr)
        call report_FITS_err(functionName, ierr, status, "- extension "//trim(this%info%tableName))

        if(nrows /= this%info%numRows) then
          ! Delete all the table rows.
          ierr = 0
          call FTDROW(this%info%fileInfo%unitNo, 1, nrows, ierr)
          call report_FITS_err(functionName, ierr, status, "- extension "//trim(this%info%tableName))
        end if
      end if

      this%info%haveAdvancedToHDU = .true.
    end if

    call this%header%write(this%info%fileInfo, status)

    if(this%info%fileInfo%isRenamed .or..not.this%haveReadColNums)&
      call this%getColNumsForWriting(status) ! sets this%haveReadColNums = .true.

    do ii = 1, nColsUsed
      if(btest(this%cols(ii)%useageFlags, writeBit)) then
        call this%cols(ii)%setupFITSnulls(this%info%fileInfo%unitNo, status)
        call this%cols(ii)%writeScaleKwds(this%info%fileInfo%unitNo, status)
      end if
    end do

    ierr = 0
    call FTFLUS(this%info%fileInfo%unitNo, ierr)
    call report_FITS_err(functionName, ierr, status, "- extension "//trim(this%info%tableName))

    ! Get the optimum number of rows to write.
    !
    ierr = 0
    call FTGRSZ(this%info%fileInfo%unitNo, rowsize, ierr)
    call report_FITS_err(functionName, ierr, status, "- extension "//trim(this%info%tableName))
    str0 = int2str(this%info%numRows)
    str1 = int2str(rowsize)
    call wrchatty(functionName, "Writing "//trim(str0)//" rows in chunks of size "//trim(str1))

    do ii = 1, nColsUsed
      if(btest(this%cols(ii)%useageFlags, writeBit)) then
        call this%cols(ii)%setNullsFromMask(status)
      end if
    end do

    do frow = 1, this%info%numRows, rowSize
      nrows = min(this%info%numRows - frow + 1, rowSize)
      lrow = frow + nrows - 1
      do ii = 1, nColsUsed
        if(btest(this%cols(ii)%useageFlags, writeBit)) then
          call this%cols(ii)%writeRows(this%info%fileInfo%unitNo, frow, lrow, nrows, status)
        end if
      end do
    end do

  end subroutine write_tBinTable

  !.....................................................................
  subroutine dealloc_cols_tBinTable(this, status)
    implicit none

    class(tBinTable) :: this
    integer, intent(inout) :: status

    integer :: stat
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "dealloc_cols_tBinTable"

    if(allocated(this%cols)) then
      call this%dealloc(status) ! deallocates arrays in each tColumn object individually.
      errmsg = ""
      deallocate(this%cols, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
    end if

    this%isInitialized = .false.
    this%useageIsSet = .false.

  end subroutine dealloc_cols_tBinTable

  !.....................................................................
  subroutine alloc_cols_tBinTable(this, nColsUsed, status)
    implicit none

    class(tBinTable) :: this
    integer, intent(in)    :: nColsUsed
    integer, intent(inout) :: status

    integer :: stat
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "alloc_cols_tBinTable"

    if(nColsUsed < 1) then
      status = c_err
      call wrerr(functionName, status, "Can't allocate column list with <1 elements.")
  return
    end if

    call this%dealloc_cols(status)
    errmsg = ""
    allocate(this%cols(nColsUsed), stat=stat, errmsg=errmsg)
    call FError(functionName, stat, errmsg, status)

  end subroutine alloc_cols_tBinTable

  !.....................................................................
  subroutine alloc_tBinTable(this, status, numRows, fillType)
    ! Allocates and initializes all columns not already allocated.
    !
    ! ***NOTE!*** This allocates all columns as non-vector columns by default. This routine should be called first for the whole table; any vector columns should be allocated subsequently via the call allocVec_tColumn().

    implicit none

    class(tBinTable) :: this
    integer,           intent(inout) :: status
    integer, optional, intent(in)    :: numRows
    integer, optional, intent(in)    :: fillType

    integer :: ii,nColsUsed
    character(*), parameter :: functionName = "alloc_tBinTable"

    nColsUsed = this%getNumCols()

    if(present(numRows)) then
      this%info%numRows = numRows
    else if(this%info%numRows<0) then
      call wrerr(functionName, warn, "You need to set the number of rows to some number >=0.")
  return
    end if

    if(nColsUsed>0 .and..not.any(this%cols%useageFlags>0)) then
      call wrerr(functionName, warn, "None of the columns has a useage flag set. Such columns will not be allocated.")
    end if

    do ii = 1, nColsUsed
      if(this%cols(ii)%useageFlags>0 .and. .not. this%cols(ii)%isAllocated()) then
        call this%cols(ii)%alloc(this%info%numRows, status, fillType)
      end if
    end do

  end subroutine alloc_tBinTable

  !.....................................................................
  subroutine dealloc_tBinTable(this, status)
    implicit none

    class(tBinTable) :: this
    integer, intent(inout) :: status

    integer :: ii,nColsUsed
    character(*), parameter :: functionName = "dealloc_tBinTable"

    nColsUsed = this%getNumCols()
    if(nColsUsed>0) then
      do ii = 1, nColsUsed
        call this%cols(ii)%dealloc(status)
      end do
    end if

  end subroutine dealloc_tBinTable

  !.....................................................................
  subroutine realloc_tBinTable(this, status, numRows, fillType)
    implicit none

    class(tBinTable) :: this
    integer,           intent(inout) :: status
    integer, optional, intent(in)    :: numRows
    integer, optional, intent(in)    :: fillType

    character(*), parameter :: functionName = "realloc_tBinTable"

    call this%dealloc(status)
    call this%alloc(status, numRows, fillType)

  end subroutine realloc_tBinTable

  !.....................................................................
  subroutine allocVec_tBinTable(this, col_name, num_vec_elements, status, fillType)
    implicit none

    class(tBinTable)                 :: this
    character(*),      intent(in)    :: col_name
    integer,           intent(in)    :: num_vec_elements
    integer,           intent(inout) :: status
    integer, optional, intent(in)    :: fillType

    character(*), parameter :: functionName = "allocVec_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)
    call this%cols(col_num)%allocVec(num_vec_elements, status, fillType)

  end subroutine allocVec_tBinTable

  !.....................................................................
  subroutine destroy_tBinTable(this, status)
    ! This is effectively the destructor method for the class.

    implicit none

    class(tBinTable) :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "destroy_tBinTable"
    integer :: ii,nColsUsed,stat
    character(buffsize) :: errmsg

    if(allocated(this%cols)) then
      nColsUsed = this%getNumCols()
      if(nColsUsed>0) then
        do ii = 1, nColsUsed
          call this%cols(ii)%destroy(status)
        end do
      end if

      call dealloc(this%sort_sequence, status)

      errmsg = ""
      deallocate(this%cols, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
    end if

    this%isInitialized = .false.
    this%useageIsSet = .false.
    this%info%haveAdvancedToHDU = .false.
    this%haveReadColNums = .false.

    call this%header%dealloc(status)

    nullify(this%info%fileInfo)

  end subroutine destroy_tBinTable

  !.....................................................................
  subroutine pack_tBinTable(this, mask, status)
    implicit none

    class(tBinTable) :: this
    logical, dimension(this%info%numRows), intent(in)    :: mask
    integer,                            intent(inout) :: status

    integer :: ii,nColsUsed
    character(*), parameter :: functionName = "pack_tBinTable"

    nColsUsed = this%getNumCols()
    if(nColsUsed>0) then
      do ii = 1, nColsUsed
        if(this%cols(ii)%isAllocated())&
          call this%cols(ii)%pack(mask, status)
      end do
    end if

    this%info%numRows = count(mask)

  end subroutine pack_tBinTable

  !.....................................................................
  subroutine reorder_tBinTable(this, order, status)
    implicit none

    class(tBinTable)                     :: this
    integer, dimension(:), intent(in)    :: order
    integer,               intent(inout) :: status

    integer :: ii,nColsUsed
    character(*), parameter :: functionName = "reorder_tBinTable"

!*** rough checks - that size(order)<=this%info%numRows, that minval(order)>0, that maxval(order)<=this%info%numRows.

    nColsUsed = this%getNumCols()
    if(nColsUsed>0) then
      do ii = 1, nColsUsed
        if(this%cols(ii)%isAllocated())&
          call this%cols(ii)%reorder(order, status)
      end do
    end if

    this%info%numRows = size(order)

  end subroutine reorder_tBinTable

  !.....................................................................
  subroutine unpack_tBinTable(this, tableB, order, status, fillType)
    implicit none

    class(tBinTable) :: this
    type(tBinTable),   intent(in)    :: tableB
    integer, dimension(:), intent(in)    :: order
    integer,               intent(inout) :: status
    integer, optional,     intent(in)    :: fillType

    integer :: ii,nColsUsed
    character(*), parameter :: functionName = "unpack_tBinTable"

    nColsUsed = this%getNumCols()
    if(nColsUsed>0) then
      do ii = 1, nColsUsed
        if(this%cols(ii)%isAllocated() .and. tableB%cols(ii)%isAllocated()) then
          call this%cols(ii)%unpack(tableB%cols(ii), order, status, fillType)
        end if
      end do
    end if

  end subroutine unpack_tBinTable

  !.....................................................................
  subroutine overwrite_section_tBinTable(this, tableB, startI, status)
    implicit none

    class(tBinTable) :: this
    class(tBinTable), intent(in)    :: tableB
    integer,             intent(in)    :: startI
    integer,             intent(inout) :: status

    integer :: finisI,ii,nColsUsed
    character(buffsize) :: errstr1,errstr2
    character(*), parameter :: functionName = "overwrite_section_tBinTable"

    finisI = startI + tableB%info%numRows - 1

    if(startI<1) then
      errStr1 = int2str(startI)
      call wrerr(functionName, s_err, "Start index "//trim(errStr1)//" is less than 1.")
      status = s_errRep
  return
    end if

    if(finisI>this%info%numRows) then
      errStr1 = int2str(finisI)
      errStr2 = int2str(this%info%numRows)
      call wrerr(functionName, s_err, "End index "//trim(errStr1)&
        //" is greater than number of elements "//trim(errStr2)//".")
      status = s_errRep
  return
    end if
 
    nColsUsed = this%getNumCols()
    if(nColsUsed>0) then
      do ii = 1, nColsUsed
        if(this%cols(ii)%isAllocated() .and. tableB%cols(ii)%isAllocated())&
          call this%cols(ii)%overwrite_section(tableB%cols(ii), startI, finisI, status)
      end do
    end if

  end subroutine overwrite_section_tBinTable

  !.....................................................................
  subroutine sort_tBinTable(this, array, status, ascend, optorder)
!*** 20210719 IMS - should we convert this to use timsort? I just added a type t_sort_scalar to erolib_timsort_mod,f90 which would allow it. However use of this type may not be the fastest, since the way it is presently written, I have to test for data type at every single step.
    implicit none

    class(tBinTable) :: this
    class(*), dimension(:), intent(in)    :: array
    integer,                intent(inout) :: status
    logical, optional,      intent(in)    :: ascend
    integer, allocatable, dimension(:), optional, intent(out) :: optorder

    integer, dimension(:), allocatable :: order
    real(r8),allocatable, dimension(:) :: tmparray_r8
    real(r4),allocatable, dimension(:) :: tmparray_r4
    integer(i4),allocatable, dimension(:) :: tmparray_i4
    integer(i2),allocatable, dimension(:) :: tmparray_i2
    integer(i1),allocatable, dimension(:) :: tmparray_i1
    character(*), parameter :: functionName = "sort_tBinTable"

    if(status > 100 .or. status == s_errRep) return

    if(this%info%numRows /= size(array)) then
      call wrerr(functionName, s_err&
        , "length of sorting array does not conform to length of type")
      status = s_errRep
  return
    end if

    select type(array)
      type is ( real(r4) )
        tmparray_r4 = array
        call sort_real(tmparray_r4, status, order)
      type is ( real(r8) )
        tmparray_r8 = array
        call sort_double(tmparray_r8, status, order)
      type is ( integer(i4) )
        tmparray_i4 = array
        call sort_i4(tmparray_i4, status, order)
      type is ( integer(i2) )
        tmparray_i2 = array
        call sort_i2(tmparray_i2, status, order)
      type is ( integer(i1) )
       tmparray_i1 = array
        call sort_i1(tmparray_i1, status, order)
      class default
call wrstop(functionName, "Sorting tBinTable with the current array type is not supported for tBinTable.")
    end select

    if(present(ascend)) then
      if(.not.ascend) call reverse_array(order)
    end if

    call this%reorder(order, status)

    if(present(optorder)) optorder = order

    call dealloc(order, status)
    call dealloc(tmparray_i1, status)
    call dealloc(tmparray_i2, status)
    call dealloc(tmparray_i4, status)
    call dealloc(tmparray_r4, status)
    call dealloc(tmparray_r8, status)

  end subroutine sort_tBinTable

  !.....................................................................
  subroutine sort_in_sequence_tBinTable(this, list_of_col_names, status)!!!, ascend)
    implicit none

    class(tBinTable)                          :: this
    character(*), dimension(:), intent(in)    :: list_of_col_names
    integer,                    intent(inout) :: status
!!!    logical, optional,               intent(in)    :: ascend

    character(*), parameter :: functionName = "sort_in_sequence_tBinTable"
    type(t_sort_bintable) :: sortable

!    status = this%check(functionName)
!    if(status/=0)&
!  return

    if(this%info%numRows<2)&
  return

    call sortable%init(this, list_of_col_names)
    call timsort(sortable)
    call this%reorder(sortable%indices, status)

  end subroutine sort_in_sequence_tBinTable

  !.....................................................................
  function getNumCols_tBinTable(this) result(nColsUsed)
    implicit none

    class(tBinTable) :: this
    integer :: nColsUsed

    if(allocated(this%cols)) then
      nColsUsed = size(this%cols)
    else
      nColsUsed = 0
!*** maybe size() returns anyway if not allocated?
    end if

  end function getNumCols_tBinTable

  !.....................................................................
  function containsData_tBinTable(this) result(containsData)
    implicit none

    class(tBinTable) :: this
    logical :: containsData

    integer :: ii
    character(*), parameter :: functionName = "containsData_tBinTable"

    containsData = .false. ! default

    if(.not.allocated(this%cols))&
  return

    do ii=1,size(this%cols)
      if(this%cols(ii)%isAllocated()) then
        containsData = .true.
    exit
      end if
    end do

  end function containsData_tBinTable

  !.....................................................................
  function getNumRows_tBinTable(this) result(numRows)
    ! This searches the columns for colName and returns the index of that column in the list this%cols.

    implicit none

    class(tBinTable) :: this
    integer :: numRows

    numRows = this%info%numRows

  end function getNumRows_tBinTable

  !.....................................................................
  subroutine set_all_useage_tBinTable(this, optionalFlag, readFlag&
    , writeFlag, useInternallyFlag, group_i)

    implicit none

    class(tBinTable) :: this
    integer, intent(in) :: optionalFlag,readFlag,writeFlag,useInternallyFlag
    integer, intent(in), optional :: group_i

    character(*), parameter :: functionName = "set_all_useage_tBinTable"
    integer :: numColsUsed,ii

    numColsUsed = this%getNumCols()
    do ii=1,numColsUsed
      if(present(group_i)) then
        if(this%cols(ii)%group_i /= group_i)&
    cycle
      end if

      call this%cols(ii)%setUseage(optionalFlag, readFlag, writeFlag, useInternallyFlag)
    end do

    this%useageIsSet = .true.

  end subroutine set_all_useage_tBinTable

  !.....................................................................
  subroutine ibset_all_useage_tBinTable(this, bitNum, group_i)
    implicit none

    class(tBinTable) :: this
    integer, intent(in) :: bitNum
    integer, intent(in), optional :: group_i

    character(*), parameter :: functionName = "ibset_all_useage_tBinTable"
    integer :: numColsUsed,ii

    numColsUsed = this%getNumCols()
    do ii=1,numColsUsed
      if(present(group_i)) then
        if(this%cols(ii)%group_i /= group_i)&
    cycle
      end if

      call this%cols(ii)%ibset_useage(bitNum)
    end do

  end subroutine ibset_all_useage_tBinTable

  !.....................................................................
  subroutine set_col_useage_tBinTable(this, col_name, optionalFlag, readFlag&
    , writeFlag, useInternallyFlag)

    implicit none

    class(tBinTable) :: this
    character(*), intent(in) :: col_name
    integer,      intent(in) :: optionalFlag,readFlag,writeFlag,useInternallyFlag

    character(*), parameter :: functionName = "set_col_useage_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)
    call this%cols(col_num)%setUseage(optionalFlag,readFlag,writeFlag,useInternallyFlag)

  end subroutine set_col_useage_tBinTable

  !.....................................................................
  subroutine ibset_col_useage_tBinTable(this, col_name, bitNum)
    implicit none

    class(tBinTable) :: this
    character(*), intent(in) :: col_name
    integer,      intent(in) :: bitNum

    character(*), parameter :: functionName = "ibset_col_useage_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)
    call this%cols(col_num)%ibset_useage(bitNum)

  end subroutine ibset_col_useage_tBinTable

  !.....................................................................
  subroutine ibclr_col_useage_tBinTable(this, col_name, bitNum)
    implicit none

    class(tBinTable) :: this
    character(*), intent(in) :: col_name
    integer,      intent(in) :: bitNum

    character(*), parameter :: functionName = "ibclr_col_useage_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)
    call this%cols(col_num)%ibclr_useage(bitNum)

  end subroutine ibclr_col_useage_tBinTable

  !.....................................................................
  logical function btest_col_useage_tBinTable(this, col_name, bitNum)
    implicit none

    class(tBinTable) :: this
    character(*), intent(in) :: col_name
    integer,      intent(in) :: bitNum

    character(*), parameter :: functionName = "btest_col_useage_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)
    btest_col_useage_tBinTable = this%cols(col_num)%btest_useage(bitNum)

  end function btest_col_useage_tBinTable

  !.....................................................................
  function getColNum_tBinTable(this, colName, ierr) result(colNum)
    ! This searches the columns for colName and returns the index of that column in the list this%cols.

    implicit none

    class(tBinTable) :: this
    character(*),      intent(in)  :: colName
    integer, optional, intent(out) :: ierr

    integer :: colNum,ii,nColsUsed
    character(*), parameter :: functionName = "getColNum_tBinTable"

    colNum = 0 ! Means no column of that name found.
    if(present(ierr)) ierr=0

    if(.not. this%isInitialized) then
      if(present(ierr)) then
        ierr = 1
  return
      else
call wrstop(functionName, "Table "//trim(this%info%tableName)&
  //": you must first initialize (set the number of columns to be used, and their names etc).")
      end if
    end if

    if(.not. allocated(this%cols)) then
      if(present(ierr)) then
        ierr = 2
  return
      else
call wrstop(functionName, "Table "//trim(this%info%tableName)&
  //": your initialization routine should allocate this%cols.")
      end if
    end if

    nColsUsed = this%getNumCols()
    if(nColsUsed<1) then
      if(present(ierr)) then
        ierr = 3
  return
      else
call wrstop(functionName, "Table "//trim(this%info%tableName)&
  //": you don't have any columns.")
      end if
    end if

    do ii=1,nColsUsed
      if(trim(this%cols(ii)%nameStr) == trim(colName)) then
        colNum = ii
    exit
      end if
    end do

    if(present(ierr)) then
      if(colNum<1) then
        ierr = 4
      else
        ierr = 0
      end if
    else if(colNum<1) then
call wrstop(functionName, "Column "//trim(colName)&
  //" not found in table "//trim(this%info%tableName))
    end if

  end function getColNum_tBinTable

  !.....................................................................
  logical function read_succeeded_tBinTable(this, col_name)
    implicit none

    class(tBinTable)         :: this
    character(*), intent(in) :: col_name

    integer :: col_num

    col_num = this%getColNum(col_name)
    read_succeeded_tBinTable = this%cols(col_num)%colNumFoundInFITS

  end function read_succeeded_tBinTable

  !.....................................................................
  subroutine detect_same_rows_tBinTable(this, ri1, ri2, nColsUsed, testThisCol, colIsSame)!, dt)
    ! It is private, so can save time by passing in pre-obtained nColsUsed.
    implicit none

    class(tBinTable) :: this
    integer,                       intent(in)  :: ri1, ri2, nColsUsed
    logical, dimension(nColsUsed), intent(in)  :: testThisCol
    logical, dimension(nColsUsed), intent(out) :: colIsSame
!logical, intent(in), optional :: dt

    integer :: ii
    character(*), parameter :: functionName = "detect_same_rows_tBinTable"

    colIsSame = .false.
    do ii=1,nColsUsed
      if(testThisCol(ii) .and. this%cols(ii)%isSame(ri1, ri2))&
        colIsSame(ii) = .true.

    end do

  end subroutine detect_same_rows_tBinTable

  !.....................................................................
  subroutine flag_duplicates_tBinTable(this, list_of_col_names&
  , status, row_is_unique, mask, do_sort)!, dotest)
    ! This sorts the table according to the columns listed, then unsets
    ! mask entries for rows which are duplicate in those columns.
    !
    ! If 'mask' is supplied, rows to omit should indicated by .false.
    ! values.
    !
    ! Note that the calling routine needs to deallocate 'row_is_unique'
    ! after use.
    !
    implicit none

    class(tBinTable)                                 :: this
    character(*), dimension(:), intent(in)           :: list_of_col_names
    integer,                    intent(inout)        :: status
    logical, allocatable, dimension(:)&
      ,                         intent(out)          :: row_is_unique
    logical, dimension(:),      intent(in), optional :: mask
    logical,                    intent(in), optional :: do_sort
!    logical,                    intent(in), optional :: dotest

    character(*), parameter :: functionName = "flag_duplicates_tBinTable"
    logical, dimension(:), allocatable :: testThisCol,colIsSame
    integer :: nColsUsed,n_cols,ii,jj,ci,col_id,ii_first_nonnull,jj_last_nonnull
    type(t_sort_bintable) :: sortable
    logical, dimension(:), allocatable :: row_is_good
    logical :: local_sort

    allocate(row_is_unique(this%info%numRows))
    row_is_unique = .true. ! default

    if(this%info%numRows < 2)&
  return

    allocate(row_is_good(this%info%numRows))
    if(present(mask)) then
      row_is_good = mask
    else
      row_is_good = .true.
    end if

    if(.not.any(row_is_good))&
  return

    if(present(do_sort)) then
      local_sort = do_sort
    else
      local_sort = .false.
    end if

    nColsUsed = this%getNumCols()
    call alloc(testThisCol, nColsUsed, status)
    call alloc(colIsSame,   nColsUsed, status)

    n_cols = size(list_of_col_names)
    testThisCol = .false. ! default
    do ci=1,n_cols
      col_id = this%getColNum(list_of_col_names(ci), status)
      if(col_id<1 .or. col_id>nColsUsed)&
call wrstop(functionName, "Impossible col_id.")

      testThisCol(col_id) = .true.
    end do

    ! Check that all the columns have data:
    !
    do ci=1,nColsUsed
      if(testThisCol(ci) .and. .not.this%cols(ci)%isAllocated()) then
        call wrerr(functionName, warn, "You asked to check column "&
          //trim(this%cols(ci)%nameStr)//" for duplicate rows, but it is not allocated.")
        testThisCol(ci) = .false.
      end if
    end do

    ! Sort the table:
    !
    call sortable%init(this, list_of_col_names)
    call timsort(sortable)

    ! Find the first good row:
    !
    do ii=1,this%info%numRows
      jj = sortable%indices(ii)
      if(row_is_good(jj)) then
        ii_first_nonnull = ii
        jj_last_nonnull = jj
    exit
      end if
    end do

!local_dotest = .false.
!if(present(dotest)) local_dotest = dotest

    do ii=ii_first_nonnull+1,this%info%numRows
      jj = sortable%indices(ii)
      if(.not.row_is_good(jj))&
    cycle

      call this%detect_same_rows(jj, jj_last_nonnull, nColsUsed, testThisCol, colIsSame)!, .true.)

      if(count(colIsSame)>=count(testThisCol))&
        row_is_unique(jj) = .false.

!if(local_dotest .and..not.row_is_unique(jj)) then
!  print *, "First non-unique row:", jj
!  do ci=1,nColsUsed
!    if(testThisCol(ci)) then
!      msg = this%cols(ci)%str(jj)
!      print *, trim(this%cols(ci)%nameStr)//": "//trim(msg)
!    end if
!  end do
!stop
!end if

      jj_last_nonnull = jj
    end do

    call dealloc(testThisCol, status)
    call dealloc(colIsSame,   status)
    call dealloc(row_is_good,   status)

    if(local_sort) then
      call this%reorder(sortable%indices, status)

      ! We also need to reorder the output:
      !
      row_is_unique = row_is_unique(sortable%indices)
    end if

    call sortable%destroy()

  end subroutine flag_duplicates_tBinTable

  !.....................................................................
  subroutine del_dup_char_tBinTable(this, list_of_col_names, status&
  , mask, do_sort)!, dotest)
    ! This sorts the table according to the columns listed, then unsets
    ! mask entries for rows which are duplicate in those columns.
    !
    ! If 'mask' is supplied, rows to omit should indicated by .false.
    ! values.
    !
    implicit none

    class(tBinTable)                                 :: this
    character(*), dimension(:), intent(in)           :: list_of_col_names
    integer,                    intent(inout)        :: status
    logical, dimension(:),      intent(in), optional :: mask
    logical,                    intent(in), optional :: do_sort
!    logical,                    intent(in), optional :: dotest

    character(*), parameter :: functionName = "del_dup_char_tBinTable"
    logical, allocatable, dimension(:) :: row_is_unique
    character(buffsize) :: numRows_dup_str

!if(present(dotest)) then
! print *, trim(functionName), trim(this%info%tableName), dotest
! stop
!end if

    call this%flag_duplicates(list_of_col_names, status, row_is_unique, mask, do_sort)!, dotest)

    if(this%info%numRows > count(row_is_unique)) then
      numRows_dup_str = int2str(this%info%numRows - count(row_is_unique), status)
      call wrerr(functionName, warn, "Number of duplicate rows removed from "&
        //trim(this%info%tableName)//": " // trim(numRows_dup_str))
      call this%pack(row_is_unique, status)
    end if

    call dealloc(row_is_unique, status)

  end subroutine del_dup_char_tBinTable

  !.....................................................................
  subroutine del_dup_seq_tBinTable(this, status, mask, do_sort)!, dotest)
    implicit none

    class(tBinTable)                            :: this
    integer,               intent(inout)        :: status
    logical, dimension(:), intent(in), optional :: mask
    logical,               intent(in), optional :: do_sort
!    logical,               intent(in), optional :: dotest

    character(*), parameter :: functionName = "del_dup_seq_tBinTable"
    integer :: nColsUsed,n_entries,ii,ci
    character(buffsize), dimension(:), allocatable :: list_of_col_names
    logical, allocatable, dimension(:) :: row_is_unique
    character(buffsize) :: numRows_dup_str

!if(present(dotest)) then
! print *, trim(functionName), trim(this%info%tableName), dotest
! stop
!end if

!!!    if(.not.allocated(this%sort_sequence))&
!!!call wrstop(functionName, "blah")

    nColsUsed = this%getNumCols()

    n_entries = size(this%sort_sequence)
    call alloc(list_of_col_names, n_entries, status)
    do ii=1,n_entries
      ci = this%sort_sequence(ii)

      if(ci<1 .or. ci>nColsUsed)&
call wrstop(functionName, "Column index is out of range.")

      list_of_col_names(ii) = this%cols(ii)%nameStr
    end do

    call this%flag_duplicates(list_of_col_names, status, row_is_unique, mask, do_sort)!, dotest)

    if(this%info%numRows > count(row_is_unique)) then
      numRows_dup_str = int2str(this%info%numRows - count(row_is_unique), status)
      call wrerr(functionName, warn, "Number of duplicate rows removed from "&
        //trim(this%info%tableName)//": " // trim(numRows_dup_str))
      call this%pack(row_is_unique, status)
    end if

    call dealloc(row_is_unique, status)
    call dealloc(list_of_col_names, status)

  end subroutine del_dup_seq_tBinTable

  !.....................................................................
  logical function isAllocated_tBinTable(this)
    ! Returns .true. if one or more columns are allocated.
    implicit none

    class(tBinTable) :: this

    character(*), parameter :: functionName = "isAllocated_tBinTable"
    integer :: numColsUsed,ii

    isAllocated_tBinTable = .false. ! default

    if(.not.this%isInitialized)&
  return

    numColsUsed = this%getNumCols()
    do ii=1,numColsUsed
      if(this%cols(ii)%isAllocated()) then
        isAllocated_tBinTable = .true.
  return
      end if
    end do

  end function isAllocated_tBinTable

  !.....................................................................
  logical function col_is_allocated_tBinTable(this, col_name)
    implicit none

    class(tBinTable) :: this
    character(*), intent(in) :: col_name

    character(*), parameter :: functionName = "col_is_allocated_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)
    col_is_allocated_tBinTable = this%cols(col_num)%isAllocated()

  end function col_is_allocated_tBinTable

  !.....................................................................
  subroutine copy_col_tBinTable(this, source_table, col_name)
    ! Copies data from columns. The source and destination columns must have the same name, size and data type.
    implicit none

    class(tBinTable) :: this,source_table
    character(*), intent(in) :: col_name

    character(*), parameter :: functionName = "copy_col_tBinTable"
    integer :: col_num_src,col_num_dest,num_rows_src,num_rows_dest
    character(buffsize) :: str_src,str_dest

    call wrchatty(functionName, "Copying column "//trim(col_name)//" data from "&
      //trim(source_table%info%tableName)//" to "//trim(this%info%tableName))

    col_num_src = source_table%getColNum(col_name)
    col_num_dest = this%getColNum(col_name)

    num_rows_src = source_table%getNumRows()
    num_rows_dest = this%getNumRows()

    if(num_rows_src /= num_rows_dest) then
      str_src = int2Str(num_rows_src)
      str_dest = int2Str(num_rows_dest)
call wrstop(functionName, "Source table has "//trim(str_src)//" rows but destination table has "//trim(str_dest))
    end if

    if(num_rows_dest<1)&
  return

    if(this%cols(col_num_dest)%num_vec_elements /= source_table%cols(col_num_src)%num_vec_elements) then
      str_src = int2Str(source_table%cols(col_num_src)%num_vec_elements)
      str_dest = int2Str(this%cols(col_num_dest)%num_vec_elements)
call wrstop(functionName, "Source table has vector size "//trim(str_src)&
  //" but destination table has "//trim(str_dest))
    end if

    call this%cols(col_num_dest)%copy_data(source_table%cols(col_num_src))

  end subroutine copy_col_tBinTable

  !.....................................................................
  subroutine get_col_unit_tBinTable(this, col_name, unit_str)
    implicit none
    class(tBinTable) :: this
    character(*), intent(in)  :: col_name
    character(*), intent(out) :: unit_str

    integer :: col_num

    col_num = this%getColNum(col_name)
    call this%cols(col_num)%get_unit(unit_str)
   
  end subroutine get_col_unit_tBinTable

  !.....................................................................
  subroutine set_col_unit_tBinTable(this, col_name, unit_str)
    implicit none
    class(tBinTable) :: this
    character(*), intent(in) :: col_name,unit_str

    integer :: col_num

    col_num = this%getColNum(col_name)
    call this%cols(col_num)%set_unit(unit_str)
   
  end subroutine set_col_unit_tBinTable

  !.....................................................................
  subroutine check_col_ptr_tBinTable(this, col_num, supposed_dataTypeChar, is_vector)
    implicit none
    class(tBinTable), target :: this
    integer,      intent(in) :: col_num
    character(1), intent(in) :: supposed_dataTypeChar
    logical,      intent(in) :: is_vector

    character(*), parameter :: functionName = "check_col_ptr_tBinTable"
    integer :: vectorSize
    character(1) :: dataTypeChar

    vectorSize = this%cols(col_num)%getVectorSize()
    dataTypeChar = this%cols(col_num)%getDataType()

    if(trim(dataTypeChar) /= supposed_dataTypeChar)&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" has data type "//trim(dataTypeChar)&
  //" not "//trim(supposed_dataTypeChar))

    if(is_vector) then
      if(vectorSize <= 0)&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is a not vector column.")
    else
      if(vectorSize > 0)&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is a vector column.")
    end if

  end subroutine check_col_ptr_tBinTable

  !.....................................................................
  subroutine get_null_col_ptr_tBinTable(this, col_num, null_ptr)
    implicit none
    class(tBinTable), target :: this
    integer,                                  intent(in)  :: col_num
    logical, pointer, dimension(:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_null_col_ptr_tBinTable"

    if(present(null_ptr)) then
      if(associated(null_ptr))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//": null pointer argument is associated.")

      null_ptr => this%cols(col_num)%isNull
    end if

  end subroutine get_null_col_ptr_tBinTable

  !.....................................................................
  subroutine get_col_ptr_str_tBinTable(this, col_name, data_ptr, null_ptr)
    implicit none
    class(tBinTable), target :: this
    character(*),                               intent(in)  :: col_name
    character(buffsize), pointer, dimension(:), intent(out) :: data_ptr
    logical, pointer, dimension(:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_col_ptr_str_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)

    call this%check_col_ptr(col_num, 'A', .false.)

    if(associated(data_ptr))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//": data pointer argument is associated.")

    if(.not.allocated(this%cols(col_num)%strData))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is not allocated.")

    data_ptr => this%cols(col_num)%strData

    call this%get_null_col_ptr(col_num, null_ptr)

  end subroutine get_col_ptr_str_tBinTable

  !.....................................................................
  subroutine get_col_ptr_i1_tBinTable(this, col_name, data_ptr, null_ptr)
    implicit none
    class(tBinTable), target :: this
    character(*),                             intent(in)  :: col_name
    integer(i1), pointer, dimension(:),       intent(out) :: data_ptr
    logical, pointer, dimension(:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_col_ptr_i1_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)

    call this%check_col_ptr(col_num, 'B', .false.)

    if(associated(data_ptr))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//": data pointer argument is associated.")

    if(.not.allocated(this%cols(col_num)%i1data))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is not allocated.")

    data_ptr => this%cols(col_num)%i1data

    call this%get_null_col_ptr(col_num, null_ptr)

  end subroutine get_col_ptr_i1_tBinTable

  !.....................................................................
  subroutine get_col_ptr_i2_tBinTable(this, col_name, data_ptr, null_ptr)
    implicit none

    class(tBinTable), target :: this
    character(*),                             intent(in)  :: col_name
    integer(i2), pointer, dimension(:),       intent(out) :: data_ptr
    logical, pointer, dimension(:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_col_ptr_i2_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)

    call this%check_col_ptr(col_num, 'I', .false.)

    if(associated(data_ptr))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//": data pointer argument is associated.")

    if(.not.allocated(this%cols(col_num)%i2data))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is not allocated.")

    data_ptr => this%cols(col_num)%i2data

    call this%get_null_col_ptr(col_num, null_ptr)

  end subroutine get_col_ptr_i2_tBinTable

  !.....................................................................
  subroutine get_col_ptr_i4_tBinTable(this, col_name, data_ptr, null_ptr)
    implicit none

    class(tBinTable), target :: this
    character(*),                             intent(in)  :: col_name
    integer(i4), pointer, dimension(:),       intent(out) :: data_ptr
    logical, pointer, dimension(:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_col_ptr_i4_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)

    call this%check_col_ptr(col_num, 'J', .false.)

    if(associated(data_ptr))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//": data pointer argument is associated.")

    if(.not.allocated(this%cols(col_num)%i4data))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is not allocated.")

    data_ptr => this%cols(col_num)%i4data

    call this%get_null_col_ptr(col_num, null_ptr)

  end subroutine get_col_ptr_i4_tBinTable

  !.....................................................................
  subroutine get_col_ptr_i8_tBinTable(this, col_name, data_ptr, null_ptr)
    implicit none

    class(tBinTable), target :: this
    character(*),                             intent(in)  :: col_name
    integer(i8), pointer, dimension(:),       intent(out) :: data_ptr
    logical, pointer, dimension(:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_col_ptr_i8_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)

    call this%check_col_ptr(col_num, 'K', .false.)

    if(associated(data_ptr))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//": data pointer argument is associated.")

    if(.not.allocated(this%cols(col_num)%i8data))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is not allocated.")

    data_ptr => this%cols(col_num)%i8data

    call this%get_null_col_ptr(col_num, null_ptr)

  end subroutine get_col_ptr_i8_tBinTable

  !.....................................................................
  subroutine get_col_ptr_r4_tBinTable(this, col_name, data_ptr, null_ptr)
    implicit none

    class(tBinTable), target :: this
    character(*),                             intent(in)  :: col_name
    real(r4), pointer, dimension(:),       intent(out) :: data_ptr
    logical, pointer, dimension(:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_col_ptr_r4_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)

    call this%check_col_ptr(col_num, 'E', .false.)

    if(associated(data_ptr))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//": data pointer argument is associated.")

    if(.not.allocated(this%cols(col_num)%r4data))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is not allocated.")

    data_ptr => this%cols(col_num)%r4data

    call this%get_null_col_ptr(col_num, null_ptr)

  end subroutine get_col_ptr_r4_tBinTable

  !.....................................................................
  subroutine get_col_ptr_r8_tBinTable(this, col_name, data_ptr, null_ptr)
    implicit none

    class(tBinTable), target :: this
    character(*),                             intent(in)  :: col_name
    real(r8), pointer, dimension(:),       intent(out) :: data_ptr
    logical, pointer, dimension(:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_col_ptr_r8_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)

    call this%check_col_ptr(col_num, 'D', .false.)

    if(associated(data_ptr))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//": data pointer argument is associated.")

    if(.not.allocated(this%cols(col_num)%r8data))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is not allocated.")

    data_ptr => this%cols(col_num)%r8data

    call this%get_null_col_ptr(col_num, null_ptr)

  end subroutine get_col_ptr_r8_tBinTable

  !.....................................................................
  subroutine get_col_ptr_bool_tBinTable(this, col_name, data_ptr, null_ptr)
    implicit none

    class(tBinTable), target :: this
    character(*),                             intent(in)  :: col_name
    logical, pointer, dimension(:),           intent(out) :: data_ptr
    logical, pointer, dimension(:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_col_ptr_bool_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)

    call this%check_col_ptr(col_num, 'L', .false.)

    if(associated(data_ptr))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//": data pointer argument is associated.")

    if(.not.allocated(this%cols(col_num)%b0data))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is not allocated.")

    data_ptr => this%cols(col_num)%b0data

    call this%get_null_col_ptr(col_num, null_ptr)

  end subroutine get_col_ptr_bool_tBinTable

  !.....................................................................
  subroutine get_vec_null_col_ptr_tBinTable(this, col_num, null_ptr)
    implicit none
    class(tBinTable), target :: this
    integer,                                    intent(in)  :: col_num
    logical, pointer, dimension(:,:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_vec_null_col_ptr_tBinTable"

    if(present(null_ptr)) then
      if(associated(null_ptr))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//": null pointer argument is associated.")

      null_ptr => this%cols(col_num)%isNullVec
    end if

  end subroutine get_vec_null_col_ptr_tBinTable

  !.....................................................................
  subroutine get_col_ptr_i1_vec_tBinTable(this, col_name, data_ptr, null_ptr)
    implicit none

    class(tBinTable), target :: this
    character(*),                               intent(in)  :: col_name
    integer(i1), pointer, dimension(:,:),       intent(out) :: data_ptr
    logical, pointer, dimension(:,:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_col_ptr_i1_vec_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)

    call this%check_col_ptr(col_num, 'B', .true.)

    if(.not.allocated(this%cols(col_num)%i1vecdata))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is not allocated.")

    data_ptr => this%cols(col_num)%i1vecdata

    call this%get_vec_null_col_ptr(col_num, null_ptr)

  end subroutine get_col_ptr_i1_vec_tBinTable

  !.....................................................................
  subroutine get_col_ptr_i2_vec_tBinTable(this, col_name, data_ptr, null_ptr)
    implicit none

    class(tBinTable), target :: this
    character(*),                               intent(in)  :: col_name
    integer(i2), pointer, dimension(:,:),       intent(out) :: data_ptr
    logical, pointer, dimension(:,:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_col_ptr_i2_vec_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)

    call this%check_col_ptr(col_num, 'I', .true.)

    if(associated(data_ptr))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//": data pointer argument is associated.")

    if(.not.allocated(this%cols(col_num)%i2vecdata))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is not allocated.")

    data_ptr => this%cols(col_num)%i2vecdata

    call this%get_vec_null_col_ptr(col_num, null_ptr)

  end subroutine get_col_ptr_i2_vec_tBinTable

  !.....................................................................
  subroutine get_col_ptr_i4_vec_tBinTable(this, col_name, data_ptr, null_ptr)
    implicit none

    class(tBinTable), target :: this
    character(*),                               intent(in)  :: col_name
    integer(i4), pointer, dimension(:,:),       intent(out) :: data_ptr
    logical, pointer, dimension(:,:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_col_ptr_i4_vec_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)

    call this%check_col_ptr(col_num, 'J', .true.)

    if(associated(data_ptr))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//": data pointer argument is associated.")

    if(.not.allocated(this%cols(col_num)%i4vecdata))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is not allocated.")

    data_ptr => this%cols(col_num)%i4vecdata

    call this%get_vec_null_col_ptr(col_num, null_ptr)

  end subroutine get_col_ptr_i4_vec_tBinTable

  !.....................................................................
  subroutine get_col_ptr_i8_vec_tBinTable(this, col_name, data_ptr, null_ptr)
    implicit none

    class(tBinTable), target :: this
    character(*),                               intent(in)  :: col_name
    integer(i8), pointer, dimension(:,:),       intent(out) :: data_ptr
    logical, pointer, dimension(:,:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_col_ptr_i8_vec_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)

    call this%check_col_ptr(col_num, 'K', .true.)

    if(associated(data_ptr))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//": data pointer argument is associated.")

    if(.not.allocated(this%cols(col_num)%i8vecdata))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is not allocated.")

    data_ptr => this%cols(col_num)%i8vecdata

    call this%get_vec_null_col_ptr(col_num, null_ptr)

  end subroutine get_col_ptr_i8_vec_tBinTable

  !.....................................................................
  subroutine get_col_ptr_r4_vec_tBinTable(this, col_name, data_ptr, null_ptr)
    implicit none

    class(tBinTable), target :: this
    character(*),                               intent(in)  :: col_name
    real(r4), pointer, dimension(:,:),          intent(out) :: data_ptr
    logical, pointer, dimension(:,:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_col_ptr_r4_vec_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)

    call this%check_col_ptr(col_num, 'E', .true.)

    if(associated(data_ptr))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//": data pointer argument is associated.")

    if(.not.allocated(this%cols(col_num)%r4vecdata))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is not allocated.")

    data_ptr => this%cols(col_num)%r4vecdata

    call this%get_vec_null_col_ptr(col_num, null_ptr)

  end subroutine get_col_ptr_r4_vec_tBinTable

  !.....................................................................
  subroutine get_col_ptr_r8_vec_tBinTable(this, col_name, data_ptr, null_ptr)
    implicit none

    class(tBinTable), target :: this
    character(*),                               intent(in)  :: col_name
    real(r8), pointer, dimension(:,:),          intent(out) :: data_ptr
    logical, pointer, dimension(:,:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_col_ptr_r8_vec_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)

    call this%check_col_ptr(col_num, 'D', .true.)

    if(associated(data_ptr))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//": data pointer argument is associated.")

    if(.not.allocated(this%cols(col_num)%r8vecdata))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is not allocated.")

    data_ptr => this%cols(col_num)%r8vecdata

    call this%get_vec_null_col_ptr(col_num, null_ptr)

  end subroutine get_col_ptr_r8_vec_tBinTable

  !.....................................................................
  subroutine get_col_ptr_bool_vec_tBinTable(this, col_name, data_ptr, null_ptr)
    implicit none

    class(tBinTable), target :: this
    character(*),                               intent(in)  :: col_name
    logical, pointer, dimension(:,:),           intent(out) :: data_ptr
    logical, pointer, dimension(:,:), optional, intent(out) :: null_ptr

    character(*), parameter :: functionName = "get_col_ptr_bool_vec_tBinTable"
    integer :: col_num

    col_num = this%getColNum(col_name)

    call this%check_col_ptr(col_num, 'L', .true.)

    if(associated(data_ptr))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//": data pointer argument is associated.")

    if(.not.allocated(this%cols(col_num)%b0vecdata))&
call wrstop(functionName, "Column "//trim(this%cols(col_num)%nameStr)&
  //" of table "//trim(this%info%tableName)//" is not allocated.")

    data_ptr => this%cols(col_num)%b0vecdata

    call this%get_vec_null_col_ptr(col_num, null_ptr)

  end subroutine get_col_ptr_bool_vec_tBinTable

  !.....................................................................
  logical function can_append_tBinTable(this, new_tab) result(theyAreMergable)
    implicit none

    class(tBinTable) :: this, new_tab

    character(*), parameter :: functionName = "can_append_tBinTable"
    integer :: ii,nColsUsedA,nColsUsedB
    character(buffsize) :: strA,strB

    nColsUsedA = this%getNumCols()
    nColsUsedB = new_tab%getNumCols()

    theyAreMergable=.false.

    if(nColsUsedA /= nColsUsedB) then
      call wrerr(functionName, s_err, "Cannot merge tables - different numbers of columns.")
  return
    end if
    if(this%isInitialized .neqv. new_tab%isInitialized) then
      call wrerr(functionName, s_err, "Cannot merge tables - they are in different states of initialization.")
  return
    end if

    if(allocated(this%cols) .and. allocated(new_tab%cols)) then
      do ii = 1, nColsUsedA
        if(this%cols(ii)%nameStr .ne. new_tab%cols(ii)%nameStr) then
          call wrerr(functionName, s_err, "Cannot merge tables - column names"&
            //trim(this%cols(ii)%nameStr)//" and "&
            //trim(new_tab%cols(ii)%nameStr)//" not equal.")
  return
        end if

        if(this%cols(ii)%dataTypeFITS .ne. new_tab%cols(ii)%dataTypeFITS) then
          call wrerr(functionName, s_err, "Cannot merge tables - data types"&
            //trim(this%cols(ii)%dataTypeFITS)//" and "&
            //trim(new_tab%cols(ii)%dataTypeFITS)//" for column "&
            //trim(this%cols(ii)%nameStr)//" are not equal.")
  return
        end if

        if(this%cols(ii)%scal%containsData .neqv. new_tab%cols(ii)%scal%containsData) then
          call wrerr(functionName, s_err, "Cannot merge tables - for column "&
            //trim(this%cols(ii)%nameStr)//" one table has scale kwds, the other not.")
  return
        end if

        if(this%cols(ii)%num_chars /= new_tab%cols(ii)%num_chars) then
          strA = int2str(this%cols(ii)%num_chars)
          strB = int2str(new_tab%cols(ii)%num_chars)
          call wrerr(functionName, s_err, "Cannot merge tables - num_chars"&
            //trim(strA)//" and "//trim(strB)//" for column "&
            //trim(this%cols(ii)%nameStr)//" are not equal.")
  return
        end if

        if(this%cols(ii)%num_vec_elements /= new_tab%cols(ii)%num_vec_elements) then
          strA = int2str(this%cols(ii)%num_vec_elements)
          strB = int2str(new_tab%cols(ii)%num_vec_elements)
          call wrerr(functionName, s_err, "Cannot merge tables - vector sizes"&
            //trim(strA)//" and "//trim(strB)//" for column "&
            //trim(this%cols(ii)%nameStr)//" are not equal.")
  return
        end if
      end do
    end if

    theyAreMergable = .true.

  end function can_append_tBinTable

  !.....................................................................
  subroutine append_tBinTable(this, new_tab, status)
    implicit none

    class(tBinTable) :: this, new_tab
    integer, intent(inout) :: status

    integer :: ii,nColsUsedA
    character(*), parameter :: functionName = "append_tBinTable"

    if(status > 100 .or. status == s_errRep)&
  return

    nColsUsedA = this%getNumCols()

    if(this%info%numRows <= 0 .and. new_tab%info%numRows <= 0) then
      call wrlog(functionName, "Tables to merge have zero elements - nothing to be done here")
  return
    end if

    if(.not.this%can_append(new_tab)) then
      call wrerr(functionName, s_err, "Tables are not mergable.")
      status = s_errRep
  return
    end if

    if(.not.allocated(this%cols))&
  return

    do ii = 1, nColsUsedA
      call this%cols(ii)%append(new_tab%cols(ii), status)
    end do

    this%info%numRows = this%info%numRows + new_tab%info%numRows

  end subroutine append_tBinTable

  !.....................................................................
  subroutine set_col_blank_tBinTable(this, col_name, fillType)
    ! Fill the column with zeros or nulls as required.
    implicit none

    class(tBinTable)                   :: this
    character(*), intent(in)           :: col_name
    integer,      intent(in), optional :: fillType

    character(*), parameter :: functionName = "set_col_blank_tBinTable"
    integer :: col_i,ierr

    ierr = 0

    col_i = this%getColNum(col_name)
    if(present(fillType)) then
      call this%cols(col_i)%blanks(fillType, ierr)
    else
      call this%cols(col_i)%blanks(TCOL_FILL_ZEROES, ierr)
    end if

  end subroutine set_col_blank_tBinTable

  !=====================================================================
  ! Methods of t_sort_bintable:

  !.....................................................................
  subroutine init_t_sort_bintable(this, bintable_obj, column_list)
    implicit none
    class(t_sort_bintable)                         :: this
    class(tBinTable),           intent(in), target :: bintable_obj
    character(*), dimension(:), intent(in)         :: column_list

    character(*), parameter :: functionName = "init_t_sort_bintable"
    integer :: ii

    this%bintable_obj => bintable_obj

    this%n_cols = size(column_list)
    allocate(this%col_comparators(this%n_cols)) !*** add error stuff
    do ii=1,this%n_cols
      call this%col_comparators(ii)%init(bintable_obj, column_list(ii))
    end do

    this%is_initialized = .true.

  end subroutine init_t_sort_bintable

  !.....................................................................
  integer function get_n_elements_t_sort_bintable(this) result(n_elements)
    implicit none
    class(t_sort_bintable) :: this

    character(*), parameter :: functionName = "get_n_elements_t_sort_bintable"

    if(.not.this%is_initialized)&
call wrstop(functionName, "You must initialize the sortable object first.")

    n_elements = this%bintable_obj%getNumRows()

  end function get_n_elements_t_sort_bintable

  !.....................................................................
  integer function compare_t_sort_bintable(this, i1, i2) result(comp)
    implicit none
    class(t_sort_bintable) :: this
    integer, intent(in)  :: i1, i2

    character(*), parameter :: functionName = "compare_t_sort_bintable"
    integer :: ii

    ! Compare column values in the name order given upon init. Only
    ! check values of column j if value pairs for all preceding columns
    ! have been equal.
    !
    do ii=1,this%n_cols
      comp = this%col_comparators(ii)%compare(i1, i2)
      if(comp/=COMP_EQ)&
    exit
    end do

  end function compare_t_sort_bintable

  !.....................................................................
  subroutine check_t_sort_bintable(this)
    ! Checks to make sure that the sort worked as intended (this is probably only needed in the early days of timsort testing).
    implicit none
    class(t_sort_bintable) :: this

    character(*), parameter :: functionName = "check_t_sort_bintable"
    integer :: n_elements,ii,comp
    character(buffsize) :: istr_0,istr_1

    if(.not.this%is_initialized)&
call wrstop(functionName, "You must initialize the sortable object first.")

    n_elements = this%get_n_elements()
    do ii=2,n_elements
      comp = this%compare(this%indices(ii-1),this%indices(ii))
      if(comp==COMP_GT) then
        istr_0 = int2str(this%indices(ii-1))
        istr_1 = int2str(this%indices(ii))
call wrstop(functionName, "Element "//trim(istr_0)//" should be <= element "//trim(istr_1))
      end if
    end do

  end subroutine check_t_sort_bintable

  !.....................................................................
  subroutine destroy_t_sort_bintable(this)
    implicit none
    class(t_sort_bintable) :: this

    integer :: ii

    do ii=1,this%n_cols
      call this%col_comparators(ii)%destroy()
    end do
    deallocate(this%col_comparators) !*** add error stuff
    this%n_cols = 0

    deallocate(this%indices) !*** add error stuff
    nullify(this%bintable_obj)

    this%is_initialized = .false.

  end subroutine destroy_t_sort_bintable

  !=====================================================================
  ! The remaining routines are not methods of any object, they're just
  ! utility routines that manipulate the objects.

  !.....................................................................
  subroutine dealloc_bintables(tables, status)
    ! This deallocates an array of tBinTable objects.
!*** this might need a bit of work.
    implicit none

    class(tBinTable), allocatable, dimension(:) :: tables
    integer, intent(inout) :: status

    integer :: stat
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "dealloc_bintables"

    if(allocated(tables)) then
      errmsg = ""
      deallocate(tables, stat=stat, errmsg=errmsg)
      call FError(functionName, stat, errmsg, status)
    end if

  end subroutine dealloc_bintables

  !.....................................................................
  subroutine alloc_bintables(tables, n, status)
    ! This allocates an array of tBinTable objects.
!*** this might need a bit of work.
    implicit none

    class(tBinTable), allocatable, dimension(:) :: tables
    integer, intent(in) :: n
    integer, intent(inout) :: status

    integer :: stat
    character(buffsize) :: errmsg
    character(*), parameter :: functionName = "alloc_bintables"

    call dealloc_bintables(tables, status)
    errmsg = ""
    allocate(tables(n), stat=stat, errmsg=errmsg)
    call FError(functionName, stat, errmsg, status)

  end subroutine alloc_bintables

  !.....................................................................
  subroutine copy_bintable_scalars(tableIn, tableOut)
!**** should make this private because we don't want the user to be encouraged to call it.
    implicit none

    class(tBinTable), intent(in)    :: tableIn
    class(tBinTable), intent(inout) :: tableOut

    if(associated(tableIn%info%fileInfo)) then
      tableOut%info%fileInfo => tableIn%info%fileInfo
    end if

    tableOut%header                 = tableIn%header
    tableOut%info%tableName         = tableIn%info%tableName
    tableOut%info%numRows           = tableIn%info%numRows
    tableOut%isInitialized          = tableIn%isInitialized
    tableOut%info%haveAdvancedToHDU = tableIn%info%haveAdvancedToHDU
    tableOut%haveReadColNums        = tableIn%haveReadColNums

  end subroutine copy_bintable_scalars

  !.....................................................................
  function checkMergable_bintables(tableInA, tableInB) result(theyAreMergable)
!**** should make this private because we don't want the user to be encouraged to call it.
!**** Hmm. I didn't make it private because it is used in evtool2_eventfile_routines_mod.f90.
    implicit none

    class(tBinTable), intent(in)  :: tableInA,tableInB

    character(*), parameter :: functionName = "checkMergable_bintables"
    logical :: theyAreMergable
    integer :: ii,nColsUsedA,nColsUsedB
    character(buffsize) :: strA,strB

    nColsUsedA = tableInA%getNumCols()
    nColsUsedB = tableInB%getNumCols()

    theyAreMergable=.false.

    if(nColsUsedA /= nColsUsedB) then
      call wrerr(functionName, s_err, "Cannot merge tables - different numbers of columns.")
  return
    end if
    if(tableInA%isInitialized .neqv. tableInB%isInitialized) then
      call wrerr(functionName, s_err, "Cannot merge tables - they are in different states of initialization.")
  return
    end if

    if(allocated(tableInA%cols) .and. allocated(tableInB%cols)) then
      do ii = 1, nColsUsedA
        if(tableInA%cols(ii)%nameStr .ne. tableInB%cols(ii)%nameStr) then
          call wrerr(functionName, s_err, "Cannot merge tables - column names"&
            //trim(tableInA%cols(ii)%nameStr)//" and "&
            //trim(tableInB%cols(ii)%nameStr)//" not equal.")
  return
        end if

        if(tableInA%cols(ii)%dataTypeFITS .ne. tableInB%cols(ii)%dataTypeFITS) then
          call wrerr(functionName, s_err, "Cannot merge tables - data types"&
            //trim(tableInA%cols(ii)%dataTypeFITS)//" and "&
            //trim(tableInB%cols(ii)%dataTypeFITS)//" for column "&
            //trim(tableInA%cols(ii)%nameStr)//" are not equal.")
  return
        end if

        if(tableInA%cols(ii)%scal%containsData .neqv. tableInB%cols(ii)%scal%containsData) then
          call wrerr(functionName, s_err, "Cannot merge tables - for column "&
            //trim(tableInA%cols(ii)%nameStr)//" one table has scale kwds, the other not.")
  return
        end if

        if(tableInA%cols(ii)%num_chars /= tableInB%cols(ii)%num_chars) then
          strA = int2str(tableInA%cols(ii)%num_chars)
          strB = int2str(tableInB%cols(ii)%num_chars)
          call wrerr(functionName, s_err, "Cannot merge tables - num_chars"&
            //trim(strA)//" and "//trim(strB)//" for column "&
            //trim(tableInA%cols(ii)%nameStr)//" are not equal.")
  return
        end if

        if(tableInA%cols(ii)%num_vec_elements /= tableInB%cols(ii)%num_vec_elements) then
          strA = int2str(tableInA%cols(ii)%num_vec_elements)
          strB = int2str(tableInB%cols(ii)%num_vec_elements)
          call wrerr(functionName, s_err, "Cannot merge tables - vector sizes"&
            //trim(strA)//" and "//trim(strB)//" for column "&
            //trim(tableInA%cols(ii)%nameStr)//" are not equal.")
  return
        end if
      end do
    end if

    theyAreMergable = .true.

  end function checkMergable_bintables

  !.....................................................................
  subroutine merge_two_bintables(tableInA, tableInB, tableOut, status)
    implicit none

    class(tBinTable), intent(in)    :: tableInA,tableInB
    class(tBinTable), intent(out)   :: tableOut
    integer,          intent(inout) :: status

    integer :: ii,nColsUsedA
    character(*), parameter :: functionName = "merge_two_bintables"

    if(status > 100 .or. status == s_errRep) return

    nColsUsedA = tableInA%getNumCols()

    if(tableInA%info%numRows <= 0 .and. tableInB%info%numRows <= 0) then
      call wrlog(functionName, "Tables to merge have zero elements - nothing to be done here")
  return
    end if

    if(.not.checkMergable_bintables(tableInA, tableInB)) then
      call wrerr(functionName, s_err, "Tables are not mergable.")
      status = s_errRep
  return
    end if

    call copy_bintable_scalars(tableInA, tableOut)
    tableOut%info%numRows = tableInA%info%numRows + tableInB%info%numRows

    if(.not.allocated(tableInA%cols))&
  return

    do ii = 1, nColsUsedA
      if(tableInA%cols(ii)%isAllocated() .and. tableInB%cols(ii)%isAllocated())&
!*** somewhere now also check that the vectority of the columns is the same.
        call merge_columns(tableInA%cols(ii), tableInB%cols(ii), tableOut%cols(ii), status)
    end do

  end subroutine merge_two_bintables

  !.....................................................................
  subroutine merge_bintable_ptrs(listOfTablePtrs, tableOut, status)
    implicit none

    type(tBinTablePtr), dimension(:), intent(in)    :: listOfTablePtrs
    class(tBinTable),             intent(out)   :: tableOut
    integer,                          intent(inout) :: status

    integer :: nfiles,fi,numRows
    character(*), parameter :: functionName = "merge_bintable_ptrs"

    if(status > 100 .or. status == s_errRep) return

    if(.not. tableOut%isInitialized) then
      call wrerr(functionName, s_err, "Output binary table is not initialized.")
      status = s_errRep
  return
    end if

    nfiles = size(listOfTablePtrs)

    numRows = listOfTablePtrs(1)%ptr%info%numRows
    do fi=2,nfiles
      if(.not.checkMergable_bintables(listOfTablePtrs(1)%ptr, listOfTablePtrs(fi)%ptr)) then
        call wrerr(functionName, s_err, "Tables are not mergable.")
        status = s_errRep
  return
      end if
      numRows = numRows + listOfTablePtrs(fi)%ptr%info%numRows
    end do

    call copy_bintable_scalars(listOfTablePtrs(1)%ptr, tableOut)
    call tableOut%alloc(status, numRows)

    numRows = 0
    do fi=1,nfiles
      call tableOut%overwrite_section(listOfTablePtrs(fi)%ptr, numRows+1, status)
      numRows = numRows + listOfTablePtrs(fi)%ptr%info%numRows
    end do

  end subroutine merge_bintable_ptrs

  !=====================================================================
  ! Methods of tFitsFile:

  subroutine open_tFitsFile(this, rwmode, status, fileName, clobber)
    implicit none

    class(tFitsFile) :: this
    integer,              intent(in)    :: rwmode
    integer,              intent(inout) :: status
    character(*),optional,intent(in)    :: fileName
    logical,     optional,intent(in)    :: clobber

    integer :: delete
    character(*), parameter :: functionName = "open_tFitsFile"

    if(this%info%fileOpenedForRead .or. this%info%fileOpenedForWrite)&
call wrstop(functionName, "File appears to be already open.")

    delete = 0
    if(present(clobber)) then
      if(clobber) delete = 1
    end if

    if(present(fileName)) then
      call this%info%setFileName(fileName)
    else if(this%info%fileName == "") then
      call wrerr(functionName, s_err, "No filename provided.")
      status = s_errRep
    end if

    call open_FITS_file(this%info%fileName, rwmode, this%info%unitNo, status, delete)

    if(status == s_errRep)&
call wrstop(functionName, "Failed to open FITS file "//trim(this%info%fileName))

    if(rwmode==FF_READONLY .or. rwmode==FF_READWRITE)&
      this%info%fileOpenedForRead = .true.
    if(rwmode==FF_CREATE .or. rwmode==FF_READWRITE)&
      this%info%fileOpenedForWrite = .true.

  end subroutine open_tFitsFile

  !.....................................................................
!  subroutine read_tFitsFile(this, status, fileName)
!    implicit none
!    class(tFitsFile)                      :: this
!    integer,                intent(inout) :: status
!    character(*), optional, intent(in)    :: fileName
!
!    character(*), parameter :: functionName = "read_tFitsFile"
!
!    if(present(fileName)) call this%open(FF_READONLY, status, filename=fileName)
!!!!    call this%read_extensions(status)
!    if(present(fileName)) call this%close(status)
!
!  end subroutine read_tFitsFile

  !.....................................................................
!  subroutine read_ext_tFitsFile(this, status)
!    implicit none
!
!    class(tFitsFile)       :: this
!    integer, intent(inout) :: status
!
!    character(*), parameter :: functionName = "read_ext_tFitsFile"
!
!call wrstop(functionName, "You must replace this method by one tailored to your particular tFitsFile subclass.")
!
!  end subroutine read_ext_tFitsFile

  !.....................................................................
  subroutine close_tFitsFile(this, status)
    implicit none

    class(tFitsFile) :: this
    integer, intent(inout) :: status

    character(*), parameter :: functionName = "close_tFitsFile"

    if(.not.this%info%fileOpenedForRead .and. .not.this%info%fileOpenedForWrite)&
      call wrerr(functionName, warn, "File "//trim(this%info%fileName)//" appears to be already closed.")

    call close_FITS_file(this%info%unitNo, status)
    this%info%fileOpenedForRead  = .false.
    this%info%fileOpenedForWrite = .false.

  end subroutine close_tFitsFile

end module erolib_bintable_mod

